/* Copyright (c) 2024 The Khronos Group Inc.
 * Copyright (c) 2024 Valve Corporation
 * Copyright (c) 2024 LunarG, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

namespace chassis {

struct CommandBufferData {
    bool is_secondary;
};

// Some handles need their own "private data" that everyone needs (even statless chassis layers)
// The goal is to have the chassis do the state tracking once, and pass down this selective information to everyone.
// Using an union, we can leverage the ErrorObject/RecordObject to get the data everywhere
union HandleData {
    CommandBufferData command_buffer;
};

}  // namespace chassis