// *** THIS FILE IS GENERATED - DO NOT EDIT ***
// See generate_spirv.py for modifications

/***************************************************************************
 *
 * Copyright (c) 2021-2026 The Khronos Group Inc.
 * Copyright (c) 2021-2026 Valve Corporation
 * Copyright (c) 2021-2026 LunarG, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ****************************************************************************/

#pragma once

#include <cstdint>

// We have found having spirv code defined the header can lead to MSVC not recognizing changes

extern const uint32_t instrumentation_buffer_device_address_comp_size;
extern const uint32_t instrumentation_buffer_device_address_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_buffer_device_address_comp_function_0_offset;
extern const uint32_t instrumentation_buffer_device_address_comp_function_1_offset;

extern const uint32_t instrumentation_descriptor_buffer_comp_size;
extern const uint32_t instrumentation_descriptor_buffer_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_descriptor_buffer_comp_function_0_offset;

extern const uint32_t instrumentation_descriptor_class_general_buffer_comp_size;
extern const uint32_t instrumentation_descriptor_class_general_buffer_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_descriptor_class_general_buffer_comp_function_0_offset;
extern const uint32_t instrumentation_descriptor_class_general_buffer_comp_function_1_offset;

extern const uint32_t instrumentation_descriptor_class_texel_buffer_comp_size;
extern const uint32_t instrumentation_descriptor_class_texel_buffer_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_descriptor_class_texel_buffer_comp_function_0_offset;

extern const uint32_t instrumentation_descriptor_heap_comp_size;
extern const uint32_t instrumentation_descriptor_heap_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_descriptor_heap_comp_function_0_offset;
extern const uint32_t instrumentation_descriptor_heap_comp_function_1_offset;
extern const uint32_t instrumentation_descriptor_heap_comp_function_2_offset;
extern const uint32_t instrumentation_descriptor_heap_comp_function_3_offset;
extern const uint32_t instrumentation_descriptor_heap_comp_function_4_offset;
extern const uint32_t instrumentation_descriptor_heap_comp_function_5_offset;
extern const uint32_t instrumentation_descriptor_heap_comp_function_6_offset;
extern const uint32_t instrumentation_descriptor_heap_comp_function_7_offset;
extern const uint32_t instrumentation_descriptor_heap_comp_function_8_offset;
extern const uint32_t instrumentation_descriptor_heap_comp_function_9_offset;

extern const uint32_t instrumentation_descriptor_indexing_oob_comp_size;
extern const uint32_t instrumentation_descriptor_indexing_oob_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_descriptor_indexing_oob_comp_function_0_offset;
extern const uint32_t instrumentation_descriptor_indexing_oob_comp_function_1_offset;
extern const uint32_t instrumentation_descriptor_indexing_oob_comp_function_2_offset;

extern const uint32_t instrumentation_log_error_comp_size;
extern const uint32_t instrumentation_log_error_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_log_error_comp_function_0_offset;

extern const uint32_t instrumentation_mesh_shading_comp_size;
extern const uint32_t instrumentation_mesh_shading_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_mesh_shading_comp_function_0_offset;
extern const uint32_t instrumentation_mesh_shading_comp_function_1_offset;

extern const uint32_t instrumentation_post_process_descriptor_index_comp_size;
extern const uint32_t instrumentation_post_process_descriptor_index_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_post_process_descriptor_index_comp_function_0_offset;

extern const uint32_t instrumentation_sanitizer_comp_size;
extern const uint32_t instrumentation_sanitizer_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_sanitizer_comp_function_0_offset;
extern const uint32_t instrumentation_sanitizer_comp_function_1_offset;
extern const uint32_t instrumentation_sanitizer_comp_function_2_offset;
extern const uint32_t instrumentation_sanitizer_comp_function_3_offset;
extern const uint32_t instrumentation_sanitizer_comp_function_4_offset;
extern const uint32_t instrumentation_sanitizer_comp_function_5_offset;

extern const uint32_t instrumentation_shared_memory_data_race_comp_size;
extern const uint32_t instrumentation_shared_memory_data_race_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_shared_memory_data_race_comp_function_0_offset;
extern const uint32_t instrumentation_shared_memory_data_race_comp_function_1_offset;
extern const uint32_t instrumentation_shared_memory_data_race_comp_function_2_offset;
extern const uint32_t instrumentation_shared_memory_data_race_comp_function_3_offset;
extern const uint32_t instrumentation_shared_memory_data_race_comp_function_4_offset;
extern const uint32_t instrumentation_shared_memory_data_race_comp_function_5_offset;

extern const uint32_t instrumentation_trace_ray_comp_size;
extern const uint32_t instrumentation_trace_ray_comp[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_trace_ray_comp_function_0_offset;
extern const uint32_t instrumentation_trace_ray_comp_function_1_offset;
extern const uint32_t instrumentation_trace_ray_comp_function_2_offset;
extern const uint32_t instrumentation_trace_ray_comp_function_3_offset;
extern const uint32_t instrumentation_trace_ray_comp_function_4_offset;
extern const uint32_t instrumentation_trace_ray_comp_function_5_offset;

extern const uint32_t instrumentation_vertex_attribute_fetch_oob_vert_size;
extern const uint32_t instrumentation_vertex_attribute_fetch_oob_vert[];
// These offset match the function in the order they are declared in the GLSL source
extern const uint32_t instrumentation_vertex_attribute_fetch_oob_vert_function_0_offset;

extern const uint32_t setup_acceleration_structure_gpu_state_update_comp_size;
extern const uint32_t setup_acceleration_structure_gpu_state_update_comp[];

extern const uint32_t setup_descriptor_encoding_update_comp_size;
extern const uint32_t setup_descriptor_encoding_update_comp[];

extern const uint32_t validation_cmd_blas_comp_size;
extern const uint32_t validation_cmd_blas_comp[];

extern const uint32_t validation_cmd_copy_buffer_to_image_comp_size;
extern const uint32_t validation_cmd_copy_buffer_to_image_comp[];

extern const uint32_t validation_cmd_copy_memory_indirect_comp_size;
extern const uint32_t validation_cmd_copy_memory_indirect_comp[];

extern const uint32_t validation_cmd_count_buffer_comp_size;
extern const uint32_t validation_cmd_count_buffer_comp[];

extern const uint32_t validation_cmd_dispatch_comp_size;
extern const uint32_t validation_cmd_dispatch_comp[];

extern const uint32_t validation_cmd_draw_indexed_indirect_index_buffer_comp_size;
extern const uint32_t validation_cmd_draw_indexed_indirect_index_buffer_comp[];

extern const uint32_t validation_cmd_draw_mesh_indirect_comp_size;
extern const uint32_t validation_cmd_draw_mesh_indirect_comp[];

extern const uint32_t validation_cmd_first_instance_comp_size;
extern const uint32_t validation_cmd_first_instance_comp[];

extern const uint32_t validation_cmd_memcmp_comp_size;
extern const uint32_t validation_cmd_memcmp_comp[];

extern const uint32_t validation_cmd_setup_draw_indexed_indirect_index_buffer_comp_size;
extern const uint32_t validation_cmd_setup_draw_indexed_indirect_index_buffer_comp[];

extern const uint32_t validation_cmd_tlas_comp_size;
extern const uint32_t validation_cmd_tlas_comp[];

extern const uint32_t validation_cmd_trace_rays_comp_size;
extern const uint32_t validation_cmd_trace_rays_comp[];
