/* Copyright (c) 2024-2026 LunarG, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#pragma once

namespace gpuav {
class Validator;
class CommandBufferSubState;

void RegisterBufferDeviceAddressValidation(Validator& gpuav, CommandBufferSubState& cb);
void RegisterPostProcessingValidation(Validator& gpuav, CommandBufferSubState& cb);
void RegisterMeshShadingValidation(Validator& gpuav, CommandBufferSubState& cb);
void RegisterSanitizer(Validator& gpuav, CommandBufferSubState& cb);
void RegisterVertexAttributeFetchOobValidation(Validator& gpuav, CommandBufferSubState& cb);
void RegisterSharedMemoryDataRaceValidation(Validator& gpuav, CommandBufferSubState& cb);
void RegisterTraceRayValidation(Validator& gpuav, CommandBufferSubState& cb);
void UpdateAccelerationStructureGpuState(Validator& gpuav, CommandBufferSubState& cb, const Location& loc, uint32_t info_count,
                                         const VkAccelerationStructureBuildGeometryInfoKHR* infos);

}  // namespace gpuav
