/* Copyright (c) 2024-2025 The Khronos Group Inc.
 * Copyright (c) 2024-2025 Valve Corporation
 * Copyright (c) 2024-2025 LunarG, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#pragma once
#include <stdint.h>

namespace vvl {

class Queue;

// References a specific submission on the given queue.
// The core validation uses the "submission" term to refer to a separate batch from the submission command.
// Synchronization validation follows more closely the specification and uses "batch" term.
struct SubmissionReference {
    Queue *queue = nullptr;
    uint64_t seq = 0;

    SubmissionReference() = default;
    SubmissionReference(Queue *q, uint64_t queue_seq) : queue(q), seq(queue_seq) {}
};

}  // namespace vvl
