// *** THIS FILE IS GENERATED - DO NOT EDIT ***
// See extended_flags_helper_generator.py for modifications

/***************************************************************************
 *
 * Copyright (c) 2015-2026 The Khronos Group Inc.
 * Copyright (c) 2015-2026 Valve Corporation
 * Copyright (c) 2015-2026 LunarG, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ****************************************************************************/
// NOLINTBEGIN

#include "error_message/error_location.h"

VkBufferUsageFlags2 GetBufferUsageFlags(const VkBufferCreateInfo& create_info);
Location GetUsageLocation(const VkBufferCreateInfo& create_info, const Location& loc);

VkImageCreateFlags2KHR GetImageCreateFlags(const VkImageCreateInfo& create_info);
Location GetFlagsLocation(const VkImageCreateInfo& create_info, const Location& loc);

VkImageUsageFlags2KHR GetImageUsageFlags(const VkImageCreateInfo& create_info);
Location GetUsageLocation(const VkImageCreateInfo& create_info, const Location& loc);

VkPipelineCreateFlags2 GetPipelineCreateFlags(const VkComputePipelineCreateInfo& create_info);
Location GetFlagsLocation(const VkComputePipelineCreateInfo& create_info, const Location& loc);

VkPipelineCreateFlags2 GetPipelineCreateFlags(const VkGraphicsPipelineCreateInfo& create_info);
Location GetFlagsLocation(const VkGraphicsPipelineCreateInfo& create_info, const Location& loc);

VkImageUsageFlags2KHR GetImageUsageFlags(const VkPhysicalDeviceImageFormatInfo2& create_info);
Location GetUsageLocation(const VkPhysicalDeviceImageFormatInfo2& create_info, const Location& loc);

VkImageCreateFlags2KHR GetImageCreateFlags(const VkPhysicalDeviceImageFormatInfo2& create_info);
Location GetFlagsLocation(const VkPhysicalDeviceImageFormatInfo2& create_info, const Location& loc);

VkImageUsageFlags2KHR GetImageUsageFlags(const VkPhysicalDeviceSparseImageFormatInfo2& create_info);
Location GetUsageLocation(const VkPhysicalDeviceSparseImageFormatInfo2& create_info, const Location& loc);

VkBufferUsageFlags2 GetBufferCreateFlags(const VkPhysicalDeviceExternalBufferInfo& create_info);
Location GetFlagsLocation(const VkPhysicalDeviceExternalBufferInfo& create_info, const Location& loc);

VkImageCreateFlags2KHR GetImageCreateFlags(const VkFramebufferAttachmentImageInfo& create_info);
Location GetFlagsLocation(const VkFramebufferAttachmentImageInfo& create_info, const Location& loc);

VkImageUsageFlags2KHR GetImageUsageFlags(const VkFramebufferAttachmentImageInfo& create_info);
Location GetUsageLocation(const VkFramebufferAttachmentImageInfo& create_info, const Location& loc);

VkImageUsageFlags2KHR GetImageUsageFlags(const VkSwapchainCreateInfoKHR& create_info);
Location GetImageUsageLocation(const VkSwapchainCreateInfoKHR& create_info, const Location& loc);

VkImageUsageFlags2KHR GetImageUsageFlags(const VkPhysicalDeviceVideoFormatInfoKHR& create_info);
Location GetImageUsageLocation(const VkPhysicalDeviceVideoFormatInfoKHR& create_info, const Location& loc);

VkPipelineCreateFlags2 GetPipelineCreateFlags(const VkRayTracingPipelineCreateInfoNV& create_info);
Location GetFlagsLocation(const VkRayTracingPipelineCreateInfoNV& create_info, const Location& loc);

VkBufferUsageFlags2 GetBufferUsageFlags(const VkDescriptorBufferBindingInfoEXT& create_info);
Location GetUsageLocation(const VkDescriptorBufferBindingInfoEXT& create_info, const Location& loc);

VkPipelineCreateFlags2 GetPipelineCreateFlags(const VkRayTracingPipelineCreateInfoKHR& create_info);
Location GetFlagsLocation(const VkRayTracingPipelineCreateInfoKHR& create_info, const Location& loc);

// NOLINTEND
