// *** THIS FILE IS GENERATED - DO NOT EDIT ***
// See error_location_helper_generator.py for modifications

/***************************************************************************
 *
 * Copyright (c) 2023-2026 The Khronos Group Inc.
 * Copyright (c) 2023-2026 Valve Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ****************************************************************************/

// NOLINTBEGIN

#include "error_location_helper.h"
#include "generated/vk_api_version.h"
#include <assert.h>
#include <string_view>

// clang-format off
namespace vvl {

const char* String(Func func) {
    static const std::string_view table[] = {
    {"INVALID_EMPTY", 15}, // Func::Empty
    {"vkAcquireDrmDisplayEXT", 23},
    {"vkAcquireFullScreenExclusiveModeEXT", 36},
    {"vkAcquireNextImage2KHR", 23},
    {"vkAcquireNextImageKHR", 22},
    {"vkAcquirePerformanceConfigurationINTEL", 39},
    {"vkAcquireProfilingLockKHR", 26},
    {"vkAcquireWinrtDisplayNV", 24},
    {"vkAcquireXlibDisplayEXT", 24},
    {"vkAllocateCommandBuffers", 25},
    {"vkAllocateDescriptorSets", 25},
    {"vkAllocateMemory", 17},
    {"vkAntiLagUpdateAMD", 19},
    {"vkBeginCommandBuffer", 21},
    {"vkBindAccelerationStructureMemoryNV", 36},
    {"vkBindBufferMemory", 19},
    {"vkBindBufferMemory2", 20},
    {"vkBindBufferMemory2KHR", 23},
    {"vkBindDataGraphPipelineSessionMemoryARM", 40},
    {"vkBindImageMemory", 18},
    {"vkBindImageMemory2", 19},
    {"vkBindImageMemory2KHR", 22},
    {"vkBindOpticalFlowSessionImageNV", 32},
    {"vkBindTensorMemoryARM", 22},
    {"vkBindVideoSessionMemoryKHR", 28},
    {"vkBuildAccelerationStructuresKHR", 33},
    {"vkBuildMicromapsEXT", 20},
    {"vkClearShaderInstrumentationMetricsARM", 39},
    {"vkCmdBeginConditionalRendering2EXT", 35},
    {"vkCmdBeginConditionalRenderingEXT", 34},
    {"vkCmdBeginCustomResolveEXT", 27},
    {"vkCmdBeginDebugUtilsLabelEXT", 29},
    {"vkCmdBeginGpaSampleAMD", 23},
    {"vkCmdBeginGpaSessionAMD", 24},
    {"vkCmdBeginPerTileExecutionQCOM", 31},
    {"vkCmdBeginQuery", 16},
    {"vkCmdBeginQueryIndexedEXT", 26},
    {"vkCmdBeginRenderPass", 21},
    {"vkCmdBeginRenderPass2", 22},
    {"vkCmdBeginRenderPass2KHR", 25},
    {"vkCmdBeginRendering", 20},
    {"vkCmdBeginRenderingKHR", 23},
    {"vkCmdBeginShaderInstrumentationARM", 35},
    {"vkCmdBeginTransformFeedback2EXT", 32},
    {"vkCmdBeginTransformFeedbackEXT", 31},
    {"vkCmdBeginVideoCodingKHR", 25},
    {"vkCmdBindDescriptorBufferEmbeddedSamplers2EXT", 46},
    {"vkCmdBindDescriptorBufferEmbeddedSamplersEXT", 45},
    {"vkCmdBindDescriptorBuffersEXT", 30},
    {"vkCmdBindDescriptorSets", 24},
    {"vkCmdBindDescriptorSets2", 25},
    {"vkCmdBindDescriptorSets2KHR", 28},
    {"vkCmdBindIndexBuffer", 21},
    {"vkCmdBindIndexBuffer2", 22},
    {"vkCmdBindIndexBuffer2KHR", 25},
    {"vkCmdBindIndexBuffer3KHR", 25},
    {"vkCmdBindInvocationMaskHUAWEI", 30},
    {"vkCmdBindPipeline", 18},
    {"vkCmdBindPipelineShaderGroupNV", 31},
    {"vkCmdBindResourceHeapEXT", 25},
    {"vkCmdBindSamplerHeapEXT", 24},
    {"vkCmdBindShadersEXT", 20},
    {"vkCmdBindShadingRateImageNV", 28},
    {"vkCmdBindTileMemoryQCOM", 24},
    {"vkCmdBindTransformFeedbackBuffers2EXT", 38},
    {"vkCmdBindTransformFeedbackBuffersEXT", 37},
    {"vkCmdBindVertexBuffers", 23},
    {"vkCmdBindVertexBuffers2", 24},
    {"vkCmdBindVertexBuffers2EXT", 27},
    {"vkCmdBindVertexBuffers3KHR", 27},
    {"vkCmdBlitImage", 15},
    {"vkCmdBlitImage2", 16},
    {"vkCmdBlitImage2KHR", 19},
    {"vkCmdBuildAccelerationStructureNV", 34},
    {"vkCmdBuildAccelerationStructuresIndirectKHR", 44},
    {"vkCmdBuildAccelerationStructuresKHR", 36},
    {"vkCmdBuildClusterAccelerationStructureIndirectNV", 49},
    {"vkCmdBuildMicromapsEXT", 23},
    {"vkCmdBuildPartitionedAccelerationStructuresNV", 46},
    {"vkCmdClearAttachments", 22},
    {"vkCmdClearColorImage", 21},
    {"vkCmdClearDepthStencilImage", 28},
    {"vkCmdControlVideoCodingKHR", 27},
    {"vkCmdConvertCooperativeVectorMatrixNV", 38},
    {"vkCmdCopyAccelerationStructureKHR", 34},
    {"vkCmdCopyAccelerationStructureNV", 33},
    {"vkCmdCopyAccelerationStructureToMemoryKHR", 42},
    {"vkCmdCopyBuffer", 16},
    {"vkCmdCopyBuffer2", 17},
    {"vkCmdCopyBuffer2KHR", 20},
    {"vkCmdCopyBufferToImage", 23},
    {"vkCmdCopyBufferToImage2", 24},
    {"vkCmdCopyBufferToImage2KHR", 27},
    {"vkCmdCopyGpaSessionResultsAMD", 30},
    {"vkCmdCopyImage", 15},
    {"vkCmdCopyImage2", 16},
    {"vkCmdCopyImage2KHR", 19},
    {"vkCmdCopyImageToBuffer", 23},
    {"vkCmdCopyImageToBuffer2", 24},
    {"vkCmdCopyImageToBuffer2KHR", 27},
    {"vkCmdCopyImageToMemoryKHR", 26},
    {"vkCmdCopyMemoryIndirectKHR", 27},
    {"vkCmdCopyMemoryIndirectNV", 26},
    {"vkCmdCopyMemoryKHR", 19},
    {"vkCmdCopyMemoryToAccelerationStructureKHR", 42},
    {"vkCmdCopyMemoryToImageIndirectKHR", 34},
    {"vkCmdCopyMemoryToImageIndirectNV", 33},
    {"vkCmdCopyMemoryToImageKHR", 26},
    {"vkCmdCopyMemoryToMicromapEXT", 29},
    {"vkCmdCopyMicromapEXT", 21},
    {"vkCmdCopyMicromapToMemoryEXT", 29},
    {"vkCmdCopyQueryPoolResults", 26},
    {"vkCmdCopyQueryPoolResultsToMemoryKHR", 37},
    {"vkCmdCopyTensorARM", 19},
    {"vkCmdCuLaunchKernelNVX", 23},
    {"vkCmdCudaLaunchKernelNV", 24},
    {"vkCmdDebugMarkerBeginEXT", 25},
    {"vkCmdDebugMarkerEndEXT", 23},
    {"vkCmdDebugMarkerInsertEXT", 26},
    {"vkCmdDecodeVideoKHR", 20},
    {"vkCmdDecompressMemoryEXT", 25},
    {"vkCmdDecompressMemoryIndirectCountEXT", 38},
    {"vkCmdDecompressMemoryIndirectCountNV", 37},
    {"vkCmdDecompressMemoryNV", 24},
    {"vkCmdDispatch", 14},
    {"vkCmdDispatchBase", 18},
    {"vkCmdDispatchBaseKHR", 21},
    {"vkCmdDispatchDataGraphARM", 26},
    {"vkCmdDispatchGraphAMDX", 23},
    {"vkCmdDispatchGraphIndirectAMDX", 31},
    {"vkCmdDispatchGraphIndirectCountAMDX", 36},
    {"vkCmdDispatchIndirect", 22},
    {"vkCmdDispatchIndirect2KHR", 26},
    {"vkCmdDispatchTileQCOM", 22},
    {"vkCmdDraw", 10},
    {"vkCmdDrawClusterHUAWEI", 23},
    {"vkCmdDrawClusterIndirectHUAWEI", 31},
    {"vkCmdDrawIndexed", 17},
    {"vkCmdDrawIndexedIndirect", 25},
    {"vkCmdDrawIndexedIndirect2KHR", 29},
    {"vkCmdDrawIndexedIndirectCount", 30},
    {"vkCmdDrawIndexedIndirectCount2KHR", 34},
    {"vkCmdDrawIndexedIndirectCountAMD", 33},
    {"vkCmdDrawIndexedIndirectCountKHR", 33},
    {"vkCmdDrawIndirect", 18},
    {"vkCmdDrawIndirect2KHR", 22},
    {"vkCmdDrawIndirectByteCount2EXT", 31},
    {"vkCmdDrawIndirectByteCountEXT", 30},
    {"vkCmdDrawIndirectCount", 23},
    {"vkCmdDrawIndirectCount2KHR", 27},
    {"vkCmdDrawIndirectCountAMD", 26},
    {"vkCmdDrawIndirectCountKHR", 26},
    {"vkCmdDrawMeshTasksEXT", 22},
    {"vkCmdDrawMeshTasksIndirect2EXT", 31},
    {"vkCmdDrawMeshTasksIndirectCount2EXT", 36},
    {"vkCmdDrawMeshTasksIndirectCountEXT", 35},
    {"vkCmdDrawMeshTasksIndirectCountNV", 34},
    {"vkCmdDrawMeshTasksIndirectEXT", 30},
    {"vkCmdDrawMeshTasksIndirectNV", 29},
    {"vkCmdDrawMeshTasksNV", 21},
    {"vkCmdDrawMultiEXT", 18},
    {"vkCmdDrawMultiIndexedEXT", 25},
    {"vkCmdEncodeVideoKHR", 20},
    {"vkCmdEndConditionalRenderingEXT", 32},
    {"vkCmdEndDebugUtilsLabelEXT", 27},
    {"vkCmdEndGpaSampleAMD", 21},
    {"vkCmdEndGpaSessionAMD", 22},
    {"vkCmdEndPerTileExecutionQCOM", 29},
    {"vkCmdEndQuery", 14},
    {"vkCmdEndQueryIndexedEXT", 24},
    {"vkCmdEndRenderPass", 19},
    {"vkCmdEndRenderPass2", 20},
    {"vkCmdEndRenderPass2KHR", 23},
    {"vkCmdEndRendering", 18},
    {"vkCmdEndRendering2EXT", 22},
    {"vkCmdEndRendering2KHR", 22},
    {"vkCmdEndRenderingKHR", 21},
    {"vkCmdEndShaderInstrumentationARM", 33},
    {"vkCmdEndTransformFeedback2EXT", 30},
    {"vkCmdEndTransformFeedbackEXT", 29},
    {"vkCmdEndVideoCodingKHR", 23},
    {"vkCmdExecuteCommands", 21},
    {"vkCmdExecuteGeneratedCommandsEXT", 33},
    {"vkCmdExecuteGeneratedCommandsNV", 32},
    {"vkCmdFillBuffer", 16},
    {"vkCmdFillMemoryKHR", 19},
    {"vkCmdInitializeGraphScratchMemoryAMDX", 38},
    {"vkCmdInsertDebugUtilsLabelEXT", 30},
    {"vkCmdNextSubpass", 17},
    {"vkCmdNextSubpass2", 18},
    {"vkCmdNextSubpass2KHR", 21},
    {"vkCmdOpticalFlowExecuteNV", 26},
    {"vkCmdPipelineBarrier", 21},
    {"vkCmdPipelineBarrier2", 22},
    {"vkCmdPipelineBarrier2KHR", 25},
    {"vkCmdPreprocessGeneratedCommandsEXT", 36},
    {"vkCmdPreprocessGeneratedCommandsNV", 35},
    {"vkCmdPushConstants", 19},
    {"vkCmdPushConstants2", 20},
    {"vkCmdPushConstants2KHR", 23},
    {"vkCmdPushDataEXT", 17},
    {"vkCmdPushDescriptorSet", 23},
    {"vkCmdPushDescriptorSet2", 24},
    {"vkCmdPushDescriptorSet2KHR", 27},
    {"vkCmdPushDescriptorSetKHR", 26},
    {"vkCmdPushDescriptorSetWithTemplate", 35},
    {"vkCmdPushDescriptorSetWithTemplate2", 36},
    {"vkCmdPushDescriptorSetWithTemplate2KHR", 39},
    {"vkCmdPushDescriptorSetWithTemplateKHR", 38},
    {"vkCmdResetEvent", 16},
    {"vkCmdResetEvent2", 17},
    {"vkCmdResetEvent2KHR", 20},
    {"vkCmdResetQueryPool", 20},
    {"vkCmdResolveImage", 18},
    {"vkCmdResolveImage2", 19},
    {"vkCmdResolveImage2KHR", 22},
    {"vkCmdSetAlphaToCoverageEnableEXT", 33},
    {"vkCmdSetAlphaToOneEnableEXT", 28},
    {"vkCmdSetAttachmentFeedbackLoopEnableEXT", 40},
    {"vkCmdSetBlendConstants", 23},
    {"vkCmdSetCheckpointNV", 21},
    {"vkCmdSetCoarseSampleOrderNV", 28},
    {"vkCmdSetColorBlendAdvancedEXT", 30},
    {"vkCmdSetColorBlendEnableEXT", 28},
    {"vkCmdSetColorBlendEquationEXT", 30},
    {"vkCmdSetColorWriteEnableEXT", 28},
    {"vkCmdSetColorWriteMaskEXT", 26},
    {"vkCmdSetComputeOccupancyPriorityNV", 35},
    {"vkCmdSetConservativeRasterizationModeEXT", 41},
    {"vkCmdSetCoverageModulationModeNV", 33},
    {"vkCmdSetCoverageModulationTableEnableNV", 40},
    {"vkCmdSetCoverageModulationTableNV", 34},
    {"vkCmdSetCoverageReductionModeNV", 32},
    {"vkCmdSetCoverageToColorEnableNV", 32},
    {"vkCmdSetCoverageToColorLocationNV", 34},
    {"vkCmdSetCullMode", 17},
    {"vkCmdSetCullModeEXT", 20},
    {"vkCmdSetDepthBias", 18},
    {"vkCmdSetDepthBias2EXT", 22},
    {"vkCmdSetDepthBiasEnable", 24},
    {"vkCmdSetDepthBiasEnableEXT", 27},
    {"vkCmdSetDepthBounds", 20},
    {"vkCmdSetDepthBoundsTestEnable", 30},
    {"vkCmdSetDepthBoundsTestEnableEXT", 33},
    {"vkCmdSetDepthClampEnableEXT", 28},
    {"vkCmdSetDepthClampRangeEXT", 27},
    {"vkCmdSetDepthClipEnableEXT", 27},
    {"vkCmdSetDepthClipNegativeOneToOneEXT", 37},
    {"vkCmdSetDepthCompareOp", 23},
    {"vkCmdSetDepthCompareOpEXT", 26},
    {"vkCmdSetDepthTestEnable", 24},
    {"vkCmdSetDepthTestEnableEXT", 27},
    {"vkCmdSetDepthWriteEnable", 25},
    {"vkCmdSetDepthWriteEnableEXT", 28},
    {"vkCmdSetDescriptorBufferOffsets2EXT", 36},
    {"vkCmdSetDescriptorBufferOffsetsEXT", 35},
    {"vkCmdSetDeviceMask", 19},
    {"vkCmdSetDeviceMaskKHR", 22},
    {"vkCmdSetDiscardRectangleEXT", 28},
    {"vkCmdSetDiscardRectangleEnableEXT", 34},
    {"vkCmdSetDiscardRectangleModeEXT", 32},
    {"vkCmdSetDispatchParametersARM", 30},
    {"vkCmdSetEvent", 14},
    {"vkCmdSetEvent2", 15},
    {"vkCmdSetEvent2KHR", 18},
    {"vkCmdSetExclusiveScissorEnableNV", 33},
    {"vkCmdSetExclusiveScissorNV", 27},
    {"vkCmdSetExtraPrimitiveOverestimationSizeEXT", 44},
    {"vkCmdSetFragmentShadingRateEnumNV", 34},
    {"vkCmdSetFragmentShadingRateKHR", 31},
    {"vkCmdSetFrontFace", 18},
    {"vkCmdSetFrontFaceEXT", 21},
    {"vkCmdSetLineRasterizationModeEXT", 33},
    {"vkCmdSetLineStipple", 20},
    {"vkCmdSetLineStippleEXT", 23},
    {"vkCmdSetLineStippleEnableEXT", 29},
    {"vkCmdSetLineStippleKHR", 23},
    {"vkCmdSetLineWidth", 18},
    {"vkCmdSetLogicOpEXT", 19},
    {"vkCmdSetLogicOpEnableEXT", 25},
    {"vkCmdSetPatchControlPointsEXT", 30},
    {"vkCmdSetPerformanceMarkerINTEL", 31},
    {"vkCmdSetPerformanceOverrideINTEL", 33},
    {"vkCmdSetPerformanceStreamMarkerINTEL", 37},
    {"vkCmdSetPolygonModeEXT", 23},
    {"vkCmdSetPrimitiveRestartEnable", 31},
    {"vkCmdSetPrimitiveRestartEnableEXT", 34},
    {"vkCmdSetPrimitiveRestartIndexEXT", 33},
    {"vkCmdSetPrimitiveTopology", 26},
    {"vkCmdSetPrimitiveTopologyEXT", 29},
    {"vkCmdSetProvokingVertexModeEXT", 31},
    {"vkCmdSetRasterizationSamplesEXT", 32},
    {"vkCmdSetRasterizationStreamEXT", 31},
    {"vkCmdSetRasterizerDiscardEnable", 32},
    {"vkCmdSetRasterizerDiscardEnableEXT", 35},
    {"vkCmdSetRayTracingPipelineStackSizeKHR", 39},
    {"vkCmdSetRenderingAttachmentLocations", 37},
    {"vkCmdSetRenderingAttachmentLocationsKHR", 40},
    {"vkCmdSetRenderingInputAttachmentIndices", 40},
    {"vkCmdSetRenderingInputAttachmentIndicesKHR", 43},
    {"vkCmdSetRepresentativeFragmentTestEnableNV", 43},
    {"vkCmdSetSampleLocationsEXT", 27},
    {"vkCmdSetSampleLocationsEnableEXT", 33},
    {"vkCmdSetSampleMaskEXT", 22},
    {"vkCmdSetScissor", 16},
    {"vkCmdSetScissorWithCount", 25},
    {"vkCmdSetScissorWithCountEXT", 28},
    {"vkCmdSetShadingRateImageEnableNV", 33},
    {"vkCmdSetStencilCompareMask", 27},
    {"vkCmdSetStencilOp", 18},
    {"vkCmdSetStencilOpEXT", 21},
    {"vkCmdSetStencilReference", 25},
    {"vkCmdSetStencilTestEnable", 26},
    {"vkCmdSetStencilTestEnableEXT", 29},
    {"vkCmdSetStencilWriteMask", 25},
    {"vkCmdSetTessellationDomainOriginEXT", 36},
    {"vkCmdSetVertexInputEXT", 23},
    {"vkCmdSetViewport", 17},
    {"vkCmdSetViewportShadingRatePaletteNV", 37},
    {"vkCmdSetViewportSwizzleNV", 26},
    {"vkCmdSetViewportWScalingEnableNV", 33},
    {"vkCmdSetViewportWScalingNV", 27},
    {"vkCmdSetViewportWithCount", 26},
    {"vkCmdSetViewportWithCountEXT", 29},
    {"vkCmdSubpassShadingHUAWEI", 26},
    {"vkCmdTraceRaysIndirect2KHR", 27},
    {"vkCmdTraceRaysIndirectKHR", 26},
    {"vkCmdTraceRaysKHR", 18},
    {"vkCmdTraceRaysNV", 17},
    {"vkCmdUpdateBuffer", 18},
    {"vkCmdUpdateMemoryKHR", 21},
    {"vkCmdUpdatePipelineIndirectBufferNV", 36},
    {"vkCmdWaitEvents", 16},
    {"vkCmdWaitEvents2", 17},
    {"vkCmdWaitEvents2KHR", 20},
    {"vkCmdWriteAccelerationStructuresPropertiesKHR", 46},
    {"vkCmdWriteAccelerationStructuresPropertiesNV", 45},
    {"vkCmdWriteBufferMarker2AMD", 27},
    {"vkCmdWriteBufferMarkerAMD", 26},
    {"vkCmdWriteMarkerToMemoryAMD", 28},
    {"vkCmdWriteMicromapsPropertiesEXT", 33},
    {"vkCmdWriteTimestamp", 20},
    {"vkCmdWriteTimestamp2", 21},
    {"vkCmdWriteTimestamp2KHR", 24},
    {"vkCompileDeferredNV", 20},
    {"vkConvertCooperativeVectorMatrixNV", 35},
    {"vkCopyAccelerationStructureKHR", 31},
    {"vkCopyAccelerationStructureToMemoryKHR", 39},
    {"vkCopyImageToImage", 19},
    {"vkCopyImageToImageEXT", 22},
    {"vkCopyImageToMemory", 20},
    {"vkCopyImageToMemoryEXT", 23},
    {"vkCopyMemoryToAccelerationStructureKHR", 39},
    {"vkCopyMemoryToImage", 20},
    {"vkCopyMemoryToImageEXT", 23},
    {"vkCopyMemoryToMicromapEXT", 26},
    {"vkCopyMicromapEXT", 18},
    {"vkCopyMicromapToMemoryEXT", 26},
    {"vkCreateAccelerationStructure2KHR", 34},
    {"vkCreateAccelerationStructureKHR", 33},
    {"vkCreateAccelerationStructureNV", 32},
    {"vkCreateAndroidSurfaceKHR", 26},
    {"vkCreateBuffer", 15},
    {"vkCreateBufferCollectionFUCHSIA", 32},
    {"vkCreateBufferView", 19},
    {"vkCreateCommandPool", 20},
    {"vkCreateComputePipelines", 25},
    {"vkCreateCuFunctionNVX", 22},
    {"vkCreateCuModuleNVX", 20},
    {"vkCreateCudaFunctionNV", 23},
    {"vkCreateCudaModuleNV", 21},
    {"vkCreateDataGraphPipelineSessionARM", 36},
    {"vkCreateDataGraphPipelinesARM", 30},
    {"vkCreateDebugReportCallbackEXT", 31},
    {"vkCreateDebugUtilsMessengerEXT", 31},
    {"vkCreateDeferredOperationKHR", 29},
    {"vkCreateDescriptorPool", 23},
    {"vkCreateDescriptorSetLayout", 28},
    {"vkCreateDescriptorUpdateTemplate", 33},
    {"vkCreateDescriptorUpdateTemplateKHR", 36},
    {"vkCreateDevice", 15},
    {"vkCreateDirectFBSurfaceEXT", 27},
    {"vkCreateDisplayModeKHR", 23},
    {"vkCreateDisplayPlaneSurfaceKHR", 31},
    {"vkCreateEvent", 14},
    {"vkCreateExecutionGraphPipelinesAMDX", 36},
    {"vkCreateExternalComputeQueueNV", 31},
    {"vkCreateFence", 14},
    {"vkCreateFramebuffer", 20},
    {"vkCreateGpaSessionAMD", 22},
    {"vkCreateGraphicsPipelines", 26},
    {"vkCreateHeadlessSurfaceEXT", 27},
    {"vkCreateIOSSurfaceMVK", 22},
    {"vkCreateImage", 14},
    {"vkCreateImagePipeSurfaceFUCHSIA", 32},
    {"vkCreateImageView", 18},
    {"vkCreateIndirectCommandsLayoutEXT", 34},
    {"vkCreateIndirectCommandsLayoutNV", 33},
    {"vkCreateIndirectExecutionSetEXT", 32},
    {"vkCreateInstance", 17},
    {"vkCreateMacOSSurfaceMVK", 24},
    {"vkCreateMetalSurfaceEXT", 24},
    {"vkCreateMicromapEXT", 20},
    {"vkCreateOpticalFlowSessionNV", 29},
    {"vkCreatePipelineBinariesKHR", 28},
    {"vkCreatePipelineCache", 22},
    {"vkCreatePipelineLayout", 23},
    {"vkCreatePrivateDataSlot", 24},
    {"vkCreatePrivateDataSlotEXT", 27},
    {"vkCreateQueryPool", 18},
    {"vkCreateRayTracingPipelinesKHR", 31},
    {"vkCreateRayTracingPipelinesNV", 30},
    {"vkCreateRenderPass", 19},
    {"vkCreateRenderPass2", 20},
    {"vkCreateRenderPass2KHR", 23},
    {"vkCreateSampler", 16},
    {"vkCreateSamplerYcbcrConversion", 31},
    {"vkCreateSamplerYcbcrConversionKHR", 34},
    {"vkCreateScreenSurfaceQNX", 25},
    {"vkCreateSemaphore", 18},
    {"vkCreateShaderInstrumentationARM", 33},
    {"vkCreateShaderModule", 21},
    {"vkCreateShadersEXT", 19},
    {"vkCreateSharedSwapchainsKHR", 28},
    {"vkCreateStreamDescriptorSurfaceGGP", 35},
    {"vkCreateSurfaceOHOS", 20},
    {"vkCreateSwapchainKHR", 21},
    {"vkCreateTensorARM", 18},
    {"vkCreateTensorViewARM", 22},
    {"vkCreateUbmSurfaceSEC", 22},
    {"vkCreateValidationCacheEXT", 27},
    {"vkCreateViSurfaceNN", 20},
    {"vkCreateVideoSessionKHR", 24},
    {"vkCreateVideoSessionParametersKHR", 34},
    {"vkCreateWaylandSurfaceKHR", 26},
    {"vkCreateWin32SurfaceKHR", 24},
    {"vkCreateXcbSurfaceKHR", 22},
    {"vkCreateXlibSurfaceKHR", 23},
    {"vkDebugMarkerSetObjectNameEXT", 30},
    {"vkDebugMarkerSetObjectTagEXT", 29},
    {"vkDebugReportMessageEXT", 24},
    {"vkDeferredOperationJoinKHR", 27},
    {"vkDestroyAccelerationStructureKHR", 34},
    {"vkDestroyAccelerationStructureNV", 33},
    {"vkDestroyBuffer", 16},
    {"vkDestroyBufferCollectionFUCHSIA", 33},
    {"vkDestroyBufferView", 20},
    {"vkDestroyCommandPool", 21},
    {"vkDestroyCuFunctionNVX", 23},
    {"vkDestroyCuModuleNVX", 21},
    {"vkDestroyCudaFunctionNV", 24},
    {"vkDestroyCudaModuleNV", 22},
    {"vkDestroyDataGraphPipelineSessionARM", 37},
    {"vkDestroyDebugReportCallbackEXT", 32},
    {"vkDestroyDebugUtilsMessengerEXT", 32},
    {"vkDestroyDeferredOperationKHR", 30},
    {"vkDestroyDescriptorPool", 24},
    {"vkDestroyDescriptorSetLayout", 29},
    {"vkDestroyDescriptorUpdateTemplate", 34},
    {"vkDestroyDescriptorUpdateTemplateKHR", 37},
    {"vkDestroyDevice", 16},
    {"vkDestroyEvent", 15},
    {"vkDestroyExternalComputeQueueNV", 32},
    {"vkDestroyFence", 15},
    {"vkDestroyFramebuffer", 21},
    {"vkDestroyGpaSessionAMD", 23},
    {"vkDestroyImage", 15},
    {"vkDestroyImageView", 19},
    {"vkDestroyIndirectCommandsLayoutEXT", 35},
    {"vkDestroyIndirectCommandsLayoutNV", 34},
    {"vkDestroyIndirectExecutionSetEXT", 33},
    {"vkDestroyInstance", 18},
    {"vkDestroyMicromapEXT", 21},
    {"vkDestroyOpticalFlowSessionNV", 30},
    {"vkDestroyPipeline", 18},
    {"vkDestroyPipelineBinaryKHR", 27},
    {"vkDestroyPipelineCache", 23},
    {"vkDestroyPipelineLayout", 24},
    {"vkDestroyPrivateDataSlot", 25},
    {"vkDestroyPrivateDataSlotEXT", 28},
    {"vkDestroyQueryPool", 19},
    {"vkDestroyRenderPass", 20},
    {"vkDestroySampler", 17},
    {"vkDestroySamplerYcbcrConversion", 32},
    {"vkDestroySamplerYcbcrConversionKHR", 35},
    {"vkDestroySemaphore", 19},
    {"vkDestroyShaderEXT", 19},
    {"vkDestroyShaderInstrumentationARM", 34},
    {"vkDestroyShaderModule", 22},
    {"vkDestroySurfaceKHR", 20},
    {"vkDestroySwapchainKHR", 22},
    {"vkDestroyTensorARM", 19},
    {"vkDestroyTensorViewARM", 23},
    {"vkDestroyValidationCacheEXT", 28},
    {"vkDestroyVideoSessionKHR", 25},
    {"vkDestroyVideoSessionParametersKHR", 35},
    {"vkDeviceWaitIdle", 17},
    {"vkDisplayPowerControlEXT", 25},
    {"vkEndCommandBuffer", 19},
    {"vkEnumerateDeviceExtensionProperties", 37},
    {"vkEnumerateDeviceLayerProperties", 33},
    {"vkEnumerateInstanceExtensionProperties", 39},
    {"vkEnumerateInstanceLayerProperties", 35},
    {"vkEnumerateInstanceVersion", 27},
    {"vkEnumeratePhysicalDeviceGroups", 32},
    {"vkEnumeratePhysicalDeviceGroupsKHR", 35},
    {"vkEnumeratePhysicalDeviceQueueFamilyPerformanceCountersByRegionARM", 67},
    {"vkEnumeratePhysicalDeviceQueueFamilyPerformanceQueryCountersKHR", 64},
    {"vkEnumeratePhysicalDeviceShaderInstrumentationMetricsARM", 57},
    {"vkEnumeratePhysicalDevices", 27},
    {"vkExportMetalObjectsEXT", 24},
    {"vkFlushMappedMemoryRanges", 26},
    {"vkFreeCommandBuffers", 21},
    {"vkFreeDescriptorSets", 21},
    {"vkFreeMemory", 13},
    {"vkGetAccelerationStructureBuildSizesKHR", 40},
    {"vkGetAccelerationStructureDeviceAddressKHR", 43},
    {"vkGetAccelerationStructureHandleNV", 35},
    {"vkGetAccelerationStructureMemoryRequirementsNV", 47},
    {"vkGetAccelerationStructureOpaqueCaptureDescriptorDataEXT", 57},
    {"vkGetAndroidHardwareBufferPropertiesANDROID", 44},
    {"vkGetBufferCollectionPropertiesFUCHSIA", 39},
    {"vkGetBufferDeviceAddress", 25},
    {"vkGetBufferDeviceAddressEXT", 28},
    {"vkGetBufferDeviceAddressKHR", 28},
    {"vkGetBufferMemoryRequirements", 30},
    {"vkGetBufferMemoryRequirements2", 31},
    {"vkGetBufferMemoryRequirements2KHR", 34},
    {"vkGetBufferOpaqueCaptureAddress", 32},
    {"vkGetBufferOpaqueCaptureAddressKHR", 35},
    {"vkGetBufferOpaqueCaptureDescriptorDataEXT", 42},
    {"vkGetCalibratedTimestampsEXT", 29},
    {"vkGetCalibratedTimestampsKHR", 29},
    {"vkGetClusterAccelerationStructureBuildSizesNV", 46},
    {"vkGetCudaModuleCacheNV", 23},
    {"vkGetDataGraphPipelineAvailablePropertiesARM", 45},
    {"vkGetDataGraphPipelinePropertiesARM", 36},
    {"vkGetDataGraphPipelineSessionBindPointRequirementsARM", 54},
    {"vkGetDataGraphPipelineSessionMemoryRequirementsARM", 51},
    {"vkGetDeferredOperationMaxConcurrencyKHR", 40},
    {"vkGetDeferredOperationResultKHR", 32},
    {"vkGetDescriptorEXT", 19},
    {"vkGetDescriptorSetHostMappingVALVE", 35},
    {"vkGetDescriptorSetLayoutBindingOffsetEXT", 41},
    {"vkGetDescriptorSetLayoutHostMappingInfoVALVE", 45},
    {"vkGetDescriptorSetLayoutSizeEXT", 32},
    {"vkGetDescriptorSetLayoutSupport", 32},
    {"vkGetDescriptorSetLayoutSupportKHR", 35},
    {"vkGetDeviceAccelerationStructureCompatibilityKHR", 49},
    {"vkGetDeviceBufferMemoryRequirements", 36},
    {"vkGetDeviceBufferMemoryRequirementsKHR", 39},
    {"vkGetDeviceCombinedImageSamplerIndexNVX", 40},
    {"vkGetDeviceFaultDebugInfoKHR", 29},
    {"vkGetDeviceFaultInfoEXT", 24},
    {"vkGetDeviceFaultReportsKHR", 27},
    {"vkGetDeviceGroupPeerMemoryFeatures", 35},
    {"vkGetDeviceGroupPeerMemoryFeaturesKHR", 38},
    {"vkGetDeviceGroupPresentCapabilitiesKHR", 39},
    {"vkGetDeviceGroupSurfacePresentModes2EXT", 40},
    {"vkGetDeviceGroupSurfacePresentModesKHR", 39},
    {"vkGetDeviceImageMemoryRequirements", 35},
    {"vkGetDeviceImageMemoryRequirementsKHR", 38},
    {"vkGetDeviceImageSparseMemoryRequirements", 41},
    {"vkGetDeviceImageSparseMemoryRequirementsKHR", 44},
    {"vkGetDeviceImageSubresourceLayout", 34},
    {"vkGetDeviceImageSubresourceLayoutKHR", 37},
    {"vkGetDeviceMemoryCommitment", 28},
    {"vkGetDeviceMemoryOpaqueCaptureAddress", 38},
    {"vkGetDeviceMemoryOpaqueCaptureAddressKHR", 41},
    {"vkGetDeviceMicromapCompatibilityEXT", 36},
    {"vkGetDeviceProcAddr", 20},
    {"vkGetDeviceQueue", 17},
    {"vkGetDeviceQueue2", 18},
    {"vkGetDeviceSubpassShadingMaxWorkgroupSizeHUAWEI", 48},
    {"vkGetDeviceTensorMemoryRequirementsARM", 39},
    {"vkGetDisplayModeProperties2KHR", 31},
    {"vkGetDisplayModePropertiesKHR", 30},
    {"vkGetDisplayPlaneCapabilities2KHR", 34},
    {"vkGetDisplayPlaneCapabilitiesKHR", 33},
    {"vkGetDisplayPlaneSupportedDisplaysKHR", 38},
    {"vkGetDrmDisplayEXT", 19},
    {"vkGetDynamicRenderingTilePropertiesQCOM", 40},
    {"vkGetEncodedVideoSessionParametersKHR", 38},
    {"vkGetEventStatus", 17},
    {"vkGetExecutionGraphPipelineNodeIndexAMDX", 41},
    {"vkGetExecutionGraphPipelineScratchSizeAMDX", 43},
    {"vkGetExternalComputeQueueDataNV", 32},
    {"vkGetFenceFdKHR", 16},
    {"vkGetFenceStatus", 17},
    {"vkGetFenceWin32HandleKHR", 25},
    {"vkGetFramebufferTilePropertiesQCOM", 35},
    {"vkGetGeneratedCommandsMemoryRequirementsEXT", 44},
    {"vkGetGeneratedCommandsMemoryRequirementsNV", 43},
    {"vkGetGpaDeviceClockInfoAMD", 27},
    {"vkGetGpaSessionResultsAMD", 26},
    {"vkGetGpaSessionStatusAMD", 25},
    {"vkGetImageDrmFormatModifierPropertiesEXT", 41},
    {"vkGetImageMemoryRequirements", 29},
    {"vkGetImageMemoryRequirements2", 30},
    {"vkGetImageMemoryRequirements2KHR", 33},
    {"vkGetImageOpaqueCaptureDataEXT", 31},
    {"vkGetImageOpaqueCaptureDescriptorDataEXT", 41},
    {"vkGetImageSparseMemoryRequirements", 35},
    {"vkGetImageSparseMemoryRequirements2", 36},
    {"vkGetImageSparseMemoryRequirements2KHR", 39},
    {"vkGetImageSubresourceLayout", 28},
    {"vkGetImageSubresourceLayout2", 29},
    {"vkGetImageSubresourceLayout2EXT", 32},
    {"vkGetImageSubresourceLayout2KHR", 32},
    {"vkGetImageViewAddressNVX", 25},
    {"vkGetImageViewHandle64NVX", 26},
    {"vkGetImageViewHandleNVX", 24},
    {"vkGetImageViewOpaqueCaptureDescriptorDataEXT", 45},
    {"vkGetInstanceProcAddr", 22},
    {"vkGetLatencyTimingsLegacyNV", 28},
    {"vkGetLatencyTimingsNV", 22},
    {"vkGetMemoryAndroidHardwareBufferANDROID", 40},
    {"vkGetMemoryFdKHR", 17},
    {"vkGetMemoryFdPropertiesKHR", 27},
    {"vkGetMemoryHostPointerPropertiesEXT", 36},
    {"vkGetMemoryMetalHandleEXT", 26},
    {"vkGetMemoryMetalHandlePropertiesEXT", 36},
    {"vkGetMemoryNativeBufferOHOS", 28},
    {"vkGetMemoryRemoteAddressNV", 27},
    {"vkGetMemoryWin32HandleKHR", 26},
    {"vkGetMemoryWin32HandleNV", 25},
    {"vkGetMemoryWin32HandlePropertiesKHR", 36},
    {"vkGetMemoryZirconHandleFUCHSIA", 31},
    {"vkGetMemoryZirconHandlePropertiesFUCHSIA", 41},
    {"vkGetMicromapBuildSizesEXT", 27},
    {"vkGetNativeBufferPropertiesOHOS", 32},
    {"vkGetPartitionedAccelerationStructuresBuildSizesNV", 51},
    {"vkGetPastPresentationTimingEXT", 31},
    {"vkGetPastPresentationTimingGOOGLE", 34},
    {"vkGetPerformanceParameterINTEL", 31},
    {"vkGetPhysicalDeviceCalibrateableTimeDomainsEXT", 47},
    {"vkGetPhysicalDeviceCalibrateableTimeDomainsKHR", 47},
    {"vkGetPhysicalDeviceCooperativeMatrixFlexibleDimensionsPropertiesNV", 67},
    {"vkGetPhysicalDeviceCooperativeMatrixPropertiesKHR", 50},
    {"vkGetPhysicalDeviceCooperativeMatrixPropertiesNV", 49},
    {"vkGetPhysicalDeviceCooperativeVectorPropertiesNV", 49},
    {"vkGetPhysicalDeviceDescriptorSizeEXT", 37},
    {"vkGetPhysicalDeviceDirectFBPresentationSupportEXT", 50},
    {"vkGetPhysicalDeviceDisplayPlaneProperties2KHR", 46},
    {"vkGetPhysicalDeviceDisplayPlanePropertiesKHR", 45},
    {"vkGetPhysicalDeviceDisplayProperties2KHR", 41},
    {"vkGetPhysicalDeviceDisplayPropertiesKHR", 40},
    {"vkGetPhysicalDeviceExternalBufferProperties", 44},
    {"vkGetPhysicalDeviceExternalBufferPropertiesKHR", 47},
    {"vkGetPhysicalDeviceExternalFenceProperties", 43},
    {"vkGetPhysicalDeviceExternalFencePropertiesKHR", 46},
    {"vkGetPhysicalDeviceExternalImageFormatPropertiesNV", 51},
    {"vkGetPhysicalDeviceExternalSemaphoreProperties", 47},
    {"vkGetPhysicalDeviceExternalSemaphorePropertiesKHR", 50},
    {"vkGetPhysicalDeviceExternalTensorPropertiesARM", 47},
    {"vkGetPhysicalDeviceFeatures", 28},
    {"vkGetPhysicalDeviceFeatures2", 29},
    {"vkGetPhysicalDeviceFeatures2KHR", 32},
    {"vkGetPhysicalDeviceFormatProperties", 36},
    {"vkGetPhysicalDeviceFormatProperties2", 37},
    {"vkGetPhysicalDeviceFormatProperties2KHR", 40},
    {"vkGetPhysicalDeviceFragmentShadingRatesKHR", 43},
    {"vkGetPhysicalDeviceImageFormatProperties", 41},
    {"vkGetPhysicalDeviceImageFormatProperties2", 42},
    {"vkGetPhysicalDeviceImageFormatProperties2KHR", 45},
    {"vkGetPhysicalDeviceMemoryProperties", 36},
    {"vkGetPhysicalDeviceMemoryProperties2", 37},
    {"vkGetPhysicalDeviceMemoryProperties2KHR", 40},
    {"vkGetPhysicalDeviceMultisamplePropertiesEXT", 44},
    {"vkGetPhysicalDeviceOpticalFlowImageFormatsNV", 45},
    {"vkGetPhysicalDevicePresentRectanglesKHR", 40},
    {"vkGetPhysicalDeviceProperties", 30},
    {"vkGetPhysicalDeviceProperties2", 31},
    {"vkGetPhysicalDeviceProperties2KHR", 34},
    {"vkGetPhysicalDeviceQueueFamilyDataGraphEngineOperationPropertiesARM", 68},
    {"vkGetPhysicalDeviceQueueFamilyDataGraphOpticalFlowImageFormatsARM", 66},
    {"vkGetPhysicalDeviceQueueFamilyDataGraphProcessingEnginePropertiesARM", 69},
    {"vkGetPhysicalDeviceQueueFamilyDataGraphPropertiesARM", 53},
    {"vkGetPhysicalDeviceQueueFamilyPerformanceQueryPassesKHR", 56},
    {"vkGetPhysicalDeviceQueueFamilyProperties", 41},
    {"vkGetPhysicalDeviceQueueFamilyProperties2", 42},
    {"vkGetPhysicalDeviceQueueFamilyProperties2KHR", 45},
    {"vkGetPhysicalDeviceScreenPresentationSupportQNX", 48},
    {"vkGetPhysicalDeviceSparseImageFormatProperties", 47},
    {"vkGetPhysicalDeviceSparseImageFormatProperties2", 48},
    {"vkGetPhysicalDeviceSparseImageFormatProperties2KHR", 51},
    {"vkGetPhysicalDeviceSupportedFramebufferMixedSamplesCombinationsNV", 66},
    {"vkGetPhysicalDeviceSurfaceCapabilities2EXT", 43},
    {"vkGetPhysicalDeviceSurfaceCapabilities2KHR", 43},
    {"vkGetPhysicalDeviceSurfaceCapabilitiesKHR", 42},
    {"vkGetPhysicalDeviceSurfaceFormats2KHR", 38},
    {"vkGetPhysicalDeviceSurfaceFormatsKHR", 37},
    {"vkGetPhysicalDeviceSurfacePresentModes2EXT", 43},
    {"vkGetPhysicalDeviceSurfacePresentModesKHR", 42},
    {"vkGetPhysicalDeviceSurfaceSupportKHR", 37},
    {"vkGetPhysicalDeviceToolProperties", 34},
    {"vkGetPhysicalDeviceToolPropertiesEXT", 37},
    {"vkGetPhysicalDeviceUbmPresentationSupportSEC", 45},
    {"vkGetPhysicalDeviceVideoCapabilitiesKHR", 40},
    {"vkGetPhysicalDeviceVideoEncodeQualityLevelPropertiesKHR", 56},
    {"vkGetPhysicalDeviceVideoFormatPropertiesKHR", 44},
    {"vkGetPhysicalDeviceWaylandPresentationSupportKHR", 49},
    {"vkGetPhysicalDeviceWin32PresentationSupportKHR", 47},
    {"vkGetPhysicalDeviceXcbPresentationSupportKHR", 45},
    {"vkGetPhysicalDeviceXlibPresentationSupportKHR", 46},
    {"vkGetPipelineBinaryDataKHR", 27},
    {"vkGetPipelineCacheData", 23},
    {"vkGetPipelineExecutableInternalRepresentationsKHR", 50},
    {"vkGetPipelineExecutablePropertiesKHR", 37},
    {"vkGetPipelineExecutableStatisticsKHR", 37},
    {"vkGetPipelineIndirectDeviceAddressNV", 37},
    {"vkGetPipelineIndirectMemoryRequirementsNV", 42},
    {"vkGetPipelineKeyKHR", 20},
    {"vkGetPipelinePropertiesEXT", 27},
    {"vkGetPrivateData", 17},
    {"vkGetPrivateDataEXT", 20},
    {"vkGetQueryPoolResults", 22},
    {"vkGetQueueCheckpointData2NV", 28},
    {"vkGetQueueCheckpointDataNV", 27},
    {"vkGetRandROutputDisplayEXT", 27},
    {"vkGetRayTracingCaptureReplayShaderGroupHandlesKHR", 50},
    {"vkGetRayTracingShaderGroupHandlesKHR", 37},
    {"vkGetRayTracingShaderGroupHandlesNV", 36},
    {"vkGetRayTracingShaderGroupStackSizeKHR", 39},
    {"vkGetRefreshCycleDurationGOOGLE", 32},
    {"vkGetRenderAreaGranularity", 27},
    {"vkGetRenderingAreaGranularity", 30},
    {"vkGetRenderingAreaGranularityKHR", 33},
    {"vkGetSamplerOpaqueCaptureDescriptorDataEXT", 43},
    {"vkGetScreenBufferPropertiesQNX", 31},
    {"vkGetSemaphoreCounterValue", 27},
    {"vkGetSemaphoreCounterValueKHR", 30},
    {"vkGetSemaphoreFdKHR", 20},
    {"vkGetSemaphoreWin32HandleKHR", 29},
    {"vkGetSemaphoreZirconHandleFUCHSIA", 34},
    {"vkGetShaderBinaryDataEXT", 25},
    {"vkGetShaderInfoAMD", 19},
    {"vkGetShaderInstrumentationValuesARM", 36},
    {"vkGetShaderModuleCreateInfoIdentifierEXT", 41},
    {"vkGetShaderModuleIdentifierEXT", 31},
    {"vkGetSleepStatusLegacyNV", 25},
    {"vkGetSwapchainCounterEXT", 25},
    {"vkGetSwapchainImagesKHR", 24},
    {"vkGetSwapchainStatusKHR", 24},
    {"vkGetSwapchainTimeDomainPropertiesEXT", 38},
    {"vkGetSwapchainTimingPropertiesEXT", 34},
    {"vkGetTensorMemoryRequirementsARM", 33},
    {"vkGetTensorOpaqueCaptureDataARM", 32},
    {"vkGetTensorOpaqueCaptureDescriptorDataARM", 42},
    {"vkGetTensorViewOpaqueCaptureDescriptorDataARM", 46},
    {"vkGetValidationCacheDataEXT", 28},
    {"vkGetVideoSessionMemoryRequirementsKHR", 39},
    {"vkGetWinrtDisplayNV", 20},
    {"vkImportFenceFdKHR", 19},
    {"vkImportFenceWin32HandleKHR", 28},
    {"vkImportSemaphoreFdKHR", 23},
    {"vkImportSemaphoreWin32HandleKHR", 32},
    {"vkImportSemaphoreZirconHandleFUCHSIA", 37},
    {"vkInitializePerformanceApiINTEL", 32},
    {"vkInvalidateMappedMemoryRanges", 31},
    {"vkLatencySleepLegacyNV", 23},
    {"vkLatencySleepNV", 17},
    {"vkMapMemory", 12},
    {"vkMapMemory2", 13},
    {"vkMapMemory2KHR", 16},
    {"vkMergePipelineCaches", 22},
    {"vkMergeValidationCachesEXT", 27},
    {"vkQueueBeginDebugUtilsLabelEXT", 31},
    {"vkQueueBindSparse", 18},
    {"vkQueueEndDebugUtilsLabelEXT", 29},
    {"vkQueueInsertDebugUtilsLabelEXT", 32},
    {"vkQueueNotifyOutOfBandLegacyNV", 31},
    {"vkQueueNotifyOutOfBandNV", 25},
    {"vkQueuePresentKHR", 18},
    {"vkQueueSetPerfHintQCOM", 23},
    {"vkQueueSetPerformanceConfigurationINTEL", 40},
    {"vkQueueSubmit", 14},
    {"vkQueueSubmit2", 15},
    {"vkQueueSubmit2KHR", 18},
    {"vkQueueWaitIdle", 16},
    {"vkRegisterCustomBorderColorEXT", 31},
    {"vkRegisterDeviceEventEXT", 25},
    {"vkRegisterDisplayEventEXT", 26},
    {"vkReleaseCapturedPipelineDataKHR", 33},
    {"vkReleaseDisplayEXT", 20},
    {"vkReleaseFullScreenExclusiveModeEXT", 36},
    {"vkReleasePerformanceConfigurationINTEL", 39},
    {"vkReleaseProfilingLockKHR", 26},
    {"vkReleaseSwapchainImagesEXT", 28},
    {"vkReleaseSwapchainImagesKHR", 28},
    {"vkResetCommandBuffer", 21},
    {"vkResetCommandPool", 19},
    {"vkResetDescriptorPool", 22},
    {"vkResetEvent", 13},
    {"vkResetFences", 14},
    {"vkResetGpaSessionAMD", 21},
    {"vkResetQueryPool", 17},
    {"vkResetQueryPoolEXT", 20},
    {"vkSetBufferCollectionBufferConstraintsFUCHSIA", 46},
    {"vkSetBufferCollectionImageConstraintsFUCHSIA", 45},
    {"vkSetDebugUtilsObjectNameEXT", 29},
    {"vkSetDebugUtilsObjectTagEXT", 28},
    {"vkSetDeviceMemoryPriorityEXT", 29},
    {"vkSetEvent", 11},
    {"vkSetGpaDeviceClockModeAMD", 27},
    {"vkSetHdrMetadataEXT", 20},
    {"vkSetLatencyMarkerLegacyNV", 27},
    {"vkSetLatencyMarkerNV", 21},
    {"vkSetLatencySleepModeLegacyNV", 30},
    {"vkSetLatencySleepModeNV", 24},
    {"vkSetLocalDimmingAMD", 21},
    {"vkSetPrivateData", 17},
    {"vkSetPrivateDataEXT", 20},
    {"vkSetSwapchainPresentTimingQueueSizeEXT", 40},
    {"vkShutdownLatencyDeviceLegacyNV", 32},
    {"vkSignalSemaphore", 18},
    {"vkSignalSemaphoreKHR", 21},
    {"vkSubmitDebugUtilsMessageEXT", 29},
    {"vkTransitionImageLayout", 24},
    {"vkTransitionImageLayoutEXT", 27},
    {"vkTrimCommandPool", 18},
    {"vkTrimCommandPoolKHR", 21},
    {"vkUninitializePerformanceApiINTEL", 34},
    {"vkUnmapMemory", 14},
    {"vkUnmapMemory2", 15},
    {"vkUnmapMemory2KHR", 18},
    {"vkUnregisterCustomBorderColorEXT", 33},
    {"vkUpdateDescriptorSetWithTemplate", 34},
    {"vkUpdateDescriptorSetWithTemplateKHR", 37},
    {"vkUpdateDescriptorSets", 23},
    {"vkUpdateIndirectExecutionSetPipelineEXT", 40},
    {"vkUpdateIndirectExecutionSetShaderEXT", 38},
    {"vkUpdateVideoSessionParametersKHR", 34},
    {"vkWaitForFences", 16},
    {"vkWaitForPresent2KHR", 21},
    {"vkWaitForPresentKHR", 20},
    {"vkWaitSemaphores", 17},
    {"vkWaitSemaphoresKHR", 20},
    {"vkWriteAccelerationStructuresPropertiesKHR", 43},
    {"vkWriteMicromapsPropertiesEXT", 30},
    {"vkWriteResourceDescriptorsEXT", 30},
    {"vkWriteSamplerDescriptorsEXT", 29},
    };
    return table[(int)func].data();
}

const char* String(Struct structure) {
    static const std::string_view table[] = {
    {"INVALID_EMPTY", 15}, // Struct::Empty
    {"VkAabbPositionsKHR", 19},
    {"VkAccelerationStructureBuildGeometryInfoKHR", 44},
    {"VkAccelerationStructureBuildRangeInfoKHR", 41},
    {"VkAccelerationStructureBuildSizesInfoKHR", 41},
    {"VkAccelerationStructureCaptureDescriptorDataInfoEXT", 52},
    {"VkAccelerationStructureCreateInfo2KHR", 38},
    {"VkAccelerationStructureCreateInfoKHR", 37},
    {"VkAccelerationStructureCreateInfoNV", 36},
    {"VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX", 60},
    {"VkAccelerationStructureDeviceAddressInfoKHR", 44},
    {"VkAccelerationStructureGeometryAabbsDataKHR", 44},
    {"VkAccelerationStructureGeometryDataKHR", 39},
    {"VkAccelerationStructureGeometryInstancesDataKHR", 48},
    {"VkAccelerationStructureGeometryKHR", 35},
    {"VkAccelerationStructureGeometryLinearSweptSpheresDataNV", 56},
    {"VkAccelerationStructureGeometryMicromapDataKHR", 47},
    {"VkAccelerationStructureGeometryMotionTrianglesDataNV", 53},
    {"VkAccelerationStructureGeometrySpheresDataNV", 45},
    {"VkAccelerationStructureGeometryTrianglesDataKHR", 48},
    {"VkAccelerationStructureInfoNV", 30},
    {"VkAccelerationStructureInstanceKHR", 35},
    {"VkAccelerationStructureMatrixMotionInstanceNV", 46},
    {"VkAccelerationStructureMemoryRequirementsInfoNV", 48},
    {"VkAccelerationStructureMotionInfoNV", 36},
    {"VkAccelerationStructureMotionInstanceDataNV", 44},
    {"VkAccelerationStructureMotionInstanceNV", 40},
    {"VkAccelerationStructureSRTMotionInstanceNV", 43},
    {"VkAccelerationStructureTrianglesDisplacementMicromapNV", 55},
    {"VkAccelerationStructureTrianglesOpacityMicromapEXT", 51},
    {"VkAccelerationStructureTrianglesOpacityMicromapKHR", 51},
    {"VkAccelerationStructureVersionInfoKHR", 38},
    {"VkAcquireNextImageInfoKHR", 26},
    {"VkAcquireProfilingLockInfoKHR", 30},
    {"VkAllocationCallbacks", 22},
    {"VkAmigoProfilingSubmitInfoSEC", 30},
    {"VkAndroidHardwareBufferFormatProperties2ANDROID", 48},
    {"VkAndroidHardwareBufferFormatPropertiesANDROID", 47},
    {"VkAndroidHardwareBufferFormatResolvePropertiesANDROID", 54},
    {"VkAndroidHardwareBufferPropertiesANDROID", 41},
    {"VkAndroidHardwareBufferUsageANDROID", 36},
    {"VkAndroidSurfaceCreateInfoKHR", 30},
    {"VkAntiLagDataAMD", 17},
    {"VkAntiLagPresentationInfoAMD", 29},
    {"VkApplicationInfo", 18},
    {"VkAttachmentDescription", 24},
    {"VkAttachmentDescription2", 25},
    {"VkAttachmentDescriptionStencilLayout", 37},
    {"VkAttachmentFeedbackLoopInfoEXT", 32},
    {"VkAttachmentReference", 22},
    {"VkAttachmentReference2", 23},
    {"VkAttachmentReferenceStencilLayout", 35},
    {"VkAttachmentSampleCountInfoAMD", 31},
    {"VkAttachmentSampleLocationsEXT", 31},
    {"VkBaseInStructure", 18},
    {"VkBaseOutStructure", 19},
    {"VkBeginCustomResolveInfoEXT", 28},
    {"VkBindAccelerationStructureMemoryInfoNV", 40},
    {"VkBindBufferMemoryDeviceGroupInfo", 34},
    {"VkBindBufferMemoryInfo", 23},
    {"VkBindDataGraphPipelineSessionMemoryInfoARM", 44},
    {"VkBindDescriptorBufferEmbeddedSamplersInfoEXT", 46},
    {"VkBindDescriptorSetsInfo", 25},
    {"VkBindHeapInfoEXT", 18},
    {"VkBindImageMemoryDeviceGroupInfo", 33},
    {"VkBindImageMemoryInfo", 22},
    {"VkBindImageMemorySwapchainInfoKHR", 34},
    {"VkBindImagePlaneMemoryInfo", 27},
    {"VkBindIndexBuffer3InfoKHR", 26},
    {"VkBindIndexBufferIndirectCommandEXT", 36},
    {"VkBindIndexBufferIndirectCommandNV", 35},
    {"VkBindMemoryStatus", 19},
    {"VkBindPipelineIndirectCommandNV", 32},
    {"VkBindShaderGroupIndirectCommandNV", 35},
    {"VkBindSparseInfo", 17},
    {"VkBindTensorMemoryInfoARM", 26},
    {"VkBindTransformFeedbackBuffer2InfoEXT", 38},
    {"VkBindVertexBuffer3InfoKHR", 27},
    {"VkBindVertexBufferIndirectCommandEXT", 37},
    {"VkBindVertexBufferIndirectCommandNV", 36},
    {"VkBindVideoSessionMemoryInfoKHR", 32},
    {"VkBlitImageCubicWeightsInfoQCOM", 32},
    {"VkBlitImageInfo2", 17},
    {"VkBufferCaptureDescriptorDataInfoEXT", 37},
    {"VkBufferCollectionBufferCreateInfoFUCHSIA", 42},
    {"VkBufferCollectionConstraintsInfoFUCHSIA", 41},
    {"VkBufferCollectionCreateInfoFUCHSIA", 36},
    {"VkBufferCollectionImageCreateInfoFUCHSIA", 41},
    {"VkBufferCollectionPropertiesFUCHSIA", 36},
    {"VkBufferConstraintsInfoFUCHSIA", 31},
    {"VkBufferCopy", 13},
    {"VkBufferCopy2", 14},
    {"VkBufferCreateInfo", 19},
    {"VkBufferDeviceAddressCreateInfoEXT", 35},
    {"VkBufferDeviceAddressInfo", 26},
    {"VkBufferImageCopy", 18},
    {"VkBufferImageCopy2", 19},
    {"VkBufferMemoryBarrier", 22},
    {"VkBufferMemoryBarrier2", 23},
    {"VkBufferMemoryRequirementsInfo2", 32},
    {"VkBufferOpaqueCaptureAddressCreateInfo", 39},
    {"VkBufferUsageFlags2CreateInfo", 30},
    {"VkBufferViewCreateInfo", 23},
    {"VkBuildPartitionedAccelerationStructureIndirectCommandNV", 57},
    {"VkBuildPartitionedAccelerationStructureInfoNV", 46},
    {"VkCalibratedTimestampInfoKHR", 29},
    {"VkCheckpointData2NV", 20},
    {"VkCheckpointDataNV", 19},
    {"VkClearAttachment", 18},
    {"VkClearColorValue", 18},
    {"VkClearDepthStencilValue", 25},
    {"VkClearRect", 12},
    {"VkClearValue", 13},
    {"VkClusterAccelerationStructureBuildClustersBottomLevelInfoNV", 61},
    {"VkClusterAccelerationStructureBuildTriangleClusterInfoNV", 57},
    {"VkClusterAccelerationStructureBuildTriangleClusterTemplateInfoNV", 65},
    {"VkClusterAccelerationStructureClustersBottomLevelInputNV", 57},
    {"VkClusterAccelerationStructureCommandsInfoNV", 45},
    {"VkClusterAccelerationStructureGeometryIndexAndGeometryFlagsNV", 62},
    {"VkClusterAccelerationStructureGetTemplateIndicesInfoNV", 55},
    {"VkClusterAccelerationStructureInputInfoNV", 42},
    {"VkClusterAccelerationStructureInstantiateClusterInfoNV", 55},
    {"VkClusterAccelerationStructureMoveObjectsInfoNV", 48},
    {"VkClusterAccelerationStructureMoveObjectsInputNV", 49},
    {"VkClusterAccelerationStructureOpInputNV", 40},
    {"VkClusterAccelerationStructureTriangleClusterInputNV", 53},
    {"VkCoarseSampleLocationNV", 25},
    {"VkCoarseSampleOrderCustomNV", 28},
    {"VkColorBlendAdvancedEXT", 24},
    {"VkColorBlendEquationEXT", 24},
    {"VkCommandBufferAllocateInfo", 28},
    {"VkCommandBufferBeginInfo", 25},
    {"VkCommandBufferInheritanceConditionalRenderingInfoEXT", 54},
    {"VkCommandBufferInheritanceDescriptorHeapInfoEXT", 48},
    {"VkCommandBufferInheritanceInfo", 31},
    {"VkCommandBufferInheritanceRenderPassTransformInfoQCOM", 54},
    {"VkCommandBufferInheritanceRenderingInfo", 40},
    {"VkCommandBufferInheritanceViewportScissorInfoNV", 48},
    {"VkCommandBufferSubmitInfo", 26},
    {"VkCommandPoolCreateInfo", 24},
    {"VkComponentMapping", 19},
    {"VkComputeOccupancyPriorityParametersNV", 39},
    {"VkComputePipelineCreateInfo", 28},
    {"VkComputePipelineIndirectBufferInfoNV", 38},
    {"VkConditionalRenderingBeginInfo2EXT", 36},
    {"VkConditionalRenderingBeginInfoEXT", 35},
    {"VkConformanceVersion", 21},
    {"VkConvertCooperativeVectorMatrixInfoNV", 39},
    {"VkCooperativeMatrixFlexibleDimensionsPropertiesNV", 50},
    {"VkCooperativeMatrixPropertiesKHR", 33},
    {"VkCooperativeMatrixPropertiesNV", 32},
    {"VkCooperativeVectorPropertiesNV", 32},
    {"VkCopyAccelerationStructureInfoKHR", 35},
    {"VkCopyAccelerationStructureToMemoryInfoKHR", 43},
    {"VkCopyBufferInfo2", 18},
    {"VkCopyBufferToImageInfo2", 25},
    {"VkCopyCommandTransformInfoQCOM", 31},
    {"VkCopyDescriptorSet", 20},
    {"VkCopyDeviceMemoryImageInfoKHR", 31},
    {"VkCopyDeviceMemoryInfoKHR", 26},
    {"VkCopyImageInfo2", 17},
    {"VkCopyImageToBufferInfo2", 25},
    {"VkCopyImageToImageInfo", 23},
    {"VkCopyImageToMemoryInfo", 24},
    {"VkCopyMemoryIndirectCommandKHR", 31},
    {"VkCopyMemoryIndirectInfoKHR", 28},
    {"VkCopyMemoryToAccelerationStructureInfoKHR", 43},
    {"VkCopyMemoryToImageIndirectCommandKHR", 38},
    {"VkCopyMemoryToImageIndirectInfoKHR", 35},
    {"VkCopyMemoryToImageInfo", 24},
    {"VkCopyMemoryToMicromapInfoEXT", 30},
    {"VkCopyMicromapInfoEXT", 22},
    {"VkCopyMicromapToMemoryInfoEXT", 30},
    {"VkCopyTensorInfoARM", 20},
    {"VkCuFunctionCreateInfoNVX", 26},
    {"VkCuLaunchInfoNVX", 18},
    {"VkCuModuleCreateInfoNVX", 24},
    {"VkCuModuleTexturingModeCreateInfoNVX", 37},
    {"VkCudaFunctionCreateInfoNV", 27},
    {"VkCudaLaunchInfoNV", 19},
    {"VkCudaModuleCreateInfoNV", 25},
    {"VkCustomResolveCreateInfoEXT", 29},
    {"VkD3D12FenceSubmitInfoKHR", 26},
    {"VkDataGraphOpticalFlowImageFormatInfoARM", 41},
    {"VkDataGraphOpticalFlowImageFormatPropertiesARM", 47},
    {"VkDataGraphPipelineBuiltinModelCreateInfoQCOM", 46},
    {"VkDataGraphPipelineCompilerControlCreateInfoARM", 48},
    {"VkDataGraphPipelineConstantARM", 31},
    {"VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM", 63},
    {"VkDataGraphPipelineCreateInfoARM", 33},
    {"VkDataGraphPipelineDispatchInfoARM", 35},
    {"VkDataGraphPipelineIdentifierCreateInfoARM", 43},
    {"VkDataGraphPipelineInfoARM", 27},
    {"VkDataGraphPipelineNeuralStatisticsCreateInfoARM", 49},
    {"VkDataGraphPipelineOpticalFlowCreateInfoARM", 44},
    {"VkDataGraphPipelineOpticalFlowDispatchInfoARM", 46},
    {"VkDataGraphPipelinePropertyQueryResultARM", 42},
    {"VkDataGraphPipelineResourceInfoARM", 35},
    {"VkDataGraphPipelineResourceInfoImageLayoutARM", 46},
    {"VkDataGraphPipelineSessionBindPointRequirementARM", 50},
    {"VkDataGraphPipelineSessionBindPointRequirementsInfoARM", 55},
    {"VkDataGraphPipelineSessionCreateInfoARM", 40},
    {"VkDataGraphPipelineSessionMemoryRequirementsInfoARM", 52},
    {"VkDataGraphPipelineSessionNeuralStatisticsCreateInfoARM", 56},
    {"VkDataGraphPipelineShaderModuleCreateInfoARM", 45},
    {"VkDataGraphPipelineSingleNodeConnectionARM", 43},
    {"VkDataGraphPipelineSingleNodeCreateInfoARM", 43},
    {"VkDataGraphProcessingEngineCreateInfoARM", 41},
    {"VkDataGraphTOSANameQualityARM", 30},
    {"VkDebugMarkerMarkerInfoEXT", 27},
    {"VkDebugMarkerObjectNameInfoEXT", 31},
    {"VkDebugMarkerObjectTagInfoEXT", 30},
    {"VkDebugReportCallbackCreateInfoEXT", 35},
    {"VkDebugUtilsLabelEXT", 21},
    {"VkDebugUtilsMessengerCallbackDataEXT", 37},
    {"VkDebugUtilsMessengerCreateInfoEXT", 35},
    {"VkDebugUtilsObjectNameInfoEXT", 30},
    {"VkDebugUtilsObjectTagInfoEXT", 29},
    {"VkDecompressMemoryInfoEXT", 26},
    {"VkDecompressMemoryRegionEXT", 28},
    {"VkDecompressMemoryRegionNV", 27},
    {"VkDedicatedAllocationBufferCreateInfoNV", 40},
    {"VkDedicatedAllocationImageCreateInfoNV", 39},
    {"VkDedicatedAllocationMemoryAllocateInfoNV", 42},
    {"VkDependencyInfo", 17},
    {"VkDepthBiasInfoEXT", 19},
    {"VkDepthBiasRepresentationInfoEXT", 33},
    {"VkDepthClampRangeEXT", 21},
    {"VkDescriptorAddressInfoEXT", 27},
    {"VkDescriptorBufferBindingInfoEXT", 33},
    {"VkDescriptorBufferBindingPushDescriptorBufferHandleEXT", 55},
    {"VkDescriptorBufferInfo", 23},
    {"VkDescriptorDataEXT", 20},
    {"VkDescriptorGetInfoEXT", 23},
    {"VkDescriptorGetTensorInfoARM", 29},
    {"VkDescriptorImageInfo", 22},
    {"VkDescriptorMappingSourceConstantOffsetEXT", 43},
    {"VkDescriptorMappingSourceDataEXT", 33},
    {"VkDescriptorMappingSourceHeapDataEXT", 37},
    {"VkDescriptorMappingSourceIndirectAddressEXT", 44},
    {"VkDescriptorMappingSourceIndirectIndexArrayEXT", 47},
    {"VkDescriptorMappingSourceIndirectIndexEXT", 42},
    {"VkDescriptorMappingSourcePushIndexEXT", 38},
    {"VkDescriptorMappingSourceShaderRecordIndexEXT", 46},
    {"VkDescriptorPoolCreateInfo", 27},
    {"VkDescriptorPoolInlineUniformBlockCreateInfo", 45},
    {"VkDescriptorPoolSize", 21},
    {"VkDescriptorSetAllocateInfo", 28},
    {"VkDescriptorSetAndBindingMappingEXT", 36},
    {"VkDescriptorSetBindingReferenceVALVE", 37},
    {"VkDescriptorSetLayoutBinding", 29},
    {"VkDescriptorSetLayoutBindingFlagsCreateInfo", 44},
    {"VkDescriptorSetLayoutCreateInfo", 32},
    {"VkDescriptorSetLayoutHostMappingInfoVALVE", 42},
    {"VkDescriptorSetLayoutSupport", 29},
    {"VkDescriptorSetVariableDescriptorCountAllocateInfo", 51},
    {"VkDescriptorSetVariableDescriptorCountLayoutSupport", 52},
    {"VkDescriptorUpdateTemplateCreateInfo", 37},
    {"VkDescriptorUpdateTemplateEntry", 32},
    {"VkDeviceAddressBindingCallbackDataEXT", 38},
    {"VkDeviceAddressRangeKHR", 24},
    {"VkDeviceBufferMemoryRequirements", 33},
    {"VkDeviceCreateInfo", 19},
    {"VkDeviceDeviceMemoryReportCreateInfoEXT", 40},
    {"VkDeviceDiagnosticsConfigCreateInfoNV", 38},
    {"VkDeviceEventInfoEXT", 21},
    {"VkDeviceFaultAddressInfoKHR", 28},
    {"VkDeviceFaultCountsEXT", 23},
    {"VkDeviceFaultDebugInfoKHR", 26},
    {"VkDeviceFaultInfoEXT", 21},
    {"VkDeviceFaultInfoKHR", 21},
    {"VkDeviceFaultShaderAbortMessageInfoKHR", 39},
    {"VkDeviceFaultVendorBinaryHeaderVersionOneKHR", 45},
    {"VkDeviceFaultVendorInfoKHR", 27},
    {"VkDeviceGroupBindSparseInfo", 28},
    {"VkDeviceGroupCommandBufferBeginInfo", 36},
    {"VkDeviceGroupDeviceCreateInfo", 30},
    {"VkDeviceGroupPresentCapabilitiesKHR", 36},
    {"VkDeviceGroupPresentInfoKHR", 28},
    {"VkDeviceGroupRenderPassBeginInfo", 33},
    {"VkDeviceGroupSubmitInfo", 24},
    {"VkDeviceGroupSwapchainCreateInfoKHR", 36},
    {"VkDeviceImageMemoryRequirements", 32},
    {"VkDeviceImageSubresourceInfo", 29},
    {"VkDeviceMemoryCopyKHR", 22},
    {"VkDeviceMemoryImageCopyKHR", 27},
    {"VkDeviceMemoryOpaqueCaptureAddressInfo", 39},
    {"VkDeviceMemoryOverallocationCreateInfoAMD", 42},
    {"VkDeviceMemoryReportCallbackDataEXT", 36},
    {"VkDeviceOrHostAddressConstAMDX", 31},
    {"VkDeviceOrHostAddressConstKHR", 30},
    {"VkDeviceOrHostAddressKHR", 25},
    {"VkDevicePipelineBinaryInternalCacheControlKHR", 46},
    {"VkDevicePrivateDataCreateInfo", 30},
    {"VkDeviceQueueCreateInfo", 24},
    {"VkDeviceQueueGlobalPriorityCreateInfo", 38},
    {"VkDeviceQueueInfo2", 19},
    {"VkDeviceQueueShaderCoreControlCreateInfoARM", 44},
    {"VkDeviceTensorMemoryRequirementsARM", 36},
    {"VkDirectDriverLoadingInfoLUNARG", 32},
    {"VkDirectDriverLoadingListLUNARG", 32},
    {"VkDirectFBSurfaceCreateInfoEXT", 31},
    {"VkDispatchGraphCountInfoAMDX", 29},
    {"VkDispatchGraphInfoAMDX", 24},
    {"VkDispatchIndirect2InfoKHR", 27},
    {"VkDispatchIndirectCommand", 26},
    {"VkDispatchParametersARM", 24},
    {"VkDispatchTileInfoQCOM", 23},
    {"VkDisplayEventInfoEXT", 22},
    {"VkDisplayModeCreateInfoKHR", 27},
    {"VkDisplayModeParametersKHR", 27},
    {"VkDisplayModeProperties2KHR", 28},
    {"VkDisplayModePropertiesKHR", 27},
    {"VkDisplayModeStereoPropertiesNV", 32},
    {"VkDisplayNativeHdrSurfaceCapabilitiesAMD", 41},
    {"VkDisplayPlaneCapabilities2KHR", 31},
    {"VkDisplayPlaneCapabilitiesKHR", 30},
    {"VkDisplayPlaneInfo2KHR", 23},
    {"VkDisplayPlaneProperties2KHR", 29},
    {"VkDisplayPlanePropertiesKHR", 28},
    {"VkDisplayPowerInfoEXT", 22},
    {"VkDisplayPresentInfoKHR", 24},
    {"VkDisplayProperties2KHR", 24},
    {"VkDisplayPropertiesKHR", 23},
    {"VkDisplaySurfaceCreateInfoKHR", 30},
    {"VkDisplaySurfaceStereoCreateInfoNV", 35},
    {"VkDrawIndexedIndirectCommand", 29},
    {"VkDrawIndirect2InfoKHR", 23},
    {"VkDrawIndirectCommand", 22},
    {"VkDrawIndirectCount2InfoKHR", 28},
    {"VkDrawIndirectCountIndirectCommandEXT", 38},
    {"VkDrawMeshTasksIndirectCommandEXT", 34},
    {"VkDrawMeshTasksIndirectCommandNV", 33},
    {"VkDrmFormatModifierProperties2EXT", 34},
    {"VkDrmFormatModifierPropertiesEXT", 33},
    {"VkDrmFormatModifierPropertiesList2EXT", 38},
    {"VkDrmFormatModifierPropertiesListEXT", 37},
    {"VkEventCreateInfo", 18},
    {"VkExecutionGraphPipelineCreateInfoAMDX", 39},
    {"VkExecutionGraphPipelineScratchSizeAMDX", 40},
    {"VkExportFenceCreateInfo", 24},
    {"VkExportFenceWin32HandleInfoKHR", 32},
    {"VkExportMemoryAllocateInfo", 27},
    {"VkExportMemoryAllocateInfoNV", 29},
    {"VkExportMemoryWin32HandleInfoKHR", 33},
    {"VkExportMemoryWin32HandleInfoNV", 32},
    {"VkExportMetalBufferInfoEXT", 27},
    {"VkExportMetalCommandQueueInfoEXT", 33},
    {"VkExportMetalDeviceInfoEXT", 27},
    {"VkExportMetalIOSurfaceInfoEXT", 30},
    {"VkExportMetalObjectCreateInfoEXT", 33},
    {"VkExportMetalObjectsInfoEXT", 28},
    {"VkExportMetalSharedEventInfoEXT", 32},
    {"VkExportMetalTextureInfoEXT", 28},
    {"VkExportSemaphoreCreateInfo", 28},
    {"VkExportSemaphoreWin32HandleInfoKHR", 36},
    {"VkExtensionProperties", 22},
    {"VkExtent2D", 11},
    {"VkExtent3D", 11},
    {"VkExternalBufferProperties", 27},
    {"VkExternalComputeQueueCreateInfoNV", 35},
    {"VkExternalComputeQueueDataParamsNV", 35},
    {"VkExternalComputeQueueDeviceCreateInfoNV", 41},
    {"VkExternalFenceProperties", 26},
    {"VkExternalFormatANDROID", 24},
    {"VkExternalFormatOHOS", 21},
    {"VkExternalFormatQNX", 20},
    {"VkExternalImageFormatProperties", 32},
    {"VkExternalImageFormatPropertiesNV", 34},
    {"VkExternalMemoryAcquireUnmodifiedEXT", 37},
    {"VkExternalMemoryBufferCreateInfo", 33},
    {"VkExternalMemoryImageCreateInfo", 32},
    {"VkExternalMemoryImageCreateInfoNV", 34},
    {"VkExternalMemoryProperties", 27},
    {"VkExternalMemoryTensorCreateInfoARM", 36},
    {"VkExternalSemaphoreProperties", 30},
    {"VkExternalTensorPropertiesARM", 30},
    {"VkFenceCreateInfo", 18},
    {"VkFenceGetFdInfoKHR", 20},
    {"VkFenceGetWin32HandleInfoKHR", 29},
    {"VkFilterCubicImageViewImageFormatPropertiesEXT", 47},
    {"VkFormatProperties", 19},
    {"VkFormatProperties2", 20},
    {"VkFormatProperties3", 20},
    {"VkFormatProperties4KHR", 23},
    {"VkFragmentShadingRateAttachmentInfoKHR", 39},
    {"VkFrameBoundaryEXT", 19},
    {"VkFrameBoundaryTensorsARM", 26},
    {"VkFramebufferAttachmentImageInfo", 33},
    {"VkFramebufferAttachmentsCreateInfo", 35},
    {"VkFramebufferCreateInfo", 24},
    {"VkFramebufferMixedSamplesCombinationNV", 39},
    {"VkGeneratedCommandsInfoEXT", 27},
    {"VkGeneratedCommandsInfoNV", 26},
    {"VkGeneratedCommandsMemoryRequirementsInfoEXT", 45},
    {"VkGeneratedCommandsMemoryRequirementsInfoNV", 44},
    {"VkGeneratedCommandsPipelineInfoEXT", 35},
    {"VkGeneratedCommandsShaderInfoEXT", 33},
    {"VkGeometryAABBNV", 17},
    {"VkGeometryDataNV", 17},
    {"VkGeometryNV", 13},
    {"VkGeometryTrianglesNV", 22},
    {"VkGetLatencyMarkerInfoNV", 25},
    {"VkGpaDeviceClockModeInfoAMD", 28},
    {"VkGpaDeviceGetClockInfoAMD", 27},
    {"VkGpaPerfBlockPropertiesAMD", 28},
    {"VkGpaPerfCounterAMD", 20},
    {"VkGpaSampleBeginInfoAMD", 24},
    {"VkGpaSessionCreateInfoAMD", 26},
    {"VkGraphicsPipelineCreateInfo", 29},
    {"VkGraphicsPipelineLibraryCreateInfoEXT", 39},
    {"VkGraphicsPipelineShaderGroupsCreateInfoNV", 43},
    {"VkGraphicsShaderGroupCreateInfoNV", 34},
    {"VkHdrMetadataEXT", 17},
    {"VkHdrVividDynamicMetadataHUAWEI", 32},
    {"VkHeadlessSurfaceCreateInfoEXT", 31},
    {"VkHostAddressRangeConstEXT", 27},
    {"VkHostAddressRangeEXT", 22},
    {"VkHostImageCopyDevicePerformanceQuery", 38},
    {"VkHostImageLayoutTransitionInfo", 32},
    {"VkIOSSurfaceCreateInfoMVK", 26},
    {"VkImageAlignmentControlCreateInfoMESA", 38},
    {"VkImageBlit", 12},
    {"VkImageBlit2", 13},
    {"VkImageCaptureDescriptorDataInfoEXT", 36},
    {"VkImageCompressionControlEXT", 29},
    {"VkImageCompressionPropertiesEXT", 32},
    {"VkImageConstraintsInfoFUCHSIA", 30},
    {"VkImageCopy", 12},
    {"VkImageCopy2", 13},
    {"VkImageCreateFlags2CreateInfoKHR", 33},
    {"VkImageCreateInfo", 18},
    {"VkImageDescriptorInfoEXT", 25},
    {"VkImageDrmFormatModifierExplicitCreateInfoEXT", 46},
    {"VkImageDrmFormatModifierListCreateInfoEXT", 42},
    {"VkImageDrmFormatModifierPropertiesEXT", 38},
    {"VkImageFormatConstraintsInfoFUCHSIA", 36},
    {"VkImageFormatListCreateInfo", 28},
    {"VkImageFormatProperties", 24},
    {"VkImageFormatProperties2", 25},
    {"VkImageMemoryBarrier", 21},
    {"VkImageMemoryBarrier2", 22},
    {"VkImageMemoryRequirementsInfo2", 31},
    {"VkImagePipeSurfaceCreateInfoFUCHSIA", 36},
    {"VkImagePlaneMemoryRequirementsInfo", 35},
    {"VkImageResolve", 15},
    {"VkImageResolve2", 16},
    {"VkImageSparseMemoryRequirementsInfo2", 37},
    {"VkImageStencilUsage2CreateInfoKHR", 34},
    {"VkImageStencilUsageCreateInfo", 30},
    {"VkImageSubresource", 19},
    {"VkImageSubresource2", 20},
    {"VkImageSubresourceLayers", 25},
    {"VkImageSubresourceRange", 24},
    {"VkImageSwapchainCreateInfoKHR", 30},
    {"VkImageToMemoryCopy", 20},
    {"VkImageUsageFlags2CreateInfoKHR", 32},
    {"VkImageViewASTCDecodeModeEXT", 29},
    {"VkImageViewAddressPropertiesNVX", 32},
    {"VkImageViewCaptureDescriptorDataInfoEXT", 40},
    {"VkImageViewCreateInfo", 22},
    {"VkImageViewHandleInfoNVX", 25},
    {"VkImageViewMinLodCreateInfoEXT", 31},
    {"VkImageViewSampleWeightCreateInfoQCOM", 38},
    {"VkImageViewSlicedCreateInfoEXT", 31},
    {"VkImageViewUsage2CreateInfoKHR", 31},
    {"VkImageViewUsageCreateInfo", 27},
    {"VkImportAndroidHardwareBufferInfoANDROID", 41},
    {"VkImportFenceFdInfoKHR", 23},
    {"VkImportFenceWin32HandleInfoKHR", 32},
    {"VkImportMemoryBufferCollectionFUCHSIA", 38},
    {"VkImportMemoryFdInfoKHR", 24},
    {"VkImportMemoryHostPointerInfoEXT", 33},
    {"VkImportMemoryMetalHandleInfoEXT", 33},
    {"VkImportMemoryWin32HandleInfoKHR", 33},
    {"VkImportMemoryWin32HandleInfoNV", 32},
    {"VkImportMemoryZirconHandleInfoFUCHSIA", 38},
    {"VkImportMetalBufferInfoEXT", 27},
    {"VkImportMetalIOSurfaceInfoEXT", 30},
    {"VkImportMetalSharedEventInfoEXT", 32},
    {"VkImportMetalTextureInfoEXT", 28},
    {"VkImportNativeBufferInfoOHOS", 29},
    {"VkImportScreenBufferInfoQNX", 28},
    {"VkImportSemaphoreFdInfoKHR", 27},
    {"VkImportSemaphoreWin32HandleInfoKHR", 36},
    {"VkImportSemaphoreZirconHandleInfoFUCHSIA", 41},
    {"VkIndirectCommandsExecutionSetTokenEXT", 39},
    {"VkIndirectCommandsIndexBufferTokenEXT", 38},
    {"VkIndirectCommandsLayoutCreateInfoEXT", 38},
    {"VkIndirectCommandsLayoutCreateInfoNV", 37},
    {"VkIndirectCommandsLayoutPushDataTokenNV", 40},
    {"VkIndirectCommandsLayoutTokenEXT", 33},
    {"VkIndirectCommandsLayoutTokenNV", 32},
    {"VkIndirectCommandsPushConstantTokenEXT", 39},
    {"VkIndirectCommandsStreamNV", 27},
    {"VkIndirectCommandsTokenDataEXT", 31},
    {"VkIndirectCommandsVertexBufferTokenEXT", 39},
    {"VkIndirectExecutionSetCreateInfoEXT", 36},
    {"VkIndirectExecutionSetInfoEXT", 30},
    {"VkIndirectExecutionSetPipelineInfoEXT", 38},
    {"VkIndirectExecutionSetShaderInfoEXT", 36},
    {"VkIndirectExecutionSetShaderLayoutInfoEXT", 42},
    {"VkInitializePerformanceApiInfoINTEL", 36},
    {"VkInputAttachmentAspectReference", 33},
    {"VkInstanceCreateInfo", 21},
    {"VkLatencySleepInfoNV", 21},
    {"VkLatencySleepModeInfoNV", 25},
    {"VkLatencySubmissionPresentIdNV", 31},
    {"VkLatencySurfaceCapabilitiesNV", 31},
    {"VkLatencyTimingsFrameReportNV", 30},
    {"VkLayerProperties", 18},
    {"VkLayerSettingEXT", 18},
    {"VkLayerSettingsCreateInfoEXT", 29},
    {"VkMacOSSurfaceCreateInfoMVK", 28},
    {"VkMappedMemoryRange", 20},
    {"VkMemoryAllocateFlagsInfo", 26},
    {"VkMemoryAllocateInfo", 21},
    {"VkMemoryBarrier", 16},
    {"VkMemoryBarrier2", 17},
    {"VkMemoryBarrierAccessFlags3KHR", 31},
    {"VkMemoryDedicatedAllocateInfo", 30},
    {"VkMemoryDedicatedAllocateInfoTensorARM", 39},
    {"VkMemoryDedicatedRequirements", 30},
    {"VkMemoryFdPropertiesKHR", 24},
    {"VkMemoryGetAndroidHardwareBufferInfoANDROID", 44},
    {"VkMemoryGetFdInfoKHR", 21},
    {"VkMemoryGetMetalHandleInfoEXT", 30},
    {"VkMemoryGetNativeBufferInfoOHOS", 32},
    {"VkMemoryGetRemoteAddressInfoNV", 31},
    {"VkMemoryGetWin32HandleInfoKHR", 30},
    {"VkMemoryGetZirconHandleInfoFUCHSIA", 35},
    {"VkMemoryHeap", 13},
    {"VkMemoryHostPointerPropertiesEXT", 33},
    {"VkMemoryMapInfo", 16},
    {"VkMemoryMapPlacedInfoEXT", 25},
    {"VkMemoryMarkerInfoAMD", 22},
    {"VkMemoryMetalHandlePropertiesEXT", 33},
    {"VkMemoryOpaqueCaptureAddressAllocateInfo", 41},
    {"VkMemoryPriorityAllocateInfoEXT", 32},
    {"VkMemoryRangeBarrierKHR", 24},
    {"VkMemoryRangeBarriersInfoKHR", 29},
    {"VkMemoryRequirements", 21},
    {"VkMemoryRequirements2", 22},
    {"VkMemoryToImageCopy", 20},
    {"VkMemoryType", 13},
    {"VkMemoryUnmapInfo", 18},
    {"VkMemoryWin32HandlePropertiesKHR", 33},
    {"VkMemoryZirconHandlePropertiesFUCHSIA", 38},
    {"VkMetalSurfaceCreateInfoEXT", 28},
    {"VkMicromapBuildInfoEXT", 23},
    {"VkMicromapBuildSizesInfoEXT", 28},
    {"VkMicromapCreateInfoEXT", 24},
    {"VkMicromapTriangleKHR", 22},
    {"VkMicromapUsageEXT", 19},
    {"VkMicromapUsageKHR", 19},
    {"VkMicromapVersionInfoEXT", 25},
    {"VkMultiDrawIndexedInfoEXT", 26},
    {"VkMultiDrawInfoEXT", 19},
    {"VkMultisamplePropertiesEXT", 27},
    {"VkMultisampledRenderToSingleSampledInfoEXT", 43},
    {"VkMultiviewPerViewAttributesInfoNVX", 36},
    {"VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM", 53},
    {"VkMutableDescriptorTypeCreateInfoEXT", 37},
    {"VkMutableDescriptorTypeListEXT", 31},
    {"VkNativeBufferFormatPropertiesOHOS", 35},
    {"VkNativeBufferPropertiesOHOS", 29},
    {"VkNativeBufferUsageOHOS", 24},
    {"VkOffset2D", 11},
    {"VkOffset3D", 11},
    {"VkOpaqueCaptureDataCreateInfoEXT", 33},
    {"VkOpaqueCaptureDescriptorDataCreateInfoEXT", 43},
    {"VkOpticalFlowExecuteInfoNV", 27},
    {"VkOpticalFlowImageFormatInfoNV", 31},
    {"VkOpticalFlowImageFormatPropertiesNV", 37},
    {"VkOpticalFlowSessionCreateInfoNV", 33},
    {"VkOpticalFlowSessionCreatePrivateDataInfoNV", 44},
    {"VkOutOfBandQueueTypeInfoNV", 27},
    {"VkPartitionedAccelerationStructureFlagsNV", 42},
    {"VkPartitionedAccelerationStructureInstancesInputNV", 51},
    {"VkPartitionedAccelerationStructureUpdateInstanceDataNV", 55},
    {"VkPartitionedAccelerationStructureWriteInstanceDataNV", 54},
    {"VkPartitionedAccelerationStructureWritePartitionTranslationDataNV", 66},
    {"VkPastPresentationTimingEXT", 28},
    {"VkPastPresentationTimingGOOGLE", 31},
    {"VkPastPresentationTimingInfoEXT", 32},
    {"VkPastPresentationTimingPropertiesEXT", 38},
    {"VkPerTileBeginInfoQCOM", 23},
    {"VkPerTileEndInfoQCOM", 21},
    {"VkPerfHintInfoQCOM", 19},
    {"VkPerformanceConfigurationAcquireInfoINTEL", 43},
    {"VkPerformanceCounterARM", 24},
    {"VkPerformanceCounterDescriptionARM", 35},
    {"VkPerformanceCounterDescriptionKHR", 35},
    {"VkPerformanceCounterKHR", 24},
    {"VkPerformanceCounterResultKHR", 30},
    {"VkPerformanceMarkerInfoINTEL", 29},
    {"VkPerformanceOverrideInfoINTEL", 31},
    {"VkPerformanceQuerySubmitInfoKHR", 32},
    {"VkPerformanceStreamMarkerInfoINTEL", 35},
    {"VkPerformanceValueDataINTEL", 28},
    {"VkPerformanceValueINTEL", 24},
    {"VkPhysicalDevice16BitStorageFeatures", 37},
    {"VkPhysicalDevice4444FormatsFeaturesEXT", 39},
    {"VkPhysicalDevice8BitStorageFeatures", 36},
    {"VkPhysicalDeviceASTCDecodeFeaturesEXT", 38},
    {"VkPhysicalDeviceAccelerationStructureFeaturesKHR", 49},
    {"VkPhysicalDeviceAccelerationStructurePropertiesKHR", 51},
    {"VkPhysicalDeviceAddressBindingReportFeaturesEXT", 48},
    {"VkPhysicalDeviceAmigoProfilingFeaturesSEC", 42},
    {"VkPhysicalDeviceAntiLagFeaturesAMD", 35},
    {"VkPhysicalDeviceAttachmentFeedbackLoopDynamicStateFeaturesEXT", 62},
    {"VkPhysicalDeviceAttachmentFeedbackLoopLayoutFeaturesEXT", 56},
    {"VkPhysicalDeviceBlendOperationAdvancedFeaturesEXT", 50},
    {"VkPhysicalDeviceBlendOperationAdvancedPropertiesEXT", 52},
    {"VkPhysicalDeviceBorderColorSwizzleFeaturesEXT", 46},
    {"VkPhysicalDeviceBufferDeviceAddressFeatures", 44},
    {"VkPhysicalDeviceBufferDeviceAddressFeaturesEXT", 47},
    {"VkPhysicalDeviceClusterAccelerationStructureFeaturesNV", 55},
    {"VkPhysicalDeviceClusterAccelerationStructurePropertiesNV", 57},
    {"VkPhysicalDeviceClusterCullingShaderFeaturesHUAWEI", 51},
    {"VkPhysicalDeviceClusterCullingShaderPropertiesHUAWEI", 53},
    {"VkPhysicalDeviceClusterCullingShaderVrsFeaturesHUAWEI", 54},
    {"VkPhysicalDeviceCoherentMemoryFeaturesAMD", 42},
    {"VkPhysicalDeviceColorWriteEnableFeaturesEXT", 44},
    {"VkPhysicalDeviceCommandBufferInheritanceFeaturesNV", 51},
    {"VkPhysicalDeviceComputeOccupancyPriorityFeaturesNV", 51},
    {"VkPhysicalDeviceComputeShaderDerivativesFeaturesKHR", 52},
    {"VkPhysicalDeviceComputeShaderDerivativesPropertiesKHR", 54},
    {"VkPhysicalDeviceConditionalRenderingFeaturesEXT", 48},
    {"VkPhysicalDeviceConservativeRasterizationPropertiesEXT", 55},
    {"VkPhysicalDeviceCooperativeMatrix2FeaturesNV", 45},
    {"VkPhysicalDeviceCooperativeMatrix2PropertiesNV", 47},
    {"VkPhysicalDeviceCooperativeMatrixConversionFeaturesQCOM", 56},
    {"VkPhysicalDeviceCooperativeMatrixDecodeVectorFeaturesNV", 56},
    {"VkPhysicalDeviceCooperativeMatrixFeaturesKHR", 45},
    {"VkPhysicalDeviceCooperativeMatrixFeaturesNV", 44},
    {"VkPhysicalDeviceCooperativeMatrixPropertiesKHR", 47},
    {"VkPhysicalDeviceCooperativeMatrixPropertiesNV", 46},
    {"VkPhysicalDeviceCooperativeVectorFeaturesNV", 44},
    {"VkPhysicalDeviceCooperativeVectorPropertiesNV", 46},
    {"VkPhysicalDeviceCopyMemoryIndirectFeaturesKHR", 46},
    {"VkPhysicalDeviceCopyMemoryIndirectFeaturesNV", 45},
    {"VkPhysicalDeviceCopyMemoryIndirectPropertiesKHR", 48},
    {"VkPhysicalDeviceCornerSampledImageFeaturesNV", 45},
    {"VkPhysicalDeviceCoverageReductionModeFeaturesNV", 48},
    {"VkPhysicalDeviceCubicClampFeaturesQCOM", 39},
    {"VkPhysicalDeviceCubicWeightsFeaturesQCOM", 41},
    {"VkPhysicalDeviceCudaKernelLaunchFeaturesNV", 43},
    {"VkPhysicalDeviceCudaKernelLaunchPropertiesNV", 45},
    {"VkPhysicalDeviceCustomBorderColorFeaturesEXT", 45},
    {"VkPhysicalDeviceCustomBorderColorPropertiesEXT", 47},
    {"VkPhysicalDeviceCustomResolveFeaturesEXT", 41},
    {"VkPhysicalDeviceDataGraphFeaturesARM", 37},
    {"VkPhysicalDeviceDataGraphModelFeaturesQCOM", 43},
    {"VkPhysicalDeviceDataGraphNeuralAcceleratorStatisticsFeaturesARM", 64},
    {"VkPhysicalDeviceDataGraphOperationSupportARM", 45},
    {"VkPhysicalDeviceDataGraphOpticalFlowFeaturesARM", 48},
    {"VkPhysicalDeviceDataGraphProcessingEngineARM", 45},
    {"VkPhysicalDeviceDedicatedAllocationImageAliasingFeaturesNV", 59},
    {"VkPhysicalDeviceDenseGeometryFormatFeaturesAMDX", 48},
    {"VkPhysicalDeviceDepthBiasControlFeaturesEXT", 44},
    {"VkPhysicalDeviceDepthClampControlFeaturesEXT", 45},
    {"VkPhysicalDeviceDepthClampZeroOneFeaturesKHR", 45},
    {"VkPhysicalDeviceDepthClipControlFeaturesEXT", 44},
    {"VkPhysicalDeviceDepthClipEnableFeaturesEXT", 43},
    {"VkPhysicalDeviceDepthStencilResolveProperties", 46},
    {"VkPhysicalDeviceDescriptorBufferDensityMapPropertiesEXT", 56},
    {"VkPhysicalDeviceDescriptorBufferFeaturesEXT", 44},
    {"VkPhysicalDeviceDescriptorBufferPropertiesEXT", 46},
    {"VkPhysicalDeviceDescriptorBufferTensorFeaturesARM", 50},
    {"VkPhysicalDeviceDescriptorBufferTensorPropertiesARM", 52},
    {"VkPhysicalDeviceDescriptorHeapFeaturesEXT", 42},
    {"VkPhysicalDeviceDescriptorHeapPropertiesEXT", 44},
    {"VkPhysicalDeviceDescriptorHeapTensorPropertiesARM", 50},
    {"VkPhysicalDeviceDescriptorIndexingFeatures", 43},
    {"VkPhysicalDeviceDescriptorIndexingProperties", 45},
    {"VkPhysicalDeviceDescriptorPoolOverallocationFeaturesNV", 55},
    {"VkPhysicalDeviceDescriptorSetHostMappingFeaturesVALVE", 54},
    {"VkPhysicalDeviceDeviceAddressCommandsFeaturesKHR", 49},
    {"VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV", 57},
    {"VkPhysicalDeviceDeviceGeneratedCommandsFeaturesEXT", 51},
    {"VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV", 50},
    {"VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT", 53},
    {"VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV", 52},
    {"VkPhysicalDeviceDeviceMemoryReportFeaturesEXT", 46},
    {"VkPhysicalDeviceDiagnosticsConfigFeaturesNV", 44},
    {"VkPhysicalDeviceDiscardRectanglePropertiesEXT", 46},
    {"VkPhysicalDeviceDisplacementMicromapFeaturesNV", 47},
    {"VkPhysicalDeviceDisplacementMicromapPropertiesNV", 49},
    {"VkPhysicalDeviceDriverProperties", 33},
    {"VkPhysicalDeviceDrmPropertiesEXT", 33},
    {"VkPhysicalDeviceDynamicRenderingFeatures", 41},
    {"VkPhysicalDeviceDynamicRenderingLocalReadFeatures", 50},
    {"VkPhysicalDeviceDynamicRenderingUnusedAttachmentsFeaturesEXT", 61},
    {"VkPhysicalDeviceElapsedTimerQueryFeaturesQCOM", 46},
    {"VkPhysicalDeviceExclusiveScissorFeaturesNV", 43},
    {"VkPhysicalDeviceExtendedDynamicState2FeaturesEXT", 49},
    {"VkPhysicalDeviceExtendedDynamicState3FeaturesEXT", 49},
    {"VkPhysicalDeviceExtendedDynamicState3PropertiesEXT", 51},
    {"VkPhysicalDeviceExtendedDynamicStateFeaturesEXT", 48},
    {"VkPhysicalDeviceExtendedFlagsFeaturesKHR", 41},
    {"VkPhysicalDeviceExtendedSparseAddressSpaceFeaturesNV", 53},
    {"VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV", 55},
    {"VkPhysicalDeviceExternalBufferInfo", 35},
    {"VkPhysicalDeviceExternalComputeQueuePropertiesNV", 49},
    {"VkPhysicalDeviceExternalFenceInfo", 34},
    {"VkPhysicalDeviceExternalFormatResolveFeaturesANDROID", 53},
    {"VkPhysicalDeviceExternalFormatResolvePropertiesANDROID", 55},
    {"VkPhysicalDeviceExternalImageFormatInfo", 40},
    {"VkPhysicalDeviceExternalMemoryHostPropertiesEXT", 48},
    {"VkPhysicalDeviceExternalMemoryRDMAFeaturesNV", 45},
    {"VkPhysicalDeviceExternalMemoryScreenBufferFeaturesQNX", 54},
    {"VkPhysicalDeviceExternalSemaphoreInfo", 38},
    {"VkPhysicalDeviceExternalTensorInfoARM", 38},
    {"VkPhysicalDeviceFaultFeaturesEXT", 33},
    {"VkPhysicalDeviceFaultFeaturesKHR", 33},
    {"VkPhysicalDeviceFaultPropertiesKHR", 35},
    {"VkPhysicalDeviceFeatures", 25},
    {"VkPhysicalDeviceFeatures2", 26},
    {"VkPhysicalDeviceFloatControlsProperties", 40},
    {"VkPhysicalDeviceFormatPackFeaturesARM", 38},
    {"VkPhysicalDeviceFragmentDensityMap2FeaturesEXT", 47},
    {"VkPhysicalDeviceFragmentDensityMap2PropertiesEXT", 49},
    {"VkPhysicalDeviceFragmentDensityMapFeaturesEXT", 46},
    {"VkPhysicalDeviceFragmentDensityMapLayeredFeaturesVALVE", 55},
    {"VkPhysicalDeviceFragmentDensityMapLayeredPropertiesVALVE", 57},
    {"VkPhysicalDeviceFragmentDensityMapOffsetFeaturesEXT", 52},
    {"VkPhysicalDeviceFragmentDensityMapOffsetPropertiesEXT", 54},
    {"VkPhysicalDeviceFragmentDensityMapPropertiesEXT", 48},
    {"VkPhysicalDeviceFragmentShaderBarycentricFeaturesKHR", 53},
    {"VkPhysicalDeviceFragmentShaderBarycentricPropertiesKHR", 55},
    {"VkPhysicalDeviceFragmentShaderInterlockFeaturesEXT", 51},
    {"VkPhysicalDeviceFragmentShadingRateEnumsFeaturesNV", 51},
    {"VkPhysicalDeviceFragmentShadingRateEnumsPropertiesNV", 53},
    {"VkPhysicalDeviceFragmentShadingRateFeaturesKHR", 47},
    {"VkPhysicalDeviceFragmentShadingRateKHR", 39},
    {"VkPhysicalDeviceFragmentShadingRatePropertiesKHR", 49},
    {"VkPhysicalDeviceFrameBoundaryFeaturesEXT", 41},
    {"VkPhysicalDeviceGlobalPriorityQueryFeatures", 44},
    {"VkPhysicalDeviceGpaFeaturesAMD", 31},
    {"VkPhysicalDeviceGpaProperties2AMD", 34},
    {"VkPhysicalDeviceGpaPropertiesAMD", 33},
    {"VkPhysicalDeviceGraphicsPipelineLibraryFeaturesEXT", 51},
    {"VkPhysicalDeviceGraphicsPipelineLibraryPropertiesEXT", 53},
    {"VkPhysicalDeviceGroupProperties", 32},
    {"VkPhysicalDeviceHdrVividFeaturesHUAWEI", 39},
    {"VkPhysicalDeviceHostImageCopyFeatures", 38},
    {"VkPhysicalDeviceHostImageCopyProperties", 40},
    {"VkPhysicalDeviceHostQueryResetFeatures", 39},
    {"VkPhysicalDeviceIDProperties", 29},
    {"VkPhysicalDeviceImage2DViewOf3DFeaturesEXT", 43},
    {"VkPhysicalDeviceImageAlignmentControlFeaturesMESA", 50},
    {"VkPhysicalDeviceImageAlignmentControlPropertiesMESA", 52},
    {"VkPhysicalDeviceImageCompressionControlFeaturesEXT", 51},
    {"VkPhysicalDeviceImageCompressionControlSwapchainFeaturesEXT", 60},
    {"VkPhysicalDeviceImageDrmFormatModifierInfoEXT", 46},
    {"VkPhysicalDeviceImageFormatInfo2", 33},
    {"VkPhysicalDeviceImageProcessing2FeaturesQCOM", 45},
    {"VkPhysicalDeviceImageProcessing2PropertiesQCOM", 47},
    {"VkPhysicalDeviceImageProcessing3FeaturesQCOM", 45},
    {"VkPhysicalDeviceImageProcessingFeaturesQCOM", 44},
    {"VkPhysicalDeviceImageProcessingPropertiesQCOM", 46},
    {"VkPhysicalDeviceImageRobustnessFeatures", 40},
    {"VkPhysicalDeviceImageSlicedViewOf3DFeaturesEXT", 47},
    {"VkPhysicalDeviceImageViewImageFormatInfoEXT", 44},
    {"VkPhysicalDeviceImageViewMinLodFeaturesEXT", 43},
    {"VkPhysicalDeviceImagelessFramebufferFeatures", 45},
    {"VkPhysicalDeviceIndexTypeUint8Features", 39},
    {"VkPhysicalDeviceInheritedViewportScissorFeaturesNV", 51},
    {"VkPhysicalDeviceInlineUniformBlockFeatures", 43},
    {"VkPhysicalDeviceInlineUniformBlockProperties", 45},
    {"VkPhysicalDeviceInternallySynchronizedQueuesFeaturesKHR", 56},
    {"VkPhysicalDeviceInvocationMaskFeaturesHUAWEI", 45},
    {"VkPhysicalDeviceLayeredApiPropertiesKHR", 40},
    {"VkPhysicalDeviceLayeredApiPropertiesListKHR", 44},
    {"VkPhysicalDeviceLayeredApiVulkanPropertiesKHR", 46},
    {"VkPhysicalDeviceLayeredDriverPropertiesMSFT", 44},
    {"VkPhysicalDeviceLegacyDitheringFeaturesEXT", 43},
    {"VkPhysicalDeviceLegacyVertexAttributesFeaturesEXT", 50},
    {"VkPhysicalDeviceLegacyVertexAttributesPropertiesEXT", 52},
    {"VkPhysicalDeviceLimits", 23},
    {"VkPhysicalDeviceLineRasterizationFeatures", 42},
    {"VkPhysicalDeviceLineRasterizationProperties", 44},
    {"VkPhysicalDeviceLinearColorAttachmentFeaturesNV", 48},
    {"VkPhysicalDeviceMaintenance10FeaturesKHR", 41},
    {"VkPhysicalDeviceMaintenance10PropertiesKHR", 43},
    {"VkPhysicalDeviceMaintenance11FeaturesKHR", 41},
    {"VkPhysicalDeviceMaintenance3Properties", 39},
    {"VkPhysicalDeviceMaintenance4Features", 37},
    {"VkPhysicalDeviceMaintenance4Properties", 39},
    {"VkPhysicalDeviceMaintenance5Features", 37},
    {"VkPhysicalDeviceMaintenance5Properties", 39},
    {"VkPhysicalDeviceMaintenance6Features", 37},
    {"VkPhysicalDeviceMaintenance6Properties", 39},
    {"VkPhysicalDeviceMaintenance7FeaturesKHR", 40},
    {"VkPhysicalDeviceMaintenance7PropertiesKHR", 42},
    {"VkPhysicalDeviceMaintenance8FeaturesKHR", 40},
    {"VkPhysicalDeviceMaintenance9FeaturesKHR", 40},
    {"VkPhysicalDeviceMaintenance9PropertiesKHR", 42},
    {"VkPhysicalDeviceMapMemoryPlacedFeaturesEXT", 43},
    {"VkPhysicalDeviceMapMemoryPlacedPropertiesEXT", 45},
    {"VkPhysicalDeviceMemoryBudgetPropertiesEXT", 42},
    {"VkPhysicalDeviceMemoryDecompressionFeaturesEXT", 47},
    {"VkPhysicalDeviceMemoryDecompressionPropertiesEXT", 49},
    {"VkPhysicalDeviceMemoryPriorityFeaturesEXT", 42},
    {"VkPhysicalDeviceMemoryProperties", 33},
    {"VkPhysicalDeviceMemoryProperties2", 34},
    {"VkPhysicalDeviceMeshShaderFeaturesEXT", 38},
    {"VkPhysicalDeviceMeshShaderFeaturesNV", 37},
    {"VkPhysicalDeviceMeshShaderPropertiesEXT", 40},
    {"VkPhysicalDeviceMeshShaderPropertiesNV", 39},
    {"VkPhysicalDeviceMultiDrawFeaturesEXT", 37},
    {"VkPhysicalDeviceMultiDrawPropertiesEXT", 39},
    {"VkPhysicalDeviceMultisampledRenderToSingleSampledFeaturesEXT", 61},
    {"VkPhysicalDeviceMultisampledRenderToSwapchainFeaturesEXT", 57},
    {"VkPhysicalDeviceMultiviewFeatures", 34},
    {"VkPhysicalDeviceMultiviewPerViewAttributesPropertiesNVX", 56},
    {"VkPhysicalDeviceMultiviewPerViewRenderAreasFeaturesQCOM", 56},
    {"VkPhysicalDeviceMultiviewPerViewViewportsFeaturesQCOM", 54},
    {"VkPhysicalDeviceMultiviewProperties", 36},
    {"VkPhysicalDeviceMutableDescriptorTypeFeaturesEXT", 49},
    {"VkPhysicalDeviceNestedCommandBufferFeaturesEXT", 47},
    {"VkPhysicalDeviceNestedCommandBufferPropertiesEXT", 49},
    {"VkPhysicalDeviceNonSeamlessCubeMapFeaturesEXT", 46},
    {"VkPhysicalDeviceOpacityMicromapFeaturesEXT", 43},
    {"VkPhysicalDeviceOpacityMicromapFeaturesKHR", 43},
    {"VkPhysicalDeviceOpacityMicromapPropertiesEXT", 45},
    {"VkPhysicalDeviceOpacityMicromapPropertiesKHR", 45},
    {"VkPhysicalDeviceOpticalFlowFeaturesNV", 38},
    {"VkPhysicalDeviceOpticalFlowPropertiesNV", 40},
    {"VkPhysicalDevicePCIBusInfoPropertiesEXT", 40},
    {"VkPhysicalDevicePageableDeviceLocalMemoryFeaturesEXT", 53},
    {"VkPhysicalDevicePartitionedAccelerationStructureFeaturesNV", 59},
    {"VkPhysicalDevicePartitionedAccelerationStructurePropertiesNV", 61},
    {"VkPhysicalDevicePerStageDescriptorSetFeaturesNV", 48},
    {"VkPhysicalDevicePerformanceCountersByRegionFeaturesARM", 55},
    {"VkPhysicalDevicePerformanceCountersByRegionPropertiesARM", 57},
    {"VkPhysicalDevicePerformanceQueryFeaturesKHR", 44},
    {"VkPhysicalDevicePerformanceQueryPropertiesKHR", 46},
    {"VkPhysicalDevicePipelineBinaryFeaturesKHR", 42},
    {"VkPhysicalDevicePipelineBinaryPropertiesKHR", 44},
    {"VkPhysicalDevicePipelineCacheIncrementalModeFeaturesSEC", 56},
    {"VkPhysicalDevicePipelineCreationCacheControlFeatures", 53},
    {"VkPhysicalDevicePipelineExecutablePropertiesFeaturesKHR", 56},
    {"VkPhysicalDevicePipelineLibraryGroupHandlesFeaturesEXT", 55},
    {"VkPhysicalDevicePipelineOpacityMicromapFeaturesARM", 51},
    {"VkPhysicalDevicePipelinePropertiesFeaturesEXT", 46},
    {"VkPhysicalDevicePipelineProtectedAccessFeatures", 48},
    {"VkPhysicalDevicePipelineRobustnessFeatures", 43},
    {"VkPhysicalDevicePipelineRobustnessProperties", 45},
    {"VkPhysicalDevicePointClippingProperties", 40},
    {"VkPhysicalDevicePortabilitySubsetFeaturesKHR", 45},
    {"VkPhysicalDevicePortabilitySubsetPropertiesKHR", 47},
    {"VkPhysicalDevicePresentBarrierFeaturesNV", 41},
    {"VkPhysicalDevicePresentId2FeaturesKHR", 38},
    {"VkPhysicalDevicePresentIdFeaturesKHR", 37},
    {"VkPhysicalDevicePresentMeteringFeaturesNV", 42},
    {"VkPhysicalDevicePresentModeFifoLatestReadyFeaturesKHR", 54},
    {"VkPhysicalDevicePresentTimingFeaturesEXT", 41},
    {"VkPhysicalDevicePresentWait2FeaturesKHR", 40},
    {"VkPhysicalDevicePresentWaitFeaturesKHR", 39},
    {"VkPhysicalDevicePrimitiveRestartIndexFeaturesEXT", 49},
    {"VkPhysicalDevicePrimitiveTopologyListRestartFeaturesEXT", 56},
    {"VkPhysicalDevicePrimitivesGeneratedQueryFeaturesEXT", 52},
    {"VkPhysicalDevicePrivateDataFeatures", 36},
    {"VkPhysicalDeviceProperties", 27},
    {"VkPhysicalDeviceProperties2", 28},
    {"VkPhysicalDeviceProtectedMemoryFeatures", 40},
    {"VkPhysicalDeviceProtectedMemoryProperties", 42},
    {"VkPhysicalDeviceProvokingVertexFeaturesEXT", 43},
    {"VkPhysicalDeviceProvokingVertexPropertiesEXT", 45},
    {"VkPhysicalDevicePushConstantBankFeaturesNV", 43},
    {"VkPhysicalDevicePushConstantBankPropertiesNV", 45},
    {"VkPhysicalDevicePushDescriptorProperties", 41},
    {"VkPhysicalDeviceQueueFamilyDataGraphProcessingEngineInfoARM", 60},
    {"VkPhysicalDeviceQueuePerfHintFeaturesQCOM", 42},
    {"VkPhysicalDeviceQueuePerfHintPropertiesQCOM", 44},
    {"VkPhysicalDeviceRGBA10X6FormatsFeaturesEXT", 43},
    {"VkPhysicalDeviceRasterizationOrderAttachmentAccessFeaturesEXT", 62},
    {"VkPhysicalDeviceRawAccessChainsFeaturesNV", 42},
    {"VkPhysicalDeviceRayQueryFeaturesKHR", 36},
    {"VkPhysicalDeviceRayTracingInvocationReorderFeaturesEXT", 55},
    {"VkPhysicalDeviceRayTracingInvocationReorderFeaturesNV", 54},
    {"VkPhysicalDeviceRayTracingInvocationReorderPropertiesEXT", 57},
    {"VkPhysicalDeviceRayTracingInvocationReorderPropertiesNV", 56},
    {"VkPhysicalDeviceRayTracingLinearSweptSpheresFeaturesNV", 55},
    {"VkPhysicalDeviceRayTracingMaintenance1FeaturesKHR", 50},
    {"VkPhysicalDeviceRayTracingMotionBlurFeaturesNV", 47},
    {"VkPhysicalDeviceRayTracingPipelineFeaturesKHR", 46},
    {"VkPhysicalDeviceRayTracingPipelinePropertiesKHR", 48},
    {"VkPhysicalDeviceRayTracingPositionFetchFeaturesKHR", 51},
    {"VkPhysicalDeviceRayTracingPropertiesNV", 39},
    {"VkPhysicalDeviceRayTracingValidationFeaturesNV", 47},
    {"VkPhysicalDeviceRelaxedLineRasterizationFeaturesIMG", 52},
    {"VkPhysicalDeviceRenderPassStripedFeaturesARM", 45},
    {"VkPhysicalDeviceRenderPassStripedPropertiesARM", 47},
    {"VkPhysicalDeviceRepresentativeFragmentTestFeaturesNV", 53},
    {"VkPhysicalDeviceRobustness2FeaturesKHR", 39},
    {"VkPhysicalDeviceRobustness2PropertiesKHR", 41},
    {"VkPhysicalDeviceSampleLocationsPropertiesEXT", 45},
    {"VkPhysicalDeviceSamplerFilterMinmaxProperties", 46},
    {"VkPhysicalDeviceSamplerYcbcrConversionFeatures", 47},
    {"VkPhysicalDeviceScalarBlockLayoutFeatures", 42},
    {"VkPhysicalDeviceSchedulingControlsDispatchParametersPropertiesARM", 66},
    {"VkPhysicalDeviceSchedulingControlsFeaturesARM", 46},
    {"VkPhysicalDeviceSchedulingControlsPropertiesARM", 48},
    {"VkPhysicalDeviceSeparateDepthStencilLayoutsFeatures", 52},
    {"VkPhysicalDeviceShader64BitIndexingFeaturesEXT", 47},
    {"VkPhysicalDeviceShaderAbortFeaturesKHR", 39},
    {"VkPhysicalDeviceShaderAbortPropertiesKHR", 41},
    {"VkPhysicalDeviceShaderAtomicFloat16VectorFeaturesNV", 52},
    {"VkPhysicalDeviceShaderAtomicFloat2FeaturesEXT", 46},
    {"VkPhysicalDeviceShaderAtomicFloatFeaturesEXT", 45},
    {"VkPhysicalDeviceShaderAtomicInt64Features", 42},
    {"VkPhysicalDeviceShaderBfloat16FeaturesKHR", 42},
    {"VkPhysicalDeviceShaderClockFeaturesKHR", 39},
    {"VkPhysicalDeviceShaderConstantDataFeaturesKHR", 46},
    {"VkPhysicalDeviceShaderCoreBuiltinsFeaturesARM", 46},
    {"VkPhysicalDeviceShaderCoreBuiltinsPropertiesARM", 48},
    {"VkPhysicalDeviceShaderCoreProperties2AMD", 41},
    {"VkPhysicalDeviceShaderCorePropertiesAMD", 40},
    {"VkPhysicalDeviceShaderCorePropertiesARM", 40},
    {"VkPhysicalDeviceShaderDemoteToHelperInvocationFeatures", 55},
    {"VkPhysicalDeviceShaderDrawParametersFeatures", 45},
    {"VkPhysicalDeviceShaderEarlyAndLateFragmentTestsFeaturesAMD", 59},
    {"VkPhysicalDeviceShaderEnqueueFeaturesAMDX", 42},
    {"VkPhysicalDeviceShaderEnqueuePropertiesAMDX", 44},
    {"VkPhysicalDeviceShaderExpectAssumeFeatures", 43},
    {"VkPhysicalDeviceShaderFloat16Int8Features", 42},
    {"VkPhysicalDeviceShaderFloat8FeaturesEXT", 40},
    {"VkPhysicalDeviceShaderFloatControls2Features", 45},
    {"VkPhysicalDeviceShaderFmaFeaturesKHR", 37},
    {"VkPhysicalDeviceShaderImageAtomicInt64FeaturesEXT", 50},
    {"VkPhysicalDeviceShaderImageFootprintFeaturesNV", 47},
    {"VkPhysicalDeviceShaderInstrumentationFeaturesARM", 49},
    {"VkPhysicalDeviceShaderInstrumentationPropertiesARM", 51},
    {"VkPhysicalDeviceShaderIntegerDotProductFeatures", 48},
    {"VkPhysicalDeviceShaderIntegerDotProductProperties", 50},
    {"VkPhysicalDeviceShaderIntegerFunctions2FeaturesINTEL", 53},
    {"VkPhysicalDeviceShaderLongVectorFeaturesEXT", 44},
    {"VkPhysicalDeviceShaderLongVectorPropertiesEXT", 46},
    {"VkPhysicalDeviceShaderMaximalReconvergenceFeaturesKHR", 54},
    {"VkPhysicalDeviceShaderMixedFloatDotProductFeaturesVALVE", 56},
    {"VkPhysicalDeviceShaderModuleIdentifierFeaturesEXT", 50},
    {"VkPhysicalDeviceShaderModuleIdentifierPropertiesEXT", 52},
    {"VkPhysicalDeviceShaderMultipleWaitQueuesFeaturesQCOM", 53},
    {"VkPhysicalDeviceShaderMultipleWaitQueuesPropertiesQCOM", 55},
    {"VkPhysicalDeviceShaderOCPMicroscalingTypesFeaturesEXT", 54},
    {"VkPhysicalDeviceShaderObjectFeaturesEXT", 40},
    {"VkPhysicalDeviceShaderObjectPropertiesEXT", 42},
    {"VkPhysicalDeviceShaderQuadControlFeaturesKHR", 45},
    {"VkPhysicalDeviceShaderRelaxedExtendedInstructionFeaturesKHR", 60},
    {"VkPhysicalDeviceShaderReplicatedCompositesFeaturesEXT", 54},
    {"VkPhysicalDeviceShaderSMBuiltinsFeaturesNV", 43},
    {"VkPhysicalDeviceShaderSMBuiltinsPropertiesNV", 45},
    {"VkPhysicalDeviceShaderSplitBarrierFeaturesEXT", 46},
    {"VkPhysicalDeviceShaderSplitBarrierPropertiesEXT", 48},
    {"VkPhysicalDeviceShaderSubgroupExtendedTypesFeatures", 52},
    {"VkPhysicalDeviceShaderSubgroupPartitionedFeaturesEXT", 53},
    {"VkPhysicalDeviceShaderSubgroupRotateFeatures", 45},
    {"VkPhysicalDeviceShaderSubgroupUniformControlFlowFeaturesKHR", 60},
    {"VkPhysicalDeviceShaderTerminateInvocationFeatures", 50},
    {"VkPhysicalDeviceShaderTileImageFeaturesEXT", 43},
    {"VkPhysicalDeviceShaderTileImagePropertiesEXT", 45},
    {"VkPhysicalDeviceShaderUniformBufferUnsizedArrayFeaturesEXT", 59},
    {"VkPhysicalDeviceShaderUntypedPointersFeaturesKHR", 49},
    {"VkPhysicalDeviceShadingRateImageFeaturesNV", 43},
    {"VkPhysicalDeviceShadingRateImagePropertiesNV", 45},
    {"VkPhysicalDeviceSparseImageFormatInfo2", 39},
    {"VkPhysicalDeviceSparseProperties", 33},
    {"VkPhysicalDeviceSubgroupProperties", 35},
    {"VkPhysicalDeviceSubgroupSizeControlFeatures", 44},
    {"VkPhysicalDeviceSubgroupSizeControlProperties", 46},
    {"VkPhysicalDeviceSubpassMergeFeedbackFeaturesEXT", 48},
    {"VkPhysicalDeviceSubpassShadingFeaturesHUAWEI", 45},
    {"VkPhysicalDeviceSubpassShadingPropertiesHUAWEI", 47},
    {"VkPhysicalDeviceSurfaceInfo2KHR", 32},
    {"VkPhysicalDeviceSwapchainMaintenance1FeaturesKHR", 49},
    {"VkPhysicalDeviceSynchronization2Features", 41},
    {"VkPhysicalDeviceTensorFeaturesARM", 34},
    {"VkPhysicalDeviceTensorPropertiesARM", 36},
    {"VkPhysicalDeviceTexelBufferAlignmentFeaturesEXT", 48},
    {"VkPhysicalDeviceTexelBufferAlignmentProperties", 47},
    {"VkPhysicalDeviceTextureCompressionASTC3DFeaturesEXT", 52},
    {"VkPhysicalDeviceTextureCompressionASTCHDRFeatures", 50},
    {"VkPhysicalDeviceThrottleHintFeaturesSEC", 40},
    {"VkPhysicalDeviceTileMemoryHeapFeaturesQCOM", 43},
    {"VkPhysicalDeviceTileMemoryHeapPropertiesQCOM", 45},
    {"VkPhysicalDeviceTilePropertiesFeaturesQCOM", 43},
    {"VkPhysicalDeviceTileShadingFeaturesQCOM", 40},
    {"VkPhysicalDeviceTileShadingPropertiesQCOM", 42},
    {"VkPhysicalDeviceTimelineSemaphoreFeatures", 42},
    {"VkPhysicalDeviceTimelineSemaphoreProperties", 44},
    {"VkPhysicalDeviceToolProperties", 31},
    {"VkPhysicalDeviceTransformFeedbackFeaturesEXT", 45},
    {"VkPhysicalDeviceTransformFeedbackPropertiesEXT", 47},
    {"VkPhysicalDeviceUnifiedImageLayoutsFeaturesKHR", 47},
    {"VkPhysicalDeviceUniformBufferStandardLayoutFeatures", 52},
    {"VkPhysicalDeviceVariablePointersFeatures", 41},
    {"VkPhysicalDeviceVertexAttributeDivisorFeatures", 47},
    {"VkPhysicalDeviceVertexAttributeDivisorProperties", 49},
    {"VkPhysicalDeviceVertexAttributeDivisorPropertiesEXT", 52},
    {"VkPhysicalDeviceVertexAttributeRobustnessFeaturesEXT", 53},
    {"VkPhysicalDeviceVertexInputDynamicStateFeaturesEXT", 51},
    {"VkPhysicalDeviceVideoDecodeVP9FeaturesKHR", 42},
    {"VkPhysicalDeviceVideoEncodeAV1FeaturesKHR", 42},
    {"VkPhysicalDeviceVideoEncodeFeedback2FeaturesKHR", 48},
    {"VkPhysicalDeviceVideoEncodeIntraRefreshFeaturesKHR", 51},
    {"VkPhysicalDeviceVideoEncodeQualityLevelInfoKHR", 47},
    {"VkPhysicalDeviceVideoEncodeQuantizationMapFeaturesKHR", 54},
    {"VkPhysicalDeviceVideoEncodeRgbConversionFeaturesVALVE", 54},
    {"VkPhysicalDeviceVideoFormatInfoKHR", 35},
    {"VkPhysicalDeviceVideoMaintenance1FeaturesKHR", 45},
    {"VkPhysicalDeviceVideoMaintenance2FeaturesKHR", 45},
    {"VkPhysicalDeviceVulkan11Features", 33},
    {"VkPhysicalDeviceVulkan11Properties", 35},
    {"VkPhysicalDeviceVulkan12Features", 33},
    {"VkPhysicalDeviceVulkan12Properties", 35},
    {"VkPhysicalDeviceVulkan13Features", 33},
    {"VkPhysicalDeviceVulkan13Properties", 35},
    {"VkPhysicalDeviceVulkan14Features", 33},
    {"VkPhysicalDeviceVulkan14Properties", 35},
    {"VkPhysicalDeviceVulkanMemoryModelFeatures", 42},
    {"VkPhysicalDeviceWorkgroupMemoryExplicitLayoutFeaturesKHR", 57},
    {"VkPhysicalDeviceYcbcr2Plane444FormatsFeaturesEXT", 49},
    {"VkPhysicalDeviceYcbcrDegammaFeaturesQCOM", 41},
    {"VkPhysicalDeviceYcbcrImageArraysFeaturesEXT", 44},
    {"VkPhysicalDeviceZeroInitializeDeviceMemoryFeaturesEXT", 54},
    {"VkPhysicalDeviceZeroInitializeWorkgroupMemoryFeatures", 54},
    {"VkPipelineBinaryCreateInfoKHR", 30},
    {"VkPipelineBinaryDataInfoKHR", 28},
    {"VkPipelineBinaryDataKHR", 24},
    {"VkPipelineBinaryHandlesInfoKHR", 31},
    {"VkPipelineBinaryInfoKHR", 24},
    {"VkPipelineBinaryKeyKHR", 23},
    {"VkPipelineBinaryKeysAndDataKHR", 31},
    {"VkPipelineCacheCreateInfo", 26},
    {"VkPipelineCacheHeaderVersionDataGraphQCOM", 42},
    {"VkPipelineCacheHeaderVersionOne", 32},
    {"VkPipelineColorBlendAdvancedStateCreateInfoEXT", 47},
    {"VkPipelineColorBlendAttachmentState", 36},
    {"VkPipelineColorBlendStateCreateInfo", 36},
    {"VkPipelineColorWriteCreateInfoEXT", 34},
    {"VkPipelineCompilerControlCreateInfoAMD", 39},
    {"VkPipelineCoverageModulationStateCreateInfoNV", 46},
    {"VkPipelineCoverageReductionStateCreateInfoNV", 45},
    {"VkPipelineCoverageToColorStateCreateInfoNV", 43},
    {"VkPipelineCreateFlags2CreateInfo", 33},
    {"VkPipelineCreateInfoKHR", 24},
    {"VkPipelineCreationFeedback", 27},
    {"VkPipelineCreationFeedbackCreateInfo", 37},
    {"VkPipelineDepthStencilStateCreateInfo", 38},
    {"VkPipelineDiscardRectangleStateCreateInfoEXT", 45},
    {"VkPipelineDynamicStateCreateInfo", 33},
    {"VkPipelineExecutableInfoKHR", 28},
    {"VkPipelineExecutableInternalRepresentationKHR", 46},
    {"VkPipelineExecutablePropertiesKHR", 34},
    {"VkPipelineExecutableStatisticKHR", 33},
    {"VkPipelineExecutableStatisticValueKHR", 38},
    {"VkPipelineFragmentDensityMapLayeredCreateInfoVALVE", 51},
    {"VkPipelineFragmentShadingRateEnumStateCreateInfoNV", 51},
    {"VkPipelineFragmentShadingRateStateCreateInfoKHR", 48},
    {"VkPipelineIndirectDeviceAddressInfoNV", 38},
    {"VkPipelineInfoKHR", 18},
    {"VkPipelineInputAssemblyStateCreateInfo", 39},
    {"VkPipelineLayoutCreateInfo", 27},
    {"VkPipelineLibraryCreateInfoKHR", 31},
    {"VkPipelineMultisampleStateCreateInfo", 37},
    {"VkPipelinePropertiesIdentifierEXT", 34},
    {"VkPipelineRasterizationConservativeStateCreateInfoEXT", 54},
    {"VkPipelineRasterizationDepthClipStateCreateInfoEXT", 51},
    {"VkPipelineRasterizationLineStateCreateInfo", 43},
    {"VkPipelineRasterizationProvokingVertexStateCreateInfoEXT", 57},
    {"VkPipelineRasterizationStateCreateInfo", 39},
    {"VkPipelineRasterizationStateRasterizationOrderAMD", 50},
    {"VkPipelineRasterizationStateStreamCreateInfoEXT", 48},
    {"VkPipelineRenderingCreateInfo", 30},
    {"VkPipelineRepresentativeFragmentTestStateCreateInfoNV", 54},
    {"VkPipelineRobustnessCreateInfo", 31},
    {"VkPipelineSampleLocationsStateCreateInfoEXT", 44},
    {"VkPipelineShaderStageCreateInfo", 32},
    {"VkPipelineShaderStageModuleIdentifierCreateInfoEXT", 51},
    {"VkPipelineShaderStageNodeCreateInfoAMDX", 40},
    {"VkPipelineShaderStageRequiredSubgroupSizeCreateInfo", 52},
    {"VkPipelineTessellationDomainOriginStateCreateInfo", 50},
    {"VkPipelineTessellationStateCreateInfo", 38},
    {"VkPipelineVertexInputDivisorStateCreateInfo", 44},
    {"VkPipelineVertexInputStateCreateInfo", 37},
    {"VkPipelineViewportCoarseSampleOrderStateCreateInfoNV", 53},
    {"VkPipelineViewportDepthClampControlCreateInfoEXT", 49},
    {"VkPipelineViewportDepthClipControlCreateInfoEXT", 48},
    {"VkPipelineViewportExclusiveScissorStateCreateInfoNV", 52},
    {"VkPipelineViewportShadingRateImageStateCreateInfoNV", 52},
    {"VkPipelineViewportStateCreateInfo", 34},
    {"VkPipelineViewportSwizzleStateCreateInfoNV", 43},
    {"VkPipelineViewportWScalingStateCreateInfoNV", 44},
    {"VkPresentFrameTokenGGP", 23},
    {"VkPresentId2KHR", 16},
    {"VkPresentIdKHR", 15},
    {"VkPresentInfoKHR", 17},
    {"VkPresentRegionKHR", 19},
    {"VkPresentRegionsKHR", 20},
    {"VkPresentStageTimeEXT", 22},
    {"VkPresentTimeGOOGLE", 20},
    {"VkPresentTimesInfoGOOGLE", 25},
    {"VkPresentTimingInfoEXT", 23},
    {"VkPresentTimingSurfaceCapabilitiesEXT", 38},
    {"VkPresentTimingsInfoEXT", 24},
    {"VkPresentWait2InfoKHR", 22},
    {"VkPrivateDataSlotCreateInfo", 28},
    {"VkProtectedSubmitInfo", 22},
    {"VkPushConstantBankInfoNV", 25},
    {"VkPushConstantRange", 20},
    {"VkPushConstantsInfo", 20},
    {"VkPushDataInfoEXT", 18},
    {"VkPushDescriptorSetInfo", 24},
    {"VkPushDescriptorSetWithTemplateInfo", 36},
    {"VkQueryLowLatencySupportNV", 27},
    {"VkQueryPoolCreateInfo", 22},
    {"VkQueryPoolPerformanceCreateInfoKHR", 36},
    {"VkQueryPoolPerformanceQueryCreateInfoINTEL", 43},
    {"VkQueryPoolVideoEncodeFeedbackCreateInfoKHR", 44},
    {"VkQueryPoolVideoEncodePerPartitionFeedbackCreateInfoKHR", 56},
    {"VkQueueFamilyCheckpointProperties2NV", 37},
    {"VkQueueFamilyCheckpointPropertiesNV", 36},
    {"VkQueueFamilyDataGraphOpticalFlowPropertiesARM", 47},
    {"VkQueueFamilyDataGraphProcessingEnginePropertiesARM", 52},
    {"VkQueueFamilyDataGraphPropertiesARM", 36},
    {"VkQueueFamilyDataGraphTOSAPropertiesARM", 40},
    {"VkQueueFamilyGlobalPriorityProperties", 38},
    {"VkQueueFamilyOptimalImageTransferGranularityPropertiesKHR", 58},
    {"VkQueueFamilyOwnershipTransferPropertiesKHR", 44},
    {"VkQueueFamilyProperties", 24},
    {"VkQueueFamilyProperties2", 25},
    {"VkQueueFamilyQueryResultStatusPropertiesKHR", 44},
    {"VkQueueFamilyVideoPropertiesKHR", 32},
    {"VkRayTracingPipelineClusterAccelerationStructureCreateInfoNV", 61},
    {"VkRayTracingPipelineCreateInfoKHR", 34},
    {"VkRayTracingPipelineCreateInfoNV", 33},
    {"VkRayTracingPipelineInterfaceCreateInfoKHR", 43},
    {"VkRayTracingShaderGroupCreateInfoKHR", 37},
    {"VkRayTracingShaderGroupCreateInfoNV", 36},
    {"VkRect2D", 9},
    {"VkRectLayerKHR", 15},
    {"VkRefreshCycleDurationGOOGLE", 29},
    {"VkReleaseCapturedPipelineDataInfoKHR", 37},
    {"VkReleaseSwapchainImagesInfoKHR", 32},
    {"VkRenderPassAttachmentBeginInfo", 32},
    {"VkRenderPassBeginInfo", 22},
    {"VkRenderPassCreateInfo", 23},
    {"VkRenderPassCreateInfo2", 24},
    {"VkRenderPassCreationControlEXT", 31},
    {"VkRenderPassCreationFeedbackCreateInfoEXT", 42},
    {"VkRenderPassCreationFeedbackInfoEXT", 36},
    {"VkRenderPassFragmentDensityMapCreateInfoEXT", 44},
    {"VkRenderPassFragmentDensityMapOffsetEndInfoEXT", 47},
    {"VkRenderPassInputAttachmentAspectCreateInfo", 44},
    {"VkRenderPassMultiviewCreateInfo", 32},
    {"VkRenderPassPerformanceCountersByRegionBeginInfoARM", 52},
    {"VkRenderPassSampleLocationsBeginInfoEXT", 40},
    {"VkRenderPassStripeBeginInfoARM", 31},
    {"VkRenderPassStripeInfoARM", 26},
    {"VkRenderPassStripeSubmitInfoARM", 32},
    {"VkRenderPassSubpassFeedbackCreateInfoEXT", 41},
    {"VkRenderPassSubpassFeedbackInfoEXT", 35},
    {"VkRenderPassTileShadingCreateInfoQCOM", 38},
    {"VkRenderPassTransformBeginInfoQCOM", 35},
    {"VkRenderingAreaInfo", 20},
    {"VkRenderingAttachmentFlagsInfoKHR", 34},
    {"VkRenderingAttachmentInfo", 26},
    {"VkRenderingAttachmentLocationInfo", 34},
    {"VkRenderingEndInfoKHR", 22},
    {"VkRenderingFragmentDensityMapAttachmentInfoEXT", 47},
    {"VkRenderingFragmentShadingRateAttachmentInfoKHR", 48},
    {"VkRenderingInfo", 16},
    {"VkRenderingInputAttachmentIndexInfo", 36},
    {"VkResolveImageInfo2", 20},
    {"VkResolveImageModeInfoKHR", 26},
    {"VkResourceDescriptorDataEXT", 28},
    {"VkResourceDescriptorInfoEXT", 28},
    {"VkSRTDataNV", 12},
    {"VkSampleLocationEXT", 20},
    {"VkSampleLocationsInfoEXT", 25},
    {"VkSamplerBlockMatchWindowCreateInfoQCOM", 40},
    {"VkSamplerBorderColorComponentMappingCreateInfoEXT", 50},
    {"VkSamplerCaptureDescriptorDataInfoEXT", 38},
    {"VkSamplerCreateInfo", 20},
    {"VkSamplerCubicWeightsCreateInfoQCOM", 36},
    {"VkSamplerCustomBorderColorCreateInfoEXT", 40},
    {"VkSamplerCustomBorderColorIndexCreateInfoEXT", 45},
    {"VkSamplerReductionModeCreateInfo", 33},
    {"VkSamplerYcbcrConversionCreateInfo", 35},
    {"VkSamplerYcbcrConversionImageFormatProperties", 46},
    {"VkSamplerYcbcrConversionInfo", 29},
    {"VkSamplerYcbcrConversionYcbcrDegammaCreateInfoQCOM", 51},
    {"VkScreenBufferFormatPropertiesQNX", 34},
    {"VkScreenBufferPropertiesQNX", 28},
    {"VkScreenSurfaceCreateInfoQNX", 29},
    {"VkSemaphoreCreateInfo", 22},
    {"VkSemaphoreGetFdInfoKHR", 24},
    {"VkSemaphoreGetWin32HandleInfoKHR", 33},
    {"VkSemaphoreGetZirconHandleInfoFUCHSIA", 38},
    {"VkSemaphoreSignalInfo", 22},
    {"VkSemaphoreSubmitInfo", 22},
    {"VkSemaphoreTypeCreateInfo", 26},
    {"VkSemaphoreWaitInfo", 20},
    {"VkSetDescriptorBufferOffsetsInfoEXT", 36},
    {"VkSetLatencyMarkerInfoNV", 25},
    {"VkSetPresentConfigNV", 21},
    {"VkSetStateFlagsIndirectCommandNV", 33},
    {"VkShaderCreateInfoEXT", 22},
    {"VkShaderDescriptorSetAndBindingMappingInfoEXT", 46},
    {"VkShaderInstrumentationCreateInfoARM", 37},
    {"VkShaderInstrumentationMetricDataHeaderARM", 43},
    {"VkShaderInstrumentationMetricDescriptionARM", 44},
    {"VkShaderModuleCreateInfo", 25},
    {"VkShaderModuleIdentifierEXT", 28},
    {"VkShaderModuleValidationCacheCreateInfoEXT", 43},
    {"VkShaderResourceUsageAMD", 25},
    {"VkShaderStatisticsInfoAMD", 26},
    {"VkShadingRatePaletteNV", 23},
    {"VkSharedPresentSurfaceCapabilities2KHR", 39},
    {"VkSharedPresentSurfaceCapabilitiesKHR", 38},
    {"VkSparseBufferMemoryBindInfo", 29},
    {"VkSparseImageFormatProperties", 30},
    {"VkSparseImageFormatProperties2", 31},
    {"VkSparseImageMemoryBind", 24},
    {"VkSparseImageMemoryBindInfo", 28},
    {"VkSparseImageMemoryRequirements", 32},
    {"VkSparseImageMemoryRequirements2", 33},
    {"VkSparseImageOpaqueMemoryBindInfo", 34},
    {"VkSparseMemoryBind", 19},
    {"VkSpecializationInfo", 21},
    {"VkSpecializationMapEntry", 25},
    {"VkStencilOpState", 17},
    {"VkStreamDescriptorSurfaceCreateInfoGGP", 39},
    {"VkStridedDeviceAddressNV", 25},
    {"VkStridedDeviceAddressRangeKHR", 31},
    {"VkStridedDeviceAddressRegionKHR", 32},
    {"VkSubmitInfo", 13},
    {"VkSubmitInfo2", 14},
    {"VkSubpassBeginInfo", 19},
    {"VkSubpassDependency", 20},
    {"VkSubpassDependency2", 21},
    {"VkSubpassDescription", 21},
    {"VkSubpassDescription2", 22},
    {"VkSubpassDescriptionDepthStencilResolve", 40},
    {"VkSubpassEndInfo", 17},
    {"VkSubpassResolvePerformanceQueryEXT", 36},
    {"VkSubpassSampleLocationsEXT", 28},
    {"VkSubpassShadingPipelineCreateInfoHUAWEI", 41},
    {"VkSubresourceHostMemcpySize", 28},
    {"VkSubresourceLayout", 20},
    {"VkSubresourceLayout2", 21},
    {"VkSubsampledImageFormatPropertiesEXT", 37},
    {"VkSurfaceCapabilities2EXT", 26},
    {"VkSurfaceCapabilities2KHR", 26},
    {"VkSurfaceCapabilitiesFullScreenExclusiveEXT", 44},
    {"VkSurfaceCapabilitiesKHR", 25},
    {"VkSurfaceCapabilitiesPresentBarrierNV", 38},
    {"VkSurfaceCapabilitiesPresentId2KHR", 35},
    {"VkSurfaceCapabilitiesPresentWait2KHR", 37},
    {"VkSurfaceCreateInfoOHOS", 24},
    {"VkSurfaceFormat2KHR", 20},
    {"VkSurfaceFormatKHR", 19},
    {"VkSurfaceFullScreenExclusiveInfoEXT", 36},
    {"VkSurfaceFullScreenExclusiveWin32InfoEXT", 41},
    {"VkSurfacePresentModeCompatibilityKHR", 37},
    {"VkSurfacePresentModeKHR", 24},
    {"VkSurfacePresentScalingCapabilitiesKHR", 39},
    {"VkSurfaceProtectedCapabilitiesKHR", 34},
    {"VkSwapchainCalibratedTimestampInfoEXT", 38},
    {"VkSwapchainCounterCreateInfoEXT", 32},
    {"VkSwapchainCreateInfoKHR", 25},
    {"VkSwapchainDisplayNativeHdrCreateInfoAMD", 41},
    {"VkSwapchainFlagsSurfaceCapabilitiesEXT", 39},
    {"VkSwapchainLatencyCreateInfoNV", 31},
    {"VkSwapchainPresentBarrierCreateInfoNV", 38},
    {"VkSwapchainPresentFenceInfoKHR", 31},
    {"VkSwapchainPresentModeInfoKHR", 30},
    {"VkSwapchainPresentModesCreateInfoKHR", 37},
    {"VkSwapchainPresentScalingCreateInfoKHR", 39},
    {"VkSwapchainTimeDomainPropertiesEXT", 35},
    {"VkSwapchainTimingPropertiesEXT", 31},
    {"VkSysmemColorSpaceFUCHSIA", 26},
    {"VkTensorCaptureDescriptorDataInfoARM", 37},
    {"VkTensorCopyARM", 16},
    {"VkTensorCreateInfoARM", 22},
    {"VkTensorDependencyInfoARM", 26},
    {"VkTensorDescriptionARM", 23},
    {"VkTensorExplicitTilingFormatPropertiesARM", 42},
    {"VkTensorFormatPropertiesARM", 28},
    {"VkTensorMemoryBarrierARM", 25},
    {"VkTensorMemoryRequirementsInfoARM", 34},
    {"VkTensorRollingBackingCreateInfoARM", 36},
    {"VkTensorViewCaptureDescriptorDataInfoARM", 41},
    {"VkTensorViewCreateInfoARM", 26},
    {"VkTexelBufferDescriptorInfoEXT", 31},
    {"VkTextureLODGatherFormatPropertiesAMD", 38},
    {"VkThrottleHintSubmitInfoSEC", 28},
    {"VkTileMemoryBindInfoQCOM", 25},
    {"VkTileMemoryRequirementsQCOM", 29},
    {"VkTileMemorySizeInfoQCOM", 25},
    {"VkTilePropertiesQCOM", 21},
    {"VkTimelineSemaphoreSubmitInfo", 30},
    {"VkTraceRaysIndirectCommand2KHR", 31},
    {"VkTraceRaysIndirectCommandKHR", 30},
    {"VkTransformMatrixKHR", 21},
    {"VkUbmSurfaceCreateInfoSEC", 26},
    {"VkValidationCacheCreateInfoEXT", 31},
    {"VkValidationFeaturesEXT", 24},
    {"VkValidationFlagsEXT", 21},
    {"VkVertexInputAttributeDescription", 34},
    {"VkVertexInputAttributeDescription2EXT", 38},
    {"VkVertexInputBindingDescription", 32},
    {"VkVertexInputBindingDescription2EXT", 36},
    {"VkVertexInputBindingDivisorDescription", 39},
    {"VkViSurfaceCreateInfoNN", 24},
    {"VkVideoBeginCodingInfoKHR", 26},
    {"VkVideoCapabilitiesKHR", 23},
    {"VkVideoCodingControlInfoKHR", 28},
    {"VkVideoDecodeAV1CapabilitiesKHR", 32},
    {"VkVideoDecodeAV1DpbSlotInfoKHR", 31},
    {"VkVideoDecodeAV1InlineSessionParametersInfoKHR", 47},
    {"VkVideoDecodeAV1PictureInfoKHR", 31},
    {"VkVideoDecodeAV1ProfileInfoKHR", 31},
    {"VkVideoDecodeAV1SessionParametersCreateInfoKHR", 47},
    {"VkVideoDecodeCapabilitiesKHR", 29},
    {"VkVideoDecodeH264CapabilitiesKHR", 33},
    {"VkVideoDecodeH264DpbSlotInfoKHR", 32},
    {"VkVideoDecodeH264InlineSessionParametersInfoKHR", 48},
    {"VkVideoDecodeH264PictureInfoKHR", 32},
    {"VkVideoDecodeH264ProfileInfoKHR", 32},
    {"VkVideoDecodeH264SessionParametersAddInfoKHR", 45},
    {"VkVideoDecodeH264SessionParametersCreateInfoKHR", 48},
    {"VkVideoDecodeH265CapabilitiesKHR", 33},
    {"VkVideoDecodeH265DpbSlotInfoKHR", 32},
    {"VkVideoDecodeH265InlineSessionParametersInfoKHR", 48},
    {"VkVideoDecodeH265PictureInfoKHR", 32},
    {"VkVideoDecodeH265ProfileInfoKHR", 32},
    {"VkVideoDecodeH265SessionParametersAddInfoKHR", 45},
    {"VkVideoDecodeH265SessionParametersCreateInfoKHR", 48},
    {"VkVideoDecodeInfoKHR", 21},
    {"VkVideoDecodeUsageInfoKHR", 26},
    {"VkVideoDecodeVP9CapabilitiesKHR", 32},
    {"VkVideoDecodeVP9PictureInfoKHR", 31},
    {"VkVideoDecodeVP9ProfileInfoKHR", 31},
    {"VkVideoEncodeAV1CapabilitiesKHR", 32},
    {"VkVideoEncodeAV1DpbSlotInfoKHR", 31},
    {"VkVideoEncodeAV1FrameSizeKHR", 29},
    {"VkVideoEncodeAV1GopRemainingFrameInfoKHR", 41},
    {"VkVideoEncodeAV1PictureInfoKHR", 31},
    {"VkVideoEncodeAV1ProfileInfoKHR", 31},
    {"VkVideoEncodeAV1QIndexKHR", 26},
    {"VkVideoEncodeAV1QualityLevelPropertiesKHR", 42},
    {"VkVideoEncodeAV1QuantizationMapCapabilitiesKHR", 47},
    {"VkVideoEncodeAV1RateControlInfoKHR", 35},
    {"VkVideoEncodeAV1RateControlLayerInfoKHR", 40},
    {"VkVideoEncodeAV1SessionCreateInfoKHR", 37},
    {"VkVideoEncodeAV1SessionParametersCreateInfoKHR", 47},
    {"VkVideoEncodeCapabilitiesKHR", 29},
    {"VkVideoEncodeFeedback2CapabilitiesKHR", 38},
    {"VkVideoEncodeH264CapabilitiesKHR", 33},
    {"VkVideoEncodeH264DpbSlotInfoKHR", 32},
    {"VkVideoEncodeH264FrameSizeKHR", 30},
    {"VkVideoEncodeH264GopRemainingFrameInfoKHR", 42},
    {"VkVideoEncodeH264NaluSliceInfoKHR", 34},
    {"VkVideoEncodeH264PictureInfoKHR", 32},
    {"VkVideoEncodeH264ProfileInfoKHR", 32},
    {"VkVideoEncodeH264QpKHR", 23},
    {"VkVideoEncodeH264QualityLevelPropertiesKHR", 43},
    {"VkVideoEncodeH264QuantizationMapCapabilitiesKHR", 48},
    {"VkVideoEncodeH264RateControlInfoKHR", 36},
    {"VkVideoEncodeH264RateControlLayerInfoKHR", 41},
    {"VkVideoEncodeH264SessionCreateInfoKHR", 38},
    {"VkVideoEncodeH264SessionParametersAddInfoKHR", 45},
    {"VkVideoEncodeH264SessionParametersCreateInfoKHR", 48},
    {"VkVideoEncodeH264SessionParametersFeedbackInfoKHR", 50},
    {"VkVideoEncodeH264SessionParametersGetInfoKHR", 45},
    {"VkVideoEncodeH265CapabilitiesKHR", 33},
    {"VkVideoEncodeH265DpbSlotInfoKHR", 32},
    {"VkVideoEncodeH265FrameSizeKHR", 30},
    {"VkVideoEncodeH265GopRemainingFrameInfoKHR", 42},
    {"VkVideoEncodeH265NaluSliceSegmentInfoKHR", 41},
    {"VkVideoEncodeH265PictureInfoKHR", 32},
    {"VkVideoEncodeH265ProfileInfoKHR", 32},
    {"VkVideoEncodeH265QpKHR", 23},
    {"VkVideoEncodeH265QualityLevelPropertiesKHR", 43},
    {"VkVideoEncodeH265QuantizationMapCapabilitiesKHR", 48},
    {"VkVideoEncodeH265RateControlInfoKHR", 36},
    {"VkVideoEncodeH265RateControlLayerInfoKHR", 41},
    {"VkVideoEncodeH265SessionCreateInfoKHR", 38},
    {"VkVideoEncodeH265SessionParametersAddInfoKHR", 45},
    {"VkVideoEncodeH265SessionParametersCreateInfoKHR", 48},
    {"VkVideoEncodeH265SessionParametersFeedbackInfoKHR", 50},
    {"VkVideoEncodeH265SessionParametersGetInfoKHR", 45},
    {"VkVideoEncodeInfoKHR", 21},
    {"VkVideoEncodeIntraRefreshCapabilitiesKHR", 41},
    {"VkVideoEncodeIntraRefreshInfoKHR", 33},
    {"VkVideoEncodeProfileRgbConversionInfoVALVE", 43},
    {"VkVideoEncodeQualityLevelInfoKHR", 33},
    {"VkVideoEncodeQualityLevelPropertiesKHR", 39},
    {"VkVideoEncodeQuantizationMapCapabilitiesKHR", 44},
    {"VkVideoEncodeQuantizationMapInfoKHR", 36},
    {"VkVideoEncodeQuantizationMapSessionParametersCreateInfoKHR", 59},
    {"VkVideoEncodeRateControlInfoKHR", 32},
    {"VkVideoEncodeRateControlLayerInfoKHR", 37},
    {"VkVideoEncodeRgbConversionCapabilitiesVALVE", 44},
    {"VkVideoEncodeSessionIntraRefreshCreateInfoKHR", 46},
    {"VkVideoEncodeSessionParametersFeedbackInfoKHR", 46},
    {"VkVideoEncodeSessionParametersGetInfoKHR", 41},
    {"VkVideoEncodeSessionRgbConversionCreateInfoVALVE", 49},
    {"VkVideoEncodeUsageInfoKHR", 26},
    {"VkVideoEndCodingInfoKHR", 24},
    {"VkVideoFormatAV1QuantizationMapPropertiesKHR", 45},
    {"VkVideoFormatH265QuantizationMapPropertiesKHR", 46},
    {"VkVideoFormatPropertiesKHR", 27},
    {"VkVideoFormatQuantizationMapPropertiesKHR", 42},
    {"VkVideoInlineQueryInfoKHR", 26},
    {"VkVideoPictureResourceInfoKHR", 30},
    {"VkVideoProfileInfoKHR", 22},
    {"VkVideoProfileListInfoKHR", 26},
    {"VkVideoReferenceIntraRefreshInfoKHR", 36},
    {"VkVideoReferenceSlotInfoKHR", 28},
    {"VkVideoSessionCreateInfoKHR", 28},
    {"VkVideoSessionMemoryRequirementsKHR", 36},
    {"VkVideoSessionParametersCreateInfoKHR", 38},
    {"VkVideoSessionParametersUpdateInfoKHR", 38},
    {"VkViewport", 11},
    {"VkViewportSwizzleNV", 20},
    {"VkViewportWScalingNV", 21},
    {"VkWaylandSurfaceCreateInfoKHR", 30},
    {"VkWin32KeyedMutexAcquireReleaseInfoKHR", 39},
    {"VkWin32KeyedMutexAcquireReleaseInfoNV", 38},
    {"VkWin32SurfaceCreateInfoKHR", 28},
    {"VkWriteDescriptorSet", 21},
    {"VkWriteDescriptorSetAccelerationStructureKHR", 45},
    {"VkWriteDescriptorSetAccelerationStructureNV", 44},
    {"VkWriteDescriptorSetInlineUniformBlock", 39},
    {"VkWriteDescriptorSetPartitionedAccelerationStructureNV", 55},
    {"VkWriteDescriptorSetTensorARM", 30},
    {"VkWriteIndirectExecutionSetPipelineEXT", 39},
    {"VkWriteIndirectExecutionSetShaderEXT", 37},
    {"VkXYColorEXT", 13},
    {"VkXcbSurfaceCreateInfoKHR", 26},
    {"VkXlibSurfaceCreateInfoKHR", 27},
    };
    return table[(int)structure].data();
}

const char* String(Field field) {
    static const std::string_view table[] = {
    {"", 1}, // Field::Empty
    {"AType", 6},
    {"BType", 6},
    {"CType", 6},
    {"DType", 6},
    {"KGranularity", 13},
    {"KSize", 6},
    {"MGranularity", 13},
    {"MSize", 6},
    {"NGranularity", 13},
    {"NSize", 6},
    {"OHOSNativeBufferUsage", 22},
    {"ResultType", 11},
    {"a", 2},
    {"aabbData", 9},
    {"aabbs", 6},
    {"accelerationStructure", 22},
    {"accelerationStructureCaptureReplay", 35},
    {"accelerationStructureCaptureReplayDescriptorDataSize", 53},
    {"accelerationStructureCount", 27},
    {"accelerationStructureDescriptorSize", 36},
    {"accelerationStructureHostCommands", 34},
    {"accelerationStructureIndirectBuild", 35},
    {"accelerationStructureNV", 24},
    {"accelerationStructureReference", 31},
    {"accelerationStructureSize", 26},
    {"acquireCount", 13},
    {"acquireUnmodifiedMemory", 24},
    {"activeComputeUnitCount", 23},
    {"actualPresentTime", 18},
    {"address", 8},
    {"addressFlags", 13},
    {"addressImageHeight", 19},
    {"addressInfoCount", 17},
    {"addressModeU", 13},
    {"addressModeV", 13},
    {"addressModeW", 13},
    {"addressOffset", 14},
    {"addressPrecision", 17},
    {"addressRange", 13},
    {"addressResolutionFlags", 23},
    {"addressRowLength", 17},
    {"addressType", 12},
    {"advancedBlendAllOperations", 27},
    {"advancedBlendCoherentOperations", 32},
    {"advancedBlendCorrelatedOverlap", 31},
    {"advancedBlendIndependentBlend", 30},
    {"advancedBlendMaxColorAttachments", 33},
    {"advancedBlendNonPremultipliedDstColor", 38},
    {"advancedBlendNonPremultipliedSrcColor", 38},
    {"advancedBlendOp", 16},
    {"alignment", 10},
    {"allocationSize", 15},
    {"allowClusterAccelerationStructure", 34},
    {"allowCommandBufferQueryCopies", 30},
    {"allowNeuralStatistics", 22},
    {"allowSamplerImageViewPostSubmitCreation", 40},
    {"alphaBlendOp", 13},
    {"alphaMode", 10},
    {"alphaToCoverageEnable", 22},
    {"alphaToOne", 11},
    {"alphaToOneEnable", 17},
    {"amigoProfiling", 15},
    {"androidHardwareBufferUsage", 27},
    {"anisotropyEnable", 17},
    {"antiLag", 8},
    {"anyHitShader", 13},
    {"apiVersion", 11},
    {"applicationNameOffset", 22},
    {"applicationVersion", 19},
    {"apronSize", 10},
    {"argCount", 9},
    {"argData", 8},
    {"arrayElement", 13},
    {"arrayLayer", 11},
    {"arrayLayers", 12},
    {"arrayOfPointers", 16},
    {"arrayPitch", 11},
    {"aspectMask", 11},
    {"aspectReferenceCount", 21},
    {"attachment", 11},
    {"attachmentCount", 16},
    {"attachmentFeedbackLoopDynamicState", 35},
    {"attachmentFeedbackLoopLayout", 29},
    {"attachmentFragmentShadingRate", 30},
    {"attachmentImageInfoCount", 25},
    {"attachmentIndex", 16},
    {"attachmentInitialSampleLocationsCount", 38},
    {"averageBitrate", 15},
    {"b", 2},
    {"b32", 4},
    {"back", 5},
    {"bank", 5},
    {"baseAddress", 12},
    {"baseArrayLayer", 15},
    {"baseGeometryIndexAndGeometryFlags", 34},
    {"baseGroupX", 11},
    {"baseGroupY", 11},
    {"baseGroupZ", 11},
    {"baseMipLevel", 13},
    {"basePipelineHandle", 19},
    {"basePipelineIndex", 18},
    {"baseTriangle", 13},
    {"basicBlockIndex", 16},
    {"biasInterpretation", 19},
    {"bidirectionalCompoundReferenceNameMask", 39},
    {"bidirectionalFlowSupported", 27},
    {"binaryCount", 12},
    {"bindCount", 10},
    {"bindInfoCount", 14},
    {"bindPoint", 10},
    {"bindPointType", 14},
    {"bindSessionMemoryInfoCount", 27},
    {"binding", 8},
    {"bindingCount", 13},
    {"bindingPoint", 13},
    {"bindingType", 12},
    {"bipredictiveFrameSize", 22},
    {"bipredictiveQIndex", 19},
    {"blendConstants", 15},
    {"blendEnable", 12},
    {"blendOverlap", 13},
    {"blockDimX", 10},
    {"blockDimY", 10},
    {"blockDimZ", 10},
    {"blockInstance", 14},
    {"blockMatchExtendedClampToEdge", 30},
    {"blockTexelViewCompatibleMultipleLayers", 39},
    {"blockType", 10},
    {"blockU64kTilingTensorFeatures", 30},
    {"blockUTilingTensorFeatures", 27},
    {"borderColor", 12},
    {"borderColorSwizzle", 19},
    {"borderColorSwizzleFromImage", 28},
    {"bresenhamLines", 15},
    {"brick16TilingTensorFeatures", 28},
    {"brick4TilingTensorFeatures", 27},
    {"brick8TilingTensorFeatures", 27},
    {"buffer", 7},
    {"bufferAddress", 14},
    {"bufferBindCount", 16},
    {"bufferCaptureReplayDescriptorDataSize", 38},
    {"bufferCollectionConstraints", 28},
    {"bufferCount", 12},
    {"bufferDescriptorAlignment", 26},
    {"bufferDescriptorSize", 21},
    {"bufferDeviceAddress", 20},
    {"bufferDeviceAddressCaptureReplay", 33},
    {"bufferDeviceAddressMultiDevice", 31},
    {"bufferFeatures", 15},
    {"bufferImageGranularity", 23},
    {"bufferImageHeight", 18},
    {"bufferMemoryBarrierCount", 25},
    {"bufferOffset", 13},
    {"bufferRowLength", 16},
    {"bufferView", 11},
    {"bufferlessPushDescriptors", 26},
    {"buildScratchSize", 17},
    {"buildType", 10},
    {"c", 2},
    {"cacheFlushOnCounterCollection", 30},
    {"cacheType", 10},
    {"cacheVersion", 13},
    {"callableShaderBindingOffset", 28},
    {"callableShaderBindingStride", 28},
    {"callableShaderBindingTableAddress", 34},
    {"callableShaderBindingTableBuffer", 33},
    {"callableShaderBindingTableSize", 31},
    {"callableShaderBindingTableStride", 33},
    {"callback", 9},
    {"capabilities", 13},
    {"category", 9},
    {"checkpointExecutionStageMask", 29},
    {"chromaBitDepth", 15},
    {"chromaFilter", 13},
    {"chromaSubsampling", 18},
    {"clampResults", 13},
    {"clearValue", 11},
    {"clearValueCount", 16},
    {"clipped", 8},
    {"clockMode", 10},
    {"clockModes", 11},
    {"closestHitShader", 17},
    {"clusterAccelerationStructure", 29},
    {"clusterBottomLevelByteAlignment", 32},
    {"clusterByteAlignment", 21},
    {"clusterFlags", 13},
    {"clusterID", 10},
    {"clusterIdOffset", 16},
    {"clusterReferences", 18},
    {"clusterReferencesCount", 23},
    {"clusterReferencesStride", 24},
    {"clusterScratchByteAlignment", 28},
    {"clusterShadingRate", 19},
    {"clusterTemplateAddress", 23},
    {"clusterTemplateBoundsByteAlignment", 35},
    {"clusterTemplateByteAlignment", 29},
    {"clustercullingShader", 21},
    {"cmdBufLabelCount", 17},
    {"codeSize", 9},
    {"codeType", 9},
    {"codedExtent", 12},
    {"codedOffset", 12},
    {"codedPictureAlignment", 22},
    {"collection", 11},
    {"collectionToken", 16},
    {"color", 6},
    {"colorAttachment", 16},
    {"colorAttachmentCount", 21},
    {"colorAttachmentFormat", 22},
    {"colorBlendOp", 13},
    {"colorSamples", 13},
    {"colorSpace", 11},
    {"colorSpaceCount", 16},
    {"colorWriteEnable", 17},
    {"colorWriteMask", 15},
    {"combinedImageSamplerDensityMapDescriptorSize", 45},
    {"combinedImageSamplerDescriptorCount", 36},
    {"combinedImageSamplerDescriptorSingleArray", 42},
    {"combinedImageSamplerDescriptorSize", 35},
    {"combinerOps", 12},
    {"commandBuffer", 14},
    {"commandBufferCount", 19},
    {"commandBufferInfoCount", 23},
    {"commandBufferInheritance", 25},
    {"commandCount", 13},
    {"commandPool", 12},
    {"compactedSize", 14},
    {"compareEnable", 14},
    {"compareMask", 12},
    {"compareOp", 10},
    {"compatibleCtbSizes", 19},
    {"compatibleHandleTypes", 22},
    {"compatibleSuperblockSizes", 26},
    {"compilerControlFlags", 21},
    {"componentMapping", 17},
    {"components", 11},
    {"compositeAlpha", 15},
    {"compressedData", 15},
    {"compressedHeaderOffset", 23},
    {"compressedSize", 15},
    {"compressionControlPlaneCount", 29},
    {"computeCapabilityMajor", 23},
    {"computeCapabilityMinor", 23},
    {"computeDerivativeGroupLinear", 29},
    {"computeDerivativeGroupQuads", 28},
    {"computeFullSubgroups", 21},
    {"computeOccupancyPriority", 25},
    {"computeUnitsPerShaderArray", 27},
    {"computeWorkGroupSize", 21},
    {"conditionalRendering", 21},
    {"conditionalRenderingEnable", 27},
    {"configuration", 14},
    {"conformanceVersion", 19},
    {"connection", 11},
    {"connectionCount", 16},
    {"connectorId", 12},
    {"consecutiveBFrameCount", 23},
    {"consecutiveBipredictiveFrameCount", 34},
    {"conservativePointAndLineRasterization", 38},
    {"conservativeRasterizationMode", 30},
    {"conservativeRasterizationPostDepthCoverage", 43},
    {"constantAlphaColorBlendFactors", 31},
    {"constantCount", 14},
    {"constantID", 11},
    {"constantOffset", 15},
    {"constantQIndex", 15},
    {"constantQp", 11},
    {"contents", 9},
    {"context", 8},
    {"conversion", 11},
    {"cooperativeMatrix", 18},
    {"cooperativeMatrixBlockLoads", 28},
    {"cooperativeMatrixConversion", 28},
    {"cooperativeMatrixConversions", 29},
    {"cooperativeMatrixDecodeVector", 30},
    {"cooperativeMatrixFlexibleDimensions", 36},
    {"cooperativeMatrixFlexibleDimensionsMaxDimension", 48},
    {"cooperativeMatrixPerElementOperations", 38},
    {"cooperativeMatrixReductions", 28},
    {"cooperativeMatrixRobustBufferAccess", 36},
    {"cooperativeMatrixSupportedStages", 33},
    {"cooperativeMatrixTensorAddressing", 34},
    {"cooperativeMatrixWorkgroupScope", 32},
    {"cooperativeMatrixWorkgroupScopeMaxWorkgroupSize", 48},
    {"cooperativeMatrixWorkgroupScopeReservedSharedMemory", 52},
    {"cooperativeVector", 18},
    {"cooperativeVectorSupportedStages", 33},
    {"cooperativeVectorTraining", 26},
    {"cooperativeVectorTrainingFloat16Accumulation", 45},
    {"cooperativeVectorTrainingFloat32Accumulation", 45},
    {"copyAddressRange", 17},
    {"copyBufferAddress", 18},
    {"copyCount", 10},
    {"copyDstLayoutCount", 19},
    {"copySrcLayoutCount", 19},
    {"cornerSampledImage", 19},
    {"correlatedViewMaskCount", 24},
    {"correlationMaskCount", 21},
    {"costFormat", 11},
    {"costSupported", 14},
    {"count", 6},
    {"countAddressFlags", 18},
    {"countAddressRange", 18},
    {"countBuffer", 12},
    {"countBufferOffset", 18},
    {"countInfo", 10},
    {"counter", 8},
    {"counterAddressCount", 20},
    {"counterBuffer", 14},
    {"counterBufferCount", 19},
    {"counterBufferOffset", 20},
    {"counterID", 10},
    {"counterIndexCount", 18},
    {"counterOffset", 14},
    {"counterPassIndex", 17},
    {"counterRangeCount", 18},
    {"coverageModulationMode", 23},
    {"coverageModulationTableCount", 29},
    {"coverageModulationTableEnable", 30},
    {"coverageReductionMode", 22},
    {"coverageToColorEnable", 22},
    {"coverageToColorLocation", 24},
    {"createFlags", 12},
    {"createInfo", 11},
    {"createInfoCount", 16},
    {"createInfoIndex", 16},
    {"ctbSizes", 9},
    {"cubicRangeClamp", 16},
    {"cubicWeights", 13},
    {"cudaKernelLaunchFeatures", 25},
    {"cullMode", 9},
    {"currentDisplay", 15},
    {"currentExtent", 14},
    {"currentStackIndex", 18},
    {"currentTransform", 17},
    {"customBorderColor", 18},
    {"customBorderColorWithoutFormat", 31},
    {"customBorderColors", 19},
    {"customResolve", 14},
    {"customSampleOrderCount", 23},
    {"data", 5},
    {"dataGraph", 10},
    {"dataGraphDescriptorBuffer", 26},
    {"dataGraphModel", 15},
    {"dataGraphNeuralAcceleratorStatistics", 37},
    {"dataGraphOpticalFlow", 21},
    {"dataGraphPipeline", 18},
    {"dataGraphShaderModule", 22},
    {"dataGraphSpecializationConstants", 33},
    {"dataGraphUpdateAfterBind", 25},
    {"dataOffset", 11},
    {"dataSize", 9},
    {"decodeMode", 11},
    {"decodeModeSharedExponent", 25},
    {"decompressRegionCount", 22},
    {"decompressedSize", 17},
    {"decompressionMethod", 20},
    {"decompressionMethods", 21},
    {"dedicatedAllocation", 20},
    {"dedicatedAllocationImageAliasing", 33},
    {"defaultRobustnessImages", 24},
    {"defaultRobustnessStorageBuffers", 32},
    {"defaultRobustnessUniformBuffers", 32},
    {"defaultRobustnessVertexInputs", 30},
    {"defaultVertexAttributeValue", 28},
    {"deferredOperation", 18},
    {"degenerateLinesRasterized", 26},
    {"degenerateTrianglesRasterized", 30},
    {"denormBehaviorIndependence", 27},
    {"denseGeometryFormat", 20},
    {"dependencyCount", 16},
    {"dependencyFlags", 16},
    {"depth", 6},
    {"depthAttachmentFormat", 22},
    {"depthBiasClamp", 15},
    {"depthBiasConstantFactor", 24},
    {"depthBiasControl", 17},
    {"depthBiasEnable", 16},
    {"depthBiasExact", 15},
    {"depthBiasRepresentation", 24},
    {"depthBiasSlopeFactor", 21},
    {"depthBounds", 12},
    {"depthBoundsTestEnable", 22},
    {"depthClamp", 11},
    {"depthClampControl", 18},
    {"depthClampEnable", 17},
    {"depthClampMode", 15},
    {"depthClampZeroOne", 18},
    {"depthClipControl", 17},
    {"depthClipEnable", 16},
    {"depthCompareOp", 15},
    {"depthFailOp", 12},
    {"depthPitch", 11},
    {"depthResolveMode", 17},
    {"depthStencil", 13},
    {"depthStencilAttachmentSamples", 30},
    {"depthStencilSamples", 20},
    {"depthStencilSwizzleOneSupport", 30},
    {"depthTestEnable", 16},
    {"depthWriteEnable", 17},
    {"description", 12},
    {"descriptorBindingAccelerationStructureUpdateAfterBind", 54},
    {"descriptorBindingInlineUniformBlockUpdateAfterBind", 51},
    {"descriptorBindingPartiallyBound", 32},
    {"descriptorBindingSampledImageUpdateAfterBind", 45},
    {"descriptorBindingStorageBufferUpdateAfterBind", 46},
    {"descriptorBindingStorageImageUpdateAfterBind", 45},
    {"descriptorBindingStorageTensorUpdateAfterBind", 46},
    {"descriptorBindingStorageTexelBufferUpdateAfterBind", 51},
    {"descriptorBindingUniformBufferUpdateAfterBind", 46},
    {"descriptorBindingUniformTexelBufferUpdateAfterBind", 51},
    {"descriptorBindingUpdateUnusedWhilePending", 42},
    {"descriptorBindingVariableDescriptorCount", 41},
    {"descriptorBuffer", 17},
    {"descriptorBufferAddressSpaceSize", 33},
    {"descriptorBufferCaptureReplay", 30},
    {"descriptorBufferImageLayoutIgnored", 35},
    {"descriptorBufferOffsetAlignment", 32},
    {"descriptorBufferPushDescriptors", 32},
    {"descriptorBufferTensorDescriptors", 34},
    {"descriptorCopyCount", 20},
    {"descriptorCount", 16},
    {"descriptorHeap", 15},
    {"descriptorHeapCaptureReplay", 28},
    {"descriptorIndexing", 19},
    {"descriptorOffset", 17},
    {"descriptorPool", 15},
    {"descriptorPoolOverallocation", 29},
    {"descriptorSet", 14},
    {"descriptorSetCount", 19},
    {"descriptorSetHostMapping", 25},
    {"descriptorSetLayout", 20},
    {"descriptorSize", 15},
    {"descriptorType", 15},
    {"descriptorTypeCount", 20},
    {"descriptorUpdateEntryCount", 27},
    {"descriptorUpdateTemplate", 25},
    {"descriptorWriteCount", 21},
    {"desiredPresentTime", 19},
    {"device", 7},
    {"deviceAddress", 14},
    {"deviceAddressCommands", 22},
    {"deviceCoherentMemory", 21},
    {"deviceEvent", 12},
    {"deviceFault", 12},
    {"deviceFaultDeviceLostOnMasked", 30},
    {"deviceFaultReportMasked", 24},
    {"deviceFaultVendorBinary", 24},
    {"deviceGeneratedCommands", 24},
    {"deviceGeneratedCommandsMultiDrawIndirectCount", 46},
    {"deviceGeneratedCommandsTransformFeedback", 41},
    {"deviceGeneratedCompute", 23},
    {"deviceGeneratedComputeCaptureReplay", 36},
    {"deviceGeneratedComputePipelines", 32},
    {"deviceID", 9},
    {"deviceIndex", 12},
    {"deviceIndexCount", 17},
    {"deviceLUID", 11},
    {"deviceLUIDValid", 16},
    {"deviceMask", 11},
    {"deviceMemoryReport", 19},
    {"deviceName", 11},
    {"deviceNodeMask", 15},
    {"deviceRelativeId", 17},
    {"deviceRenderAreaCount", 22},
    {"deviceType", 11},
    {"deviceUUID", 11},
    {"dfb", 4},
    {"diagnosticsConfig", 18},
    {"dimension", 10},
    {"dimensionCount", 15},
    {"dirtyIntraRefreshRegions", 25},
    {"disableInternalCache", 21},
    {"disabledValidationCheckCount", 29},
    {"disabledValidationFeatureCount", 31},
    {"disallowMerging", 16},
    {"discardRectangleCount", 22},
    {"discardRectangleEnable", 23},
    {"discardRectangleMode", 21},
    {"discardable", 12},
    {"discreteQueuePriorities", 24},
    {"displacedMicromapPrimitiveFlags", 32},
    {"displacedMicromapPrimitiveFlagsStride", 38},
    {"displacementBiasAndScaleBuffer", 31},
    {"displacementBiasAndScaleFormat", 31},
    {"displacementBiasAndScaleStride", 31},
    {"displacementMicromap", 21},
    {"displacementVectorBuffer", 25},
    {"displacementVectorFormat", 25},
    {"displacementVectorStride", 25},
    {"display", 8},
    {"displayEvent", 13},
    {"displayMode", 12},
    {"displayModeProperties", 22},
    {"displayName", 12},
    {"displayPlaneProperties", 23},
    {"displayPrimaryBlue", 19},
    {"displayPrimaryGreen", 20},
    {"displayPrimaryRed", 18},
    {"displayProperties", 18},
    {"divisor", 8},
    {"domainOrigin", 13},
    {"dpy", 4},
    {"drawCount", 10},
    {"drawIndirectCount", 18},
    {"drawIndirectFirstInstance", 26},
    {"driverCount", 12},
    {"driverEndTimeUs", 16},
    {"driverID", 9},
    {"driverInfo", 11},
    {"driverName", 11},
    {"driverStartTimeUs", 18},
    {"driverUUID", 11},
    {"driverVersion", 14},
    {"drmFd", 6},
    {"drmFormatModifier", 18},
    {"drmFormatModifierCount", 23},
    {"drmFormatModifierPlaneCount", 28},
    {"drmFormatModifierTilingFeatures", 32},
    {"dst", 4},
    {"dstAccelerationStructure", 25},
    {"dstAccelerationStructureData", 29},
    {"dstAccessMask", 14},
    {"dstAccessMask3", 15},
    {"dstAddress", 11},
    {"dstAddressesArray", 18},
    {"dstAlphaBlendFactor", 20},
    {"dstArrayElement", 16},
    {"dstBinding", 11},
    {"dstBuffer", 10},
    {"dstBufferOffset", 16},
    {"dstBufferRange", 15},
    {"dstCache", 9},
    {"dstColorBlendFactor", 20},
    {"dstComponentType", 17},
    {"dstCopyFlags", 13},
    {"dstData", 8},
    {"dstFlags", 9},
    {"dstImage", 9},
    {"dstImageLayout", 15},
    {"dstImplicitData", 16},
    {"dstLayout", 10},
    {"dstMicromap", 12},
    {"dstOffset", 10},
    {"dstOffsets", 11},
    {"dstPictureResource", 19},
    {"dstPremultiplied", 17},
    {"dstQueueFamilyIndex", 20},
    {"dstRange", 9},
    {"dstRect", 8},
    {"dstSet", 7},
    {"dstSizesArray", 14},
    {"dstStageMask", 13},
    {"dstStride", 10},
    {"dstSubpass", 11},
    {"dstSubresource", 15},
    {"dstTensor", 10},
    {"dualSrcBlend", 13},
    {"duration", 9},
    {"dwAccess", 9},
    {"dynamicGeneratedPipelineLayout", 31},
    {"dynamicMetadataSize", 20},
    {"dynamicOffsetCount", 19},
    {"dynamicPipelineLayout", 22},
    {"dynamicPrimitiveTopologyUnrestricted", 37},
    {"dynamicRendering", 17},
    {"dynamicRenderingLocalRead", 26},
    {"dynamicRenderingLocalReadDepthStencilAttachments", 49},
    {"dynamicRenderingLocalReadMultisampledAttachments", 49},
    {"dynamicRenderingUnusedAttachments", 34},
    {"dynamicStateCount", 18},
    {"earliestPresentTime", 20},
    {"earlyFragmentMultisampleCoverageAfterSampleCounting", 52},
    {"earlyFragmentSampleMaskTestBeforeSampleCounting", 48},
    {"elapsedTimerQuery", 18},
    {"enable", 7},
    {"enableCbCrDegamma", 18},
    {"enablePartitionTranslation", 27},
    {"enableYDegamma", 15},
    {"enabledExtensionCount", 22},
    {"enabledLayerCount", 18},
    {"enabledValidationFeatureCount", 30},
    {"encodeFeedbackFlags", 20},
    {"encodeInputPictureGranularity", 30},
    {"endCapsMode", 12},
    {"engine", 7},
    {"engineClockFrequency", 21},
    {"engineClockRatioToPeak", 23},
    {"engineNameOffset", 17},
    {"engineType", 11},
    {"engineVersion", 14},
    {"event", 6},
    {"eventCount", 11},
    {"eventID", 8},
    {"events", 7},
    {"exclusiveScissor", 17},
    {"exclusiveScissorCount", 22},
    {"executableIndex", 16},
    {"executionGraph", 15},
    {"executionGraphDispatchAddressAlignment", 39},
    {"executionSetWriteCount", 23},
    {"expectDyadicTemporalLayerPattern", 33},
    {"expectDyadicTemporalSubLayerPattern", 36},
    {"explicitAABB", 13},
    {"exportFromImportedHandleTypes", 30},
    {"exportObjectType", 17},
    {"extendedDynamicState", 21},
    {"extendedDynamicState2", 22},
    {"extendedDynamicState2LogicOp", 29},
    {"extendedDynamicState2PatchControlPoints", 40},
    {"extendedDynamicState3AlphaToCoverageEnable", 43},
    {"extendedDynamicState3AlphaToOneEnable", 38},
    {"extendedDynamicState3ColorBlendAdvanced", 40},
    {"extendedDynamicState3ColorBlendEnable", 38},
    {"extendedDynamicState3ColorBlendEquation", 40},
    {"extendedDynamicState3ColorWriteMask", 36},
    {"extendedDynamicState3ConservativeRasterizationMode", 51},
    {"extendedDynamicState3CoverageModulationMode", 44},
    {"extendedDynamicState3CoverageModulationTable", 45},
    {"extendedDynamicState3CoverageModulationTableEnable", 51},
    {"extendedDynamicState3CoverageReductionMode", 43},
    {"extendedDynamicState3CoverageToColorEnable", 43},
    {"extendedDynamicState3CoverageToColorLocation", 45},
    {"extendedDynamicState3DepthClampEnable", 38},
    {"extendedDynamicState3DepthClipEnable", 37},
    {"extendedDynamicState3DepthClipNegativeOneToOne", 47},
    {"extendedDynamicState3ExtraPrimitiveOverestimationSize", 54},
    {"extendedDynamicState3LineRasterizationMode", 43},
    {"extendedDynamicState3LineStippleEnable", 39},
    {"extendedDynamicState3LogicOpEnable", 35},
    {"extendedDynamicState3PolygonMode", 33},
    {"extendedDynamicState3ProvokingVertexMode", 41},
    {"extendedDynamicState3RasterizationSamples", 42},
    {"extendedDynamicState3RasterizationStream", 41},
    {"extendedDynamicState3RepresentativeFragmentTestEnable", 54},
    {"extendedDynamicState3SampleLocationsEnable", 43},
    {"extendedDynamicState3SampleMask", 32},
    {"extendedDynamicState3ShadingRateImageEnable", 44},
    {"extendedDynamicState3TessellationDomainOrigin", 46},
    {"extendedDynamicState3ViewportSwizzle", 37},
    {"extendedDynamicState3ViewportWScalingEnable", 44},
    {"extendedFlags", 14},
    {"extendedSparseAddressSpace", 27},
    {"extendedSparseAddressSpaceSize", 31},
    {"extendedSparseBufferUsageFlags", 31},
    {"extendedSparseImageUsageFlags", 30},
    {"extensionCount", 15},
    {"extensionName", 14},
    {"extent", 7},
    {"externalDataSize", 17},
    {"externalFenceFeatures", 22},
    {"externalFormat", 15},
    {"externalFormatResolve", 22},
    {"externalFormatResolveChromaOffsetX", 35},
    {"externalFormatResolveChromaOffsetY", 35},
    {"externalHandleType", 19},
    {"externalMemoryFeatures", 23},
    {"externalMemoryProperties", 25},
    {"externalMemoryRDMA", 19},
    {"externalQueue", 14},
    {"externalSemaphoreFeatures", 26},
    {"extraCount", 11},
    {"extraPrimitiveOverestimationSize", 33},
    {"extraPrimitiveOverestimationSizeGranularity", 44},
    {"f64", 4},
    {"faceMask", 9},
    {"failOp", 7},
    {"faultAddressInfo", 17},
    {"fd", 3},
    {"features", 9},
    {"feedbackLoopEnable", 19},
    {"fence", 6},
    {"fenceCount", 11},
    {"fieldOffsetGranularity", 23},
    {"fillModeNonSolid", 17},
    {"filmGrainSupport", 17},
    {"filter", 7},
    {"filterCenter", 13},
    {"filterCubic", 12},
    {"filterCubicMinmax", 18},
    {"filterMinmaxImageComponentMapping", 34},
    {"filterMinmaxSingleComponentFormats", 35},
    {"filterSize", 11},
    {"finalLayout", 12},
    {"firstAttachment", 16},
    {"firstBinding", 13},
    {"firstCounterBuffer", 19},
    {"firstCounterRange", 18},
    {"firstDiscardRectangle", 22},
    {"firstDrawTimestamp", 19},
    {"firstExclusiveScissor", 22},
    {"firstGroup", 11},
    {"firstIndex", 11},
    {"firstInstance", 14},
    {"firstQuery", 11},
    {"firstScissor", 13},
    {"firstSet", 9},
    {"firstTask", 10},
    {"firstVertex", 12},
    {"firstViewport", 14},
    {"flags", 6},
    {"float32", 8},
    {"float64", 8},
    {"floatRepresentation", 20},
    {"flowVectorFormat", 17},
    {"fmaRate", 8},
    {"forceExplicitReconstruction", 28},
    {"foreignMemoryHandleTypes", 25},
    {"foreignSemaphoreHandleTypes", 28},
    {"format", 7},
    {"formatA4B4G4R4", 15},
    {"formatA4R4G4B4", 15},
    {"formatConstraintsCount", 23},
    {"formatFeatures", 15},
    {"formatPack", 11},
    {"formatProperties", 17},
    {"formatRgba10x6WithoutYCbCrSampler", 34},
    {"fragmentDensityInvocations", 27},
    {"fragmentDensityMap", 19},
    {"fragmentDensityMapAttachment", 29},
    {"fragmentDensityMapDeferred", 27},
    {"fragmentDensityMapDynamic", 26},
    {"fragmentDensityMapLayered", 26},
    {"fragmentDensityMapNonSubsampledImages", 38},
    {"fragmentDensityMapOffset", 25},
    {"fragmentDensityOffsetCount", 27},
    {"fragmentDensityOffsetGranularity", 33},
    {"fragmentShaderBarycentric", 26},
    {"fragmentShaderPixelInterlock", 29},
    {"fragmentShaderSampleInterlock", 30},
    {"fragmentShaderShadingRateInterlock", 35},
    {"fragmentShadingRateClampCombinerInputs", 39},
    {"fragmentShadingRateEnums", 25},
    {"fragmentShadingRateNonTrivialCombinerOps", 41},
    {"fragmentShadingRateStrictMultiplyCombiner", 42},
    {"fragmentShadingRateWithConservativeRasterization", 49},
    {"fragmentShadingRateWithCustomSampleLocations", 45},
    {"fragmentShadingRateWithFragmentShaderInterlock", 47},
    {"fragmentShadingRateWithSampleMask", 34},
    {"fragmentShadingRateWithShaderDepthStencilWrites", 48},
    {"fragmentShadingRateWithShaderSampleMask", 40},
    {"fragmentSize", 13},
    {"fragmentStoresAndAtomics", 25},
    {"frameBSize", 11},
    {"frameBoundary", 14},
    {"frameHeaderOffset", 18},
    {"frameID", 8},
    {"frameISize", 11},
    {"frameIndex", 11},
    {"framePSize", 11},
    {"frameRateDenominator", 21},
    {"frameRateNumerator", 19},
    {"frameToken", 11},
    {"framebuffer", 12},
    {"framebufferColorSampleCounts", 29},
    {"framebufferDepthSampleCounts", 29},
    {"framebufferIntegerColorSampleCounts", 36},
    {"framebufferNoAttachmentsSampleCounts", 37},
    {"framebufferStencilSampleCounts", 31},
    {"front", 6},
    {"frontFace", 10},
    {"fullDrawIndexUint32", 20},
    {"fullScreenExclusive", 20},
    {"fullScreenExclusiveSupported", 29},
    {"fullyCoveredFragmentShaderInputVariable", 40},
    {"function", 9},
    {"g", 2},
    {"generalShader", 14},
    {"generateObuExtensionHeader", 27},
    {"generatePrefixNalu", 19},
    {"geometry", 9},
    {"geometryCount", 14},
    {"geometryFlags", 14},
    {"geometryIndex", 14},
    {"geometryIndexAndFlagsBuffer", 28},
    {"geometryIndexAndFlagsBufferStride", 34},
    {"geometryIndexOffset", 20},
    {"geometryShader", 15},
    {"geometryStreams", 16},
    {"geometryType", 13},
    {"globalAlpha", 12},
    {"globalFlowSupported", 20},
    {"globalPriority", 15},
    {"globalPriorityQuery", 20},
    {"gopFrameCount", 14},
    {"gopRemainingB", 14},
    {"gopRemainingBipredictive", 25},
    {"gopRemainingI", 14},
    {"gopRemainingIntra", 18},
    {"gopRemainingP", 14},
    {"gopRemainingPredictive", 23},
    {"gpaSession", 11},
    {"gpuRenderEndTimeUs", 19},
    {"gpuRenderStartTimeUs", 21},
    {"graphicsPipelineLibrary", 24},
    {"graphicsPipelineLibraryFastLinking", 35},
    {"graphicsPipelineLibraryIndependentInterpolationDecoration", 58},
    {"gridDimX", 9},
    {"gridDimY", 9},
    {"gridDimZ", 9},
    {"group", 6},
    {"groupCount", 11},
    {"groupCountX", 12},
    {"groupCountY", 12},
    {"groupCountZ", 12},
    {"groupId", 8},
    {"groupIndex", 11},
    {"groupShader", 12},
    {"groupSize", 10},
    {"handle", 7},
    {"handleType", 11},
    {"handleTypes", 12},
    {"hasOverrides", 13},
    {"hasPrimary", 11},
    {"hasRender", 10},
    {"hasStdPPSOverrides", 19},
    {"hasStdSPSOverrides", 19},
    {"hasStdVPSOverrides", 19},
    {"hdmi3DSupported", 16},
    {"hdrVivid", 9},
    {"headerSize", 11},
    {"headerVersion", 14},
    {"heapArrayStride", 16},
    {"heapBudget", 11},
    {"heapData", 9},
    {"heapIndex", 10},
    {"heapIndexStride", 16},
    {"heapOffset", 11},
    {"heapRange", 10},
    {"heapUsage", 10},
    {"height", 7},
    {"hinstance", 10},
    {"hintGridSize", 13},
    {"hintSupported", 14},
    {"hitShaderBindingOffset", 23},
    {"hitShaderBindingStride", 23},
    {"hitShaderBindingTableAddress", 29},
    {"hitShaderBindingTableBuffer", 28},
    {"hitShaderBindingTableSize", 26},
    {"hitShaderBindingTableStride", 28},
    {"hmonitor", 9},
    {"hostAddress", 12},
    {"hostImageCopy", 14},
    {"hostQueryReset", 15},
    {"hwnd", 5},
    {"i64", 4},
    {"id", 3},
    {"identicalMemoryLayout", 22},
    {"identicalMemoryTypeRequirements", 32},
    {"identifier", 11},
    {"identifierSize", 15},
    {"identityTransformOrder", 23},
    {"idrPeriod", 10},
    {"image", 6},
    {"image2DViewOf3D", 16},
    {"image2DViewOf3DSparse", 22},
    {"imageAlignmentControl", 22},
    {"imageArrayLayers", 17},
    {"imageBindCount", 15},
    {"imageCaptureReplayDescriptorDataSize", 37},
    {"imageCaptureReplayOpaqueDataSize", 33},
    {"imageColorSpace", 16},
    {"imageCompressionControl", 24},
    {"imageCompressionControlSwapchain", 33},
    {"imageCompressionFixedRateFlags", 31},
    {"imageCompressionFlags", 22},
    {"imageCount", 11},
    {"imageCreateFlags", 17},
    {"imageCreateInfo", 16},
    {"imageCubeArray", 15},
    {"imageDescriptorAlignment", 25},
    {"imageDescriptorSize", 20},
    {"imageExtent", 12},
    {"imageFootprint", 15},
    {"imageFormat", 12},
    {"imageFormatProperties", 22},
    {"imageGatherExtendedModes", 25},
    {"imageGatherLinear", 18},
    {"imageGranularity", 17},
    {"imageIndex", 11},
    {"imageIndexCount", 16},
    {"imageLayout", 12},
    {"imageMemoryBarrierCount", 24},
    {"imageMipTailFirstLod", 21},
    {"imageMipTailOffset", 19},
    {"imageMipTailSize", 17},
    {"imageMipTailStride", 19},
    {"imageOffset", 12},
    {"imageOpaqueBindCount", 21},
    {"imagePipeHandle", 16},
    {"imageSharingMode", 17},
    {"imageSlicedViewOf3D", 20},
    {"imageSubresource", 17},
    {"imageTiling", 12},
    {"imageType", 10},
    {"imageUsage", 11},
    {"imageUsageFlags", 16},
    {"imageView", 10},
    {"imageView2DOn3DImage", 21},
    {"imageViewBinding", 17},
    {"imageViewCaptureReplayDescriptorDataSize", 41},
    {"imageViewFormatReinterpretation", 32},
    {"imageViewFormatSwizzle", 23},
    {"imageViewIndex", 15},
    {"imageViewType", 14},
    {"imagelessFramebuffer", 21},
    {"images", 7},
    {"implementationVersion", 22},
    {"independentBlend", 17},
    {"independentResolve", 19},
    {"independentResolveNone", 23},
    {"index", 6},
    {"indexBuffer", 12},
    {"indexBufferStride", 18},
    {"indexCount", 11},
    {"indexData", 10},
    {"indexOffset", 12},
    {"indexStride", 12},
    {"indexType", 10},
    {"indexTypeCount", 15},
    {"indexTypeUint8", 15},
    {"indexingMode", 13},
    {"indirectAddress", 16},
    {"indirectAddressSize", 20},
    {"indirectBufferOffsetAlignment", 30},
    {"indirectCommandsAddress", 24},
    {"indirectCommandsCountAddress", 29},
    {"indirectCommandsLayout", 23},
    {"indirectCopy", 13},
    {"indirectDeviceAddress", 22},
    {"indirectExecutionSet", 21},
    {"indirectIndex", 14},
    {"indirectIndexArray", 19},
    {"indirectMemoryCopy", 19},
    {"indirectMemoryToImageCopy", 26},
    {"indirectStateFlags", 19},
    {"indirectStride", 15},
    {"info", 5},
    {"infoCount", 10},
    {"infoType", 9},
    {"infos", 6},
    {"inheritedConditionalRendering", 30},
    {"inheritedQueries", 17},
    {"inheritedViewportScissor2D", 27},
    {"initialDataSize", 16},
    {"initialLayout", 14},
    {"initialPipeline", 16},
    {"initialValue", 13},
    {"initialVirtualBufferSizeInMs", 29},
    {"inlineUniformBlock", 19},
    {"input", 6},
    {"inputAttachmentCount", 21},
    {"inputAttachmentDescriptorSize", 30},
    {"inputAttachmentIndex", 21},
    {"inputInterpretation", 20},
    {"inputRate", 10},
    {"inputSampleTimeUs", 18},
    {"inputType", 10},
    {"instance", 9},
    {"instanceContributionToHitGroupIndex", 36},
    {"instanceCount", 14},
    {"instanceCustomIndex", 20},
    {"instanceData", 13},
    {"instanceFlags", 14},
    {"instanceID", 11},
    {"instanceIndex", 14},
    {"instanceMask", 13},
    {"instanceOffset", 15},
    {"instanceShaderBindingTableRecordOffset", 39},
    {"instances", 10},
    {"instantiationBoundingBoxLimit", 30},
    {"instructionAddressInfo", 23},
    {"instrumentation", 16},
    {"int32", 6},
    {"int64", 6},
    {"integerDotProduct16BitMixedSignednessAccelerated", 49},
    {"integerDotProduct16BitSignedAccelerated", 40},
    {"integerDotProduct16BitUnsignedAccelerated", 42},
    {"integerDotProduct32BitMixedSignednessAccelerated", 49},
    {"integerDotProduct32BitSignedAccelerated", 40},
    {"integerDotProduct32BitUnsignedAccelerated", 42},
    {"integerDotProduct4x8BitPackedMixedSignednessAccelerated", 56},
    {"integerDotProduct4x8BitPackedSignedAccelerated", 47},
    {"integerDotProduct4x8BitPackedUnsignedAccelerated", 49},
    {"integerDotProduct64BitMixedSignednessAccelerated", 49},
    {"integerDotProduct64BitSignedAccelerated", 40},
    {"integerDotProduct64BitUnsignedAccelerated", 42},
    {"integerDotProduct8BitMixedSignednessAccelerated", 48},
    {"integerDotProduct8BitSignedAccelerated", 39},
    {"integerDotProduct8BitUnsignedAccelerated", 41},
    {"integerDotProductAccumulatingSaturating16BitMixedSignednessAccelerated", 71},
    {"integerDotProductAccumulatingSaturating16BitSignedAccelerated", 62},
    {"integerDotProductAccumulatingSaturating16BitUnsignedAccelerated", 64},
    {"integerDotProductAccumulatingSaturating32BitMixedSignednessAccelerated", 71},
    {"integerDotProductAccumulatingSaturating32BitSignedAccelerated", 62},
    {"integerDotProductAccumulatingSaturating32BitUnsignedAccelerated", 64},
    {"integerDotProductAccumulatingSaturating4x8BitPackedMixedSignednessAccelerated", 78},
    {"integerDotProductAccumulatingSaturating4x8BitPackedSignedAccelerated", 69},
    {"integerDotProductAccumulatingSaturating4x8BitPackedUnsignedAccelerated", 71},
    {"integerDotProductAccumulatingSaturating64BitMixedSignednessAccelerated", 71},
    {"integerDotProductAccumulatingSaturating64BitSignedAccelerated", 62},
    {"integerDotProductAccumulatingSaturating64BitUnsignedAccelerated", 64},
    {"integerDotProductAccumulatingSaturating8BitMixedSignednessAccelerated", 70},
    {"integerDotProductAccumulatingSaturating8BitSignedAccelerated", 61},
    {"integerDotProductAccumulatingSaturating8BitUnsignedAccelerated", 63},
    {"internallySynchronizedQueues", 29},
    {"intersectionShader", 19},
    {"intraFrameSize", 15},
    {"intraQIndex", 12},
    {"intraRefreshCycleDuration", 26},
    {"intraRefreshIndex", 18},
    {"intraRefreshMode", 17},
    {"intraRefreshModes", 18},
    {"invocationMask", 15},
    {"ioSurface", 10},
    {"isForeign", 10},
    {"isPreprocessed", 15},
    {"isText", 7},
    {"key", 4},
    {"keyFramePeriod", 15},
    {"keySize", 8},
    {"largePoints", 12},
    {"latencyModeEnable", 18},
    {"layer", 6},
    {"layerCount", 11},
    {"layerName", 10},
    {"layeredAPI", 11},
    {"layeredApiCount", 16},
    {"layeredShadingRateAttachments", 30},
    {"layers", 7},
    {"layout", 7},
    {"ldsSizePerLocalWorkGroup", 25},
    {"ldsUsageSizeInBytes", 20},
    {"leastRepresentableValueForceUnormRepresentation", 48},
    {"legacyDithering", 16},
    {"legacyVertexAttributes", 23},
    {"level", 6},
    {"levelCount", 11},
    {"libraryCount", 13},
    {"limits", 7},
    {"lineRasterizationMode", 22},
    {"lineStippleFactor", 18},
    {"lineStipplePattern", 19},
    {"lineSubPixelPrecisionBits", 26},
    {"lineWidth", 10},
    {"lineWidthGranularity", 21},
    {"lineWidthRange", 15},
    {"linearColorAttachment", 22},
    {"linearSweptSpheres", 19},
    {"linearTilingFeatures", 21},
    {"linearTilingTensorFeatures", 27},
    {"loadOp", 7},
    {"localDeviceIndex", 17},
    {"localDimmingEnable", 19},
    {"localDimmingSupport", 20},
    {"location", 9},
    {"logicOp", 8},
    {"logicOpEnable", 14},
    {"longVector", 11},
    {"lowLatencyBoost", 16},
    {"lowLatencyMode", 15},
    {"lumaBitDepth", 13},
    {"magFilter", 10},
    {"maintenance10", 14},
    {"maintenance11", 14},
    {"maintenance4", 13},
    {"maintenance5", 13},
    {"maintenance6", 13},
    {"maintenance7", 13},
    {"maintenance8", 13},
    {"maintenance9", 13},
    {"major", 6},
    {"mapEntryCount", 14},
    {"mappingCount", 13},
    {"marker", 7},
    {"mask", 5},
    {"matrix", 7},
    {"matrixInterpretation", 21},
    {"matrixMotionInstance", 21},
    {"maxAccelerationStructureCount", 30},
    {"maxActiveReferencePictures", 27},
    {"maxAnisotropy", 14},
    {"maxApronSize", 13},
    {"maxArrayLayers", 15},
    {"maxBPictureL0ReferenceCount", 28},
    {"maxBidirectionalCompoundGroup1ReferenceCount", 45},
    {"maxBidirectionalCompoundGroup2ReferenceCount", 45},
    {"maxBidirectionalCompoundReferenceCount", 39},
    {"maxBitrate", 11},
    {"maxBlockMatchRegion", 20},
    {"maxBlockMatchWindow", 20},
    {"maxBoundDescriptorSets", 23},
    {"maxBoxFilterBlockSize", 22},
    {"maxBufferCount", 15},
    {"maxBufferSize", 14},
    {"maxClipDistances", 17},
    {"maxClusterCountPerAccelerationStructure", 40},
    {"maxClusterGeometryIndex", 24},
    {"maxClusterTriangleCount", 24},
    {"maxClusterUniqueGeometryCount", 30},
    {"maxClusterVertexCount", 22},
    {"maxCodedExtent", 15},
    {"maxColorAttachments", 20},
    {"maxCombinedClipAndCullDistances", 32},
    {"maxCombinedImageSamplerDescriptorCount", 39},
    {"maxCommandBufferNestingLevel", 29},
    {"maxComputePushConstantBanks", 28},
    {"maxComputePushDataBanks", 24},
    {"maxComputeSharedMemorySize", 27},
    {"maxComputeWorkGroupCount", 25},
    {"maxComputeWorkGroupInvocations", 31},
    {"maxComputeWorkGroupSize", 24},
    {"maxComputeWorkgroupSubgroups", 29},
    {"maxContentLightLevel", 21},
    {"maxCooperativeVectorComponents", 31},
    {"maxCullDistances", 17},
    {"maxCustomBorderColorSamplers", 29},
    {"maxDecompressionCount", 22},
    {"maxDecompressionIndirectCount", 30},
    {"maxDepth", 9},
    {"maxDepthBounds", 15},
    {"maxDepthClamp", 14},
    {"maxDescriptorBufferBindings", 28},
    {"maxDescriptorHeapEmbeddedSamplers", 34},
    {"maxDescriptorSetAccelerationStructures", 39},
    {"maxDescriptorSetInlineUniformBlocks", 36},
    {"maxDescriptorSetInputAttachments", 33},
    {"maxDescriptorSetSampledImages", 30},
    {"maxDescriptorSetSamplers", 25},
    {"maxDescriptorSetStorageBuffers", 31},
    {"maxDescriptorSetStorageBuffersDynamic", 38},
    {"maxDescriptorSetStorageImages", 30},
    {"maxDescriptorSetStorageTensors", 31},
    {"maxDescriptorSetSubsampledSamplers", 35},
    {"maxDescriptorSetTotalBuffersDynamic", 36},
    {"maxDescriptorSetTotalStorageBuffersDynamic", 43},
    {"maxDescriptorSetTotalUniformBuffersDynamic", 43},
    {"maxDescriptorSetUniformBuffers", 31},
    {"maxDescriptorSetUniformBuffersDynamic", 38},
    {"maxDescriptorSetUpdateAfterBindAccelerationStructures", 54},
    {"maxDescriptorSetUpdateAfterBindInlineUniformBlocks", 51},
    {"maxDescriptorSetUpdateAfterBindInputAttachments", 48},
    {"maxDescriptorSetUpdateAfterBindSampledImages", 45},
    {"maxDescriptorSetUpdateAfterBindSamplers", 40},
    {"maxDescriptorSetUpdateAfterBindStorageBuffers", 46},
    {"maxDescriptorSetUpdateAfterBindStorageBuffersDynamic", 53},
    {"maxDescriptorSetUpdateAfterBindStorageImages", 45},
    {"maxDescriptorSetUpdateAfterBindStorageTensors", 46},
    {"maxDescriptorSetUpdateAfterBindTotalBuffersDynamic", 51},
    {"maxDescriptorSetUpdateAfterBindTotalStorageBuffersDynamic", 58},
    {"maxDescriptorSetUpdateAfterBindTotalUniformBuffersDynamic", 58},
    {"maxDescriptorSetUpdateAfterBindUniformBuffers", 46},
    {"maxDescriptorSetUpdateAfterBindUniformBuffersDynamic", 53},
    {"maxDeviceFaultCount", 20},
    {"maxDiscardRectangles", 21},
    {"maxDisplacementMicromapSubdivisionLevel", 40},
    {"maxDpbSlots", 12},
    {"maxDrawCount", 13},
    {"maxDrawIndexedIndexValue", 25},
    {"maxDrawIndirectCount", 21},
    {"maxDrawMeshTasksCount", 22},
    {"maxDstExtent", 13},
    {"maxDstPosition", 15},
    {"maxEmbeddedImmutableSamplerBindings", 36},
    {"maxEmbeddedImmutableSamplers", 29},
    {"maxEventID", 11},
    {"maxExecutionGraphDepth", 23},
    {"maxExecutionGraphShaderOutputNodes", 35},
    {"maxExecutionGraphShaderPayloadCount", 36},
    {"maxExecutionGraphShaderPayloadSize", 35},
    {"maxExecutionGraphWorkgroupCount", 32},
    {"maxExecutionGraphWorkgroups", 28},
    {"maxExtent", 10},
    {"maxExternalQueues", 18},
    {"maxExtraPrimitiveOverestimationSize", 36},
    {"maxFPS", 7},
    {"maxFragmentCombinedOutputResources", 35},
    {"maxFragmentDensityMapLayers", 28},
    {"maxFragmentDensityTexelSize", 28},
    {"maxFragmentDualSrcAttachments", 30},
    {"maxFragmentInputComponents", 27},
    {"maxFragmentOutputAttachments", 29},
    {"maxFragmentShadingRateAttachmentTexelSize", 42},
    {"maxFragmentShadingRateAttachmentTexelSizeAspectRatio", 53},
    {"maxFragmentShadingRateCoverageSamples", 38},
    {"maxFragmentShadingRateInvocationCount", 38},
    {"maxFragmentShadingRateRasterizationSamples", 43},
    {"maxFragmentSize", 16},
    {"maxFragmentSizeAspectRatio", 27},
    {"maxFrameAverageLightLevel", 26},
    {"maxFrameSize", 13},
    {"maxFramebufferHeight", 21},
    {"maxFramebufferLayers", 21},
    {"maxFramebufferWidth", 20},
    {"maxGeometryCount", 17},
    {"maxGeometryIndex", 17},
    {"maxGeometryIndexValue", 22},
    {"maxGeometryInputComponents", 27},
    {"maxGeometryOutputComponents", 28},
    {"maxGeometryOutputVertices", 26},
    {"maxGeometryShaderInvocations", 29},
    {"maxGeometryTotalOutputComponents", 33},
    {"maxGlobalOnlyCounters", 22},
    {"maxGlobalSharedCounters", 24},
    {"maxGraphicsPushConstantBanks", 29},
    {"maxGraphicsPushDataBanks", 25},
    {"maxGraphicsShaderGroupCount", 28},
    {"maxHeight", 10},
    {"maxImageArrayLayers", 20},
    {"maxImageCount", 14},
    {"maxImageDimension1D", 20},
    {"maxImageDimension2D", 20},
    {"maxImageDimension3D", 20},
    {"maxImageDimensionCube", 22},
    {"maxImageExtent", 15},
    {"maxIndirectCommandsIndirectStride", 34},
    {"maxIndirectCommandsStreamCount", 31},
    {"maxIndirectCommandsStreamStride", 32},
    {"maxIndirectCommandsTokenCount", 30},
    {"maxIndirectCommandsTokenOffset", 31},
    {"maxIndirectPipelineCount", 25},
    {"maxIndirectSequenceCount", 25},
    {"maxIndirectShaderObjectCount", 29},
    {"maxInlineUniformBlockBindings", 30},
    {"maxInlineUniformBlockSize", 26},
    {"maxInlineUniformTotalSize", 26},
    {"maxInstanceCount", 17},
    {"maxInstanceInGlobalPartitionCount", 34},
    {"maxInstancePerPartitionCount", 29},
    {"maxInstances", 13},
    {"maxInterpolationOffset", 23},
    {"maxIntraRefreshActiveReferencePictures", 39},
    {"maxIntraRefreshCycleDuration", 29},
    {"maxL1ReferenceCount", 20},
    {"maxLevel", 9},
    {"maxLevelIdc", 12},
    {"maxLod", 7},
    {"maxLuminance", 13},
    {"maxMemoryAllocationCount", 25},
    {"maxMemoryAllocationSize", 24},
    {"maxMeshMultiviewViewCount", 26},
    {"maxMeshOutputComponents", 24},
    {"maxMeshOutputLayers", 20},
    {"maxMeshOutputMemorySize", 24},
    {"maxMeshOutputPrimitives", 24},
    {"maxMeshOutputVertices", 22},
    {"maxMeshPayloadAndOutputMemorySize", 34},
    {"maxMeshPayloadAndSharedMemorySize", 34},
    {"maxMeshSharedMemorySize", 24},
    {"maxMeshTotalMemorySize", 23},
    {"maxMeshWorkGroupCount", 22},
    {"maxMeshWorkGroupInvocations", 28},
    {"maxMeshWorkGroupSize", 21},
    {"maxMeshWorkGroupTotalCount", 27},
    {"maxMicromapTriangles", 21},
    {"maxMipLevels", 13},
    {"maxMovedBytes", 14},
    {"maxMultiDrawCount", 18},
    {"maxMultiviewInstanceIndex", 26},
    {"maxMultiviewViewCount", 22},
    {"maxNumRegionsOfInterest", 24},
    {"maxOpacity2StateSubdivisionLevel", 33},
    {"maxOpacity4StateSubdivisionLevel", 33},
    {"maxOpacityLossy4StateSubdivisionLevel", 38},
    {"maxOperatingPoints", 19},
    {"maxOutputClusterCount", 22},
    {"maxPPictureL0ReferenceCount", 28},
    {"maxPartitionCount", 18},
    {"maxPerDimensionTensorElements", 30},
    {"maxPerPartitionFeedbackEntries", 31},
    {"maxPerRegionPerformanceCounters", 32},
    {"maxPerSetDescriptors", 21},
    {"maxPerStageDescriptorAccelerationStructures", 44},
    {"maxPerStageDescriptorInlineUniformBlocks", 41},
    {"maxPerStageDescriptorInputAttachments", 38},
    {"maxPerStageDescriptorSampledImages", 35},
    {"maxPerStageDescriptorSamplers", 30},
    {"maxPerStageDescriptorSetStorageTensors", 39},
    {"maxPerStageDescriptorStorageBuffers", 36},
    {"maxPerStageDescriptorStorageImages", 35},
    {"maxPerStageDescriptorUniformBuffers", 36},
    {"maxPerStageDescriptorUpdateAfterBindAccelerationStructures", 59},
    {"maxPerStageDescriptorUpdateAfterBindInlineUniformBlocks", 56},
    {"maxPerStageDescriptorUpdateAfterBindInputAttachments", 53},
    {"maxPerStageDescriptorUpdateAfterBindSampledImages", 50},
    {"maxPerStageDescriptorUpdateAfterBindSamplers", 45},
    {"maxPerStageDescriptorUpdateAfterBindStorageBuffers", 51},
    {"maxPerStageDescriptorUpdateAfterBindStorageImages", 50},
    {"maxPerStageDescriptorUpdateAfterBindStorageTensors", 51},
    {"maxPerStageDescriptorUpdateAfterBindUniformBuffers", 51},
    {"maxPerStageResources", 21},
    {"maxPerStageUpdateAfterBindResources", 36},
    {"maxPipelineCount", 17},
    {"maxPipelineRayHitAttributeSize", 31},
    {"maxPipelineRayPayloadSize", 26},
    {"maxPipelineRayRecursionDepth", 29},
    {"maxPreferredMeshWorkGroupInvocations", 37},
    {"maxPreferredTaskWorkGroupInvocations", 37},
    {"maxPrimitiveCount", 18},
    {"maxPrimitiveIndex", 18},
    {"maxPushConstantsSize", 21},
    {"maxPushDataSize", 16},
    {"maxPushDescriptors", 19},
    {"maxQIndex", 10},
    {"maxQIndexDelta", 15},
    {"maxQp", 6},
    {"maxQpDelta", 11},
    {"maxQualityLevels", 17},
    {"maxQuantizationMapExtent", 25},
    {"maxQueuedWorkGroupBatches", 26},
    {"maxRateControlLayers", 21},
    {"maxRayDispatchInvocationCount", 30},
    {"maxRayHitAttributeSize", 23},
    {"maxRayRecursionDepth", 21},
    {"maxRecursionDepth", 18},
    {"maxRenderPassStripes", 21},
    {"maxResourceDescriptorBufferBindings", 36},
    {"maxResourceDescriptorBufferRange", 33},
    {"maxResourceHeapSize", 20},
    {"maxResourceSize", 16},
    {"maxSampleLocationGridSize", 26},
    {"maxSampleMaskWords", 19},
    {"maxSamplerAllocationCount", 26},
    {"maxSamplerAnisotropy", 21},
    {"maxSamplerDescriptorBufferBindings", 35},
    {"maxSamplerDescriptorBufferRange", 32},
    {"maxSamplerHeapSize", 19},
    {"maxSamplerLodBias", 18},
    {"maxScaledImageExtent", 21},
    {"maxSequenceCount", 17},
    {"maxSequencesCount", 18},
    {"maxSets", 8},
    {"maxSgprAllocation", 18},
    {"maxShaderAbortMessageSize", 26},
    {"maxShaderBindingTableRecordIndex", 33},
    {"maxShaderCount", 15},
    {"maxShaderGroupStride", 21},
    {"maxShaderWaitQueues", 20},
    {"maxSingleReferenceCount", 24},
    {"maxSize", 8},
    {"maxSliceCount", 14},
    {"maxSliceSegmentCount", 21},
    {"maxSpatialLayerCount", 21},
    {"maxSqttSeBufferSize", 20},
    {"maxSrcExtent", 13},
    {"maxSrcPosition", 15},
    {"maxStdPPSCount", 15},
    {"maxStdSPSCount", 15},
    {"maxStdVPSCount", 15},
    {"maxStorageBufferRange", 22},
    {"maxStreamingCounters", 21},
    {"maxSubLayerCount", 17},
    {"maxSubgroupSize", 16},
    {"maxSubpassShadingWorkgroupSizeAspectRatio", 42},
    {"maxSubsampledArrayLayers", 25},
    {"maxTaskOutputCount", 19},
    {"maxTaskPayloadAndSharedMemorySize", 34},
    {"maxTaskPayloadSize", 19},
    {"maxTaskSharedMemorySize", 24},
    {"maxTaskTotalMemorySize", 23},
    {"maxTaskWorkGroupCount", 22},
    {"maxTaskWorkGroupInvocations", 28},
    {"maxTaskWorkGroupSize", 21},
    {"maxTaskWorkGroupTotalCount", 27},
    {"maxTemporalLayerCount", 22},
    {"maxTensorDimensionCount", 24},
    {"maxTensorElements", 18},
    {"maxTensorShaderAccessArrayLength", 33},
    {"maxTensorShaderAccessSize", 26},
    {"maxTensorSize", 14},
    {"maxTensorStride", 16},
    {"maxTessellationControlPerPatchOutputComponents", 47},
    {"maxTessellationControlPerVertexInputComponents", 47},
    {"maxTessellationControlPerVertexOutputComponents", 48},
    {"maxTessellationControlTotalOutputComponents", 44},
    {"maxTessellationEvaluationInputComponents", 41},
    {"maxTessellationEvaluationOutputComponents", 42},
    {"maxTessellationGenerationLevel", 31},
    {"maxTessellationPatchSize", 25},
    {"maxTexelBufferElements", 23},
    {"maxTexelGatherOffset", 21},
    {"maxTexelOffset", 15},
    {"maxTileShadingRate", 19},
    {"maxTileSize", 12},
    {"maxTiles", 9},
    {"maxTimelineSemaphoreValueDifference", 36},
    {"maxTotalClusterCount", 21},
    {"maxTotalTriangleCount", 22},
    {"maxTotalVertexCount", 20},
    {"maxTransformFeedbackBufferDataSize", 35},
    {"maxTransformFeedbackBufferDataStride", 37},
    {"maxTransformFeedbackBufferSize", 31},
    {"maxTransformFeedbackBuffers", 28},
    {"maxTransformFeedbackStreamDataSize", 35},
    {"maxTransformFeedbackStreams", 28},
    {"maxTriangleCount", 17},
    {"maxTrianglesPerCluster", 23},
    {"maxUnidirectionalCompoundGroup1ReferenceCount", 46},
    {"maxUnidirectionalCompoundReferenceCount", 40},
    {"maxUniformBufferRange", 22},
    {"maxUpdateAfterBindDescriptorsInAllPools", 40},
    {"maxVariableDescriptorCount", 27},
    {"maxVectorComponents", 20},
    {"maxVertex", 10},
    {"maxVertexAttribDivisor", 23},
    {"maxVertexInputAttributeOffset", 30},
    {"maxVertexInputAttributes", 25},
    {"maxVertexInputBindingStride", 28},
    {"maxVertexInputBindings", 23},
    {"maxVertexOutputComponents", 26},
    {"maxVerticesPerCluster", 22},
    {"maxVgprAllocation", 18},
    {"maxViewportDimensions", 22},
    {"maxViewports", 13},
    {"maxWarpsPerShaderCore", 22},
    {"maxWeightFilterDimension", 25},
    {"maxWeightFilterPhases", 22},
    {"maxWidth", 9},
    {"maxWorkGroupCount", 18},
    {"maxWorkGroupSize", 17},
    {"maxX", 5},
    {"maxY", 5},
    {"maxZ", 5},
    {"maximumRequestedAlignment", 26},
    {"meanFlowL1NormHint", 19},
    {"memory", 7},
    {"memoryBarrierCount", 19},
    {"memoryBindIndex", 16},
    {"memoryClockFrequency", 21},
    {"memoryClockRatioToPeak", 23},
    {"memoryDecompression", 20},
    {"memoryDeviceIndex", 18},
    {"memoryHeapCount", 16},
    {"memoryHeaps", 12},
    {"memoryImageHeight", 18},
    {"memoryMapPlaced", 16},
    {"memoryMapRangePlaced", 21},
    {"memoryObjectId", 15},
    {"memoryOffset", 13},
    {"memoryPriority", 15},
    {"memoryProperties", 17},
    {"memoryRangeBarrierCount", 24},
    {"memoryRangeCount", 17},
    {"memoryRequirements", 19},
    {"memoryRowLength", 16},
    {"memorySize", 11},
    {"memoryTypeBits", 15},
    {"memoryTypeCount", 16},
    {"memoryTypeIndex", 16},
    {"memoryTypes", 12},
    {"memoryUnmapReserve", 19},
    {"meshAndTaskShaderDerivatives", 29},
    {"meshOutputPerPrimitiveGranularity", 34},
    {"meshOutputPerVertexGranularity", 31},
    {"meshShader", 11},
    {"meshShaderQueries", 18},
    {"messageCode", 12},
    {"messageDataSize", 16},
    {"messageIdNumber", 16},
    {"messageSeverity", 16},
    {"messageType", 12},
    {"messageTypes", 13},
    {"messenger", 10},
    {"micromap", 9},
    {"micromapCaptureReplay", 22},
    {"micromapCount", 14},
    {"micromapHostCommands", 21},
    {"micromapSize", 13},
    {"minAccelerationStructureScratchOffsetAlignment", 47},
    {"minBitstreamBufferOffsetAlignment", 34},
    {"minBitstreamBufferSizeAlignment", 32},
    {"minBufferCount", 15},
    {"minBufferCountForCamping", 25},
    {"minBufferCountForDedicatedSlack", 32},
    {"minBufferCountForSharedSlack", 29},
    {"minCodedExtent", 15},
    {"minDepth", 9},
    {"minDepthBounds", 15},
    {"minDepthClamp", 14},
    {"minDstExtent", 13},
    {"minDstPosition", 15},
    {"minFilter", 10},
    {"minFragmentDensityTexelSize", 28},
    {"minFragmentShadingRateAttachmentTexelSize", 42},
    {"minHeight", 10},
    {"minImageCount", 14},
    {"minImageExtent", 15},
    {"minImageTransferGranularity", 28},
    {"minImportedHostPointerAlignment", 32},
    {"minIndirectCommandsBufferOffsetAlignment", 41},
    {"minInterpolationOffset", 23},
    {"minLod", 7},
    {"minLuminance", 13},
    {"minMemoryMapAlignment", 22},
    {"minPlacedMemoryMapAlignment", 28},
    {"minPositionTruncateBitCount", 28},
    {"minQIndex", 10},
    {"minQIndexDelta", 15},
    {"minQp", 6},
    {"minQpDelta", 11},
    {"minResourceHeapReservedRange", 29},
    {"minSampleShading", 17},
    {"minSamplerHeapReservedRange", 28},
    {"minSamplerHeapReservedRangeWithEmbedded", 40},
    {"minScaledImageExtent", 21},
    {"minSequencesCountBufferOffsetAlignment", 39},
    {"minSequencesIndexBufferOffsetAlignment", 39},
    {"minSgprAllocation", 18},
    {"minSize", 8},
    {"minSrcExtent", 13},
    {"minSrcPosition", 15},
    {"minStorageBufferOffsetAlignment", 32},
    {"minSubgroupSize", 16},
    {"minTexelBufferOffsetAlignment", 30},
    {"minTexelGatherOffset", 21},
    {"minTexelOffset", 15},
    {"minTileSize", 12},
    {"minUniformBufferOffsetAlignment", 32},
    {"minVertexInputBindingStrideAlignment", 37},
    {"minVgprAllocation", 18},
    {"minWidth", 9},
    {"minX", 5},
    {"minY", 5},
    {"minZ", 5},
    {"minimumIntervalUs", 18},
    {"minor", 6},
    {"mipLevel", 9},
    {"mipLevels", 10},
    {"mipLodBias", 11},
    {"mipmapMode", 11},
    {"mipmapPrecisionBits", 20},
    {"missShaderBindingOffset", 24},
    {"missShaderBindingStride", 24},
    {"missShaderBindingTableAddress", 30},
    {"missShaderBindingTableBuffer", 29},
    {"missShaderBindingTableSize", 27},
    {"missShaderBindingTableStride", 29},
    {"mode", 5},
    {"modes", 6},
    {"module", 7},
    {"mtlBuffer", 10},
    {"mtlCommandQueue", 16},
    {"mtlDevice", 10},
    {"mtlSharedEvent", 15},
    {"mtlTexture", 11},
    {"multiDraw", 10},
    {"multiDrawIndirect", 18},
    {"multiViewport", 14},
    {"multisampleArrayImage", 22},
    {"multisampledRenderToSingleSampled", 34},
    {"multisampledRenderToSingleSampledEnable", 40},
    {"multisampledRenderToSwapchain", 30},
    {"multiview", 10},
    {"multiviewClusterCullingShader", 30},
    {"multiviewGeometryShader", 24},
    {"multiviewMeshShader", 20},
    {"multiviewPerViewRenderAreas", 28},
    {"multiviewPerViewViewports", 26},
    {"multiviewTessellationShader", 28},
    {"mutableComparisonSamplers", 26},
    {"mutableDescriptorType", 22},
    {"mutableDescriptorTypeListCount", 31},
    {"naluSliceEntryCount", 20},
    {"naluSliceSegmentEntryCount", 27},
    {"name", 5},
    {"nativeUnalignedPerformance", 27},
    {"negativeOneToOne", 17},
    {"nestedCommandBuffer", 20},
    {"nestedCommandBufferRendering", 29},
    {"nestedCommandBufferSimultaneousUse", 35},
    {"newLayout", 10},
    {"nextStage", 10},
    {"noInvocationFragmentShadingRates", 33},
    {"noMoveOverlap", 14},
    {"nodeIndex", 10},
    {"nodeType", 9},
    {"nonCoherentAtomSize", 20},
    {"nonRectangularIntraRefreshRegions", 34},
    {"nonSeamlessCubeMap", 19},
    {"nonStrictSinglePixelWideLinesUseParallelogram", 46},
    {"nonStrictWideLinesUseParallelogram", 35},
    {"nullColorAttachmentWithExternalFormatResolve", 45},
    {"nullDescriptor", 15},
    {"numAABBs", 9},
    {"numAvailableSgprs", 18},
    {"numAvailableVgprs", 18},
    {"numColumns", 11},
    {"numFramesPerBatch", 18},
    {"numMetrics", 11},
    {"numObjects", 11},
    {"numPhases", 10},
    {"numPhysicalSgprs", 17},
    {"numPhysicalVgprs", 17},
    {"numRows", 8},
    {"numTriangles", 13},
    {"numUsedSgprs", 13},
    {"numUsedVgprs", 13},
    {"numVertices", 12},
    {"object", 7},
    {"objectCount", 12},
    {"objectHandle", 13},
    {"objectIndex", 12},
    {"objectType", 11},
    {"occlusionQueryEnable", 21},
    {"occlusionQueryPrecise", 22},
    {"occupancyPriority", 18},
    {"occupancyThrottling", 20},
    {"offset", 7},
    {"oldLayout", 10},
    {"oldSwapchain", 13},
    {"opInput", 8},
    {"opMode", 7},
    {"opType", 7},
    {"opacityMicromapArray", 21},
    {"opacityMicromapIndexBuffer", 27},
    {"opacityMicromapIndexBufferStride", 33},
    {"opacityMicromapIndexType", 25},
    {"opaqueCaptureAddress", 21},
    {"opaqueCaptureDescriptorData", 28},
    {"operation", 10},
    {"operationType", 14},
    {"opticalFlow", 12},
    {"optimal", 8},
    {"optimalBufferCopyOffsetAlignment", 33},
    {"optimalBufferCopyRowPitchAlignment", 35},
    {"optimalDeviceAccess", 20},
    {"optimalImageTransferGranularity", 32},
    {"optimalImageTransferToQueueFamilies", 36},
    {"optimalTilingFeatures", 22},
    {"optimalTilingLayoutUUID", 24},
    {"optimalTilingTensorFeatures", 28},
    {"origin", 7},
    {"osRenderQueueEndTimeUs", 23},
    {"osRenderQueueStartTimeUs", 25},
    {"outputGridSize", 15},
    {"overallocationBehavior", 23},
    {"pAccelerationStructure", 23},
    {"pAccelerationStructures", 24},
    {"pAcquireInfo", 13},
    {"pAcquireKeys", 13},
    {"pAcquireSyncs", 14},
    {"pAcquireTimeoutMilliseconds", 28},
    {"pAcquireTimeouts", 17},
    {"pAddress", 9},
    {"pAddressInfos", 14},
    {"pAddressRange", 14},
    {"pAllocateInfo", 14},
    {"pAllocator", 11},
    {"pApiVersion", 12},
    {"pApplicationInfo", 17},
    {"pApplicationName", 17},
    {"pAspectReferences", 18},
    {"pAttachmentImageInfos", 22},
    {"pAttachmentInitialSampleLocations", 34},
    {"pAttachments", 13},
    {"pAttributes", 12},
    {"pBeginCustomResolveInfo", 24},
    {"pBeginInfo", 11},
    {"pBinaries", 10},
    {"pBindDescriptorBufferEmbeddedSamplersInfo", 42},
    {"pBindDescriptorSetsInfo", 24},
    {"pBindInfo", 10},
    {"pBindInfos", 11},
    {"pBindPointRequirementCount", 27},
    {"pBindPointRequirements", 23},
    {"pBindSessionMemoryInfos", 24},
    {"pBindingFlags", 14},
    {"pBindingInfos", 14},
    {"pBindingReference", 18},
    {"pBindings", 10},
    {"pBinds", 7},
    {"pBlitImageInfo", 15},
    {"pBorderColor", 13},
    {"pBuffer", 8},
    {"pBufferBinds", 13},
    {"pBufferConstraintsInfo", 23},
    {"pBufferIndices", 15},
    {"pBufferInfo", 12},
    {"pBufferMemoryBarriers", 22},
    {"pBuffers", 9},
    {"pBuildInfo", 11},
    {"pCacheData", 11},
    {"pCacheSize", 11},
    {"pCallableShaderBindingTable", 28},
    {"pCallback", 10},
    {"pCallbackData", 14},
    {"pCapabilities", 14},
    {"pCheckpointData", 16},
    {"pCheckpointDataCount", 21},
    {"pCheckpointMarker", 18},
    {"pClearValues", 13},
    {"pClustersBottomLevel", 21},
    {"pCmdBufLabels", 14},
    {"pCode", 6},
    {"pCodingControlInfo", 19},
    {"pCollection", 12},
    {"pColor", 7},
    {"pColorAttachmentFormats", 24},
    {"pColorAttachmentInputIndices", 29},
    {"pColorAttachmentLocations", 26},
    {"pColorAttachmentSamples", 24},
    {"pColorAttachments", 18},
    {"pColorBlendAdvanced", 20},
    {"pColorBlendEnables", 19},
    {"pColorBlendEquations", 21},
    {"pColorBlendState", 17},
    {"pColorSpaces", 13},
    {"pColorWriteEnables", 19},
    {"pColorWriteMasks", 17},
    {"pCombinationCount", 18},
    {"pCombinations", 14},
    {"pCombinedImageSampler", 22},
    {"pCommandBufferDeviceMasks", 26},
    {"pCommandBufferInfos", 20},
    {"pCommandBuffers", 16},
    {"pCommandInfos", 14},
    {"pCommandPool", 13},
    {"pCommittedMemoryInBytes", 24},
    {"pCompatibility", 15},
    {"pConditionalRenderingBegin", 27},
    {"pConfiguration", 15},
    {"pConnections", 13},
    {"pConstantData", 14},
    {"pConstants", 11},
    {"pCopyBufferInfo", 16},
    {"pCopyBufferToImageInfo", 23},
    {"pCopyDstLayouts", 16},
    {"pCopyImageInfo", 15},
    {"pCopyImageToBufferInfo", 23},
    {"pCopyImageToImageInfo", 22},
    {"pCopyImageToMemoryInfo", 23},
    {"pCopyMemoryIndirectInfo", 24},
    {"pCopyMemoryInfo", 16},
    {"pCopyMemoryToImageIndirectInfo", 31},
    {"pCopyMemoryToImageInfo", 23},
    {"pCopySrcLayouts", 16},
    {"pCopyTensorInfo", 16},
    {"pCorrelatedViewMasks", 21},
    {"pCorrelationMasks", 18},
    {"pCountInfo", 11},
    {"pCounterAddresses", 18},
    {"pCounterBufferOffsets", 22},
    {"pCounterBuffers", 16},
    {"pCounterCount", 14},
    {"pCounterDescriptions", 21},
    {"pCounterIndices", 16},
    {"pCounterInfo", 13},
    {"pCounterInfos", 14},
    {"pCounterValue", 14},
    {"pCounters", 10},
    {"pCoverageModulationTable", 25},
    {"pCreateInfo", 12},
    {"pCreateInfos", 13},
    {"pCustomSampleOrders", 20},
    {"pData", 6},
    {"pDataSize", 10},
    {"pDatas", 7},
    {"pDebugInfo", 11},
    {"pDecodeInfo", 12},
    {"pDecompressMemoryInfoEXT", 25},
    {"pDecompressMemoryRegions", 25},
    {"pDeferredOperation", 19},
    {"pDependencies", 14},
    {"pDependencyInfo", 16},
    {"pDependencyInfos", 17},
    {"pDepthAttachment", 17},
    {"pDepthBiasInfo", 15},
    {"pDepthClampRange", 17},
    {"pDepthInputAttachmentIndex", 27},
    {"pDepthStencil", 14},
    {"pDepthStencilAttachment", 24},
    {"pDepthStencilResolveAttachment", 31},
    {"pDepthStencilState", 19},
    {"pDescription", 13},
    {"pDescriptionCount", 18},
    {"pDescriptions", 14},
    {"pDescriptor", 12},
    {"pDescriptorCopies", 18},
    {"pDescriptorCounts", 18},
    {"pDescriptorInfo", 16},
    {"pDescriptorPool", 16},
    {"pDescriptorSets", 16},
    {"pDescriptorTypes", 17},
    {"pDescriptorUpdateEntries", 25},
    {"pDescriptorUpdateTemplate", 26},
    {"pDescriptorWrites", 18},
    {"pDescriptors", 13},
    {"pDevice", 8},
    {"pDeviceEventInfo", 17},
    {"pDeviceGroupPresentCapabilities", 32},
    {"pDeviceIndices", 15},
    {"pDeviceMasks", 13},
    {"pDeviceRenderAreas", 19},
    {"pDimensions", 12},
    {"pDisabledValidationChecks", 26},
    {"pDisabledValidationFeatures", 28},
    {"pDiscardRectangles", 19},
    {"pDispatchParameters", 20},
    {"pDispatchTileInfo", 18},
    {"pDisplay", 9},
    {"pDisplayCount", 14},
    {"pDisplayEventInfo", 18},
    {"pDisplayPlaneInfo", 18},
    {"pDisplayPowerInfo", 18},
    {"pDisplayTimingProperties", 25},
    {"pDisplays", 10},
    {"pDrivers", 9},
    {"pDrmFormatModifierProperties", 29},
    {"pDrmFormatModifiers", 20},
    {"pDstOffset", 11},
    {"pDstRange", 10},
    {"pDstSize", 9},
    {"pDynamicMetadata", 17},
    {"pDynamicOffsets", 16},
    {"pDynamicState", 14},
    {"pDynamicStates", 15},
    {"pEmbeddedSampler", 17},
    {"pEnabledFeatures", 17},
    {"pEnabledValidationFeatures", 27},
    {"pEncodeInfo", 12},
    {"pEndCodingInfo", 15},
    {"pEngineName", 12},
    {"pEvent", 7},
    {"pEvents", 8},
    {"pExclusiveScissorEnables", 25},
    {"pExclusiveScissors", 19},
    {"pExecutableCount", 17},
    {"pExecutableInfo", 16},
    {"pExecuteInfo", 13},
    {"pExecutionSet", 14},
    {"pExecutionSetWrites", 20},
    {"pExtensions", 12},
    {"pExtent", 8},
    {"pExternalBufferInfo", 20},
    {"pExternalBufferProperties", 26},
    {"pExternalFenceInfo", 19},
    {"pExternalFenceProperties", 25},
    {"pExternalImageFormatProperties", 31},
    {"pExternalQueue", 15},
    {"pExternalSemaphoreInfo", 23},
    {"pExternalSemaphoreProperties", 29},
    {"pExternalTensorInfo", 20},
    {"pExternalTensorProperties", 26},
    {"pExtras", 8},
    {"pFaultCounts", 13},
    {"pFaultInfo", 11},
    {"pFd", 4},
    {"pFeatures", 10},
    {"pFeedbackInfo", 14},
    {"pFence", 7},
    {"pFences", 8},
    {"pFixedRateFlags", 16},
    {"pFormatConstraints", 19},
    {"pFormatCount", 13},
    {"pFormatInfo", 12},
    {"pFormatProperties", 18},
    {"pFragmentDensityOffsets", 24},
    {"pFragmentShadingRateAttachment", 31},
    {"pFragmentShadingRateCount", 26},
    {"pFragmentShadingRates", 22},
    {"pFragmentSize", 14},
    {"pFramebuffer", 13},
    {"pFunction", 10},
    {"pGeneratedCommandsInfo", 23},
    {"pGeometries", 12},
    {"pGetFdInfo", 11},
    {"pGetMetalHandleInfo", 20},
    {"pGetWin32HandleInfo", 20},
    {"pGetZirconHandleInfo", 21},
    {"pGpaSampleBeginInfo", 20},
    {"pGpaSession", 12},
    {"pGranularity", 13},
    {"pGroups", 8},
    {"pHandle", 8},
    {"pHitShaderBindingTable", 23},
    {"pHostMapping", 13},
    {"pHostPointer", 13},
    {"pIdentifier", 12},
    {"pImage", 7},
    {"pImageBinds", 12},
    {"pImageConstraintsInfo", 22},
    {"pImageFormatInfo", 17},
    {"pImageFormatProperties", 23},
    {"pImageIndex", 12},
    {"pImageIndices", 14},
    {"pImageInfo", 11},
    {"pImageMemoryBarriers", 21},
    {"pImageOpaqueBinds", 18},
    {"pImageSubresources", 19},
    {"pImages", 8},
    {"pImmutableSamplers", 19},
    {"pImportFenceFdInfo", 19},
    {"pImportFenceWin32HandleInfo", 28},
    {"pImportSemaphoreFdInfo", 23},
    {"pImportSemaphoreWin32HandleInfo", 32},
    {"pImportSemaphoreZirconHandleInfo", 33},
    {"pIndex", 7},
    {"pIndexBuffer", 13},
    {"pIndexInfo", 11},
    {"pIndexTypeValues", 17},
    {"pIndexTypes", 12},
    {"pIndirectCommandsLayout", 24},
    {"pIndirectDeviceAddresses", 25},
    {"pIndirectExecutionSet", 22},
    {"pIndirectStrides", 17},
    {"pInfo", 6},
    {"pInfoSize", 10},
    {"pInfos", 7},
    {"pInheritanceInfo", 17},
    {"pInitialData", 13},
    {"pInitialShaders", 16},
    {"pInitializeInfo", 16},
    {"pInputAssemblyState", 20},
    {"pInputAttachmentImage", 22},
    {"pInputAttachmentIndexInfo", 26},
    {"pInputAttachments", 18},
    {"pInstance", 10},
    {"pInstrumentation", 17},
    {"pInternalRepresentationCount", 29},
    {"pInternalRepresentations", 25},
    {"pKeysAndDataInfo", 17},
    {"pLabelInfo", 11},
    {"pLabelName", 11},
    {"pLatencyMarkerInfo", 19},
    {"pLaunchInfo", 12},
    {"pLayer", 7},
    {"pLayerName", 11},
    {"pLayerPrefix", 13},
    {"pLayeredApis", 13},
    {"pLayers", 8},
    {"pLayout", 8},
    {"pLayoutSizeInBytes", 19},
    {"pLibraries", 11},
    {"pLibraryInfo", 13},
    {"pLibraryInterface", 18},
    {"pLocationInfo", 14},
    {"pLowLatencyMode", 16},
    {"pMapEntries", 12},
    {"pMappings", 10},
    {"pMarkerInfo", 12},
    {"pMarkerName", 12},
    {"pMaxDeviation", 14},
    {"pMaxPrimitiveCounts", 20},
    {"pMaxWorkgroupSize", 18},
    {"pMemory", 8},
    {"pMemoryBarriers", 16},
    {"pMemoryFdProperties", 20},
    {"pMemoryGetRemoteAddressInfo", 28},
    {"pMemoryHostPointerProperties", 29},
    {"pMemoryMapInfo", 15},
    {"pMemoryMetalHandleProperties", 29},
    {"pMemoryProperties", 18},
    {"pMemoryRangeBarriers", 21},
    {"pMemoryRanges", 14},
    {"pMemoryRequirements", 20},
    {"pMemoryRequirementsCount", 25},
    {"pMemoryUnmapInfo", 17},
    {"pMemoryWin32HandleProperties", 29},
    {"pMemoryZirconHandleProperties", 30},
    {"pMessage", 9},
    {"pMessageData", 13},
    {"pMessageIdName", 15},
    {"pMessenger", 11},
    {"pMetadata", 10},
    {"pMetalObjectsInfo", 18},
    {"pMetricBlockCount", 18},
    {"pMetricValues", 14},
    {"pMicromap", 10},
    {"pMicromaps", 11},
    {"pMissShaderBindingTable", 24},
    {"pMode", 6},
    {"pModes", 7},
    {"pModule", 8},
    {"pMoveObjects", 13},
    {"pMultisampleProperties", 23},
    {"pMultisampleState", 18},
    {"pMutableDescriptorTypeLists", 28},
    {"pNaluSliceEntries", 18},
    {"pNaluSliceSegmentEntries", 25},
    {"pName", 6},
    {"pNameInfo", 10},
    {"pNext", 6},
    {"pNodeIndex", 11},
    {"pNodeInfo", 10},
    {"pNumPasses", 11},
    {"pObjectName", 12},
    {"pObjects", 9},
    {"pOffset", 8},
    {"pOffsets", 9},
    {"pOperation", 11},
    {"pOpticalFlowImageFormatInfo", 28},
    {"pOverrideInfo", 14},
    {"pParameters", 12},
    {"pParametersAddInfo", 19},
    {"pParams", 8},
    {"pPastPresentationTimingInfo", 28},
    {"pPastPresentationTimingProperties", 34},
    {"pPeerMemoryFeatures", 20},
    {"pPerTileBeginInfo", 18},
    {"pPerTileEndInfo", 16},
    {"pPerViewRenderAreas", 20},
    {"pPerfBlocks", 12},
    {"pPerfCounters", 14},
    {"pPerfHintInfo", 14},
    {"pPerformanceQueryCreateInfo", 28},
    {"pPhysicalDeviceCount", 21},
    {"pPhysicalDeviceGroupCount", 26},
    {"pPhysicalDeviceGroupProperties", 31},
    {"pPhysicalDevices", 17},
    {"pPictureResource", 17},
    {"pPipelineBinaries", 18},
    {"pPipelineBinaryData", 20},
    {"pPipelineBinaryDataSize", 24},
    {"pPipelineBinaryKey", 19},
    {"pPipelineBinaryKeys", 20},
    {"pPipelineCache", 15},
    {"pPipelineCreateInfo", 20},
    {"pPipelineCreationFeedback", 26},
    {"pPipelineInfo", 14},
    {"pPipelineKey", 13},
    {"pPipelineLayout", 16},
    {"pPipelineProperties", 20},
    {"pPipelineStageCreationFeedbacks", 32},
    {"pPipelines", 11},
    {"pPlacedAddress", 15},
    {"pPlaneLayouts", 14},
    {"pPoolSizes", 11},
    {"pPostSubpassSampleLocations", 28},
    {"pPresentIds", 12},
    {"pPresentInfo", 13},
    {"pPresentModeCount", 18},
    {"pPresentModes", 14},
    {"pPresentStages", 15},
    {"pPresentWait2Info", 18},
    {"pPresentationInfo", 18},
    {"pPresentationTimingCount", 25},
    {"pPresentationTimings", 21},
    {"pPreserveAttachments", 21},
    {"pPrivateData", 13},
    {"pPrivateDataSlot", 17},
    {"pProcessingEngines", 19},
    {"pProfiles", 10},
    {"pProperties", 12},
    {"pPropertiesCount", 17},
    {"pPropertyCount", 15},
    {"pPushConstant", 14},
    {"pPushConstantRanges", 20},
    {"pPushConstantsInfo", 19},
    {"pPushDataInfo", 14},
    {"pPushDescriptorSetInfo", 23},
    {"pPushDescriptorSetWithTemplateInfo", 35},
    {"pQualityLevelInfo", 18},
    {"pQualityLevelProperties", 24},
    {"pQueriedLowLatencyData", 23},
    {"pQueryPool", 11},
    {"pQueue", 7},
    {"pQueueCreateInfos", 18},
    {"pQueueFamilyDataGraphProcessingEngineInfo", 42},
    {"pQueueFamilyDataGraphProcessingEngineProperties", 48},
    {"pQueueFamilyDataGraphProperties", 32},
    {"pQueueFamilyDataGraphPropertyCount", 35},
    {"pQueueFamilyIndices", 20},
    {"pQueueFamilyProperties", 23},
    {"pQueueFamilyPropertyCount", 26},
    {"pQueueInfo", 11},
    {"pQueueLabels", 13},
    {"pQueuePriorities", 17},
    {"pQueueTypeInfo", 15},
    {"pRanges", 8},
    {"pRasterizationState", 20},
    {"pRaygenShaderBindingTable", 26},
    {"pRectCount", 11},
    {"pRectangles", 12},
    {"pRects", 7},
    {"pReferenceSlots", 16},
    {"pRegions", 9},
    {"pReleaseInfo", 13},
    {"pReleaseKeys", 13},
    {"pReleaseSyncs", 14},
    {"pRenderPass", 12},
    {"pRenderPassBegin", 17},
    {"pRenderPassFeedback", 20},
    {"pRenderingAreaInfo", 19},
    {"pRenderingEndInfo", 18},
    {"pRenderingInfo", 15},
    {"pResolveAttachments", 20},
    {"pResolveImageInfo", 18},
    {"pResourceHeapBindInfo", 22},
    {"pResourceInfos", 15},
    {"pResources", 11},
    {"pResult", 8},
    {"pResults", 9},
    {"pSampleID", 10},
    {"pSampleLocations", 17},
    {"pSampleLocationsInfo", 21},
    {"pSampleMask", 12},
    {"pSampledImage", 14},
    {"pSampler", 9},
    {"pSamplerHeapBindInfo", 21},
    {"pSamplers", 10},
    {"pScissors", 10},
    {"pSemaphore", 11},
    {"pSemaphores", 12},
    {"pSession", 9},
    {"pSetDescriptorBufferOffsetsInfo", 32},
    {"pSetLayout", 11},
    {"pSetLayoutInfos", 16},
    {"pSetLayouts", 12},
    {"pSettingName", 13},
    {"pSettings", 10},
    {"pSetupReferenceSlot", 20},
    {"pShaderGroupCaptureReplayHandle", 32},
    {"pShaderInfo", 12},
    {"pShaderModule", 14},
    {"pShaders", 9},
    {"pShadingRatePaletteEntries", 27},
    {"pShadingRatePalettes", 21},
    {"pSignalInfo", 12},
    {"pSignalSemaphoreDeviceIndices", 30},
    {"pSignalSemaphoreInfos", 22},
    {"pSignalSemaphoreValues", 23},
    {"pSignalSemaphores", 18},
    {"pSizeInBytes", 13},
    {"pSizeInfo", 10},
    {"pSizes", 7},
    {"pSleepInfo", 11},
    {"pSleepModeInfo", 15},
    {"pSliceOffsets", 14},
    {"pSliceSegmentOffsets", 21},
    {"pSparseMemoryRequirementCount", 30},
    {"pSparseMemoryRequirements", 26},
    {"pSpecializationInfo", 20},
    {"pSplitInstanceBindRegions", 26},
    {"pSrcCaches", 11},
    {"pSrcOffset", 11},
    {"pStages", 8},
    {"pStatisticCount", 16},
    {"pStatistics", 12},
    {"pStdDecoderModelInfo", 21},
    {"pStdHeaderVersion", 18},
    {"pStdOperatingPoints", 20},
    {"pStdPPS", 8},
    {"pStdPPSs", 9},
    {"pStdPictureInfo", 16},
    {"pStdReferenceInfo", 18},
    {"pStdSPS", 8},
    {"pStdSPSs", 9},
    {"pStdSequenceHeader", 19},
    {"pStdSliceHeader", 16},
    {"pStdSliceSegmentHeader", 23},
    {"pStdVPS", 8},
    {"pStdVPSs", 9},
    {"pStencilAttachment", 19},
    {"pStencilInputAttachmentIndex", 29},
    {"pStorageBuffer", 15},
    {"pStorageImage", 14},
    {"pStorageTexelBuffer", 20},
    {"pStreamStrides", 15},
    {"pStreams", 9},
    {"pStrides", 9},
    {"pStripeInfos", 13},
    {"pStripeSemaphoreInfos", 22},
    {"pSubmits", 9},
    {"pSubpassBeginInfo", 18},
    {"pSubpassEndInfo", 16},
    {"pSubpassFeedback", 17},
    {"pSubpasses", 11},
    {"pSubresource", 13},
    {"pSupport", 9},
    {"pSupported", 11},
    {"pSurface", 9},
    {"pSurfaceCapabilities", 21},
    {"pSurfaceFormatCount", 20},
    {"pSurfaceFormats", 16},
    {"pSurfaceInfo", 13},
    {"pSwapchain", 11},
    {"pSwapchainImageCount", 21},
    {"pSwapchainImages", 17},
    {"pSwapchainTimeDomainProperties", 31},
    {"pSwapchainTimingProperties", 27},
    {"pSwapchainTimingPropertiesCounter", 34},
    {"pSwapchains", 12},
    {"pTag", 5},
    {"pTagInfo", 9},
    {"pTensor", 8},
    {"pTensorARM", 11},
    {"pTensorMemoryBarriers", 22},
    {"pTensorViews", 13},
    {"pTensors", 9},
    {"pTessellationState", 19},
    {"pTexelBuffer", 13},
    {"pTexelBufferView", 17},
    {"pTileMemoryBindInfo", 20},
    {"pTileOffsets", 13},
    {"pTileSizes", 11},
    {"pTimeDomainCount", 17},
    {"pTimeDomainIds", 15},
    {"pTimeDomains", 13},
    {"pTimeDomainsCounter", 20},
    {"pTimes", 7},
    {"pTimestampInfos", 16},
    {"pTimestamps", 12},
    {"pTimingInfos", 13},
    {"pTimings", 9},
    {"pTokens", 8},
    {"pToolCount", 11},
    {"pToolProperties", 16},
    {"pTransitions", 13},
    {"pTriangleClusters", 18},
    {"pUniformBuffer", 15},
    {"pUniformTexelBuffer", 20},
    {"pUpdateInfo", 12},
    {"pUsageCounts", 13},
    {"pUserData", 10},
    {"pValidationCache", 17},
    {"pValue", 7},
    {"pValues", 8},
    {"pVendorBinaryData", 18},
    {"pVendorInfos", 13},
    {"pVendorOptions", 15},
    {"pVersionData", 13},
    {"pVersionInfo", 13},
    {"pVertexAttributeDescriptions", 29},
    {"pVertexBindingDescriptions", 27},
    {"pVertexBindingDivisors", 23},
    {"pVertexBuffer", 14},
    {"pVertexInfo", 12},
    {"pVertexInputState", 18},
    {"pVertexOffset", 14},
    {"pVideoFormatInfo", 17},
    {"pVideoFormatProperties", 23},
    {"pVideoFormatPropertyCount", 26},
    {"pVideoProfile", 14},
    {"pVideoSession", 14},
    {"pVideoSessionParameters", 24},
    {"pVideoSessionParametersInfo", 28},
    {"pView", 6},
    {"pViewFormats", 13},
    {"pViewMasks", 11},
    {"pViewOffsets", 13},
    {"pViewportDepths", 16},
    {"pViewportState", 15},
    {"pViewportSwizzles", 18},
    {"pViewportWScalings", 19},
    {"pViewports", 11},
    {"pWaitDstStageMask", 18},
    {"pWaitInfo", 10},
    {"pWaitSemaphoreDeviceIndices", 28},
    {"pWaitSemaphoreInfos", 20},
    {"pWaitSemaphoreValues", 21},
    {"pWaitSemaphores", 16},
    {"pYcbcrConversion", 17},
    {"pZirconHandle", 14},
    {"pageableDeviceLocalMemory", 26},
    {"paramCount", 11},
    {"parameter", 10},
    {"parameters", 11},
    {"params", 7},
    {"partitionCount", 15},
    {"partitionIndependentIntraRefreshRegions", 40},
    {"partitionIndex", 15},
    {"partitionTranslation", 21},
    {"partitionedAccelerationStructure", 33},
    {"passOp", 7},
    {"patch", 6},
    {"patchControlPoints", 19},
    {"payloadCount", 13},
    {"payloadStride", 14},
    {"payloads", 9},
    {"pciBus", 7},
    {"pciDevice", 10},
    {"pciDomain", 10},
    {"pciFunction", 12},
    {"perBasicBlockGranularity", 25},
    {"perPartitionEncodeFeedbackFlags", 32},
    {"perStageDescriptorSet", 22},
    {"perViewAttributes", 18},
    {"perViewAttributesPositionXOnly", 31},
    {"perViewPositionAllComponents", 29},
    {"perViewRenderAreaCount", 23},
    {"perfBlockCount", 15},
    {"perfCounterCount", 17},
    {"perfCounterDeviceMemoryLimit", 29},
    {"perfCounters", 13},
    {"performEncodeRgbConversion", 27},
    {"performanceCounterMultipleQueryPools", 37},
    {"performanceCounterQueryPools", 29},
    {"performanceCounterRegionSize", 29},
    {"performanceCountersByRegion", 28},
    {"performanceCountersSampling", 28},
    {"performanceLevel", 17},
    {"persistent", 11},
    {"persistentContent", 18},
    {"pfnAllocation", 14},
    {"pfnCallback", 12},
    {"pfnFree", 8},
    {"pfnGetInstanceProcAddr", 23},
    {"pfnInternalAllocation", 22},
    {"pfnInternalFree", 16},
    {"pfnReallocation", 16},
    {"pfnUserCallback", 16},
    {"physicalDevice", 15},
    {"physicalDeviceCount", 20},
    {"physicalDevices", 16},
    {"physicalDimensions", 19},
    {"physicalResolution", 19},
    {"pictureAccessGranularity", 25},
    {"pictureFormat", 14},
    {"pictureLayout", 14},
    {"pipeline", 9},
    {"pipelineAddress", 16},
    {"pipelineBinaries", 17},
    {"pipelineBinary", 15},
    {"pipelineBinaryCompressedData", 29},
    {"pipelineBinaryCount", 20},
    {"pipelineBinaryInternalCache", 28},
    {"pipelineBinaryInternalCacheControl", 35},
    {"pipelineBinaryPrecompiledInternalCache", 39},
    {"pipelineBinaryPrefersInternalCache", 35},
    {"pipelineBindPoint", 18},
    {"pipelineCache", 14},
    {"pipelineCacheIncrementalMode", 29},
    {"pipelineCacheUUID", 18},
    {"pipelineCount", 14},
    {"pipelineCreationCacheControl", 29},
    {"pipelineDeviceAddressCaptureReplay", 35},
    {"pipelineExecutableInfo", 23},
    {"pipelineFragmentShadingRate", 28},
    {"pipelineIdentifier", 19},
    {"pipelineLayout", 15},
    {"pipelineLibraryGroupHandles", 28},
    {"pipelineOpacityMicromap", 24},
    {"pipelinePropertiesIdentifier", 29},
    {"pipelineProtectedAccess", 24},
    {"pipelineRobustness", 19},
    {"pipelineStackSize", 18},
    {"pipelineStage", 14},
    {"pipelineStageCreationFeedbackCount", 35},
    {"pipelineStatistics", 19},
    {"pipelineStatisticsQuery", 24},
    {"pixelRate", 10},
    {"pixelX", 7},
    {"pixelY", 7},
    {"plane", 6},
    {"planeAspect", 12},
    {"planeIndex", 11},
    {"planeReorderPossible", 21},
    {"planeStackIndex", 16},
    {"pointClippingBehavior", 22},
    {"pointPolygons", 14},
    {"pointSizeGranularity", 21},
    {"pointSizeRange", 15},
    {"polygonMode", 12},
    {"polygonModePointSize", 21},
    {"poolSizeCount", 14},
    {"positionTruncateBitCount", 25},
    {"postMergeIndex", 15},
    {"postMergeSubpassCount", 22},
    {"postSubpassSampleLocationsCount", 32},
    {"powerState", 11},
    {"ppBuildRangeInfos", 18},
    {"ppData", 7},
    {"ppEnabledExtensionNames", 24},
    {"ppEnabledLayerNames", 20},
    {"ppGeometries", 13},
    {"ppMaxPrimitiveCounts", 21},
    {"ppUsageCounts", 14},
    {"preTransform", 13},
    {"precedingExternallyEncodedBytes", 32},
    {"predictionMode", 15},
    {"predictiveFrameSize", 20},
    {"predictiveQIndex", 17},
    {"preferNonCoherent", 18},
    {"preferredBidirectionalCompoundReferenceNameMask", 48},
    {"preferredConsecutiveBFrameCount", 32},
    {"preferredConsecutiveBipredictiveFrameCount", 43},
    {"preferredConstantQIndex", 24},
    {"preferredConstantQp", 20},
    {"preferredGopFrameCount", 23},
    {"preferredIdrPeriod", 19},
    {"preferredKeyFramePeriod", 24},
    {"preferredMaxBidirectionalCompoundGroup1ReferenceCount", 54},
    {"preferredMaxBidirectionalCompoundGroup2ReferenceCount", 54},
    {"preferredMaxBidirectionalCompoundReferenceCount", 48},
    {"preferredMaxL0ReferenceCount", 29},
    {"preferredMaxL1ReferenceCount", 29},
    {"preferredMaxSingleReferenceCount", 33},
    {"preferredMaxUnidirectionalCompoundGroup1ReferenceCount", 55},
    {"preferredMaxUnidirectionalCompoundReferenceCount", 49},
    {"preferredQueue", 15},
    {"preferredRateControlFlags", 26},
    {"preferredRateControlLayerCount", 31},
    {"preferredRateControlMode", 25},
    {"preferredSingleReferenceNameMask", 33},
    {"preferredStdEntropyCodingModeFlag", 34},
    {"preferredSubLayerCount", 23},
    {"preferredTemporalLayerCount", 28},
    {"preferredUnidirectionalCompoundReferenceNameMask", 49},
    {"prefersCompactPrimitiveOutput", 30},
    {"prefersCompactVertexOutput", 27},
    {"prefersDedicatedAllocation", 27},
    {"prefersGopRemainingFrames", 26},
    {"prefersLocalInvocationPrimitiveOutput", 38},
    {"prefersLocalInvocationVertexOutput", 35},
    {"preprocessAddress", 18},
    {"preprocessBuffer", 17},
    {"preprocessOffset", 17},
    {"preprocessSize", 15},
    {"presentAtAbsoluteTime", 22},
    {"presentAtAbsoluteTimeSupported", 31},
    {"presentAtRelativeTime", 22},
    {"presentAtRelativeTimeSupported", 31},
    {"presentBarrier", 15},
    {"presentBarrierEnable", 21},
    {"presentBarrierSupported", 24},
    {"presentConfigFeedback", 22},
    {"presentEndTimeUs", 17},
    {"presentGravityX", 16},
    {"presentGravityY", 16},
    {"presentID", 10},
    {"presentId", 10},
    {"presentId2", 11},
    {"presentId2Supported", 20},
    {"presentMargin", 14},
    {"presentMask", 12},
    {"presentMetering", 16},
    {"presentMode", 12},
    {"presentModeCount", 17},
    {"presentModeFifoLatestReady", 27},
    {"presentStage", 13},
    {"presentStageCount", 18},
    {"presentStageQueries", 20},
    {"presentStartTimeUs", 19},
    {"presentTiming", 14},
    {"presentTimingSupported", 23},
    {"presentWait", 12},
    {"presentWait2", 13},
    {"presentWait2Supported", 22},
    {"presentationTimingCount", 24},
    {"preserveAttachmentCount", 24},
    {"primaryMajor", 13},
    {"primaryMinor", 13},
    {"primaryReferenceCdfOnly", 24},
    {"primitiveCount", 15},
    {"primitiveFragmentShadingRate", 29},
    {"primitiveFragmentShadingRateMeshShader", 39},
    {"primitiveFragmentShadingRateWithMultipleViewports", 50},
    {"primitiveOffset", 16},
    {"primitiveOverestimationSize", 28},
    {"primitiveRestartEnable", 23},
    {"primitiveRestartIndex", 22},
    {"primitiveTopology", 18},
    {"primitiveTopologyListRestart", 29},
    {"primitiveTopologyPatchListRestart", 34},
    {"primitiveUnderestimation", 25},
    {"primitivesGeneratedQuery", 25},
    {"primitivesGeneratedQueryWithNonZeroStreams", 43},
    {"primitivesGeneratedQueryWithRasterizerDiscard", 46},
    {"priorities", 11},
    {"priority", 9},
    {"priorityCount", 14},
    {"privateData", 12},
    {"privateDataSlot", 16},
    {"privateDataSlotRequestCount", 28},
    {"processingEngineCount", 22},
    {"profileCount", 13},
    {"properties", 11},
    {"propertiesCount", 16},
    {"property", 9},
    {"propertyFlags", 14},
    {"protectedDescriptorHeaps", 25},
    {"protectedMemory", 16},
    {"protectedNoFault", 17},
    {"protectedSubmit", 16},
    {"provokingVertexLast", 20},
    {"provokingVertexMode", 20},
    {"provokingVertexModePerPipeline", 31},
    {"purposes", 9},
    {"pushAddressOffset", 18},
    {"pushConstantBank", 17},
    {"pushConstantRangeCount", 23},
    {"pushDataOffset", 15},
    {"pushDataSize", 13},
    {"pushDescriptor", 15},
    {"pushIndex", 10},
    {"pushOffset", 11},
    {"pushconstantOffset", 19},
    {"pushconstantPipelineLayout", 27},
    {"pushconstantShaderStageFlags", 29},
    {"pushconstantSize", 17},
    {"pvx", 4},
    {"pvy", 4},
    {"pvz", 4},
    {"qpB", 4},
    {"qpI", 4},
    {"qpP", 4},
    {"quadDivergentImplicitLod", 25},
    {"quadOperationsInAllStages", 26},
    {"qualityFlags", 13},
    {"qualityLevel", 13},
    {"quantizationMap", 16},
    {"quantizationMapExtent", 22},
    {"quantizationMapTexelSize", 25},
    {"query", 6},
    {"queryCount", 11},
    {"queryFlags", 11},
    {"queryPool", 10},
    {"queryResultFlags", 17},
    {"queryResultStatusSupport", 25},
    {"queryType", 10},
    {"queue", 6},
    {"queueCount", 11},
    {"queueCreateInfoCount", 21},
    {"queueFamilyIndex", 17},
    {"queueFamilyIndexCount", 22},
    {"queueFamilyProperties", 22},
    {"queueFlags", 11},
    {"queueIndex", 11},
    {"queueLabelCount", 16},
    {"queuePerfHint", 14},
    {"queueSubmitBoundary", 20},
    {"queueType", 10},
    {"qw", 3},
    {"qx", 3},
    {"qy", 3},
    {"qz", 3},
    {"r", 2},
    {"radiusData", 11},
    {"radiusFormat", 13},
    {"radiusStride", 13},
    {"range", 6},
    {"rangeCount", 11},
    {"rasterizationOrder", 19},
    {"rasterizationOrderColorAttachmentAccess", 40},
    {"rasterizationOrderDepthAttachmentAccess", 40},
    {"rasterizationOrderStencilAttachmentAccess", 42},
    {"rasterizationSamples", 21},
    {"rasterizationStream", 20},
    {"rasterizerDiscardEnable", 24},
    {"rateControlGroup", 17},
    {"rateControlMode", 16},
    {"rateControlModes", 17},
    {"rayQuery", 9},
    {"rayTracingInvocationReorder", 28},
    {"rayTracingInvocationReorderReorderingHint", 42},
    {"rayTracingMaintenance1", 23},
    {"rayTracingMotionBlur", 21},
    {"rayTracingMotionBlurPipelineTraceRaysIndirect", 46},
    {"rayTracingPipeline", 19},
    {"rayTracingPipelineShaderGroupHandleCaptureReplay", 49},
    {"rayTracingPipelineShaderGroupHandleCaptureReplayMixed", 54},
    {"rayTracingPipelineTraceRaysIndirect", 36},
    {"rayTracingPipelineTraceRaysIndirect2", 37},
    {"rayTracingPositionFetch", 24},
    {"rayTracingValidation", 21},
    {"rayTraversalPrimitiveCulling", 29},
    {"raygenShaderBindingOffset", 26},
    {"raygenShaderBindingTableBuffer", 31},
    {"raygenShaderRecordAddress", 26},
    {"raygenShaderRecordSize", 23},
    {"rect", 5},
    {"rectCount", 10},
    {"rectangleCount", 15},
    {"rectangularLines", 17},
    {"reductionMode", 14},
    {"reference", 10},
    {"referenceNameSlotIndices", 25},
    {"referencePictureFormat", 23},
    {"referenceSlotCount", 19},
    {"refreshDuration", 16},
    {"refreshInterval", 16},
    {"refreshRate", 12},
    {"regionAlignment", 16},
    {"regionCount", 12},
    {"relaxedLineRasterization", 25},
    {"releaseCount", 13},
    {"remoteDeviceIndex", 18},
    {"renderArea", 11},
    {"renderMajor", 12},
    {"renderMinor", 12},
    {"renderPass", 11},
    {"renderPassStripeGranularity", 28},
    {"renderPassStriped", 18},
    {"renderSubmitEndTimeUs", 22},
    {"renderSubmitStartTimeUs", 24},
    {"renderpass", 11},
    {"reportAddressBinding", 21},
    {"reportComplete", 15},
    {"reportedAddress", 16},
    {"representativeFragmentTest", 27},
    {"representativeFragmentTestEnable", 33},
    {"requestIndex", 13},
    {"requiredFormatFeatures", 23},
    {"requiredSubgroupSize", 21},
    {"requiredSubgroupSizeStages", 27},
    {"requiresDedicatedAllocation", 28},
    {"requiresGopRemainingFrames", 27},
    {"reserved", 9},
    {"reservedExternalQueues", 23},
    {"reservedRangeOffset", 20},
    {"reservedRangeSize", 18},
    {"residencyAlignedMipSize", 24},
    {"residencyNonResidentStrict", 27},
    {"residencyStandard2DBlockShape", 30},
    {"residencyStandard2DMultisampleBlockShape", 41},
    {"residencyStandard3DBlockShape", 30},
    {"resolveImageLayout", 19},
    {"resolveImageView", 17},
    {"resolveMode", 12},
    {"resolveSrgbFormatAppliesTransferFunction", 41},
    {"resolveSrgbFormatSupportsTransferFunctionControl", 49},
    {"resourceCount", 14},
    {"resourceDescriptorBufferAddressSpaceSize", 41},
    {"resourceDeviceIndex", 20},
    {"resourceHeapAlignment", 22},
    {"resourceInfoCount", 18},
    {"resourceMask", 13},
    {"resourceOffset", 15},
    {"resourceUsage", 14},
    {"resultIndex", 12},
    {"resultSubIndex", 15},
    {"resultType", 11},
    {"revisionId", 11},
    {"rgbModel", 9},
    {"rgbModels", 10},
    {"rgbRange", 9},
    {"rgbRanges", 10},
    {"rgba4OpaqueBlackSwizzled", 25},
    {"robustBufferAccess", 19},
    {"robustBufferAccess2", 20},
    {"robustBufferAccessUpdateAfterBind", 34},
    {"robustFragmentShadingRateAttachmentAccess", 42},
    {"robustImageAccess", 18},
    {"robustImageAccess2", 19},
    {"robustStorageBufferAccessSizeAlignment", 39},
    {"robustStorageBufferDescriptorSize", 34},
    {"robustStorageTexelBufferDescriptorSize", 39},
    {"robustUniformBufferAccessSizeAlignment", 39},
    {"robustUniformBufferDescriptorSize", 34},
    {"robustUniformTexelBufferDescriptorSize", 39},
    {"roundingModeIndependence", 25},
    {"rowPitch", 9},
    {"rowStrideAlignment", 19},
    {"rrOutput", 9},
    {"runtimeDescriptorArray", 23},
    {"sType", 6},
    {"sample", 7},
    {"sampleCount", 12},
    {"sampleCounts", 13},
    {"sampleID", 9},
    {"sampleInternalOperations", 25},
    {"sampleLocationCoordinateRange", 30},
    {"sampleLocationCount", 20},
    {"sampleLocationGridSize", 23},
    {"sampleLocationSampleCounts", 27},
    {"sampleLocationSubPixelBits", 27},
    {"sampleLocationsCount", 21},
    {"sampleLocationsEnable", 22},
    {"sampleLocationsInfo", 20},
    {"sampleLocationsPerPixel", 24},
    {"sampleOrderType", 16},
    {"sampleRateShading", 18},
    {"sampleShadingEnable", 20},
    {"sampleType", 11},
    {"sampledImageColorSampleCounts", 30},
    {"sampledImageDepthSampleCounts", 30},
    {"sampledImageDescriptorSize", 27},
    {"sampledImageIntegerSampleCounts", 32},
    {"sampledImageStencilSampleCounts", 32},
    {"sampler", 8},
    {"sampler2DViewOf3D", 18},
    {"samplerAddressOffset", 21},
    {"samplerAnisotropy", 18},
    {"samplerCaptureReplayDescriptorDataSize", 39},
    {"samplerCount", 13},
    {"samplerDescriptorAlignment", 27},
    {"samplerDescriptorBufferAddressSpaceSize", 40},
    {"samplerDescriptorSize", 22},
    {"samplerFilterMinmax", 20},
    {"samplerHeapAlignment", 21},
    {"samplerHeapArrayStride", 23},
    {"samplerHeapIndexStride", 23},
    {"samplerHeapOffset", 18},
    {"samplerIndex", 13},
    {"samplerMipLodBias", 18},
    {"samplerMirrorClampToEdge", 25},
    {"samplerPushOffset", 18},
    {"samplerShaderRecordOffset", 26},
    {"samplerYcbcrConversion", 23},
    {"samplerYcbcrConversionComponents", 33},
    {"samplerYcbcrConversionCount", 28},
    {"samples", 8},
    {"saturatingAccumulation", 23},
    {"scalarBlockLayout", 18},
    {"scale", 6},
    {"scalingBehavior", 16},
    {"schedulingControls", 19},
    {"schedulingControlsFlags", 24},
    {"schedulingControlsMaxQueuedBatchesCount", 40},
    {"schedulingControlsMaxWarpsCount", 32},
    {"schedulingControlsMaxWorkGroupBatchSize", 40},
    {"scissorCount", 13},
    {"scope", 6},
    {"scratch", 8},
    {"scratchData", 12},
    {"scratchMemUsageInBytes", 23},
    {"scratchOffset", 14},
    {"scratchSize", 12},
    {"screenBufferImport", 19},
    {"screenUsage", 12},
    {"secondaryCopySource", 20},
    {"selectableCubicWeights", 23},
    {"semaphore", 10},
    {"semaphoreCount", 15},
    {"semaphoreType", 14},
    {"separateDepthStencilAttachmentAccess", 37},
    {"separateDepthStencilLayouts", 28},
    {"separateStencilMaskRef", 23},
    {"sequenceCountAddress", 21},
    {"sequencesCount", 15},
    {"sequencesCountBuffer", 21},
    {"sequencesCountOffset", 21},
    {"sequencesIndexBuffer", 21},
    {"sequencesIndexOffset", 21},
    {"serializeRegions", 17},
    {"session", 8},
    {"set", 4},
    {"setCount", 9},
    {"setLayoutCount", 15},
    {"setStride", 10},
    {"settingCount", 13},
    {"sgprAllocationGranularity", 26},
    {"sgprsPerSimd", 13},
    {"shader", 7},
    {"shader64BitIndexing", 20},
    {"shaderAbort", 12},
    {"shaderArraysPerEngineCount", 27},
    {"shaderBFloat16CooperativeMatrix", 32},
    {"shaderBFloat16DotProduct", 25},
    {"shaderBFloat16Type", 19},
    {"shaderBinaryUUID", 17},
    {"shaderBinaryVersion", 20},
    {"shaderBufferFloat16AtomicAdd", 29},
    {"shaderBufferFloat16AtomicMinMax", 32},
    {"shaderBufferFloat16Atomics", 27},
    {"shaderBufferFloat32AtomicAdd", 29},
    {"shaderBufferFloat32AtomicMinMax", 32},
    {"shaderBufferFloat32Atomics", 27},
    {"shaderBufferFloat64AtomicAdd", 29},
    {"shaderBufferFloat64AtomicMinMax", 32},
    {"shaderBufferFloat64Atomics", 27},
    {"shaderBufferInt64Atomics", 25},
    {"shaderClipDistance", 19},
    {"shaderConstantData", 19},
    {"shaderCoreBuiltins", 19},
    {"shaderCoreCount", 16},
    {"shaderCoreFeatures", 19},
    {"shaderCoreMask", 15},
    {"shaderCount", 12},
    {"shaderCullDistance", 19},
    {"shaderDemoteToHelperInvocation", 31},
    {"shaderDenormFlushToZeroFloat16", 31},
    {"shaderDenormFlushToZeroFloat32", 31},
    {"shaderDenormFlushToZeroFloat64", 31},
    {"shaderDenormPreserveFloat16", 28},
    {"shaderDenormPreserveFloat32", 28},
    {"shaderDenormPreserveFloat64", 28},
    {"shaderDeviceClock", 18},
    {"shaderDrawParameters", 21},
    {"shaderEarlyAndLateFragmentTests", 32},
    {"shaderEngineCount", 18},
    {"shaderEnqueue", 14},
    {"shaderExpectAssume", 19},
    {"shaderFloat16", 14},
    {"shaderFloat16VectorAtomics", 27},
    {"shaderFloat4", 13},
    {"shaderFloat6", 13},
    {"shaderFloat64", 14},
    {"shaderFloat8", 13},
    {"shaderFloat8CooperativeMatrix", 30},
    {"shaderFloat8UnsignedE8M0", 25},
    {"shaderFloatControls2", 21},
    {"shaderFmaFloat16", 17},
    {"shaderFmaFloat32", 17},
    {"shaderFmaFloat64", 17},
    {"shaderGroupBaseAlignment", 25},
    {"shaderGroupHandleAlignment", 27},
    {"shaderGroupHandleCaptureReplaySize", 35},
    {"shaderGroupHandleSize", 22},
    {"shaderImageFloat32AtomicAdd", 28},
    {"shaderImageFloat32AtomicMinMax", 31},
    {"shaderImageFloat32Atomics", 26},
    {"shaderImageGatherExtended", 26},
    {"shaderImageInt64Atomics", 24},
    {"shaderInputAttachmentArrayDynamicIndexing", 42},
    {"shaderInputAttachmentArrayNonUniformIndexing", 45},
    {"shaderInputAttachmentArrayNonUniformIndexingNative", 51},
    {"shaderInstrumentation", 22},
    {"shaderInt16", 12},
    {"shaderInt64", 12},
    {"shaderInt8", 11},
    {"shaderIntegerDotProduct", 24},
    {"shaderIntegerFunctions2", 24},
    {"shaderMXInt8", 13},
    {"shaderMaximalReconvergence", 27},
    {"shaderMeshEnqueue", 18},
    {"shaderMixedFloatDotProductBFloat16Acc", 38},
    {"shaderMixedFloatDotProductFloat16AccFloat16", 44},
    {"shaderMixedFloatDotProductFloat16AccFloat32", 44},
    {"shaderMixedFloatDotProductFloat8AccFloat32", 43},
    {"shaderModule", 13},
    {"shaderModuleIdentifier", 23},
    {"shaderModuleIdentifierAlgorithmUUID", 36},
    {"shaderMultipleWaitQueues", 25},
    {"shaderObject", 13},
    {"shaderOutputLayer", 18},
    {"shaderOutputViewportIndex", 26},
    {"shaderQuadControl", 18},
    {"shaderRawAccessChains", 22},
    {"shaderRecordAddressOffset", 26},
    {"shaderRecordDataOffset", 23},
    {"shaderRecordIndex", 18},
    {"shaderRecordOffset", 19},
    {"shaderRelaxedExtendedInstruction", 33},
    {"shaderReplicatedComposites", 27},
    {"shaderResourceMinLod", 21},
    {"shaderResourceResidency", 24},
    {"shaderRoundingModeRTEFloat16", 29},
    {"shaderRoundingModeRTEFloat32", 29},
    {"shaderRoundingModeRTEFloat64", 29},
    {"shaderRoundingModeRTZFloat16", 29},
    {"shaderRoundingModeRTZFloat32", 29},
    {"shaderRoundingModeRTZFloat64", 29},
    {"shaderSMBuiltins", 17},
    {"shaderSMCount", 14},
    {"shaderSampleRateInterpolationFunctions", 39},
    {"shaderSampledImageArrayDynamicIndexing", 39},
    {"shaderSampledImageArrayNonUniformIndexing", 42},
    {"shaderSampledImageArrayNonUniformIndexingNative", 48},
    {"shaderSharedFloat16AtomicAdd", 29},
    {"shaderSharedFloat16AtomicMinMax", 32},
    {"shaderSharedFloat16Atomics", 27},
    {"shaderSharedFloat32AtomicAdd", 29},
    {"shaderSharedFloat32AtomicMinMax", 32},
    {"shaderSharedFloat32Atomics", 27},
    {"shaderSharedFloat64AtomicAdd", 29},
    {"shaderSharedFloat64AtomicMinMax", 32},
    {"shaderSharedFloat64Atomics", 27},
    {"shaderSharedInt64Atomics", 25},
    {"shaderSignedZeroInfNanPreserveFloat16", 38},
    {"shaderSignedZeroInfNanPreserveFloat32", 38},
    {"shaderSignedZeroInfNanPreserveFloat64", 38},
    {"shaderSplitBarrier", 19},
    {"shaderStage", 12},
    {"shaderStageMask", 16},
    {"shaderStages", 13},
    {"shaderStorageBufferArrayDynamicIndexing", 40},
    {"shaderStorageBufferArrayNonUniformIndexing", 43},
    {"shaderStorageBufferArrayNonUniformIndexingNative", 49},
    {"shaderStorageImageArrayDynamicIndexing", 39},
    {"shaderStorageImageArrayNonUniformIndexing", 42},
    {"shaderStorageImageArrayNonUniformIndexingNative", 48},
    {"shaderStorageImageExtendedFormats", 34},
    {"shaderStorageImageMultisample", 30},
    {"shaderStorageImageReadWithoutFormat", 36},
    {"shaderStorageImageWriteWithoutFormat", 37},
    {"shaderStorageTensorArrayDynamicIndexing", 40},
    {"shaderStorageTensorArrayNonUniformIndexing", 43},
    {"shaderStorageTensorArrayNonUniformIndexingNative", 49},
    {"shaderStorageTexelBufferArrayDynamicIndexing", 45},
    {"shaderStorageTexelBufferArrayNonUniformIndexing", 48},
    {"shaderSubgroupClock", 20},
    {"shaderSubgroupExtendedTypes", 28},
    {"shaderSubgroupPartitioned", 26},
    {"shaderSubgroupRotate", 21},
    {"shaderSubgroupRotateClustered", 30},
    {"shaderSubgroupUniformControlFlow", 33},
    {"shaderTensorAccess", 19},
    {"shaderTensorSupportedStages", 28},
    {"shaderTerminateInvocation", 26},
    {"shaderTessellationAndGeometryPointSize", 39},
    {"shaderTileImageCoherentReadAccelerated", 39},
    {"shaderTileImageColorReadAccess", 31},
    {"shaderTileImageDepthReadAccess", 31},
    {"shaderTileImageReadFromHelperInvocation", 40},
    {"shaderTileImageReadSampleFromPixelRateInvocation", 49},
    {"shaderTileImageStencilReadAccess", 33},
    {"shaderUniformBufferArrayDynamicIndexing", 40},
    {"shaderUniformBufferArrayNonUniformIndexing", 43},
    {"shaderUniformBufferArrayNonUniformIndexingNative", 49},
    {"shaderUniformBufferUnsizedArray", 32},
    {"shaderUniformTexelBufferArrayDynamicIndexing", 45},
    {"shaderUniformTexelBufferArrayNonUniformIndexing", 48},
    {"shaderUntypedPointers", 22},
    {"shaderWarpsPerCore", 19},
    {"shaderWarpsPerSM", 17},
    {"shaderZeroInitializeWorkgroupMemory", 36},
    {"shadingRate", 12},
    {"shadingRateAttachmentTexelSize", 31},
    {"shadingRateCoarseSampleOrder", 29},
    {"shadingRateImage", 17},
    {"shadingRateImageEnable", 23},
    {"shadingRateMaxCoarseSamples", 28},
    {"shadingRatePaletteEntryCount", 29},
    {"shadingRatePaletteSize", 23},
    {"shadingRateTexelSize", 21},
    {"shadingRateType", 16},
    {"sharedMemBytes", 15},
    {"sharedPresentSupportedUsageFlags", 33},
    {"sharingMode", 12},
    {"signalSemaphore", 16},
    {"signalSemaphoreCount", 21},
    {"signalSemaphoreInfoCount", 25},
    {"signalSemaphoreValueCount", 26},
    {"signalSemaphoreValuesCount", 27},
    {"simEndTimeUs", 13},
    {"simStartTimeUs", 15},
    {"simdPerComputeUnit", 19},
    {"singleReferenceNameMask", 24},
    {"size", 5},
    {"sizeGranularity", 16},
    {"sliceCount", 11},
    {"sliceOffset", 12},
    {"sliceSegmentCount", 18},
    {"slotIndex", 10},
    {"smoothLines", 12},
    {"source", 7},
    {"sourceData", 11},
    {"sparseAddressSpaceSize", 23},
    {"sparseBinding", 14},
    {"sparseDescriptorHeaps", 22},
    {"sparseImageFloat32AtomicAdd", 28},
    {"sparseImageFloat32AtomicMinMax", 31},
    {"sparseImageFloat32Atomics", 26},
    {"sparseImageInt64Atomics", 24},
    {"sparseProperties", 17},
    {"sparseResidency16Samples", 25},
    {"sparseResidency2Samples", 24},
    {"sparseResidency4Samples", 24},
    {"sparseResidency8Samples", 24},
    {"sparseResidencyAliased", 23},
    {"sparseResidencyBuffer", 22},
    {"sparseResidencyImage2D", 23},
    {"sparseResidencyImage3D", 23},
    {"specVersion", 12},
    {"spheres", 8},
    {"splitBarrierReservedSharedMemory", 33},
    {"splitInstanceBindRegionCount", 29},
    {"sqShaderMask", 13},
    {"sqShaderMaskEnable", 19},
    {"sqThreadTraceDeviceMemoryLimit", 31},
    {"sqThreadTraceEnable", 20},
    {"sqThreadTraceSuppressInstructionTokens", 39},
    {"sqThreadTracing", 16},
    {"src", 4},
    {"srcAccelerationStructure", 25},
    {"srcAccelerationStructureData", 29},
    {"srcAccessMask", 14},
    {"srcAccessMask3", 15},
    {"srcAddress", 11},
    {"srcAlphaBlendFactor", 20},
    {"srcArrayElement", 16},
    {"srcBinding", 11},
    {"srcBuffer", 10},
    {"srcBufferOffset", 16},
    {"srcBufferRange", 15},
    {"srcCacheCount", 14},
    {"srcColorBlendFactor", 20},
    {"srcComponentType", 17},
    {"srcCopyFlags", 13},
    {"srcData", 8},
    {"srcFlags", 9},
    {"srcImage", 9},
    {"srcImageLayout", 15},
    {"srcInfos", 9},
    {"srcInfosArray", 14},
    {"srcInfosCount", 14},
    {"srcLayout", 10},
    {"srcOffset", 10},
    {"srcOffsets", 11},
    {"srcPictureResource", 19},
    {"srcPremultiplied", 17},
    {"srcQueueFamilyIndex", 20},
    {"srcRange", 9},
    {"srcRect", 8},
    {"srcSet", 7},
    {"srcSize", 8},
    {"srcStageMask", 13},
    {"srcStride", 10},
    {"srcSubpass", 11},
    {"srcSubresource", 15},
    {"srcTensor", 10},
    {"srgb", 5},
    {"srtMotionInstance", 18},
    {"stage", 6},
    {"stageCount", 11},
    {"stageFlags", 11},
    {"stageMask", 10},
    {"stages", 7},
    {"standardSampleLocations", 24},
    {"startAddress", 13},
    {"stateCommandBuffer", 19},
    {"staticInstance", 15},
    {"stdHeaderVersion", 17},
    {"stdOperatingPointCount", 23},
    {"stdPPSCount", 12},
    {"stdPPSId", 9},
    {"stdProfile", 11},
    {"stdProfileIdc", 14},
    {"stdSPSCount", 12},
    {"stdSPSId", 9},
    {"stdSyntaxFlags", 15},
    {"stdVPSCount", 12},
    {"stdVPSId", 9},
    {"stencil", 8},
    {"stencilAttachmentFormat", 24},
    {"stencilFinalLayout", 19},
    {"stencilInitialLayout", 21},
    {"stencilLayout", 14},
    {"stencilLoadOp", 14},
    {"stencilResolveMode", 19},
    {"stencilStoreOp", 15},
    {"stencilTestEnable", 18},
    {"stencilUsage", 13},
    {"stereoType", 11},
    {"stippledBresenhamLines", 23},
    {"stippledLineEnable", 19},
    {"stippledRectangularLines", 25},
    {"stippledSmoothLines", 20},
    {"storage", 8},
    {"storageBuffer16BitAccess", 25},
    {"storageBuffer8BitAccess", 24},
    {"storageBufferDescriptorSize", 28},
    {"storageBuffers", 15},
    {"storageImageDescriptorSize", 27},
    {"storageImageSampleCounts", 25},
    {"storageInputOutput16", 21},
    {"storagePushConstant16", 22},
    {"storagePushConstant8", 21},
    {"storageTexelBufferDescriptorSize", 33},
    {"storageTexelBufferOffsetAlignmentBytes", 39},
    {"storageTexelBufferOffsetSingleTexelAlignment", 45},
    {"storeOp", 8},
    {"stream", 7},
    {"streamCount", 12},
    {"streamDescriptor", 17},
    {"streamingPerfCounters", 22},
    {"streamingPerfTraceSampleInterval", 33},
    {"strictLines", 12},
    {"stride", 7},
    {"strideInBytes", 14},
    {"stripeArea", 11},
    {"stripeInfoCount", 16},
    {"stripeSemaphoreInfoCount", 25},
    {"subLayerCount", 14},
    {"subPixelInterpolationOffsetBits", 32},
    {"subPixelPrecisionBits", 22},
    {"subTexelPrecisionBits", 22},
    {"subdivisionLevel", 17},
    {"subgroupBroadcastDynamicId", 27},
    {"subgroupQuadOperationsInAllStages", 34},
    {"subgroupSize", 13},
    {"subgroupSizeControl", 20},
    {"subgroupSupportedOperations", 28},
    {"subgroupSupportedStages", 24},
    {"subminor", 9},
    {"submitCount", 12},
    {"subpass", 8},
    {"subpassCount", 13},
    {"subpassIndex", 13},
    {"subpassMergeFeedback", 21},
    {"subpassMergeStatus", 19},
    {"subpassShading", 15},
    {"subresource", 12},
    {"subresourceLayout", 18},
    {"subresourceRange", 17},
    {"subsampledCoarseReconstructionEarlyAccess", 42},
    {"subsampledImageDescriptorCount", 31},
    {"subsampledLoads", 16},
    {"subsetAllocation", 17},
    {"suggestedXChromaOffset", 23},
    {"suggestedYChromaOffset", 23},
    {"suggestedYcbcrModel", 20},
    {"suggestedYcbcrRange", 20},
    {"superblockSizes", 16},
    {"supersampleFragmentShadingRates", 32},
    {"supported", 10},
    {"supportedAlpha", 15},
    {"supportedCompositeAlpha", 24},
    {"supportedDepthResolveModes", 27},
    {"supportedEncodeFeedbackFlags", 29},
    {"supportedHintGridSizes", 23},
    {"supportedImageAlignmentMask", 28},
    {"supportedIndirectCommandsInputModes", 36},
    {"supportedIndirectCommandsShaderStages", 38},
    {"supportedIndirectCommandsShaderStagesPipelineBinding", 53},
    {"supportedIndirectCommandsShaderStagesShaderBinding", 51},
    {"supportedOperations", 20},
    {"supportedOutputGridSizes", 25},
    {"supportedPerPartitionEncodeFeedbackFlags", 41},
    {"supportedPresentGravityX", 25},
    {"supportedPresentGravityY", 25},
    {"supportedPresentScaling", 24},
    {"supportedQueues", 16},
    {"supportedStages", 16},
    {"supportedStencilResolveModes", 29},
    {"supportedSurfaceCounters", 25},
    {"supportedTransforms", 20},
    {"supportedUsageFlags", 20},
    {"supportsNonZeroFirstInstance", 29},
    {"supportsProtected", 18},
    {"supportsTextureGatherLODBiasAMD", 32},
    {"surface", 8},
    {"surfaceCapabilities", 20},
    {"surfaceCounters", 16},
    {"surfaceFormat", 14},
    {"swapBufferTimestamp", 20},
    {"swapChain", 10},
    {"swapchain", 10},
    {"swapchainCount", 15},
    {"swapchainMaintenance1", 22},
    {"swapchainSupportedFlags", 24},
    {"sx", 3},
    {"sy", 3},
    {"synchronization2", 17},
    {"sysmemColorSpaceIndex", 22},
    {"sysmemPixelFormat", 18},
    {"sz", 3},
    {"tagName", 8},
    {"tagSize", 8},
    {"targetTime", 11},
    {"targetTimeDomainPresentStage", 29},
    {"taskCount", 10},
    {"taskShader", 11},
    {"templateType", 13},
    {"temporalLayerCount", 19},
    {"tensor", 7},
    {"tensorCaptureReplayDescriptorDataSize", 38},
    {"tensorCaptureReplayOpaqueDataSize", 34},
    {"tensorCount", 12},
    {"tensorDescriptorAlignment", 26},
    {"tensorDescriptorSize", 21},
    {"tensorMemoryBarrierCount", 25},
    {"tensorNonPacked", 16},
    {"tensorView", 11},
    {"tensorViewCaptureReplayDescriptorDataSize", 42},
    {"tensorViewCount", 16},
    {"tensors", 8},
    {"tessellationIsolines", 21},
    {"tessellationPointMode", 22},
    {"tessellationShader", 19},
    {"texelBufferAlignment", 21},
    {"texelRate", 10},
    {"textureBlockMatch", 18},
    {"textureBlockMatch2", 19},
    {"textureBoxFilter", 17},
    {"textureCompressionASTC_3D", 26},
    {"textureCompressionASTC_HDR", 27},
    {"textureCompressionASTC_LDR", 27},
    {"textureCompressionBC", 21},
    {"textureCompressionETC2", 23},
    {"textureSampleWeighted", 22},
    {"throttleHint", 13},
    {"tileApronSize", 14},
    {"tileBufferTransfers", 20},
    {"tileCount", 10},
    {"tileGranularity", 16},
    {"tileMemoryHeap", 15},
    {"tileProperties", 15},
    {"tileShading", 12},
    {"tileShadingAnisotropicApron", 28},
    {"tileShadingApron", 17},
    {"tileShadingAtomicOps", 21},
    {"tileShadingColorAttachments", 28},
    {"tileShadingDepthAttachments", 28},
    {"tileShadingDispatchTile", 24},
    {"tileShadingFragmentStage", 25},
    {"tileShadingImageProcessing", 27},
    {"tileShadingInputAttachments", 28},
    {"tileShadingPerTileDispatch", 27},
    {"tileShadingPerTileDraw", 23},
    {"tileShadingSampledAttachments", 30},
    {"tileShadingStencilAttachments", 30},
    {"tileSize", 9},
    {"tilesOffset", 12},
    {"tiling", 7},
    {"time", 5},
    {"timeDomain", 11},
    {"timeDomainCount", 16},
    {"timeDomainId", 13},
    {"timeDomainsCounter", 19},
    {"timelineSemaphore", 18},
    {"timeout", 8},
    {"timestampComputeAndGraphics", 28},
    {"timestampCount", 15},
    {"timestampPeriod", 16},
    {"timestampValidBits", 19},
    {"timingCount", 12},
    {"timingPostSample", 17},
    {"timingPreSample", 16},
    {"timingPropertiesCounter", 24},
    {"tokenCount", 11},
    {"tokenType", 10},
    {"toolchainVersion", 17},
    {"topology", 9},
    {"transform", 10},
    {"transformBlockSizes", 20},
    {"transformData", 14},
    {"transformFeedback", 18},
    {"transformFeedbackDraw", 22},
    {"transformFeedbackPreservesProvokingVertex", 42},
    {"transformFeedbackPreservesTriangleFanProvokingVertex", 53},
    {"transformFeedbackQueries", 25},
    {"transformFeedbackRasterizationStreamSelect", 43},
    {"transformFeedbackStreamsLinesTriangles", 39},
    {"transformOffset", 16},
    {"transformT0", 12},
    {"transformT1", 12},
    {"transitionCount", 16},
    {"transpose", 10},
    {"triStripVertexOrderIndependentOfProvokingVertex", 48},
    {"triangleArray", 14},
    {"triangleArrayStride", 20},
    {"triangleCount", 14},
    {"triangleFans", 13},
    {"triangles", 10},
    {"tuningMode", 11},
    {"tx", 3},
    {"ty", 3},
    {"type", 5},
    {"tz", 3},
    {"u64", 4},
    {"uint32", 7},
    {"uint64", 7},
    {"uncompressedHeaderOffset", 25},
    {"underlyingAPI", 14},
    {"unidirectionalCompoundReferenceNameMask", 40},
    {"unifiedImageLayouts", 20},
    {"unifiedImageLayoutsVideo", 25},
    {"uniformAndStorageBuffer16BitAccess", 35},
    {"uniformAndStorageBuffer8BitAccess", 34},
    {"uniformBufferDescriptorSize", 28},
    {"uniformBufferStandardLayout", 28},
    {"uniformBuffers", 15},
    {"uniformTexelBufferDescriptorSize", 33},
    {"uniformTexelBufferOffsetAlignmentBytes", 39},
    {"uniformTexelBufferOffsetSingleTexelAlignment", 45},
    {"unit", 5},
    {"unnormalizedCoordinates", 24},
    {"update", 7},
    {"updateRange", 12},
    {"updateScratchSize", 18},
    {"updateSequenceCount", 20},
    {"usage", 6},
    {"usageCountsCount", 17},
    {"use64bitTexturing", 18},
    {"useCombinedImageSamplerIndex", 29},
    {"useGopRemainingFrames", 22},
    {"useMaxFrameSize", 16},
    {"useMaxLevel", 12},
    {"useMaxLevelIdc", 15},
    {"useMaxQIndex", 13},
    {"useMaxQp", 9},
    {"useMinQIndex", 13},
    {"useMinQp", 9},
    {"uuid", 5},
    {"validationCache", 16},
    {"value", 6},
    {"value32", 8},
    {"value64", 8},
    {"valueBool", 10},
    {"valueCount", 11},
    {"valueFloat", 11},
    {"valueString", 12},
    {"variableMultisampleRate", 24},
    {"variablePointers", 17},
    {"variablePointersStorageBuffer", 30},
    {"variableSampleLocations", 24},
    {"vendorBinarySize", 17},
    {"vendorFaultCode", 16},
    {"vendorFaultData", 16},
    {"vendorID", 9},
    {"vendorInfo", 11},
    {"vendorInfoCount", 16},
    {"version", 8},
    {"vertexAttributeAccessBeyondStride", 34},
    {"vertexAttributeDescriptionCount", 32},
    {"vertexAttributeInstanceRateDivisor", 35},
    {"vertexAttributeInstanceRateZeroDivisor", 39},
    {"vertexAttributeRobustness", 26},
    {"vertexBindingDescriptionCount", 30},
    {"vertexBindingDivisorCount", 26},
    {"vertexBindingUnit", 18},
    {"vertexBuffer", 13},
    {"vertexBufferStride", 19},
    {"vertexCount", 12},
    {"vertexData", 11},
    {"vertexDynamicStride", 20},
    {"vertexFormat", 13},
    {"vertexInputDynamicState", 24},
    {"vertexInputs", 13},
    {"vertexOffset", 13},
    {"vertexPipelineStoresAndAtomics", 31},
    {"vertexStride", 13},
    {"vgprAllocationGranularity", 26},
    {"vgprsPerSimd", 13},
    {"videoCodecOperation", 20},
    {"videoCodecOperations", 21},
    {"videoContentHints", 18},
    {"videoDecodeVP9", 15},
    {"videoEncodeAV1", 15},
    {"videoEncodeFeedback2", 21},
    {"videoEncodeIntraRefresh", 24},
    {"videoEncodeQuantizationMap", 27},
    {"videoEncodeRgbConversion", 25},
    {"videoMaintenance1", 18},
    {"videoMaintenance2", 18},
    {"videoSession", 13},
    {"videoSessionParameters", 23},
    {"videoSessionParametersTemplate", 31},
    {"videoUsageHints", 16},
    {"view", 5},
    {"viewFormatCount", 16},
    {"viewMask", 9},
    {"viewOffset", 11},
    {"viewType", 9},
    {"viewportBoundsRange", 20},
    {"viewportCount", 14},
    {"viewportDepthCount", 19},
    {"viewportScissor2D", 18},
    {"viewportSubPixelBits", 21},
    {"viewportWScalingEnable", 23},
    {"virtualBufferSizeInMs", 22},
    {"visibleRegion", 14},
    {"visualID", 9},
    {"visual_id", 10},
    {"vulkanMemoryModel", 18},
    {"vulkanMemoryModelAvailabilityVisibilityChains", 46},
    {"vulkanMemoryModelDeviceScope", 29},
    {"w", 2},
    {"waitAll", 8},
    {"waitSemaphoreCount", 19},
    {"waitSemaphoreInfoCount", 23},
    {"waitSemaphoreValueCount", 24},
    {"waitSemaphoreValuesCount", 25},
    {"wavefrontSize", 14},
    {"wavefrontsPerSimd", 18},
    {"whitePoint", 11},
    {"wideLines", 10},
    {"width", 6},
    {"window", 7},
    {"windowCompareMode", 18},
    {"windowExtent", 13},
    {"workGroupBatchSize", 19},
    {"workgroupInvocations", 21},
    {"workgroupMemoryExplicitLayout", 30},
    {"workgroupMemoryExplicitLayout16BitAccess", 41},
    {"workgroupMemoryExplicitLayout8BitAccess", 40},
    {"workgroupMemoryExplicitLayoutScalarBlockLayout", 47},
    {"wraps", 6},
    {"writeMask", 10},
    {"writeStdPPS", 12},
    {"writeStdSPS", 12},
    {"writeStdVPS", 12},
    {"x", 2},
    {"xChromaOffset", 14},
    {"xChromaOffsets", 15},
    {"xcoeff", 7},
    {"y", 2},
    {"yChromaOffset", 14},
    {"yChromaOffsets", 15},
    {"ycbcr2plane444Formats", 22},
    {"ycbcrConversion", 16},
    {"ycbcrDegamma", 13},
    {"ycbcrImageArrays", 17},
    {"ycbcrModel", 11},
    {"ycbcrRange", 11},
    {"ycoeff", 7},
    {"z", 2},
    {"zeroCount", 10},
    {"zeroInitializeDeviceMemory", 27},
    {"zirconHandle", 13},
    };
    return table[(int)field].data();
}

const char* String(Enum value) {
    static const std::string_view table[] = {
    {"INVALID_EMPTY", 15}, // Enum::Empty
    {"VkAccelerationStructureBuildTypeKHR", 36},
    {"VkAccelerationStructureCompatibilityKHR", 40},
    {"VkAccelerationStructureMemoryRequirementsTypeNV", 48},
    {"VkAccelerationStructureMotionInstanceTypeNV", 44},
    {"VkAccelerationStructureSerializedBlockTypeKHR", 46},
    {"VkAccelerationStructureTypeKHR", 31},
    {"VkAntiLagModeAMD", 17},
    {"VkAntiLagStageAMD", 18},
    {"VkAttachmentLoadOp", 19},
    {"VkAttachmentStoreOp", 20},
    {"VkBlendFactor", 14},
    {"VkBlendOp", 10},
    {"VkBlendOverlapEXT", 18},
    {"VkBlockMatchWindowCompareModeQCOM", 34},
    {"VkBorderColor", 14},
    {"VkBuildAccelerationStructureModeKHR", 36},
    {"VkBuildMicromapModeEXT", 23},
    {"VkChromaLocation", 17},
    {"VkClusterAccelerationStructureOpModeNV", 39},
    {"VkClusterAccelerationStructureOpTypeNV", 39},
    {"VkClusterAccelerationStructureTypeNV", 37},
    {"VkCoarseSampleOrderTypeNV", 26},
    {"VkColorSpaceKHR", 16},
    {"VkCommandBufferLevel", 21},
    {"VkCompareOp", 12},
    {"VkComponentSwizzle", 19},
    {"VkComponentTypeKHR", 19},
    {"VkCompressedTriangleFormatAMDX", 31},
    {"VkConservativeRasterizationModeEXT", 35},
    {"VkCooperativeVectorMatrixLayoutNV", 34},
    {"VkCopyAccelerationStructureModeKHR", 35},
    {"VkCopyMicromapModeEXT", 22},
    {"VkCoverageModulationModeNV", 27},
    {"VkCoverageReductionModeNV", 26},
    {"VkCubicFilterWeightsQCOM", 25},
    {"VkDataGraphModelCacheTypeQCOM", 30},
    {"VkDataGraphOpticalFlowPerformanceLevelARM", 42},
    {"VkDataGraphPipelineNodeConnectionTypeARM", 41},
    {"VkDataGraphPipelineNodeTypeARM", 31},
    {"VkDataGraphPipelinePropertyARM", 31},
    {"VkDataGraphPipelineSessionBindPointARM", 39},
    {"VkDataGraphPipelineSessionBindPointTypeARM", 43},
    {"VkDataGraphTOSALevelARM", 24},
    {"VkDebugReportObjectTypeEXT", 27},
    {"VkDefaultVertexAttributeValueKHR", 33},
    {"VkDepthBiasRepresentationEXT", 29},
    {"VkDepthClampModeEXT", 20},
    {"VkDescriptorMappingSourceEXT", 29},
    {"VkDescriptorType", 17},
    {"VkDescriptorUpdateTemplateType", 31},
    {"VkDeviceAddressBindingTypeEXT", 30},
    {"VkDeviceEventTypeEXT", 21},
    {"VkDeviceFaultAddressTypeKHR", 28},
    {"VkDeviceFaultVendorBinaryHeaderVersionKHR", 42},
    {"VkDeviceMemoryReportEventTypeEXT", 33},
    {"VkDirectDriverLoadingModeLUNARG", 32},
    {"VkDiscardRectangleModeEXT", 26},
    {"VkDisplacementMicromapFormatNV", 31},
    {"VkDisplayEventTypeEXT", 22},
    {"VkDisplayPowerStateEXT", 23},
    {"VkDisplaySurfaceStereoTypeNV", 29},
    {"VkDriverId", 11},
    {"VkDynamicState", 15},
    {"VkFilter", 9},
    {"VkFormat", 9},
    {"VkFragmentShadingRateCombinerOpKHR", 35},
    {"VkFragmentShadingRateNV", 24},
    {"VkFragmentShadingRateTypeNV", 28},
    {"VkFrontFace", 12},
    {"VkFullScreenExclusiveEXT", 25},
    {"VkGeometryTypeKHR", 18},
    {"VkGpaDeviceClockModeAMD", 24},
    {"VkGpaPerfBlockAMD", 18},
    {"VkGpaSampleTypeAMD", 19},
    {"VkImageLayout", 14},
    {"VkImageTiling", 14},
    {"VkImageType", 12},
    {"VkImageViewType", 16},
    {"VkIndexType", 12},
    {"VkIndirectCommandsTokenTypeEXT", 31},
    {"VkIndirectCommandsTokenTypeNV", 30},
    {"VkIndirectExecutionSetInfoTypeEXT", 34},
    {"VkInternalAllocationType", 25},
    {"VkLatencyMarkerNV", 18},
    {"VkLayerSettingTypeEXT", 22},
    {"VkLayeredDriverUnderlyingApiMSFT", 33},
    {"VkLineRasterizationMode", 24},
    {"VkLogicOp", 10},
    {"VkMemoryOverallocationBehaviorAMD", 34},
    {"VkMicromapTypeEXT", 18},
    {"VkNeuralAcceleratorStatisticsModeARM", 37},
    {"VkObjectType", 13},
    {"VkOpacityMicromapFormatKHR", 27},
    {"VkOpacityMicromapSpecialIndexKHR", 33},
    {"VkOpticalFlowPerformanceLevelNV", 32},
    {"VkOpticalFlowSessionBindingPointNV", 35},
    {"VkOutOfBandQueueTypeNV", 23},
    {"VkPartitionedAccelerationStructureOpTypeNV", 43},
    {"VkPerfHintTypeQCOM", 19},
    {"VkPerformanceConfigurationTypeINTEL", 36},
    {"VkPerformanceCounterScopeKHR", 29},
    {"VkPerformanceCounterStorageKHR", 31},
    {"VkPerformanceCounterUnitKHR", 28},
    {"VkPerformanceOverrideTypeINTEL", 31},
    {"VkPerformanceParameterTypeINTEL", 32},
    {"VkPerformanceValueTypeINTEL", 28},
    {"VkPhysicalDeviceDataGraphOperationTypeARM", 42},
    {"VkPhysicalDeviceDataGraphProcessingEngineTypeARM", 49},
    {"VkPhysicalDeviceLayeredApiKHR", 30},
    {"VkPhysicalDeviceType", 21},
    {"VkPipelineBindPoint", 20},
    {"VkPipelineCacheHeaderVersion", 29},
    {"VkPipelineExecutableStatisticFormatKHR", 39},
    {"VkPipelineRobustnessBufferBehavior", 35},
    {"VkPipelineRobustnessImageBehavior", 34},
    {"VkPointClippingBehavior", 24},
    {"VkPolygonMode", 14},
    {"VkPresentModeKHR", 17},
    {"VkPrimitiveTopology", 20},
    {"VkProvokingVertexModeEXT", 25},
    {"VkQueryPoolSamplingModeINTEL", 29},
    {"VkQueryResultStatusKHR", 23},
    {"VkQueryType", 12},
    {"VkQueueGlobalPriority", 22},
    {"VkRasterizationOrderAMD", 24},
    {"VkRayTracingInvocationReorderModeEXT", 37},
    {"VkRayTracingLssIndexingModeNV", 30},
    {"VkRayTracingLssPrimitiveEndCapsModeNV", 38},
    {"VkRayTracingShaderGroupTypeKHR", 31},
    {"VkResult", 9},
    {"VkSamplerAddressMode", 21},
    {"VkSamplerMipmapMode", 20},
    {"VkSamplerReductionMode", 23},
    {"VkSamplerYcbcrModelConversion", 30},
    {"VkSamplerYcbcrRange", 20},
    {"VkScopeKHR", 11},
    {"VkSemaphoreType", 16},
    {"VkShaderCodeTypeEXT", 20},
    {"VkShaderFloatControlsIndependence", 34},
    {"VkShaderGroupShaderKHR", 23},
    {"VkShaderInfoTypeAMD", 20},
    {"VkShadingRatePaletteEntryNV", 28},
    {"VkSharingMode", 14},
    {"VkStencilOp", 12},
    {"VkStructureType", 16},
    {"VkSubpassContents", 18},
    {"VkSubpassMergeStatusEXT", 24},
    {"VkSystemAllocationScope", 24},
    {"VkTensorTilingARM", 18},
    {"VkTessellationDomainOrigin", 27},
    {"VkThrottleHintTypeSEC", 22},
    {"VkTimeDomainKHR", 16},
    {"VkValidationCacheHeaderVersionEXT", 34},
    {"VkValidationCheckEXT", 21},
    {"VkValidationFeatureDisableEXT", 30},
    {"VkValidationFeatureEnableEXT", 29},
    {"VkVendorId", 11},
    {"VkVertexInputRate", 18},
    {"VkVideoEncodeAV1PredictionModeKHR", 34},
    {"VkVideoEncodeAV1RateControlGroupKHR", 36},
    {"VkVideoEncodeTuningModeKHR", 27},
    {"VkViewportCoordinateSwizzleNV", 30},
    };
    return table[(int)value].data();
}

const char* String(FlagBitmask value) {
    static const std::string_view table[] = {
    {"INVALID_EMPTY", 15}, // FlagBitmask::Empty
    {"VkAccelerationStructureCreateFlagBitsKHR", 41},
    {"VkAccessFlagBits", 17},
    {"VkAccessFlagBits2", 18},
    {"VkAccessFlagBits3KHR", 21},
    {"VkAcquireProfilingLockFlagBitsKHR", 34},
    {"VkAddressCommandFlagBitsKHR", 28},
    {"VkAddressCopyFlagBitsKHR", 25},
    {"VkAttachmentDescriptionFlagBits", 32},
    {"VkBufferCreateFlagBits", 23},
    {"VkBufferUsageFlagBits", 22},
    {"VkBufferUsageFlagBits2", 23},
    {"VkBuildAccelerationStructureFlagBitsKHR", 40},
    {"VkBuildMicromapFlagBitsEXT", 27},
    {"VkClusterAccelerationStructureAddressResolutionFlagBitsNV", 58},
    {"VkClusterAccelerationStructureClusterFlagBitsNV", 48},
    {"VkClusterAccelerationStructureGeometryFlagBitsNV", 49},
    {"VkClusterAccelerationStructureIndexFormatFlagBitsNV", 52},
    {"VkColorComponentFlagBits", 25},
    {"VkCommandBufferResetFlagBits", 29},
    {"VkCommandBufferUsageFlagBits", 29},
    {"VkCommandPoolCreateFlagBits", 28},
    {"VkCommandPoolResetFlagBits", 27},
    {"VkCompositeAlphaFlagBitsKHR", 28},
    {"VkConditionalRenderingFlagBitsEXT", 34},
    {"VkCullModeFlagBits", 19},
    {"VkDataGraphOpticalFlowCreateFlagBitsARM", 40},
    {"VkDataGraphOpticalFlowExecuteFlagBitsARM", 41},
    {"VkDataGraphOpticalFlowGridSizeFlagBitsARM", 42},
    {"VkDataGraphOpticalFlowImageUsageFlagBitsARM", 44},
    {"VkDataGraphPipelineDispatchFlagBitsARM", 39},
    {"VkDataGraphPipelineSessionCreateFlagBitsARM", 44},
    {"VkDataGraphTOSAQualityFlagBitsARM", 34},
    {"VkDebugReportFlagBitsEXT", 25},
    {"VkDebugUtilsMessageSeverityFlagBitsEXT", 39},
    {"VkDebugUtilsMessageTypeFlagBitsEXT", 35},
    {"VkDependencyFlagBits", 21},
    {"VkDescriptorBindingFlagBits", 28},
    {"VkDescriptorPoolCreateFlagBits", 31},
    {"VkDescriptorSetLayoutCreateFlagBits", 36},
    {"VkDeviceAddressBindingFlagBitsEXT", 34},
    {"VkDeviceDiagnosticsConfigFlagBitsNV", 36},
    {"VkDeviceFaultFlagBitsKHR", 25},
    {"VkDeviceGroupPresentModeFlagBitsKHR", 36},
    {"VkDeviceQueueCreateFlagBits", 28},
    {"VkDisplayPlaneAlphaFlagBitsKHR", 31},
    {"VkEventCreateFlagBits", 22},
    {"VkExportMetalObjectTypeFlagBitsEXT", 35},
    {"VkExternalFenceFeatureFlagBits", 31},
    {"VkExternalFenceHandleTypeFlagBits", 34},
    {"VkExternalMemoryFeatureFlagBits", 32},
    {"VkExternalMemoryFeatureFlagBitsNV", 34},
    {"VkExternalMemoryHandleTypeFlagBits", 35},
    {"VkExternalMemoryHandleTypeFlagBitsNV", 37},
    {"VkExternalSemaphoreFeatureFlagBits", 35},
    {"VkExternalSemaphoreHandleTypeFlagBits", 38},
    {"VkFenceCreateFlagBits", 22},
    {"VkFenceImportFlagBits", 22},
    {"VkFormatFeatureFlagBits", 24},
    {"VkFormatFeatureFlagBits2", 25},
    {"VkFormatFeatureFlagBits4KHR", 28},
    {"VkFrameBoundaryFlagBitsEXT", 27},
    {"VkFramebufferCreateFlagBits", 28},
    {"VkGeometryFlagBitsKHR", 22},
    {"VkGeometryInstanceFlagBitsKHR", 30},
    {"VkGpaSqShaderStageFlagBitsAMD", 30},
    {"VkGraphicsPipelineLibraryFlagBitsEXT", 37},
    {"VkHostImageCopyFlagBits", 24},
    {"VkImageAspectFlagBits", 22},
    {"VkImageCompressionFixedRateFlagBitsEXT", 39},
    {"VkImageCompressionFlagBitsEXT", 30},
    {"VkImageConstraintsInfoFlagBitsFUCHSIA", 38},
    {"VkImageCreateFlagBits", 22},
    {"VkImageCreateFlagBits2KHR", 26},
    {"VkImageUsageFlagBits", 21},
    {"VkImageUsageFlagBits2KHR", 25},
    {"VkImageViewCreateFlagBits", 26},
    {"VkIndirectCommandsInputModeFlagBitsEXT", 39},
    {"VkIndirectCommandsLayoutUsageFlagBitsEXT", 41},
    {"VkIndirectCommandsLayoutUsageFlagBitsNV", 40},
    {"VkIndirectStateFlagBitsNV", 26},
    {"VkInstanceCreateFlagBits", 25},
    {"VkMemoryAllocateFlagBits", 25},
    {"VkMemoryDecompressionMethodFlagBitsEXT", 39},
    {"VkMemoryHeapFlagBits", 21},
    {"VkMemoryMapFlagBits", 20},
    {"VkMemoryPropertyFlagBits", 25},
    {"VkMemoryUnmapFlagBits", 22},
    {"VkMicromapCreateFlagBitsEXT", 28},
    {"VkOpticalFlowExecuteFlagBitsNV", 31},
    {"VkOpticalFlowGridSizeFlagBitsNV", 32},
    {"VkOpticalFlowSessionCreateFlagBitsNV", 37},
    {"VkOpticalFlowUsageFlagBitsNV", 29},
    {"VkPartitionedAccelerationStructureInstanceFlagBitsNV", 53},
    {"VkPastPresentationTimingFlagBitsEXT", 36},
    {"VkPeerMemoryFeatureFlagBits", 28},
    {"VkPerformanceCounterDescriptionFlagBitsKHR", 43},
    {"VkPhysicalDeviceSchedulingControlsFlagBitsARM", 46},
    {"VkPipelineCacheCreateFlagBits", 30},
    {"VkPipelineColorBlendStateCreateFlagBits", 40},
    {"VkPipelineCompilerControlFlagBitsAMD", 37},
    {"VkPipelineCreateFlagBits", 25},
    {"VkPipelineCreateFlagBits2", 26},
    {"VkPipelineCreationFeedbackFlagBits", 35},
    {"VkPipelineDepthStencilStateCreateFlagBits", 42},
    {"VkPipelineLayoutCreateFlagBits", 31},
    {"VkPipelineShaderStageCreateFlagBits", 36},
    {"VkPipelineStageFlagBits", 24},
    {"VkPipelineStageFlagBits2", 25},
    {"VkPresentGravityFlagBitsKHR", 28},
    {"VkPresentScalingFlagBitsKHR", 28},
    {"VkPresentStageFlagBitsEXT", 26},
    {"VkPresentTimingInfoFlagBitsEXT", 31},
    {"VkQueryControlFlagBits", 23},
    {"VkQueryPipelineStatisticFlagBits", 33},
    {"VkQueryPoolCreateFlagBits", 26},
    {"VkQueryResultFlagBits", 22},
    {"VkQueueFlagBits", 16},
    {"VkRenderPassCreateFlagBits", 27},
    {"VkRenderingAttachmentFlagBitsKHR", 33},
    {"VkRenderingFlagBits", 20},
    {"VkResolveImageFlagBitsKHR", 26},
    {"VkResolveModeFlagBits", 22},
    {"VkSampleCountFlagBits", 22},
    {"VkSamplerCreateFlagBits", 24},
    {"VkSemaphoreImportFlagBits", 26},
    {"VkSemaphoreWaitFlagBits", 24},
    {"VkShaderCorePropertiesFlagBitsAMD", 34},
    {"VkShaderCreateFlagBitsEXT", 26},
    {"VkShaderStageFlagBits", 22},
    {"VkSparseImageFormatFlagBits", 28},
    {"VkSparseMemoryBindFlagBits", 27},
    {"VkSpirvResourceTypeFlagBitsEXT", 31},
    {"VkStencilFaceFlagBits", 22},
    {"VkSubgroupFeatureFlagBits", 26},
    {"VkSubmitFlagBits", 17},
    {"VkSubpassDescriptionFlagBits", 29},
    {"VkSurfaceCounterFlagBitsEXT", 28},
    {"VkSurfaceTransformFlagBitsKHR", 30},
    {"VkSwapchainCreateFlagBitsKHR", 29},
    {"VkTensorCreateFlagBitsARM", 26},
    {"VkTensorUsageFlagBitsARM", 25},
    {"VkTensorViewCreateFlagBitsARM", 30},
    {"VkTileShadingRenderPassFlagBitsQCOM", 36},
    {"VkToolPurposeFlagBits", 22},
    {"VkVideoCapabilityFlagBitsKHR", 29},
    {"VkVideoChromaSubsamplingFlagBitsKHR", 36},
    {"VkVideoCodecOperationFlagBitsKHR", 33},
    {"VkVideoCodingControlFlagBitsKHR", 32},
    {"VkVideoComponentBitDepthFlagBitsKHR", 36},
    {"VkVideoDecodeCapabilityFlagBitsKHR", 35},
    {"VkVideoDecodeH264PictureLayoutFlagBitsKHR", 42},
    {"VkVideoDecodeUsageFlagBitsKHR", 30},
    {"VkVideoEncodeAV1CapabilityFlagBitsKHR", 38},
    {"VkVideoEncodeAV1RateControlFlagBitsKHR", 39},
    {"VkVideoEncodeAV1StdFlagBitsKHR", 31},
    {"VkVideoEncodeAV1SuperblockSizeFlagBitsKHR", 42},
    {"VkVideoEncodeCapabilityFlagBitsKHR", 35},
    {"VkVideoEncodeContentFlagBitsKHR", 32},
    {"VkVideoEncodeFeedbackFlagBitsKHR", 33},
    {"VkVideoEncodeFlagBitsKHR", 25},
    {"VkVideoEncodeH264CapabilityFlagBitsKHR", 39},
    {"VkVideoEncodeH264RateControlFlagBitsKHR", 40},
    {"VkVideoEncodeH264StdFlagBitsKHR", 32},
    {"VkVideoEncodeH265CapabilityFlagBitsKHR", 39},
    {"VkVideoEncodeH265CtbSizeFlagBitsKHR", 36},
    {"VkVideoEncodeH265RateControlFlagBitsKHR", 40},
    {"VkVideoEncodeH265StdFlagBitsKHR", 32},
    {"VkVideoEncodeH265TransformBlockSizeFlagBitsKHR", 47},
    {"VkVideoEncodeIntraRefreshModeFlagBitsKHR", 41},
    {"VkVideoEncodePerPartitionFeedbackFlagBitsKHR", 45},
    {"VkVideoEncodeRateControlModeFlagBitsKHR", 40},
    {"VkVideoEncodeRgbChromaOffsetFlagBitsVALVE", 42},
    {"VkVideoEncodeRgbModelConversionFlagBitsVALVE", 45},
    {"VkVideoEncodeRgbRangeCompressionFlagBitsVALVE", 46},
    {"VkVideoEncodeUsageFlagBitsKHR", 30},
    {"VkVideoSessionCreateFlagBitsKHR", 32},
    {"VkVideoSessionParametersCreateFlagBitsKHR", 42},
    };
    return table[(int)value].data();
}

const char* String(Extension extension) {
    static const std::string_view table[] = {
    {"INVALID_EMPTY", 15}, // Extension::Empty
    {"VK_AMDX_dense_geometry_format", 30},
    {"VK_AMDX_shader_enqueue", 23},
    {"VK_AMD_anti_lag", 16},
    {"VK_AMD_buffer_marker", 21},
    {"VK_AMD_device_coherent_memory", 30},
    {"VK_AMD_display_native_hdr", 26},
    {"VK_AMD_draw_indirect_count", 27},
    {"VK_AMD_gcn_shader", 18},
    {"VK_AMD_gpa_interface", 21},
    {"VK_AMD_gpu_shader_half_float", 29},
    {"VK_AMD_gpu_shader_int16", 24},
    {"VK_AMD_memory_overallocation_behavior", 38},
    {"VK_AMD_mixed_attachment_samples", 32},
    {"VK_AMD_negative_viewport_height", 32},
    {"VK_AMD_pipeline_compiler_control", 33},
    {"VK_AMD_rasterization_order", 27},
    {"VK_AMD_shader_ballot", 21},
    {"VK_AMD_shader_core_properties", 30},
    {"VK_AMD_shader_core_properties2", 31},
    {"VK_AMD_shader_early_and_late_fragment_tests", 44},
    {"VK_AMD_shader_explicit_vertex_parameter", 40},
    {"VK_AMD_shader_fragment_mask", 28},
    {"VK_AMD_shader_image_load_store_lod", 35},
    {"VK_AMD_shader_info", 19},
    {"VK_AMD_shader_trinary_minmax", 29},
    {"VK_AMD_texture_gather_bias_lod", 31},
    {"VK_ANDROID_external_format_resolve", 35},
    {"VK_ANDROID_external_memory_android_hardware_buffer", 51},
    {"VK_ARM_data_graph", 18},
    {"VK_ARM_data_graph_instruction_set_tosa", 39},
    {"VK_ARM_data_graph_neural_accelerator_statistics", 48},
    {"VK_ARM_data_graph_optical_flow", 31},
    {"VK_ARM_format_pack", 19},
    {"VK_ARM_performance_counters_by_region", 38},
    {"VK_ARM_pipeline_opacity_micromap", 33},
    {"VK_ARM_rasterization_order_attachment_access", 45},
    {"VK_ARM_render_pass_striped", 27},
    {"VK_ARM_scheduling_controls", 27},
    {"VK_ARM_shader_core_builtins", 28},
    {"VK_ARM_shader_core_properties", 30},
    {"VK_ARM_shader_instrumentation", 30},
    {"VK_ARM_tensor_controls", 23},
    {"VK_ARM_tensors", 15},
    {"VK_EXT_4444_formats", 20},
    {"VK_EXT_acquire_drm_display", 27},
    {"VK_EXT_acquire_xlib_display", 28},
    {"VK_EXT_astc_decode_mode", 24},
    {"VK_EXT_attachment_feedback_loop_dynamic_state", 46},
    {"VK_EXT_attachment_feedback_loop_layout", 39},
    {"VK_EXT_blend_operation_advanced", 32},
    {"VK_EXT_border_color_swizzle", 28},
    {"VK_EXT_buffer_device_address", 29},
    {"VK_EXT_calibrated_timestamps", 29},
    {"VK_EXT_color_write_enable", 26},
    {"VK_EXT_conditional_rendering", 29},
    {"VK_EXT_conservative_rasterization", 34},
    {"VK_EXT_custom_border_color", 27},
    {"VK_EXT_custom_resolve", 22},
    {"VK_EXT_debug_marker", 20},
    {"VK_EXT_debug_report", 20},
    {"VK_EXT_debug_utils", 19},
    {"VK_EXT_depth_bias_control", 26},
    {"VK_EXT_depth_clamp_control", 27},
    {"VK_EXT_depth_clamp_zero_one", 28},
    {"VK_EXT_depth_clip_control", 26},
    {"VK_EXT_depth_clip_enable", 25},
    {"VK_EXT_depth_range_unrestricted", 32},
    {"VK_EXT_descriptor_buffer", 25},
    {"VK_EXT_descriptor_heap", 23},
    {"VK_EXT_descriptor_indexing", 27},
    {"VK_EXT_device_address_binding_report", 37},
    {"VK_EXT_device_fault", 20},
    {"VK_EXT_device_generated_commands", 33},
    {"VK_EXT_device_memory_report", 28},
    {"VK_EXT_direct_mode_display", 27},
    {"VK_EXT_directfb_surface", 24},
    {"VK_EXT_discard_rectangles", 26},
    {"VK_EXT_display_control", 23},
    {"VK_EXT_display_surface_counter", 31},
    {"VK_EXT_dynamic_rendering_unused_attachments", 44},
    {"VK_EXT_extended_dynamic_state", 30},
    {"VK_EXT_extended_dynamic_state2", 31},
    {"VK_EXT_extended_dynamic_state3", 31},
    {"VK_EXT_external_memory_acquire_unmodified", 42},
    {"VK_EXT_external_memory_dma_buf", 31},
    {"VK_EXT_external_memory_host", 28},
    {"VK_EXT_external_memory_metal", 29},
    {"VK_EXT_filter_cubic", 20},
    {"VK_EXT_fragment_density_map", 28},
    {"VK_EXT_fragment_density_map2", 29},
    {"VK_EXT_fragment_density_map_offset", 35},
    {"VK_EXT_fragment_shader_interlock", 33},
    {"VK_EXT_frame_boundary", 22},
    {"VK_EXT_full_screen_exclusive", 29},
    {"VK_EXT_global_priority", 23},
    {"VK_EXT_global_priority_query", 29},
    {"VK_EXT_graphics_pipeline_library", 33},
    {"VK_EXT_hdr_metadata", 20},
    {"VK_EXT_headless_surface", 24},
    {"VK_EXT_host_image_copy", 23},
    {"VK_EXT_host_query_reset", 24},
    {"VK_EXT_image_2d_view_of_3d", 27},
    {"VK_EXT_image_compression_control", 33},
    {"VK_EXT_image_compression_control_swapchain", 43},
    {"VK_EXT_image_drm_format_modifier", 33},
    {"VK_EXT_image_robustness", 24},
    {"VK_EXT_image_sliced_view_of_3d", 31},
    {"VK_EXT_image_view_min_lod", 26},
    {"VK_EXT_index_type_uint8", 24},
    {"VK_EXT_inline_uniform_block", 28},
    {"VK_EXT_layer_settings", 22},
    {"VK_EXT_legacy_dithering", 24},
    {"VK_EXT_legacy_vertex_attributes", 32},
    {"VK_EXT_line_rasterization", 26},
    {"VK_EXT_load_store_op_none", 26},
    {"VK_EXT_map_memory_placed", 25},
    {"VK_EXT_memory_budget", 21},
    {"VK_EXT_memory_decompression", 28},
    {"VK_EXT_memory_priority", 23},
    {"VK_EXT_mesh_shader", 19},
    {"VK_EXT_metal_objects", 21},
    {"VK_EXT_metal_surface", 21},
    {"VK_EXT_multi_draw", 18},
    {"VK_EXT_multisampled_render_to_single_sampled", 45},
    {"VK_EXT_multisampled_render_to_swapchain", 40},
    {"VK_EXT_mutable_descriptor_type", 31},
    {"VK_EXT_nested_command_buffer", 29},
    {"VK_EXT_non_seamless_cube_map", 29},
    {"VK_EXT_opacity_micromap", 24},
    {"VK_EXT_pageable_device_local_memory", 36},
    {"VK_EXT_pci_bus_info", 20},
    {"VK_EXT_physical_device_drm", 27},
    {"VK_EXT_pipeline_creation_cache_control", 39},
    {"VK_EXT_pipeline_creation_feedback", 34},
    {"VK_EXT_pipeline_library_group_handles", 38},
    {"VK_EXT_pipeline_properties", 27},
    {"VK_EXT_pipeline_protected_access", 33},
    {"VK_EXT_pipeline_robustness", 27},
    {"VK_EXT_post_depth_coverage", 27},
    {"VK_EXT_present_mode_fifo_latest_ready", 38},
    {"VK_EXT_present_timing", 22},
    {"VK_EXT_primitive_restart_index", 31},
    {"VK_EXT_primitive_topology_list_restart", 39},
    {"VK_EXT_primitives_generated_query", 34},
    {"VK_EXT_private_data", 20},
    {"VK_EXT_provoking_vertex", 24},
    {"VK_EXT_queue_family_foreign", 28},
    {"VK_EXT_rasterization_order_attachment_access", 45},
    {"VK_EXT_ray_tracing_invocation_reorder", 38},
    {"VK_EXT_rgba10x6_formats", 24},
    {"VK_EXT_robustness2", 19},
    {"VK_EXT_sample_locations", 24},
    {"VK_EXT_sampler_filter_minmax", 29},
    {"VK_EXT_scalar_block_layout", 27},
    {"VK_EXT_separate_stencil_usage", 30},
    {"VK_EXT_shader_64bit_indexing", 29},
    {"VK_EXT_shader_atomic_float", 27},
    {"VK_EXT_shader_atomic_float2", 28},
    {"VK_EXT_shader_demote_to_helper_invocation", 42},
    {"VK_EXT_shader_float8", 21},
    {"VK_EXT_shader_image_atomic_int64", 33},
    {"VK_EXT_shader_long_vector", 26},
    {"VK_EXT_shader_module_identifier", 32},
    {"VK_EXT_shader_object", 21},
    {"VK_EXT_shader_ocp_microscaling_types", 37},
    {"VK_EXT_shader_replicated_composites", 36},
    {"VK_EXT_shader_split_barrier", 28},
    {"VK_EXT_shader_stencil_export", 29},
    {"VK_EXT_shader_subgroup_ballot", 30},
    {"VK_EXT_shader_subgroup_partitioned", 35},
    {"VK_EXT_shader_subgroup_vote", 28},
    {"VK_EXT_shader_tile_image", 25},
    {"VK_EXT_shader_uniform_buffer_unsized_array", 43},
    {"VK_EXT_shader_viewport_index_layer", 35},
    {"VK_EXT_subgroup_size_control", 29},
    {"VK_EXT_subpass_merge_feedback", 30},
    {"VK_EXT_surface_maintenance1", 28},
    {"VK_EXT_swapchain_colorspace", 28},
    {"VK_EXT_swapchain_maintenance1", 30},
    {"VK_EXT_texel_buffer_alignment", 30},
    {"VK_EXT_texture_compression_astc_3d", 35},
    {"VK_EXT_texture_compression_astc_hdr", 36},
    {"VK_EXT_tooling_info", 20},
    {"VK_EXT_transform_feedback", 26},
    {"VK_EXT_validation_cache", 24},
    {"VK_EXT_validation_features", 27},
    {"VK_EXT_validation_flags", 24},
    {"VK_EXT_vertex_attribute_divisor", 32},
    {"VK_EXT_vertex_attribute_robustness", 35},
    {"VK_EXT_vertex_input_dynamic_state", 34},
    {"VK_EXT_ycbcr_2plane_444_formats", 32},
    {"VK_EXT_ycbcr_image_arrays", 26},
    {"VK_EXT_zero_initialize_device_memory", 37},
    {"VK_FUCHSIA_buffer_collection", 29},
    {"VK_FUCHSIA_external_memory", 27},
    {"VK_FUCHSIA_external_semaphore", 30},
    {"VK_FUCHSIA_imagepipe_surface", 29},
    {"VK_GGP_frame_token", 19},
    {"VK_GGP_stream_descriptor_surface", 33},
    {"VK_GOOGLE_decorate_string", 26},
    {"VK_GOOGLE_display_timing", 25},
    {"VK_GOOGLE_hlsl_functionality1", 30},
    {"VK_GOOGLE_surfaceless_query", 28},
    {"VK_GOOGLE_user_type", 20},
    {"VK_HUAWEI_cluster_culling_shader", 33},
    {"VK_HUAWEI_hdr_vivid", 20},
    {"VK_HUAWEI_invocation_mask", 26},
    {"VK_HUAWEI_subpass_shading", 26},
    {"VK_IMG_filter_cubic", 20},
    {"VK_IMG_filter_linear_2d", 24},
    {"VK_IMG_format_pvrtc", 20},
    {"VK_IMG_relaxed_line_rasterization", 34},
    {"VK_INTEL_performance_query", 27},
    {"VK_INTEL_shader_integer_functions2", 35},
    {"VK_KHR_16bit_storage", 21},
    {"VK_KHR_8bit_storage", 20},
    {"VK_KHR_acceleration_structure", 30},
    {"VK_KHR_android_surface", 23},
    {"VK_KHR_bind_memory2", 20},
    {"VK_KHR_buffer_device_address", 29},
    {"VK_KHR_calibrated_timestamps", 29},
    {"VK_KHR_compute_shader_derivatives", 34},
    {"VK_KHR_cooperative_matrix", 26},
    {"VK_KHR_copy_commands2", 22},
    {"VK_KHR_copy_memory_indirect", 28},
    {"VK_KHR_create_renderpass2", 26},
    {"VK_KHR_dedicated_allocation", 28},
    {"VK_KHR_deferred_host_operations", 32},
    {"VK_KHR_depth_clamp_zero_one", 28},
    {"VK_KHR_depth_stencil_resolve", 29},
    {"VK_KHR_descriptor_update_template", 34},
    {"VK_KHR_device_address_commands", 31},
    {"VK_KHR_device_fault", 20},
    {"VK_KHR_device_group", 20},
    {"VK_KHR_device_group_creation", 29},
    {"VK_KHR_display", 15},
    {"VK_KHR_display_swapchain", 25},
    {"VK_KHR_draw_indirect_count", 27},
    {"VK_KHR_driver_properties", 25},
    {"VK_KHR_dynamic_rendering", 25},
    {"VK_KHR_dynamic_rendering_local_read", 36},
    {"VK_KHR_extended_flags", 22},
    {"VK_KHR_external_fence", 22},
    {"VK_KHR_external_fence_capabilities", 35},
    {"VK_KHR_external_fence_fd", 25},
    {"VK_KHR_external_fence_win32", 28},
    {"VK_KHR_external_memory", 23},
    {"VK_KHR_external_memory_capabilities", 36},
    {"VK_KHR_external_memory_fd", 26},
    {"VK_KHR_external_memory_win32", 29},
    {"VK_KHR_external_semaphore", 26},
    {"VK_KHR_external_semaphore_capabilities", 39},
    {"VK_KHR_external_semaphore_fd", 29},
    {"VK_KHR_external_semaphore_win32", 32},
    {"VK_KHR_format_feature_flags2", 29},
    {"VK_KHR_fragment_shader_barycentric", 35},
    {"VK_KHR_fragment_shading_rate", 29},
    {"VK_KHR_get_display_properties2", 31},
    {"VK_KHR_get_memory_requirements2", 32},
    {"VK_KHR_get_physical_device_properties2", 39},
    {"VK_KHR_get_surface_capabilities2", 33},
    {"VK_KHR_global_priority", 23},
    {"VK_KHR_image_format_list", 25},
    {"VK_KHR_imageless_framebuffer", 29},
    {"VK_KHR_incremental_present", 27},
    {"VK_KHR_index_type_uint8", 24},
    {"VK_KHR_internally_synchronized_queues", 38},
    {"VK_KHR_line_rasterization", 26},
    {"VK_KHR_load_store_op_none", 26},
    {"VK_KHR_maintenance1", 20},
    {"VK_KHR_maintenance10", 21},
    {"VK_KHR_maintenance11", 21},
    {"VK_KHR_maintenance2", 20},
    {"VK_KHR_maintenance3", 20},
    {"VK_KHR_maintenance4", 20},
    {"VK_KHR_maintenance5", 20},
    {"VK_KHR_maintenance6", 20},
    {"VK_KHR_maintenance7", 20},
    {"VK_KHR_maintenance8", 20},
    {"VK_KHR_maintenance9", 20},
    {"VK_KHR_map_memory2", 19},
    {"VK_KHR_multiview", 17},
    {"VK_KHR_opacity_micromap", 24},
    {"VK_KHR_performance_query", 25},
    {"VK_KHR_pipeline_binary", 23},
    {"VK_KHR_pipeline_executable_properties", 38},
    {"VK_KHR_pipeline_library", 24},
    {"VK_KHR_portability_enumeration", 31},
    {"VK_KHR_portability_subset", 26},
    {"VK_KHR_present_id", 18},
    {"VK_KHR_present_id2", 19},
    {"VK_KHR_present_mode_fifo_latest_ready", 38},
    {"VK_KHR_present_wait", 20},
    {"VK_KHR_present_wait2", 21},
    {"VK_KHR_push_descriptor", 23},
    {"VK_KHR_ray_query", 17},
    {"VK_KHR_ray_tracing_maintenance1", 32},
    {"VK_KHR_ray_tracing_pipeline", 28},
    {"VK_KHR_ray_tracing_position_fetch", 34},
    {"VK_KHR_relaxed_block_layout", 28},
    {"VK_KHR_robustness2", 19},
    {"VK_KHR_sampler_mirror_clamp_to_edge", 36},
    {"VK_KHR_sampler_ycbcr_conversion", 32},
    {"VK_KHR_separate_depth_stencil_layouts", 38},
    {"VK_KHR_shader_abort", 20},
    {"VK_KHR_shader_atomic_int64", 27},
    {"VK_KHR_shader_bfloat16", 23},
    {"VK_KHR_shader_clock", 20},
    {"VK_KHR_shader_constant_data", 28},
    {"VK_KHR_shader_draw_parameters", 30},
    {"VK_KHR_shader_expect_assume", 28},
    {"VK_KHR_shader_float16_int8", 27},
    {"VK_KHR_shader_float_controls", 29},
    {"VK_KHR_shader_float_controls2", 30},
    {"VK_KHR_shader_fma", 18},
    {"VK_KHR_shader_integer_dot_product", 34},
    {"VK_KHR_shader_maximal_reconvergence", 36},
    {"VK_KHR_shader_non_semantic_info", 32},
    {"VK_KHR_shader_quad_control", 27},
    {"VK_KHR_shader_relaxed_extended_instruction", 43},
    {"VK_KHR_shader_subgroup_extended_types", 38},
    {"VK_KHR_shader_subgroup_rotate", 30},
    {"VK_KHR_shader_subgroup_uniform_control_flow", 44},
    {"VK_KHR_shader_terminate_invocation", 35},
    {"VK_KHR_shader_untyped_pointers", 31},
    {"VK_KHR_shared_presentable_image", 32},
    {"VK_KHR_spirv_1_4", 17},
    {"VK_KHR_storage_buffer_storage_class", 36},
    {"VK_KHR_surface", 15},
    {"VK_KHR_surface_maintenance1", 28},
    {"VK_KHR_surface_protected_capabilities", 38},
    {"VK_KHR_swapchain", 17},
    {"VK_KHR_swapchain_maintenance1", 30},
    {"VK_KHR_swapchain_mutable_format", 32},
    {"VK_KHR_synchronization2", 24},
    {"VK_KHR_timeline_semaphore", 26},
    {"VK_KHR_unified_image_layouts", 29},
    {"VK_KHR_uniform_buffer_standard_layout", 38},
    {"VK_KHR_variable_pointers", 25},
    {"VK_KHR_vertex_attribute_divisor", 32},
    {"VK_KHR_video_decode_av1", 24},
    {"VK_KHR_video_decode_h264", 25},
    {"VK_KHR_video_decode_h265", 25},
    {"VK_KHR_video_decode_queue", 26},
    {"VK_KHR_video_decode_vp9", 24},
    {"VK_KHR_video_encode_av1", 24},
    {"VK_KHR_video_encode_feedback2", 30},
    {"VK_KHR_video_encode_h264", 25},
    {"VK_KHR_video_encode_h265", 25},
    {"VK_KHR_video_encode_intra_refresh", 34},
    {"VK_KHR_video_encode_quantization_map", 37},
    {"VK_KHR_video_encode_queue", 26},
    {"VK_KHR_video_maintenance1", 26},
    {"VK_KHR_video_maintenance2", 26},
    {"VK_KHR_video_queue", 19},
    {"VK_KHR_vulkan_memory_model", 27},
    {"VK_KHR_wayland_surface", 23},
    {"VK_KHR_win32_keyed_mutex", 25},
    {"VK_KHR_win32_surface", 21},
    {"VK_KHR_workgroup_memory_explicit_layout", 40},
    {"VK_KHR_xcb_surface", 19},
    {"VK_KHR_xlib_surface", 20},
    {"VK_KHR_zero_initialize_workgroup_memory", 40},
    {"VK_LUNARG_direct_driver_loading", 32},
    {"VK_MESA_image_alignment_control", 32},
    {"VK_MSFT_layered_driver", 23},
    {"VK_MVK_ios_surface", 19},
    {"VK_MVK_macos_surface", 21},
    {"VK_NN_vi_surface", 17},
    {"VK_NVX_binary_import", 21},
    {"VK_NVX_image_view_handle", 25},
    {"VK_NVX_multiview_per_view_attributes", 37},
    {"VK_NV_acquire_winrt_display", 28},
    {"VK_NV_clip_space_w_scaling", 27},
    {"VK_NV_cluster_acceleration_structure", 37},
    {"VK_NV_command_buffer_inheritance", 33},
    {"VK_NV_compute_occupancy_priority", 33},
    {"VK_NV_compute_shader_derivatives", 33},
    {"VK_NV_cooperative_matrix", 25},
    {"VK_NV_cooperative_matrix2", 26},
    {"VK_NV_cooperative_matrix_decode_vector", 39},
    {"VK_NV_cooperative_vector", 25},
    {"VK_NV_copy_memory_indirect", 27},
    {"VK_NV_corner_sampled_image", 27},
    {"VK_NV_coverage_reduction_mode", 30},
    {"VK_NV_cuda_kernel_launch", 25},
    {"VK_NV_dedicated_allocation", 27},
    {"VK_NV_dedicated_allocation_image_aliasing", 42},
    {"VK_NV_descriptor_pool_overallocation", 37},
    {"VK_NV_device_diagnostic_checkpoints", 36},
    {"VK_NV_device_diagnostics_config", 32},
    {"VK_NV_device_generated_commands", 32},
    {"VK_NV_device_generated_commands_compute", 40},
    {"VK_NV_displacement_micromap", 28},
    {"VK_NV_display_stereo", 21},
    {"VK_NV_extended_sparse_address_space", 36},
    {"VK_NV_external_compute_queue", 29},
    {"VK_NV_external_memory", 22},
    {"VK_NV_external_memory_capabilities", 35},
    {"VK_NV_external_memory_rdma", 27},
    {"VK_NV_external_memory_win32", 28},
    {"VK_NV_fill_rectangle", 21},
    {"VK_NV_fragment_coverage_to_color", 33},
    {"VK_NV_fragment_shader_barycentric", 34},
    {"VK_NV_fragment_shading_rate_enums", 34},
    {"VK_NV_framebuffer_mixed_samples", 32},
    {"VK_NV_geometry_shader_passthrough", 34},
    {"VK_NV_glsl_shader", 18},
    {"VK_NV_inherited_viewport_scissor", 33},
    {"VK_NV_linear_color_attachment", 30},
    {"VK_NV_low_latency", 18},
    {"VK_NV_low_latency2", 19},
    {"VK_NV_memory_decompression", 27},
    {"VK_NV_mesh_shader", 18},
    {"VK_NV_optical_flow", 19},
    {"VK_NV_partitioned_acceleration_structure", 41},
    {"VK_NV_per_stage_descriptor_set", 31},
    {"VK_NV_present_barrier", 22},
    {"VK_NV_present_metering", 23},
    {"VK_NV_push_constant_bank", 25},
    {"VK_NV_raw_access_chains", 24},
    {"VK_NV_ray_tracing", 18},
    {"VK_NV_ray_tracing_invocation_reorder", 37},
    {"VK_NV_ray_tracing_linear_swept_spheres", 39},
    {"VK_NV_ray_tracing_motion_blur", 30},
    {"VK_NV_ray_tracing_validation", 29},
    {"VK_NV_representative_fragment_test", 35},
    {"VK_NV_sample_mask_override_coverage", 36},
    {"VK_NV_scissor_exclusive", 24},
    {"VK_NV_shader_atomic_float16_vector", 35},
    {"VK_NV_shader_image_footprint", 29},
    {"VK_NV_shader_sm_builtins", 25},
    {"VK_NV_shader_subgroup_partitioned", 34},
    {"VK_NV_shading_rate_image", 25},
    {"VK_NV_viewport_array2", 22},
    {"VK_NV_viewport_swizzle", 23},
    {"VK_NV_win32_keyed_mutex", 24},
    {"VK_OHOS_external_memory", 24},
    {"VK_OHOS_surface", 16},
    {"VK_QCOM_cooperative_matrix_conversion", 38},
    {"VK_QCOM_data_graph_model", 25},
    {"VK_QCOM_elapsed_timer_query", 28},
    {"VK_QCOM_filter_cubic_clamp", 27},
    {"VK_QCOM_filter_cubic_weights", 29},
    {"VK_QCOM_fragment_density_map_offset", 36},
    {"VK_QCOM_image_processing", 25},
    {"VK_QCOM_image_processing2", 26},
    {"VK_QCOM_image_processing3", 26},
    {"VK_QCOM_multiview_per_view_render_areas", 40},
    {"VK_QCOM_multiview_per_view_viewports", 37},
    {"VK_QCOM_queue_perf_hint", 24},
    {"VK_QCOM_render_pass_shader_resolve", 35},
    {"VK_QCOM_render_pass_store_ops", 30},
    {"VK_QCOM_render_pass_transform", 30},
    {"VK_QCOM_rotated_copy_commands", 30},
    {"VK_QCOM_shader_multiple_wait_queues", 36},
    {"VK_QCOM_tile_memory_heap", 25},
    {"VK_QCOM_tile_properties", 24},
    {"VK_QCOM_tile_shading", 21},
    {"VK_QCOM_ycbcr_degamma", 22},
    {"VK_QNX_external_memory_screen_buffer", 37},
    {"VK_QNX_screen_surface", 22},
    {"VK_SEC_amigo_profiling", 23},
    {"VK_SEC_pipeline_cache_incremental_mode", 39},
    {"VK_SEC_throttle_hint", 21},
    {"VK_SEC_ubm_surface", 19},
    {"VK_VALVE_descriptor_set_host_mapping", 37},
    {"VK_VALVE_fragment_density_map_layered", 38},
    {"VK_VALVE_mutable_descriptor_type", 33},
    {"VK_VALVE_shader_mixed_float_dot_product", 40},
    {"VK_VALVE_video_encode_rgb_conversion", 37},
    };
    return table[(int)extension].data();
}

bool IsFieldPointer(Field field) {
    switch (field) {
    case Field::address:
    case Field::connection:
    case Field::context:
    case Field::device:
    case Field::dfb:
    case Field::displayName:
    case Field::dpy:
    case Field::handle:
    case Field::hostAddress:
    case Field::opaqueCaptureDescriptorData:
    case Field::pAccelerationStructure:
    case Field::pAccelerationStructures:
    case Field::pAcquireInfo:
    case Field::pAcquireKeys:
    case Field::pAcquireSyncs:
    case Field::pAcquireTimeoutMilliseconds:
    case Field::pAcquireTimeouts:
    case Field::pAddress:
    case Field::pAddressInfos:
    case Field::pAddressRange:
    case Field::pAllocateInfo:
    case Field::pAllocator:
    case Field::pApiVersion:
    case Field::pApplicationInfo:
    case Field::pApplicationName:
    case Field::pAspectReferences:
    case Field::pAttachmentImageInfos:
    case Field::pAttachmentInitialSampleLocations:
    case Field::pAttachments:
    case Field::pAttributes:
    case Field::pBeginCustomResolveInfo:
    case Field::pBeginInfo:
    case Field::pBinaries:
    case Field::pBindDescriptorBufferEmbeddedSamplersInfo:
    case Field::pBindDescriptorSetsInfo:
    case Field::pBindInfo:
    case Field::pBindInfos:
    case Field::pBindPointRequirementCount:
    case Field::pBindPointRequirements:
    case Field::pBindSessionMemoryInfos:
    case Field::pBindingFlags:
    case Field::pBindingInfos:
    case Field::pBindingReference:
    case Field::pBindings:
    case Field::pBinds:
    case Field::pBlitImageInfo:
    case Field::pBorderColor:
    case Field::pBuffer:
    case Field::pBufferBinds:
    case Field::pBufferConstraintsInfo:
    case Field::pBufferIndices:
    case Field::pBufferInfo:
    case Field::pBufferMemoryBarriers:
    case Field::pBuffers:
    case Field::pBuildInfo:
    case Field::pCacheData:
    case Field::pCacheSize:
    case Field::pCallableShaderBindingTable:
    case Field::pCallback:
    case Field::pCallbackData:
    case Field::pCapabilities:
    case Field::pCheckpointData:
    case Field::pCheckpointDataCount:
    case Field::pCheckpointMarker:
    case Field::pClearValues:
    case Field::pClustersBottomLevel:
    case Field::pCmdBufLabels:
    case Field::pCode:
    case Field::pCodingControlInfo:
    case Field::pCollection:
    case Field::pColor:
    case Field::pColorAttachmentFormats:
    case Field::pColorAttachmentInputIndices:
    case Field::pColorAttachmentLocations:
    case Field::pColorAttachmentSamples:
    case Field::pColorAttachments:
    case Field::pColorBlendAdvanced:
    case Field::pColorBlendEnables:
    case Field::pColorBlendEquations:
    case Field::pColorBlendState:
    case Field::pColorSpaces:
    case Field::pColorWriteEnables:
    case Field::pColorWriteMasks:
    case Field::pCombinationCount:
    case Field::pCombinations:
    case Field::pCombinedImageSampler:
    case Field::pCommandBufferDeviceMasks:
    case Field::pCommandBufferInfos:
    case Field::pCommandBuffers:
    case Field::pCommandInfos:
    case Field::pCommandPool:
    case Field::pCommittedMemoryInBytes:
    case Field::pCompatibility:
    case Field::pConditionalRenderingBegin:
    case Field::pConfiguration:
    case Field::pConnections:
    case Field::pConstantData:
    case Field::pConstants:
    case Field::pCopyBufferInfo:
    case Field::pCopyBufferToImageInfo:
    case Field::pCopyDstLayouts:
    case Field::pCopyImageInfo:
    case Field::pCopyImageToBufferInfo:
    case Field::pCopyImageToImageInfo:
    case Field::pCopyImageToMemoryInfo:
    case Field::pCopyMemoryIndirectInfo:
    case Field::pCopyMemoryInfo:
    case Field::pCopyMemoryToImageIndirectInfo:
    case Field::pCopyMemoryToImageInfo:
    case Field::pCopySrcLayouts:
    case Field::pCopyTensorInfo:
    case Field::pCorrelatedViewMasks:
    case Field::pCorrelationMasks:
    case Field::pCountInfo:
    case Field::pCounterAddresses:
    case Field::pCounterBufferOffsets:
    case Field::pCounterBuffers:
    case Field::pCounterCount:
    case Field::pCounterDescriptions:
    case Field::pCounterIndices:
    case Field::pCounterInfo:
    case Field::pCounterInfos:
    case Field::pCounterValue:
    case Field::pCounters:
    case Field::pCoverageModulationTable:
    case Field::pCreateInfo:
    case Field::pCreateInfos:
    case Field::pCustomSampleOrders:
    case Field::pData:
    case Field::pDataSize:
    case Field::pDatas:
    case Field::pDebugInfo:
    case Field::pDecodeInfo:
    case Field::pDecompressMemoryInfoEXT:
    case Field::pDecompressMemoryRegions:
    case Field::pDeferredOperation:
    case Field::pDependencies:
    case Field::pDependencyInfo:
    case Field::pDependencyInfos:
    case Field::pDepthAttachment:
    case Field::pDepthBiasInfo:
    case Field::pDepthClampRange:
    case Field::pDepthInputAttachmentIndex:
    case Field::pDepthStencil:
    case Field::pDepthStencilAttachment:
    case Field::pDepthStencilResolveAttachment:
    case Field::pDepthStencilState:
    case Field::pDescription:
    case Field::pDescriptionCount:
    case Field::pDescriptions:
    case Field::pDescriptor:
    case Field::pDescriptorCopies:
    case Field::pDescriptorCounts:
    case Field::pDescriptorInfo:
    case Field::pDescriptorPool:
    case Field::pDescriptorSets:
    case Field::pDescriptorTypes:
    case Field::pDescriptorUpdateEntries:
    case Field::pDescriptorUpdateTemplate:
    case Field::pDescriptorWrites:
    case Field::pDescriptors:
    case Field::pDevice:
    case Field::pDeviceEventInfo:
    case Field::pDeviceGroupPresentCapabilities:
    case Field::pDeviceIndices:
    case Field::pDeviceMasks:
    case Field::pDeviceRenderAreas:
    case Field::pDimensions:
    case Field::pDisabledValidationChecks:
    case Field::pDisabledValidationFeatures:
    case Field::pDiscardRectangles:
    case Field::pDispatchParameters:
    case Field::pDispatchTileInfo:
    case Field::pDisplay:
    case Field::pDisplayCount:
    case Field::pDisplayEventInfo:
    case Field::pDisplayPlaneInfo:
    case Field::pDisplayPowerInfo:
    case Field::pDisplayTimingProperties:
    case Field::pDisplays:
    case Field::pDrivers:
    case Field::pDrmFormatModifierProperties:
    case Field::pDrmFormatModifiers:
    case Field::pDstOffset:
    case Field::pDstRange:
    case Field::pDstSize:
    case Field::pDynamicMetadata:
    case Field::pDynamicOffsets:
    case Field::pDynamicState:
    case Field::pDynamicStates:
    case Field::pEmbeddedSampler:
    case Field::pEnabledFeatures:
    case Field::pEnabledValidationFeatures:
    case Field::pEncodeInfo:
    case Field::pEndCodingInfo:
    case Field::pEngineName:
    case Field::pEvent:
    case Field::pEvents:
    case Field::pExclusiveScissorEnables:
    case Field::pExclusiveScissors:
    case Field::pExecutableCount:
    case Field::pExecutableInfo:
    case Field::pExecuteInfo:
    case Field::pExecutionSet:
    case Field::pExecutionSetWrites:
    case Field::pExtensions:
    case Field::pExtent:
    case Field::pExternalBufferInfo:
    case Field::pExternalBufferProperties:
    case Field::pExternalFenceInfo:
    case Field::pExternalFenceProperties:
    case Field::pExternalImageFormatProperties:
    case Field::pExternalQueue:
    case Field::pExternalSemaphoreInfo:
    case Field::pExternalSemaphoreProperties:
    case Field::pExternalTensorInfo:
    case Field::pExternalTensorProperties:
    case Field::pExtras:
    case Field::pFaultCounts:
    case Field::pFaultInfo:
    case Field::pFd:
    case Field::pFeatures:
    case Field::pFeedbackInfo:
    case Field::pFence:
    case Field::pFences:
    case Field::pFixedRateFlags:
    case Field::pFormatConstraints:
    case Field::pFormatCount:
    case Field::pFormatInfo:
    case Field::pFormatProperties:
    case Field::pFragmentDensityOffsets:
    case Field::pFragmentShadingRateAttachment:
    case Field::pFragmentShadingRateCount:
    case Field::pFragmentShadingRates:
    case Field::pFragmentSize:
    case Field::pFramebuffer:
    case Field::pFunction:
    case Field::pGeneratedCommandsInfo:
    case Field::pGeometries:
    case Field::pGetFdInfo:
    case Field::pGetMetalHandleInfo:
    case Field::pGetWin32HandleInfo:
    case Field::pGetZirconHandleInfo:
    case Field::pGpaSampleBeginInfo:
    case Field::pGpaSession:
    case Field::pGranularity:
    case Field::pGroups:
    case Field::pHandle:
    case Field::pHitShaderBindingTable:
    case Field::pHostMapping:
    case Field::pHostPointer:
    case Field::pIdentifier:
    case Field::pImage:
    case Field::pImageBinds:
    case Field::pImageConstraintsInfo:
    case Field::pImageFormatInfo:
    case Field::pImageFormatProperties:
    case Field::pImageIndex:
    case Field::pImageIndices:
    case Field::pImageInfo:
    case Field::pImageMemoryBarriers:
    case Field::pImageOpaqueBinds:
    case Field::pImageSubresources:
    case Field::pImages:
    case Field::pImmutableSamplers:
    case Field::pImportFenceFdInfo:
    case Field::pImportFenceWin32HandleInfo:
    case Field::pImportSemaphoreFdInfo:
    case Field::pImportSemaphoreWin32HandleInfo:
    case Field::pImportSemaphoreZirconHandleInfo:
    case Field::pIndex:
    case Field::pIndexBuffer:
    case Field::pIndexInfo:
    case Field::pIndexTypeValues:
    case Field::pIndexTypes:
    case Field::pIndirectCommandsLayout:
    case Field::pIndirectDeviceAddresses:
    case Field::pIndirectExecutionSet:
    case Field::pIndirectStrides:
    case Field::pInfo:
    case Field::pInfoSize:
    case Field::pInfos:
    case Field::pInheritanceInfo:
    case Field::pInitialData:
    case Field::pInitialShaders:
    case Field::pInitializeInfo:
    case Field::pInputAssemblyState:
    case Field::pInputAttachmentImage:
    case Field::pInputAttachmentIndexInfo:
    case Field::pInputAttachments:
    case Field::pInstance:
    case Field::pInstrumentation:
    case Field::pInternalRepresentationCount:
    case Field::pInternalRepresentations:
    case Field::pKeysAndDataInfo:
    case Field::pLabelInfo:
    case Field::pLabelName:
    case Field::pLatencyMarkerInfo:
    case Field::pLaunchInfo:
    case Field::pLayer:
    case Field::pLayerName:
    case Field::pLayerPrefix:
    case Field::pLayeredApis:
    case Field::pLayers:
    case Field::pLayout:
    case Field::pLayoutSizeInBytes:
    case Field::pLibraries:
    case Field::pLibraryInfo:
    case Field::pLibraryInterface:
    case Field::pLocationInfo:
    case Field::pLowLatencyMode:
    case Field::pMapEntries:
    case Field::pMappings:
    case Field::pMarkerInfo:
    case Field::pMarkerName:
    case Field::pMaxDeviation:
    case Field::pMaxPrimitiveCounts:
    case Field::pMaxWorkgroupSize:
    case Field::pMemory:
    case Field::pMemoryBarriers:
    case Field::pMemoryFdProperties:
    case Field::pMemoryGetRemoteAddressInfo:
    case Field::pMemoryHostPointerProperties:
    case Field::pMemoryMapInfo:
    case Field::pMemoryMetalHandleProperties:
    case Field::pMemoryProperties:
    case Field::pMemoryRangeBarriers:
    case Field::pMemoryRanges:
    case Field::pMemoryRequirements:
    case Field::pMemoryRequirementsCount:
    case Field::pMemoryUnmapInfo:
    case Field::pMemoryWin32HandleProperties:
    case Field::pMemoryZirconHandleProperties:
    case Field::pMessage:
    case Field::pMessageData:
    case Field::pMessageIdName:
    case Field::pMessenger:
    case Field::pMetadata:
    case Field::pMetalObjectsInfo:
    case Field::pMetricBlockCount:
    case Field::pMetricValues:
    case Field::pMicromap:
    case Field::pMicromaps:
    case Field::pMissShaderBindingTable:
    case Field::pMode:
    case Field::pModes:
    case Field::pModule:
    case Field::pMoveObjects:
    case Field::pMultisampleProperties:
    case Field::pMultisampleState:
    case Field::pMutableDescriptorTypeLists:
    case Field::pNaluSliceEntries:
    case Field::pNaluSliceSegmentEntries:
    case Field::pName:
    case Field::pNameInfo:
    case Field::pNext:
    case Field::pNodeIndex:
    case Field::pNodeInfo:
    case Field::pNumPasses:
    case Field::pObjectName:
    case Field::pObjects:
    case Field::pOffset:
    case Field::pOffsets:
    case Field::pOperation:
    case Field::pOpticalFlowImageFormatInfo:
    case Field::pOverrideInfo:
    case Field::pParameters:
    case Field::pParametersAddInfo:
    case Field::pParams:
    case Field::pPastPresentationTimingInfo:
    case Field::pPastPresentationTimingProperties:
    case Field::pPeerMemoryFeatures:
    case Field::pPerTileBeginInfo:
    case Field::pPerTileEndInfo:
    case Field::pPerViewRenderAreas:
    case Field::pPerfBlocks:
    case Field::pPerfCounters:
    case Field::pPerfHintInfo:
    case Field::pPerformanceQueryCreateInfo:
    case Field::pPhysicalDeviceCount:
    case Field::pPhysicalDeviceGroupCount:
    case Field::pPhysicalDeviceGroupProperties:
    case Field::pPhysicalDevices:
    case Field::pPictureResource:
    case Field::pPipelineBinaries:
    case Field::pPipelineBinaryData:
    case Field::pPipelineBinaryDataSize:
    case Field::pPipelineBinaryKey:
    case Field::pPipelineBinaryKeys:
    case Field::pPipelineCache:
    case Field::pPipelineCreateInfo:
    case Field::pPipelineCreationFeedback:
    case Field::pPipelineInfo:
    case Field::pPipelineKey:
    case Field::pPipelineLayout:
    case Field::pPipelineProperties:
    case Field::pPipelineStageCreationFeedbacks:
    case Field::pPipelines:
    case Field::pPlacedAddress:
    case Field::pPlaneLayouts:
    case Field::pPoolSizes:
    case Field::pPostSubpassSampleLocations:
    case Field::pPresentIds:
    case Field::pPresentInfo:
    case Field::pPresentModeCount:
    case Field::pPresentModes:
    case Field::pPresentStages:
    case Field::pPresentWait2Info:
    case Field::pPresentationInfo:
    case Field::pPresentationTimingCount:
    case Field::pPresentationTimings:
    case Field::pPreserveAttachments:
    case Field::pPrivateData:
    case Field::pPrivateDataSlot:
    case Field::pProcessingEngines:
    case Field::pProfiles:
    case Field::pProperties:
    case Field::pPropertiesCount:
    case Field::pPropertyCount:
    case Field::pPushConstant:
    case Field::pPushConstantRanges:
    case Field::pPushConstantsInfo:
    case Field::pPushDataInfo:
    case Field::pPushDescriptorSetInfo:
    case Field::pPushDescriptorSetWithTemplateInfo:
    case Field::pQualityLevelInfo:
    case Field::pQualityLevelProperties:
    case Field::pQueriedLowLatencyData:
    case Field::pQueryPool:
    case Field::pQueue:
    case Field::pQueueCreateInfos:
    case Field::pQueueFamilyDataGraphProcessingEngineInfo:
    case Field::pQueueFamilyDataGraphProcessingEngineProperties:
    case Field::pQueueFamilyDataGraphProperties:
    case Field::pQueueFamilyDataGraphPropertyCount:
    case Field::pQueueFamilyIndices:
    case Field::pQueueFamilyProperties:
    case Field::pQueueFamilyPropertyCount:
    case Field::pQueueInfo:
    case Field::pQueueLabels:
    case Field::pQueuePriorities:
    case Field::pQueueTypeInfo:
    case Field::pRanges:
    case Field::pRasterizationState:
    case Field::pRaygenShaderBindingTable:
    case Field::pRectCount:
    case Field::pRectangles:
    case Field::pRects:
    case Field::pReferenceSlots:
    case Field::pRegions:
    case Field::pReleaseInfo:
    case Field::pReleaseKeys:
    case Field::pReleaseSyncs:
    case Field::pRenderPass:
    case Field::pRenderPassBegin:
    case Field::pRenderPassFeedback:
    case Field::pRenderingAreaInfo:
    case Field::pRenderingEndInfo:
    case Field::pRenderingInfo:
    case Field::pResolveAttachments:
    case Field::pResolveImageInfo:
    case Field::pResourceHeapBindInfo:
    case Field::pResourceInfos:
    case Field::pResources:
    case Field::pResult:
    case Field::pResults:
    case Field::pSampleID:
    case Field::pSampleLocations:
    case Field::pSampleLocationsInfo:
    case Field::pSampleMask:
    case Field::pSampledImage:
    case Field::pSampler:
    case Field::pSamplerHeapBindInfo:
    case Field::pSamplers:
    case Field::pScissors:
    case Field::pSemaphore:
    case Field::pSemaphores:
    case Field::pSession:
    case Field::pSetDescriptorBufferOffsetsInfo:
    case Field::pSetLayout:
    case Field::pSetLayoutInfos:
    case Field::pSetLayouts:
    case Field::pSettingName:
    case Field::pSettings:
    case Field::pSetupReferenceSlot:
    case Field::pShaderGroupCaptureReplayHandle:
    case Field::pShaderInfo:
    case Field::pShaderModule:
    case Field::pShaders:
    case Field::pShadingRatePaletteEntries:
    case Field::pShadingRatePalettes:
    case Field::pSignalInfo:
    case Field::pSignalSemaphoreDeviceIndices:
    case Field::pSignalSemaphoreInfos:
    case Field::pSignalSemaphoreValues:
    case Field::pSignalSemaphores:
    case Field::pSizeInBytes:
    case Field::pSizeInfo:
    case Field::pSizes:
    case Field::pSleepInfo:
    case Field::pSleepModeInfo:
    case Field::pSliceOffsets:
    case Field::pSliceSegmentOffsets:
    case Field::pSparseMemoryRequirementCount:
    case Field::pSparseMemoryRequirements:
    case Field::pSpecializationInfo:
    case Field::pSplitInstanceBindRegions:
    case Field::pSrcCaches:
    case Field::pSrcOffset:
    case Field::pStages:
    case Field::pStatisticCount:
    case Field::pStatistics:
    case Field::pStdDecoderModelInfo:
    case Field::pStdHeaderVersion:
    case Field::pStdOperatingPoints:
    case Field::pStdPPS:
    case Field::pStdPPSs:
    case Field::pStdPictureInfo:
    case Field::pStdReferenceInfo:
    case Field::pStdSPS:
    case Field::pStdSPSs:
    case Field::pStdSequenceHeader:
    case Field::pStdSliceHeader:
    case Field::pStdSliceSegmentHeader:
    case Field::pStdVPS:
    case Field::pStdVPSs:
    case Field::pStencilAttachment:
    case Field::pStencilInputAttachmentIndex:
    case Field::pStorageBuffer:
    case Field::pStorageImage:
    case Field::pStorageTexelBuffer:
    case Field::pStreamStrides:
    case Field::pStreams:
    case Field::pStrides:
    case Field::pStripeInfos:
    case Field::pStripeSemaphoreInfos:
    case Field::pSubmits:
    case Field::pSubpassBeginInfo:
    case Field::pSubpassEndInfo:
    case Field::pSubpassFeedback:
    case Field::pSubpasses:
    case Field::pSubresource:
    case Field::pSupport:
    case Field::pSupported:
    case Field::pSurface:
    case Field::pSurfaceCapabilities:
    case Field::pSurfaceFormatCount:
    case Field::pSurfaceFormats:
    case Field::pSurfaceInfo:
    case Field::pSwapchain:
    case Field::pSwapchainImageCount:
    case Field::pSwapchainImages:
    case Field::pSwapchainTimeDomainProperties:
    case Field::pSwapchainTimingProperties:
    case Field::pSwapchainTimingPropertiesCounter:
    case Field::pSwapchains:
    case Field::pTag:
    case Field::pTagInfo:
    case Field::pTensor:
    case Field::pTensorARM:
    case Field::pTensorMemoryBarriers:
    case Field::pTensorViews:
    case Field::pTensors:
    case Field::pTessellationState:
    case Field::pTexelBuffer:
    case Field::pTexelBufferView:
    case Field::pTileMemoryBindInfo:
    case Field::pTileOffsets:
    case Field::pTileSizes:
    case Field::pTimeDomainCount:
    case Field::pTimeDomainIds:
    case Field::pTimeDomains:
    case Field::pTimeDomainsCounter:
    case Field::pTimes:
    case Field::pTimestampInfos:
    case Field::pTimestamps:
    case Field::pTimingInfos:
    case Field::pTimings:
    case Field::pTokens:
    case Field::pToolCount:
    case Field::pToolProperties:
    case Field::pTransitions:
    case Field::pTriangleClusters:
    case Field::pUniformBuffer:
    case Field::pUniformTexelBuffer:
    case Field::pUpdateInfo:
    case Field::pUsageCounts:
    case Field::pUserData:
    case Field::pValidationCache:
    case Field::pValue:
    case Field::pValues:
    case Field::pVendorBinaryData:
    case Field::pVendorInfos:
    case Field::pVendorOptions:
    case Field::pVersionData:
    case Field::pVersionInfo:
    case Field::pVertexAttributeDescriptions:
    case Field::pVertexBindingDescriptions:
    case Field::pVertexBindingDivisors:
    case Field::pVertexBuffer:
    case Field::pVertexInfo:
    case Field::pVertexInputState:
    case Field::pVertexOffset:
    case Field::pVideoFormatInfo:
    case Field::pVideoFormatProperties:
    case Field::pVideoFormatPropertyCount:
    case Field::pVideoProfile:
    case Field::pVideoSession:
    case Field::pVideoSessionParameters:
    case Field::pVideoSessionParametersInfo:
    case Field::pView:
    case Field::pViewFormats:
    case Field::pViewMasks:
    case Field::pViewOffsets:
    case Field::pViewportDepths:
    case Field::pViewportState:
    case Field::pViewportSwizzles:
    case Field::pViewportWScalings:
    case Field::pViewports:
    case Field::pWaitDstStageMask:
    case Field::pWaitInfo:
    case Field::pWaitSemaphoreDeviceIndices:
    case Field::pWaitSemaphoreInfos:
    case Field::pWaitSemaphoreValues:
    case Field::pWaitSemaphores:
    case Field::pYcbcrConversion:
    case Field::pZirconHandle:
    case Field::params:
    case Field::pfnAllocation:
    case Field::pfnCallback:
    case Field::pfnFree:
    case Field::pfnGetInstanceProcAddr:
    case Field::pfnInternalAllocation:
    case Field::pfnInternalFree:
    case Field::pfnReallocation:
    case Field::pfnUserCallback:
    case Field::ppBuildRangeInfos:
    case Field::ppData:
    case Field::ppEnabledExtensionNames:
    case Field::ppEnabledLayerNames:
    case Field::ppGeometries:
    case Field::ppMaxPrimitiveCounts:
    case Field::ppUsageCounts:
    case Field::valueString:
        return true;
    default:
        return false;
    }
}

Struct StypeToStruct(VkStructureType stype) {
    switch (stype) {
    case VK_STRUCTURE_TYPE_APPLICATION_INFO:
       return Struct::VkApplicationInfo;
    case VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO:
       return Struct::VkInstanceCreateInfo;
    case VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO:
       return Struct::VkDeviceQueueCreateInfo;
    case VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO:
       return Struct::VkDeviceCreateInfo;
    case VK_STRUCTURE_TYPE_SUBMIT_INFO:
       return Struct::VkSubmitInfo;
    case VK_STRUCTURE_TYPE_MAPPED_MEMORY_RANGE:
       return Struct::VkMappedMemoryRange;
    case VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO:
       return Struct::VkMemoryAllocateInfo;
    case VK_STRUCTURE_TYPE_BIND_SPARSE_INFO:
       return Struct::VkBindSparseInfo;
    case VK_STRUCTURE_TYPE_FENCE_CREATE_INFO:
       return Struct::VkFenceCreateInfo;
    case VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO:
       return Struct::VkSemaphoreCreateInfo;
    case VK_STRUCTURE_TYPE_QUERY_POOL_CREATE_INFO:
       return Struct::VkQueryPoolCreateInfo;
    case VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO:
       return Struct::VkBufferCreateInfo;
    case VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO:
       return Struct::VkImageCreateInfo;
    case VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO:
       return Struct::VkImageViewCreateInfo;
    case VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO:
       return Struct::VkCommandPoolCreateInfo;
    case VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO:
       return Struct::VkCommandBufferAllocateInfo;
    case VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_INFO:
       return Struct::VkCommandBufferInheritanceInfo;
    case VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO:
       return Struct::VkCommandBufferBeginInfo;
    case VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER:
       return Struct::VkBufferMemoryBarrier;
    case VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER:
       return Struct::VkImageMemoryBarrier;
    case VK_STRUCTURE_TYPE_MEMORY_BARRIER:
       return Struct::VkMemoryBarrier;
    case VK_STRUCTURE_TYPE_EVENT_CREATE_INFO:
       return Struct::VkEventCreateInfo;
    case VK_STRUCTURE_TYPE_BUFFER_VIEW_CREATE_INFO:
       return Struct::VkBufferViewCreateInfo;
    case VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO:
       return Struct::VkShaderModuleCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_CACHE_CREATE_INFO:
       return Struct::VkPipelineCacheCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO:
       return Struct::VkPipelineShaderStageCreateInfo;
    case VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_CREATE_INFO:
       return Struct::VkComputePipelineCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO:
       return Struct::VkPipelineLayoutCreateInfo;
    case VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO:
       return Struct::VkSamplerCreateInfo;
    case VK_STRUCTURE_TYPE_COPY_DESCRIPTOR_SET:
       return Struct::VkCopyDescriptorSet;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO:
       return Struct::VkDescriptorPoolCreateInfo;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO:
       return Struct::VkDescriptorSetAllocateInfo;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO:
       return Struct::VkDescriptorSetLayoutCreateInfo;
    case VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET:
       return Struct::VkWriteDescriptorSet;
    case VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO:
       return Struct::VkPipelineColorBlendStateCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO:
       return Struct::VkPipelineDepthStencilStateCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO:
       return Struct::VkPipelineDynamicStateCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO:
       return Struct::VkPipelineInputAssemblyStateCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO:
       return Struct::VkPipelineMultisampleStateCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO:
       return Struct::VkPipelineRasterizationStateCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_TESSELLATION_STATE_CREATE_INFO:
       return Struct::VkPipelineTessellationStateCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO:
       return Struct::VkPipelineVertexInputStateCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO:
       return Struct::VkPipelineViewportStateCreateInfo;
    case VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO:
       return Struct::VkGraphicsPipelineCreateInfo;
    case VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO:
       return Struct::VkFramebufferCreateInfo;
    case VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO:
       return Struct::VkRenderPassCreateInfo;
    case VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO:
       return Struct::VkRenderPassBeginInfo;
    case VK_STRUCTURE_TYPE_BIND_BUFFER_MEMORY_INFO:
       return Struct::VkBindBufferMemoryInfo;
    case VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_INFO:
       return Struct::VkBindImageMemoryInfo;
    case VK_STRUCTURE_TYPE_MEMORY_DEDICATED_REQUIREMENTS:
       return Struct::VkMemoryDedicatedRequirements;
    case VK_STRUCTURE_TYPE_MEMORY_DEDICATED_ALLOCATE_INFO:
       return Struct::VkMemoryDedicatedAllocateInfo;
    case VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_FLAGS_INFO:
       return Struct::VkMemoryAllocateFlagsInfo;
    case VK_STRUCTURE_TYPE_DEVICE_GROUP_COMMAND_BUFFER_BEGIN_INFO:
       return Struct::VkDeviceGroupCommandBufferBeginInfo;
    case VK_STRUCTURE_TYPE_DEVICE_GROUP_SUBMIT_INFO:
       return Struct::VkDeviceGroupSubmitInfo;
    case VK_STRUCTURE_TYPE_DEVICE_GROUP_BIND_SPARSE_INFO:
       return Struct::VkDeviceGroupBindSparseInfo;
    case VK_STRUCTURE_TYPE_BIND_BUFFER_MEMORY_DEVICE_GROUP_INFO:
       return Struct::VkBindBufferMemoryDeviceGroupInfo;
    case VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_DEVICE_GROUP_INFO:
       return Struct::VkBindImageMemoryDeviceGroupInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GROUP_PROPERTIES:
       return Struct::VkPhysicalDeviceGroupProperties;
    case VK_STRUCTURE_TYPE_DEVICE_GROUP_DEVICE_CREATE_INFO:
       return Struct::VkDeviceGroupDeviceCreateInfo;
    case VK_STRUCTURE_TYPE_BUFFER_MEMORY_REQUIREMENTS_INFO_2:
       return Struct::VkBufferMemoryRequirementsInfo2;
    case VK_STRUCTURE_TYPE_IMAGE_MEMORY_REQUIREMENTS_INFO_2:
       return Struct::VkImageMemoryRequirementsInfo2;
    case VK_STRUCTURE_TYPE_IMAGE_SPARSE_MEMORY_REQUIREMENTS_INFO_2:
       return Struct::VkImageSparseMemoryRequirementsInfo2;
    case VK_STRUCTURE_TYPE_MEMORY_REQUIREMENTS_2:
       return Struct::VkMemoryRequirements2;
    case VK_STRUCTURE_TYPE_SPARSE_IMAGE_MEMORY_REQUIREMENTS_2:
       return Struct::VkSparseImageMemoryRequirements2;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2:
       return Struct::VkPhysicalDeviceFeatures2;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROPERTIES_2:
       return Struct::VkPhysicalDeviceProperties2;
    case VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_2:
       return Struct::VkFormatProperties2;
    case VK_STRUCTURE_TYPE_IMAGE_FORMAT_PROPERTIES_2:
       return Struct::VkImageFormatProperties2;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_FORMAT_INFO_2:
       return Struct::VkPhysicalDeviceImageFormatInfo2;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_PROPERTIES_2:
       return Struct::VkQueueFamilyProperties2;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_PROPERTIES_2:
       return Struct::VkPhysicalDeviceMemoryProperties2;
    case VK_STRUCTURE_TYPE_SPARSE_IMAGE_FORMAT_PROPERTIES_2:
       return Struct::VkSparseImageFormatProperties2;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SPARSE_IMAGE_FORMAT_INFO_2:
       return Struct::VkPhysicalDeviceSparseImageFormatInfo2;
    case VK_STRUCTURE_TYPE_IMAGE_VIEW_USAGE_CREATE_INFO:
       return Struct::VkImageViewUsageCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROTECTED_MEMORY_FEATURES:
       return Struct::VkPhysicalDeviceProtectedMemoryFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROTECTED_MEMORY_PROPERTIES:
       return Struct::VkPhysicalDeviceProtectedMemoryProperties;
    case VK_STRUCTURE_TYPE_DEVICE_QUEUE_INFO_2:
       return Struct::VkDeviceQueueInfo2;
    case VK_STRUCTURE_TYPE_PROTECTED_SUBMIT_INFO:
       return Struct::VkProtectedSubmitInfo;
    case VK_STRUCTURE_TYPE_BIND_IMAGE_PLANE_MEMORY_INFO:
       return Struct::VkBindImagePlaneMemoryInfo;
    case VK_STRUCTURE_TYPE_IMAGE_PLANE_MEMORY_REQUIREMENTS_INFO:
       return Struct::VkImagePlaneMemoryRequirementsInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_IMAGE_FORMAT_INFO:
       return Struct::VkPhysicalDeviceExternalImageFormatInfo;
    case VK_STRUCTURE_TYPE_EXTERNAL_IMAGE_FORMAT_PROPERTIES:
       return Struct::VkExternalImageFormatProperties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_BUFFER_INFO:
       return Struct::VkPhysicalDeviceExternalBufferInfo;
    case VK_STRUCTURE_TYPE_EXTERNAL_BUFFER_PROPERTIES:
       return Struct::VkExternalBufferProperties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ID_PROPERTIES:
       return Struct::VkPhysicalDeviceIDProperties;
    case VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_IMAGE_CREATE_INFO:
       return Struct::VkExternalMemoryImageCreateInfo;
    case VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_BUFFER_CREATE_INFO:
       return Struct::VkExternalMemoryBufferCreateInfo;
    case VK_STRUCTURE_TYPE_EXPORT_MEMORY_ALLOCATE_INFO:
       return Struct::VkExportMemoryAllocateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FENCE_INFO:
       return Struct::VkPhysicalDeviceExternalFenceInfo;
    case VK_STRUCTURE_TYPE_EXTERNAL_FENCE_PROPERTIES:
       return Struct::VkExternalFenceProperties;
    case VK_STRUCTURE_TYPE_EXPORT_FENCE_CREATE_INFO:
       return Struct::VkExportFenceCreateInfo;
    case VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_CREATE_INFO:
       return Struct::VkExportSemaphoreCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_SEMAPHORE_INFO:
       return Struct::VkPhysicalDeviceExternalSemaphoreInfo;
    case VK_STRUCTURE_TYPE_EXTERNAL_SEMAPHORE_PROPERTIES:
       return Struct::VkExternalSemaphoreProperties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_PROPERTIES:
       return Struct::VkPhysicalDeviceSubgroupProperties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_16BIT_STORAGE_FEATURES:
       return Struct::VkPhysicalDevice16BitStorageFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VARIABLE_POINTERS_FEATURES:
       return Struct::VkPhysicalDeviceVariablePointersFeatures;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_UPDATE_TEMPLATE_CREATE_INFO:
       return Struct::VkDescriptorUpdateTemplateCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_3_PROPERTIES:
       return Struct::VkPhysicalDeviceMaintenance3Properties;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_SUPPORT:
       return Struct::VkDescriptorSetLayoutSupport;
    case VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_CREATE_INFO:
       return Struct::VkSamplerYcbcrConversionCreateInfo;
    case VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_INFO:
       return Struct::VkSamplerYcbcrConversionInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLER_YCBCR_CONVERSION_FEATURES:
       return Struct::VkPhysicalDeviceSamplerYcbcrConversionFeatures;
    case VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_IMAGE_FORMAT_PROPERTIES:
       return Struct::VkSamplerYcbcrConversionImageFormatProperties;
    case VK_STRUCTURE_TYPE_DEVICE_GROUP_RENDER_PASS_BEGIN_INFO:
       return Struct::VkDeviceGroupRenderPassBeginInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_POINT_CLIPPING_PROPERTIES:
       return Struct::VkPhysicalDevicePointClippingProperties;
    case VK_STRUCTURE_TYPE_RENDER_PASS_INPUT_ATTACHMENT_ASPECT_CREATE_INFO:
       return Struct::VkRenderPassInputAttachmentAspectCreateInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_TESSELLATION_DOMAIN_ORIGIN_STATE_CREATE_INFO:
       return Struct::VkPipelineTessellationDomainOriginStateCreateInfo;
    case VK_STRUCTURE_TYPE_RENDER_PASS_MULTIVIEW_CREATE_INFO:
       return Struct::VkRenderPassMultiviewCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_FEATURES:
       return Struct::VkPhysicalDeviceMultiviewFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PROPERTIES:
       return Struct::VkPhysicalDeviceMultiviewProperties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_DRAW_PARAMETERS_FEATURES:
       return Struct::VkPhysicalDeviceShaderDrawParametersFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DRIVER_PROPERTIES:
       return Struct::VkPhysicalDeviceDriverProperties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_FEATURES:
       return Struct::VkPhysicalDeviceVulkan11Features;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_PROPERTIES:
       return Struct::VkPhysicalDeviceVulkan11Properties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES:
       return Struct::VkPhysicalDeviceVulkan12Features;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_PROPERTIES:
       return Struct::VkPhysicalDeviceVulkan12Properties;
    case VK_STRUCTURE_TYPE_IMAGE_FORMAT_LIST_CREATE_INFO:
       return Struct::VkImageFormatListCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_MEMORY_MODEL_FEATURES:
       return Struct::VkPhysicalDeviceVulkanMemoryModelFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES:
       return Struct::VkPhysicalDeviceHostQueryResetFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_FEATURES:
       return Struct::VkPhysicalDeviceTimelineSemaphoreFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_PROPERTIES:
       return Struct::VkPhysicalDeviceTimelineSemaphoreProperties;
    case VK_STRUCTURE_TYPE_SEMAPHORE_TYPE_CREATE_INFO:
       return Struct::VkSemaphoreTypeCreateInfo;
    case VK_STRUCTURE_TYPE_TIMELINE_SEMAPHORE_SUBMIT_INFO:
       return Struct::VkTimelineSemaphoreSubmitInfo;
    case VK_STRUCTURE_TYPE_SEMAPHORE_WAIT_INFO:
       return Struct::VkSemaphoreWaitInfo;
    case VK_STRUCTURE_TYPE_SEMAPHORE_SIGNAL_INFO:
       return Struct::VkSemaphoreSignalInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BUFFER_DEVICE_ADDRESS_FEATURES:
       return Struct::VkPhysicalDeviceBufferDeviceAddressFeatures;
    case VK_STRUCTURE_TYPE_BUFFER_DEVICE_ADDRESS_INFO:
       return Struct::VkBufferDeviceAddressInfo;
    case VK_STRUCTURE_TYPE_BUFFER_OPAQUE_CAPTURE_ADDRESS_CREATE_INFO:
       return Struct::VkBufferOpaqueCaptureAddressCreateInfo;
    case VK_STRUCTURE_TYPE_MEMORY_OPAQUE_CAPTURE_ADDRESS_ALLOCATE_INFO:
       return Struct::VkMemoryOpaqueCaptureAddressAllocateInfo;
    case VK_STRUCTURE_TYPE_DEVICE_MEMORY_OPAQUE_CAPTURE_ADDRESS_INFO:
       return Struct::VkDeviceMemoryOpaqueCaptureAddressInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_8BIT_STORAGE_FEATURES:
       return Struct::VkPhysicalDevice8BitStorageFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_INT64_FEATURES:
       return Struct::VkPhysicalDeviceShaderAtomicInt64Features;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_FLOAT16_INT8_FEATURES:
       return Struct::VkPhysicalDeviceShaderFloat16Int8Features;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FLOAT_CONTROLS_PROPERTIES:
       return Struct::VkPhysicalDeviceFloatControlsProperties;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_BINDING_FLAGS_CREATE_INFO:
       return Struct::VkDescriptorSetLayoutBindingFlagsCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_INDEXING_FEATURES:
       return Struct::VkPhysicalDeviceDescriptorIndexingFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_INDEXING_PROPERTIES:
       return Struct::VkPhysicalDeviceDescriptorIndexingProperties;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_VARIABLE_DESCRIPTOR_COUNT_ALLOCATE_INFO:
       return Struct::VkDescriptorSetVariableDescriptorCountAllocateInfo;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_VARIABLE_DESCRIPTOR_COUNT_LAYOUT_SUPPORT:
       return Struct::VkDescriptorSetVariableDescriptorCountLayoutSupport;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCALAR_BLOCK_LAYOUT_FEATURES:
       return Struct::VkPhysicalDeviceScalarBlockLayoutFeatures;
    case VK_STRUCTURE_TYPE_SAMPLER_REDUCTION_MODE_CREATE_INFO:
       return Struct::VkSamplerReductionModeCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLER_FILTER_MINMAX_PROPERTIES:
       return Struct::VkPhysicalDeviceSamplerFilterMinmaxProperties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_UNIFORM_BUFFER_STANDARD_LAYOUT_FEATURES:
       return Struct::VkPhysicalDeviceUniformBufferStandardLayoutFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_EXTENDED_TYPES_FEATURES:
       return Struct::VkPhysicalDeviceShaderSubgroupExtendedTypesFeatures;
    case VK_STRUCTURE_TYPE_ATTACHMENT_DESCRIPTION_2:
       return Struct::VkAttachmentDescription2;
    case VK_STRUCTURE_TYPE_ATTACHMENT_REFERENCE_2:
       return Struct::VkAttachmentReference2;
    case VK_STRUCTURE_TYPE_SUBPASS_DESCRIPTION_2:
       return Struct::VkSubpassDescription2;
    case VK_STRUCTURE_TYPE_SUBPASS_DEPENDENCY_2:
       return Struct::VkSubpassDependency2;
    case VK_STRUCTURE_TYPE_SUBPASS_BEGIN_INFO:
       return Struct::VkSubpassBeginInfo;
    case VK_STRUCTURE_TYPE_SUBPASS_END_INFO:
       return Struct::VkSubpassEndInfo;
    case VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO_2:
       return Struct::VkRenderPassCreateInfo2;
    case VK_STRUCTURE_TYPE_SUBPASS_DESCRIPTION_DEPTH_STENCIL_RESOLVE:
       return Struct::VkSubpassDescriptionDepthStencilResolve;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_STENCIL_RESOLVE_PROPERTIES:
       return Struct::VkPhysicalDeviceDepthStencilResolveProperties;
    case VK_STRUCTURE_TYPE_IMAGE_STENCIL_USAGE_CREATE_INFO:
       return Struct::VkImageStencilUsageCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGELESS_FRAMEBUFFER_FEATURES:
       return Struct::VkPhysicalDeviceImagelessFramebufferFeatures;
    case VK_STRUCTURE_TYPE_FRAMEBUFFER_ATTACHMENT_IMAGE_INFO:
       return Struct::VkFramebufferAttachmentImageInfo;
    case VK_STRUCTURE_TYPE_RENDER_PASS_ATTACHMENT_BEGIN_INFO:
       return Struct::VkRenderPassAttachmentBeginInfo;
    case VK_STRUCTURE_TYPE_FRAMEBUFFER_ATTACHMENTS_CREATE_INFO:
       return Struct::VkFramebufferAttachmentsCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SEPARATE_DEPTH_STENCIL_LAYOUTS_FEATURES:
       return Struct::VkPhysicalDeviceSeparateDepthStencilLayoutsFeatures;
    case VK_STRUCTURE_TYPE_ATTACHMENT_REFERENCE_STENCIL_LAYOUT:
       return Struct::VkAttachmentReferenceStencilLayout;
    case VK_STRUCTURE_TYPE_ATTACHMENT_DESCRIPTION_STENCIL_LAYOUT:
       return Struct::VkAttachmentDescriptionStencilLayout;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_FEATURES:
       return Struct::VkPhysicalDeviceVulkan13Features;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_PROPERTIES:
       return Struct::VkPhysicalDeviceVulkan13Properties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TOOL_PROPERTIES:
       return Struct::VkPhysicalDeviceToolProperties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIVATE_DATA_FEATURES:
       return Struct::VkPhysicalDevicePrivateDataFeatures;
    case VK_STRUCTURE_TYPE_DEVICE_PRIVATE_DATA_CREATE_INFO:
       return Struct::VkDevicePrivateDataCreateInfo;
    case VK_STRUCTURE_TYPE_PRIVATE_DATA_SLOT_CREATE_INFO:
       return Struct::VkPrivateDataSlotCreateInfo;
    case VK_STRUCTURE_TYPE_MEMORY_BARRIER_2:
       return Struct::VkMemoryBarrier2;
    case VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER_2:
       return Struct::VkBufferMemoryBarrier2;
    case VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER_2:
       return Struct::VkImageMemoryBarrier2;
    case VK_STRUCTURE_TYPE_DEPENDENCY_INFO:
       return Struct::VkDependencyInfo;
    case VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO:
       return Struct::VkSemaphoreSubmitInfo;
    case VK_STRUCTURE_TYPE_COMMAND_BUFFER_SUBMIT_INFO:
       return Struct::VkCommandBufferSubmitInfo;
    case VK_STRUCTURE_TYPE_SUBMIT_INFO_2:
       return Struct::VkSubmitInfo2;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SYNCHRONIZATION_2_FEATURES:
       return Struct::VkPhysicalDeviceSynchronization2Features;
    case VK_STRUCTURE_TYPE_BUFFER_COPY_2:
       return Struct::VkBufferCopy2;
    case VK_STRUCTURE_TYPE_COPY_BUFFER_INFO_2:
       return Struct::VkCopyBufferInfo2;
    case VK_STRUCTURE_TYPE_IMAGE_COPY_2:
       return Struct::VkImageCopy2;
    case VK_STRUCTURE_TYPE_COPY_IMAGE_INFO_2:
       return Struct::VkCopyImageInfo2;
    case VK_STRUCTURE_TYPE_BUFFER_IMAGE_COPY_2:
       return Struct::VkBufferImageCopy2;
    case VK_STRUCTURE_TYPE_COPY_BUFFER_TO_IMAGE_INFO_2:
       return Struct::VkCopyBufferToImageInfo2;
    case VK_STRUCTURE_TYPE_COPY_IMAGE_TO_BUFFER_INFO_2:
       return Struct::VkCopyImageToBufferInfo2;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXTURE_COMPRESSION_ASTC_HDR_FEATURES:
       return Struct::VkPhysicalDeviceTextureCompressionASTCHDRFeatures;
    case VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_3:
       return Struct::VkFormatProperties3;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_FEATURES:
       return Struct::VkPhysicalDeviceMaintenance4Features;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_PROPERTIES:
       return Struct::VkPhysicalDeviceMaintenance4Properties;
    case VK_STRUCTURE_TYPE_DEVICE_BUFFER_MEMORY_REQUIREMENTS:
       return Struct::VkDeviceBufferMemoryRequirements;
    case VK_STRUCTURE_TYPE_DEVICE_IMAGE_MEMORY_REQUIREMENTS:
       return Struct::VkDeviceImageMemoryRequirements;
    case VK_STRUCTURE_TYPE_PIPELINE_CREATION_FEEDBACK_CREATE_INFO:
       return Struct::VkPipelineCreationFeedbackCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TERMINATE_INVOCATION_FEATURES:
       return Struct::VkPhysicalDeviceShaderTerminateInvocationFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_DEMOTE_TO_HELPER_INVOCATION_FEATURES:
       return Struct::VkPhysicalDeviceShaderDemoteToHelperInvocationFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_CREATION_CACHE_CONTROL_FEATURES:
       return Struct::VkPhysicalDevicePipelineCreationCacheControlFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ZERO_INITIALIZE_WORKGROUP_MEMORY_FEATURES:
       return Struct::VkPhysicalDeviceZeroInitializeWorkgroupMemoryFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_ROBUSTNESS_FEATURES:
       return Struct::VkPhysicalDeviceImageRobustnessFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_SIZE_CONTROL_FEATURES:
       return Struct::VkPhysicalDeviceSubgroupSizeControlFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBGROUP_SIZE_CONTROL_PROPERTIES:
       return Struct::VkPhysicalDeviceSubgroupSizeControlProperties;
    case VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_REQUIRED_SUBGROUP_SIZE_CREATE_INFO:
       return Struct::VkPipelineShaderStageRequiredSubgroupSizeCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INLINE_UNIFORM_BLOCK_FEATURES:
       return Struct::VkPhysicalDeviceInlineUniformBlockFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INLINE_UNIFORM_BLOCK_PROPERTIES:
       return Struct::VkPhysicalDeviceInlineUniformBlockProperties;
    case VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_INLINE_UNIFORM_BLOCK:
       return Struct::VkWriteDescriptorSetInlineUniformBlock;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_INLINE_UNIFORM_BLOCK_CREATE_INFO:
       return Struct::VkDescriptorPoolInlineUniformBlockCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_DOT_PRODUCT_FEATURES:
       return Struct::VkPhysicalDeviceShaderIntegerDotProductFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_DOT_PRODUCT_PROPERTIES:
       return Struct::VkPhysicalDeviceShaderIntegerDotProductProperties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXEL_BUFFER_ALIGNMENT_PROPERTIES:
       return Struct::VkPhysicalDeviceTexelBufferAlignmentProperties;
    case VK_STRUCTURE_TYPE_IMAGE_BLIT_2:
       return Struct::VkImageBlit2;
    case VK_STRUCTURE_TYPE_BLIT_IMAGE_INFO_2:
       return Struct::VkBlitImageInfo2;
    case VK_STRUCTURE_TYPE_IMAGE_RESOLVE_2:
       return Struct::VkImageResolve2;
    case VK_STRUCTURE_TYPE_RESOLVE_IMAGE_INFO_2:
       return Struct::VkResolveImageInfo2;
    case VK_STRUCTURE_TYPE_RENDERING_ATTACHMENT_INFO:
       return Struct::VkRenderingAttachmentInfo;
    case VK_STRUCTURE_TYPE_RENDERING_INFO:
       return Struct::VkRenderingInfo;
    case VK_STRUCTURE_TYPE_PIPELINE_RENDERING_CREATE_INFO:
       return Struct::VkPipelineRenderingCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DYNAMIC_RENDERING_FEATURES:
       return Struct::VkPhysicalDeviceDynamicRenderingFeatures;
    case VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_RENDERING_INFO:
       return Struct::VkCommandBufferInheritanceRenderingInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_4_FEATURES:
       return Struct::VkPhysicalDeviceVulkan14Features;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_4_PROPERTIES:
       return Struct::VkPhysicalDeviceVulkan14Properties;
    case VK_STRUCTURE_TYPE_DEVICE_QUEUE_GLOBAL_PRIORITY_CREATE_INFO:
       return Struct::VkDeviceQueueGlobalPriorityCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GLOBAL_PRIORITY_QUERY_FEATURES:
       return Struct::VkPhysicalDeviceGlobalPriorityQueryFeatures;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_GLOBAL_PRIORITY_PROPERTIES:
       return Struct::VkQueueFamilyGlobalPriorityProperties;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INDEX_TYPE_UINT8_FEATURES:
       return Struct::VkPhysicalDeviceIndexTypeUint8Features;
    case VK_STRUCTURE_TYPE_MEMORY_MAP_INFO:
       return Struct::VkMemoryMapInfo;
    case VK_STRUCTURE_TYPE_MEMORY_UNMAP_INFO:
       return Struct::VkMemoryUnmapInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_5_FEATURES:
       return Struct::VkPhysicalDeviceMaintenance5Features;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_5_PROPERTIES:
       return Struct::VkPhysicalDeviceMaintenance5Properties;
    case VK_STRUCTURE_TYPE_SUBRESOURCE_LAYOUT_2:
       return Struct::VkSubresourceLayout2;
    case VK_STRUCTURE_TYPE_IMAGE_SUBRESOURCE_2:
       return Struct::VkImageSubresource2;
    case VK_STRUCTURE_TYPE_DEVICE_IMAGE_SUBRESOURCE_INFO:
       return Struct::VkDeviceImageSubresourceInfo;
    case VK_STRUCTURE_TYPE_BUFFER_USAGE_FLAGS_2_CREATE_INFO:
       return Struct::VkBufferUsageFlags2CreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_6_FEATURES:
       return Struct::VkPhysicalDeviceMaintenance6Features;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_6_PROPERTIES:
       return Struct::VkPhysicalDeviceMaintenance6Properties;
    case VK_STRUCTURE_TYPE_BIND_MEMORY_STATUS:
       return Struct::VkBindMemoryStatus;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_IMAGE_COPY_FEATURES:
       return Struct::VkPhysicalDeviceHostImageCopyFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_IMAGE_COPY_PROPERTIES:
       return Struct::VkPhysicalDeviceHostImageCopyProperties;
    case VK_STRUCTURE_TYPE_MEMORY_TO_IMAGE_COPY:
       return Struct::VkMemoryToImageCopy;
    case VK_STRUCTURE_TYPE_IMAGE_TO_MEMORY_COPY:
       return Struct::VkImageToMemoryCopy;
    case VK_STRUCTURE_TYPE_COPY_MEMORY_TO_IMAGE_INFO:
       return Struct::VkCopyMemoryToImageInfo;
    case VK_STRUCTURE_TYPE_COPY_IMAGE_TO_MEMORY_INFO:
       return Struct::VkCopyImageToMemoryInfo;
    case VK_STRUCTURE_TYPE_COPY_IMAGE_TO_IMAGE_INFO:
       return Struct::VkCopyImageToImageInfo;
    case VK_STRUCTURE_TYPE_HOST_IMAGE_LAYOUT_TRANSITION_INFO:
       return Struct::VkHostImageLayoutTransitionInfo;
    case VK_STRUCTURE_TYPE_SUBRESOURCE_HOST_MEMCPY_SIZE:
       return Struct::VkSubresourceHostMemcpySize;
    case VK_STRUCTURE_TYPE_HOST_IMAGE_COPY_DEVICE_PERFORMANCE_QUERY:
       return Struct::VkHostImageCopyDevicePerformanceQuery;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_ROTATE_FEATURES:
       return Struct::VkPhysicalDeviceShaderSubgroupRotateFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_FLOAT_CONTROLS_2_FEATURES:
       return Struct::VkPhysicalDeviceShaderFloatControls2Features;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_EXPECT_ASSUME_FEATURES:
       return Struct::VkPhysicalDeviceShaderExpectAssumeFeatures;
    case VK_STRUCTURE_TYPE_PIPELINE_CREATE_FLAGS_2_CREATE_INFO:
       return Struct::VkPipelineCreateFlags2CreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PUSH_DESCRIPTOR_PROPERTIES:
       return Struct::VkPhysicalDevicePushDescriptorProperties;
    case VK_STRUCTURE_TYPE_BIND_DESCRIPTOR_SETS_INFO:
       return Struct::VkBindDescriptorSetsInfo;
    case VK_STRUCTURE_TYPE_PUSH_CONSTANTS_INFO:
       return Struct::VkPushConstantsInfo;
    case VK_STRUCTURE_TYPE_PUSH_DESCRIPTOR_SET_INFO:
       return Struct::VkPushDescriptorSetInfo;
    case VK_STRUCTURE_TYPE_PUSH_DESCRIPTOR_SET_WITH_TEMPLATE_INFO:
       return Struct::VkPushDescriptorSetWithTemplateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_PROTECTED_ACCESS_FEATURES:
       return Struct::VkPhysicalDevicePipelineProtectedAccessFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_ROBUSTNESS_FEATURES:
       return Struct::VkPhysicalDevicePipelineRobustnessFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_ROBUSTNESS_PROPERTIES:
       return Struct::VkPhysicalDevicePipelineRobustnessProperties;
    case VK_STRUCTURE_TYPE_PIPELINE_ROBUSTNESS_CREATE_INFO:
       return Struct::VkPipelineRobustnessCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINE_RASTERIZATION_FEATURES:
       return Struct::VkPhysicalDeviceLineRasterizationFeatures;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINE_RASTERIZATION_PROPERTIES:
       return Struct::VkPhysicalDeviceLineRasterizationProperties;
    case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_LINE_STATE_CREATE_INFO:
       return Struct::VkPipelineRasterizationLineStateCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_PROPERTIES:
       return Struct::VkPhysicalDeviceVertexAttributeDivisorProperties;
    case VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_DIVISOR_STATE_CREATE_INFO:
       return Struct::VkPipelineVertexInputDivisorStateCreateInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_FEATURES:
       return Struct::VkPhysicalDeviceVertexAttributeDivisorFeatures;
    case VK_STRUCTURE_TYPE_RENDERING_AREA_INFO:
       return Struct::VkRenderingAreaInfo;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DYNAMIC_RENDERING_LOCAL_READ_FEATURES:
       return Struct::VkPhysicalDeviceDynamicRenderingLocalReadFeatures;
    case VK_STRUCTURE_TYPE_RENDERING_ATTACHMENT_LOCATION_INFO:
       return Struct::VkRenderingAttachmentLocationInfo;
    case VK_STRUCTURE_TYPE_RENDERING_INPUT_ATTACHMENT_INDEX_INFO:
       return Struct::VkRenderingInputAttachmentIndexInfo;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR:
       return Struct::VkSwapchainCreateInfoKHR;
    case VK_STRUCTURE_TYPE_PRESENT_INFO_KHR:
       return Struct::VkPresentInfoKHR;
    case VK_STRUCTURE_TYPE_IMAGE_SWAPCHAIN_CREATE_INFO_KHR:
       return Struct::VkImageSwapchainCreateInfoKHR;
    case VK_STRUCTURE_TYPE_BIND_IMAGE_MEMORY_SWAPCHAIN_INFO_KHR:
       return Struct::VkBindImageMemorySwapchainInfoKHR;
    case VK_STRUCTURE_TYPE_ACQUIRE_NEXT_IMAGE_INFO_KHR:
       return Struct::VkAcquireNextImageInfoKHR;
    case VK_STRUCTURE_TYPE_DEVICE_GROUP_PRESENT_CAPABILITIES_KHR:
       return Struct::VkDeviceGroupPresentCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_DEVICE_GROUP_PRESENT_INFO_KHR:
       return Struct::VkDeviceGroupPresentInfoKHR;
    case VK_STRUCTURE_TYPE_DEVICE_GROUP_SWAPCHAIN_CREATE_INFO_KHR:
       return Struct::VkDeviceGroupSwapchainCreateInfoKHR;
    case VK_STRUCTURE_TYPE_DISPLAY_MODE_CREATE_INFO_KHR:
       return Struct::VkDisplayModeCreateInfoKHR;
    case VK_STRUCTURE_TYPE_DISPLAY_SURFACE_CREATE_INFO_KHR:
       return Struct::VkDisplaySurfaceCreateInfoKHR;
    case VK_STRUCTURE_TYPE_DISPLAY_PRESENT_INFO_KHR:
       return Struct::VkDisplayPresentInfoKHR;
    case VK_STRUCTURE_TYPE_XLIB_SURFACE_CREATE_INFO_KHR:
       return Struct::VkXlibSurfaceCreateInfoKHR;
    case VK_STRUCTURE_TYPE_XCB_SURFACE_CREATE_INFO_KHR:
       return Struct::VkXcbSurfaceCreateInfoKHR;
    case VK_STRUCTURE_TYPE_WAYLAND_SURFACE_CREATE_INFO_KHR:
       return Struct::VkWaylandSurfaceCreateInfoKHR;
    case VK_STRUCTURE_TYPE_ANDROID_SURFACE_CREATE_INFO_KHR:
       return Struct::VkAndroidSurfaceCreateInfoKHR;
    case VK_STRUCTURE_TYPE_WIN32_SURFACE_CREATE_INFO_KHR:
       return Struct::VkWin32SurfaceCreateInfoKHR;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_QUERY_RESULT_STATUS_PROPERTIES_KHR:
       return Struct::VkQueueFamilyQueryResultStatusPropertiesKHR;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_VIDEO_PROPERTIES_KHR:
       return Struct::VkQueueFamilyVideoPropertiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_PROFILE_INFO_KHR:
       return Struct::VkVideoProfileInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_PROFILE_LIST_INFO_KHR:
       return Struct::VkVideoProfileListInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_CAPABILITIES_KHR:
       return Struct::VkVideoCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_FORMAT_INFO_KHR:
       return Struct::VkPhysicalDeviceVideoFormatInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_FORMAT_PROPERTIES_KHR:
       return Struct::VkVideoFormatPropertiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_PICTURE_RESOURCE_INFO_KHR:
       return Struct::VkVideoPictureResourceInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_REFERENCE_SLOT_INFO_KHR:
       return Struct::VkVideoReferenceSlotInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_SESSION_MEMORY_REQUIREMENTS_KHR:
       return Struct::VkVideoSessionMemoryRequirementsKHR;
    case VK_STRUCTURE_TYPE_BIND_VIDEO_SESSION_MEMORY_INFO_KHR:
       return Struct::VkBindVideoSessionMemoryInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_SESSION_CREATE_INFO_KHR:
       return Struct::VkVideoSessionCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_SESSION_PARAMETERS_CREATE_INFO_KHR:
       return Struct::VkVideoSessionParametersCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_SESSION_PARAMETERS_UPDATE_INFO_KHR:
       return Struct::VkVideoSessionParametersUpdateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_BEGIN_CODING_INFO_KHR:
       return Struct::VkVideoBeginCodingInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_END_CODING_INFO_KHR:
       return Struct::VkVideoEndCodingInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_CODING_CONTROL_INFO_KHR:
       return Struct::VkVideoCodingControlInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_CAPABILITIES_KHR:
       return Struct::VkVideoDecodeCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_USAGE_INFO_KHR:
       return Struct::VkVideoDecodeUsageInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_INFO_KHR:
       return Struct::VkVideoDecodeInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_CAPABILITIES_KHR:
       return Struct::VkVideoEncodeH264CapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_QUALITY_LEVEL_PROPERTIES_KHR:
       return Struct::VkVideoEncodeH264QualityLevelPropertiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_CREATE_INFO_KHR:
       return Struct::VkVideoEncodeH264SessionCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_ADD_INFO_KHR:
       return Struct::VkVideoEncodeH264SessionParametersAddInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_CREATE_INFO_KHR:
       return Struct::VkVideoEncodeH264SessionParametersCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_GET_INFO_KHR:
       return Struct::VkVideoEncodeH264SessionParametersGetInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_SESSION_PARAMETERS_FEEDBACK_INFO_KHR:
       return Struct::VkVideoEncodeH264SessionParametersFeedbackInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_NALU_SLICE_INFO_KHR:
       return Struct::VkVideoEncodeH264NaluSliceInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_PICTURE_INFO_KHR:
       return Struct::VkVideoEncodeH264PictureInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_DPB_SLOT_INFO_KHR:
       return Struct::VkVideoEncodeH264DpbSlotInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_PROFILE_INFO_KHR:
       return Struct::VkVideoEncodeH264ProfileInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_RATE_CONTROL_INFO_KHR:
       return Struct::VkVideoEncodeH264RateControlInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_RATE_CONTROL_LAYER_INFO_KHR:
       return Struct::VkVideoEncodeH264RateControlLayerInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_GOP_REMAINING_FRAME_INFO_KHR:
       return Struct::VkVideoEncodeH264GopRemainingFrameInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_CAPABILITIES_KHR:
       return Struct::VkVideoEncodeH265CapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_CREATE_INFO_KHR:
       return Struct::VkVideoEncodeH265SessionCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_QUALITY_LEVEL_PROPERTIES_KHR:
       return Struct::VkVideoEncodeH265QualityLevelPropertiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_ADD_INFO_KHR:
       return Struct::VkVideoEncodeH265SessionParametersAddInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_CREATE_INFO_KHR:
       return Struct::VkVideoEncodeH265SessionParametersCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_GET_INFO_KHR:
       return Struct::VkVideoEncodeH265SessionParametersGetInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_SESSION_PARAMETERS_FEEDBACK_INFO_KHR:
       return Struct::VkVideoEncodeH265SessionParametersFeedbackInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_NALU_SLICE_SEGMENT_INFO_KHR:
       return Struct::VkVideoEncodeH265NaluSliceSegmentInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_PICTURE_INFO_KHR:
       return Struct::VkVideoEncodeH265PictureInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_DPB_SLOT_INFO_KHR:
       return Struct::VkVideoEncodeH265DpbSlotInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_PROFILE_INFO_KHR:
       return Struct::VkVideoEncodeH265ProfileInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_RATE_CONTROL_INFO_KHR:
       return Struct::VkVideoEncodeH265RateControlInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_RATE_CONTROL_LAYER_INFO_KHR:
       return Struct::VkVideoEncodeH265RateControlLayerInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_GOP_REMAINING_FRAME_INFO_KHR:
       return Struct::VkVideoEncodeH265GopRemainingFrameInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_PROFILE_INFO_KHR:
       return Struct::VkVideoDecodeH264ProfileInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_CAPABILITIES_KHR:
       return Struct::VkVideoDecodeH264CapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_SESSION_PARAMETERS_ADD_INFO_KHR:
       return Struct::VkVideoDecodeH264SessionParametersAddInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_SESSION_PARAMETERS_CREATE_INFO_KHR:
       return Struct::VkVideoDecodeH264SessionParametersCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_PICTURE_INFO_KHR:
       return Struct::VkVideoDecodeH264PictureInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_DPB_SLOT_INFO_KHR:
       return Struct::VkVideoDecodeH264DpbSlotInfoKHR;
    case VK_STRUCTURE_TYPE_IMPORT_MEMORY_WIN32_HANDLE_INFO_KHR:
       return Struct::VkImportMemoryWin32HandleInfoKHR;
    case VK_STRUCTURE_TYPE_EXPORT_MEMORY_WIN32_HANDLE_INFO_KHR:
       return Struct::VkExportMemoryWin32HandleInfoKHR;
    case VK_STRUCTURE_TYPE_MEMORY_WIN32_HANDLE_PROPERTIES_KHR:
       return Struct::VkMemoryWin32HandlePropertiesKHR;
    case VK_STRUCTURE_TYPE_MEMORY_GET_WIN32_HANDLE_INFO_KHR:
       return Struct::VkMemoryGetWin32HandleInfoKHR;
    case VK_STRUCTURE_TYPE_IMPORT_MEMORY_FD_INFO_KHR:
       return Struct::VkImportMemoryFdInfoKHR;
    case VK_STRUCTURE_TYPE_MEMORY_FD_PROPERTIES_KHR:
       return Struct::VkMemoryFdPropertiesKHR;
    case VK_STRUCTURE_TYPE_MEMORY_GET_FD_INFO_KHR:
       return Struct::VkMemoryGetFdInfoKHR;
    case VK_STRUCTURE_TYPE_WIN32_KEYED_MUTEX_ACQUIRE_RELEASE_INFO_KHR:
       return Struct::VkWin32KeyedMutexAcquireReleaseInfoKHR;
    case VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_WIN32_HANDLE_INFO_KHR:
       return Struct::VkImportSemaphoreWin32HandleInfoKHR;
    case VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_WIN32_HANDLE_INFO_KHR:
       return Struct::VkExportSemaphoreWin32HandleInfoKHR;
    case VK_STRUCTURE_TYPE_D3D12_FENCE_SUBMIT_INFO_KHR:
       return Struct::VkD3D12FenceSubmitInfoKHR;
    case VK_STRUCTURE_TYPE_SEMAPHORE_GET_WIN32_HANDLE_INFO_KHR:
       return Struct::VkSemaphoreGetWin32HandleInfoKHR;
    case VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_FD_INFO_KHR:
       return Struct::VkImportSemaphoreFdInfoKHR;
    case VK_STRUCTURE_TYPE_SEMAPHORE_GET_FD_INFO_KHR:
       return Struct::VkSemaphoreGetFdInfoKHR;
    case VK_STRUCTURE_TYPE_PRESENT_REGIONS_KHR:
       return Struct::VkPresentRegionsKHR;
    case VK_STRUCTURE_TYPE_SHARED_PRESENT_SURFACE_CAPABILITIES_KHR:
       return Struct::VkSharedPresentSurfaceCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_IMPORT_FENCE_WIN32_HANDLE_INFO_KHR:
       return Struct::VkImportFenceWin32HandleInfoKHR;
    case VK_STRUCTURE_TYPE_EXPORT_FENCE_WIN32_HANDLE_INFO_KHR:
       return Struct::VkExportFenceWin32HandleInfoKHR;
    case VK_STRUCTURE_TYPE_FENCE_GET_WIN32_HANDLE_INFO_KHR:
       return Struct::VkFenceGetWin32HandleInfoKHR;
    case VK_STRUCTURE_TYPE_IMPORT_FENCE_FD_INFO_KHR:
       return Struct::VkImportFenceFdInfoKHR;
    case VK_STRUCTURE_TYPE_FENCE_GET_FD_INFO_KHR:
       return Struct::VkFenceGetFdInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_QUERY_FEATURES_KHR:
       return Struct::VkPhysicalDevicePerformanceQueryFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_QUERY_PROPERTIES_KHR:
       return Struct::VkPhysicalDevicePerformanceQueryPropertiesKHR;
    case VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_KHR:
       return Struct::VkPerformanceCounterKHR;
    case VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_DESCRIPTION_KHR:
       return Struct::VkPerformanceCounterDescriptionKHR;
    case VK_STRUCTURE_TYPE_QUERY_POOL_PERFORMANCE_CREATE_INFO_KHR:
       return Struct::VkQueryPoolPerformanceCreateInfoKHR;
    case VK_STRUCTURE_TYPE_ACQUIRE_PROFILING_LOCK_INFO_KHR:
       return Struct::VkAcquireProfilingLockInfoKHR;
    case VK_STRUCTURE_TYPE_PERFORMANCE_QUERY_SUBMIT_INFO_KHR:
       return Struct::VkPerformanceQuerySubmitInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SURFACE_INFO_2_KHR:
       return Struct::VkPhysicalDeviceSurfaceInfo2KHR;
    case VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_2_KHR:
       return Struct::VkSurfaceCapabilities2KHR;
    case VK_STRUCTURE_TYPE_SURFACE_FORMAT_2_KHR:
       return Struct::VkSurfaceFormat2KHR;
    case VK_STRUCTURE_TYPE_DISPLAY_PROPERTIES_2_KHR:
       return Struct::VkDisplayProperties2KHR;
    case VK_STRUCTURE_TYPE_DISPLAY_PLANE_PROPERTIES_2_KHR:
       return Struct::VkDisplayPlaneProperties2KHR;
    case VK_STRUCTURE_TYPE_DISPLAY_MODE_PROPERTIES_2_KHR:
       return Struct::VkDisplayModeProperties2KHR;
    case VK_STRUCTURE_TYPE_DISPLAY_PLANE_INFO_2_KHR:
       return Struct::VkDisplayPlaneInfo2KHR;
    case VK_STRUCTURE_TYPE_DISPLAY_PLANE_CAPABILITIES_2_KHR:
       return Struct::VkDisplayPlaneCapabilities2KHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_BFLOAT16_FEATURES_KHR:
       return Struct::VkPhysicalDeviceShaderBfloat16FeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PORTABILITY_SUBSET_FEATURES_KHR:
       return Struct::VkPhysicalDevicePortabilitySubsetFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PORTABILITY_SUBSET_PROPERTIES_KHR:
       return Struct::VkPhysicalDevicePortabilitySubsetPropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CLOCK_FEATURES_KHR:
       return Struct::VkPhysicalDeviceShaderClockFeaturesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_PROFILE_INFO_KHR:
       return Struct::VkVideoDecodeH265ProfileInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_CAPABILITIES_KHR:
       return Struct::VkVideoDecodeH265CapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_SESSION_PARAMETERS_ADD_INFO_KHR:
       return Struct::VkVideoDecodeH265SessionParametersAddInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_SESSION_PARAMETERS_CREATE_INFO_KHR:
       return Struct::VkVideoDecodeH265SessionParametersCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_PICTURE_INFO_KHR:
       return Struct::VkVideoDecodeH265PictureInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_DPB_SLOT_INFO_KHR:
       return Struct::VkVideoDecodeH265DpbSlotInfoKHR;
    case VK_STRUCTURE_TYPE_FRAGMENT_SHADING_RATE_ATTACHMENT_INFO_KHR:
       return Struct::VkFragmentShadingRateAttachmentInfoKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_FRAGMENT_SHADING_RATE_STATE_CREATE_INFO_KHR:
       return Struct::VkPipelineFragmentShadingRateStateCreateInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_FEATURES_KHR:
       return Struct::VkPhysicalDeviceFragmentShadingRateFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceFragmentShadingRatePropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_KHR:
       return Struct::VkPhysicalDeviceFragmentShadingRateKHR;
    case VK_STRUCTURE_TYPE_RENDERING_FRAGMENT_SHADING_RATE_ATTACHMENT_INFO_KHR:
       return Struct::VkRenderingFragmentShadingRateAttachmentInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CONSTANT_DATA_FEATURES_KHR:
       return Struct::VkPhysicalDeviceShaderConstantDataFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ABORT_FEATURES_KHR:
       return Struct::VkPhysicalDeviceShaderAbortFeaturesKHR;
    case VK_STRUCTURE_TYPE_DEVICE_FAULT_SHADER_ABORT_MESSAGE_INFO_KHR:
       return Struct::VkDeviceFaultShaderAbortMessageInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ABORT_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceShaderAbortPropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_QUAD_CONTROL_FEATURES_KHR:
       return Struct::VkPhysicalDeviceShaderQuadControlFeaturesKHR;
    case VK_STRUCTURE_TYPE_SURFACE_PROTECTED_CAPABILITIES_KHR:
       return Struct::VkSurfaceProtectedCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_WAIT_FEATURES_KHR:
       return Struct::VkPhysicalDevicePresentWaitFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_EXECUTABLE_PROPERTIES_FEATURES_KHR:
       return Struct::VkPhysicalDevicePipelineExecutablePropertiesFeaturesKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_INFO_KHR:
       return Struct::VkPipelineInfoKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_PROPERTIES_KHR:
       return Struct::VkPipelineExecutablePropertiesKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_INFO_KHR:
       return Struct::VkPipelineExecutableInfoKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_STATISTIC_KHR:
       return Struct::VkPipelineExecutableStatisticKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_EXECUTABLE_INTERNAL_REPRESENTATION_KHR:
       return Struct::VkPipelineExecutableInternalRepresentationKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_LIBRARY_CREATE_INFO_KHR:
       return Struct::VkPipelineLibraryCreateInfoKHR;
    case VK_STRUCTURE_TYPE_PRESENT_ID_KHR:
       return Struct::VkPresentIdKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_ID_FEATURES_KHR:
       return Struct::VkPhysicalDevicePresentIdFeaturesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_INFO_KHR:
       return Struct::VkVideoEncodeInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_CAPABILITIES_KHR:
       return Struct::VkVideoEncodeCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_QUERY_POOL_VIDEO_ENCODE_FEEDBACK_CREATE_INFO_KHR:
       return Struct::VkQueryPoolVideoEncodeFeedbackCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_USAGE_INFO_KHR:
       return Struct::VkVideoEncodeUsageInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_RATE_CONTROL_LAYER_INFO_KHR:
       return Struct::VkVideoEncodeRateControlLayerInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_RATE_CONTROL_INFO_KHR:
       return Struct::VkVideoEncodeRateControlInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_QUALITY_LEVEL_INFO_KHR:
       return Struct::VkPhysicalDeviceVideoEncodeQualityLevelInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUALITY_LEVEL_PROPERTIES_KHR:
       return Struct::VkVideoEncodeQualityLevelPropertiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUALITY_LEVEL_INFO_KHR:
       return Struct::VkVideoEncodeQualityLevelInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_PARAMETERS_GET_INFO_KHR:
       return Struct::VkVideoEncodeSessionParametersGetInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_PARAMETERS_FEEDBACK_INFO_KHR:
       return Struct::VkVideoEncodeSessionParametersFeedbackInfoKHR;
    case VK_STRUCTURE_TYPE_DEVICE_MEMORY_COPY_KHR:
       return Struct::VkDeviceMemoryCopyKHR;
    case VK_STRUCTURE_TYPE_COPY_DEVICE_MEMORY_INFO_KHR:
       return Struct::VkCopyDeviceMemoryInfoKHR;
    case VK_STRUCTURE_TYPE_DEVICE_MEMORY_IMAGE_COPY_KHR:
       return Struct::VkDeviceMemoryImageCopyKHR;
    case VK_STRUCTURE_TYPE_COPY_DEVICE_MEMORY_IMAGE_INFO_KHR:
       return Struct::VkCopyDeviceMemoryImageInfoKHR;
    case VK_STRUCTURE_TYPE_MEMORY_RANGE_BARRIER_KHR:
       return Struct::VkMemoryRangeBarrierKHR;
    case VK_STRUCTURE_TYPE_MEMORY_RANGE_BARRIERS_INFO_KHR:
       return Struct::VkMemoryRangeBarriersInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_ADDRESS_COMMANDS_FEATURES_KHR:
       return Struct::VkPhysicalDeviceDeviceAddressCommandsFeaturesKHR;
    case VK_STRUCTURE_TYPE_BIND_INDEX_BUFFER_3_INFO_KHR:
       return Struct::VkBindIndexBuffer3InfoKHR;
    case VK_STRUCTURE_TYPE_BIND_VERTEX_BUFFER_3_INFO_KHR:
       return Struct::VkBindVertexBuffer3InfoKHR;
    case VK_STRUCTURE_TYPE_DRAW_INDIRECT_2_INFO_KHR:
       return Struct::VkDrawIndirect2InfoKHR;
    case VK_STRUCTURE_TYPE_DRAW_INDIRECT_COUNT_2_INFO_KHR:
       return Struct::VkDrawIndirectCount2InfoKHR;
    case VK_STRUCTURE_TYPE_DISPATCH_INDIRECT_2_INFO_KHR:
       return Struct::VkDispatchIndirect2InfoKHR;
    case VK_STRUCTURE_TYPE_CONDITIONAL_RENDERING_BEGIN_INFO_2_EXT:
       return Struct::VkConditionalRenderingBeginInfo2EXT;
    case VK_STRUCTURE_TYPE_BIND_TRANSFORM_FEEDBACK_BUFFER_2_INFO_EXT:
       return Struct::VkBindTransformFeedbackBuffer2InfoEXT;
    case VK_STRUCTURE_TYPE_MEMORY_MARKER_INFO_AMD:
       return Struct::VkMemoryMarkerInfoAMD;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CREATE_INFO_2_KHR:
       return Struct::VkAccelerationStructureCreateInfo2KHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_BARYCENTRIC_FEATURES_KHR:
       return Struct::VkPhysicalDeviceFragmentShaderBarycentricFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_BARYCENTRIC_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceFragmentShaderBarycentricPropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_UNIFORM_CONTROL_FLOW_FEATURES_KHR:
       return Struct::VkPhysicalDeviceShaderSubgroupUniformControlFlowFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_WORKGROUP_MEMORY_EXPLICIT_LAYOUT_FEATURES_KHR:
       return Struct::VkPhysicalDeviceWorkgroupMemoryExplicitLayoutFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_MAINTENANCE_1_FEATURES_KHR:
       return Struct::VkPhysicalDeviceRayTracingMaintenance1FeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_UNTYPED_POINTERS_FEATURES_KHR:
       return Struct::VkPhysicalDeviceShaderUntypedPointersFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MAXIMAL_RECONVERGENCE_FEATURES_KHR:
       return Struct::VkPhysicalDeviceShaderMaximalReconvergenceFeaturesKHR;
    case VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_PRESENT_ID_2_KHR:
       return Struct::VkSurfaceCapabilitiesPresentId2KHR;
    case VK_STRUCTURE_TYPE_PRESENT_ID_2_KHR:
       return Struct::VkPresentId2KHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_ID_2_FEATURES_KHR:
       return Struct::VkPhysicalDevicePresentId2FeaturesKHR;
    case VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_PRESENT_WAIT_2_KHR:
       return Struct::VkSurfaceCapabilitiesPresentWait2KHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_WAIT_2_FEATURES_KHR:
       return Struct::VkPhysicalDevicePresentWait2FeaturesKHR;
    case VK_STRUCTURE_TYPE_PRESENT_WAIT_2_INFO_KHR:
       return Struct::VkPresentWait2InfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_POSITION_FETCH_FEATURES_KHR:
       return Struct::VkPhysicalDeviceRayTracingPositionFetchFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_BINARY_FEATURES_KHR:
       return Struct::VkPhysicalDevicePipelineBinaryFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_BINARY_PROPERTIES_KHR:
       return Struct::VkPhysicalDevicePipelineBinaryPropertiesKHR;
    case VK_STRUCTURE_TYPE_DEVICE_PIPELINE_BINARY_INTERNAL_CACHE_CONTROL_KHR:
       return Struct::VkDevicePipelineBinaryInternalCacheControlKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_BINARY_KEY_KHR:
       return Struct::VkPipelineBinaryKeyKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_CREATE_INFO_KHR:
       return Struct::VkPipelineCreateInfoKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_BINARY_CREATE_INFO_KHR:
       return Struct::VkPipelineBinaryCreateInfoKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_BINARY_INFO_KHR:
       return Struct::VkPipelineBinaryInfoKHR;
    case VK_STRUCTURE_TYPE_RELEASE_CAPTURED_PIPELINE_DATA_INFO_KHR:
       return Struct::VkReleaseCapturedPipelineDataInfoKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_BINARY_DATA_INFO_KHR:
       return Struct::VkPipelineBinaryDataInfoKHR;
    case VK_STRUCTURE_TYPE_PIPELINE_BINARY_HANDLES_INFO_KHR:
       return Struct::VkPipelineBinaryHandlesInfoKHR;
    case VK_STRUCTURE_TYPE_SURFACE_PRESENT_MODE_KHR:
       return Struct::VkSurfacePresentModeKHR;
    case VK_STRUCTURE_TYPE_SURFACE_PRESENT_SCALING_CAPABILITIES_KHR:
       return Struct::VkSurfacePresentScalingCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_SURFACE_PRESENT_MODE_COMPATIBILITY_KHR:
       return Struct::VkSurfacePresentModeCompatibilityKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SWAPCHAIN_MAINTENANCE_1_FEATURES_KHR:
       return Struct::VkPhysicalDeviceSwapchainMaintenance1FeaturesKHR;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_FENCE_INFO_KHR:
       return Struct::VkSwapchainPresentFenceInfoKHR;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_MODES_CREATE_INFO_KHR:
       return Struct::VkSwapchainPresentModesCreateInfoKHR;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_MODE_INFO_KHR:
       return Struct::VkSwapchainPresentModeInfoKHR;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_SCALING_CREATE_INFO_KHR:
       return Struct::VkSwapchainPresentScalingCreateInfoKHR;
    case VK_STRUCTURE_TYPE_RELEASE_SWAPCHAIN_IMAGES_INFO_KHR:
       return Struct::VkReleaseSwapchainImagesInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INTERNALLY_SYNCHRONIZED_QUEUES_FEATURES_KHR:
       return Struct::VkPhysicalDeviceInternallySynchronizedQueuesFeaturesKHR;
    case VK_STRUCTURE_TYPE_COOPERATIVE_MATRIX_PROPERTIES_KHR:
       return Struct::VkCooperativeMatrixPropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_FEATURES_KHR:
       return Struct::VkPhysicalDeviceCooperativeMatrixFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceCooperativeMatrixPropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COMPUTE_SHADER_DERIVATIVES_FEATURES_KHR:
       return Struct::VkPhysicalDeviceComputeShaderDerivativesFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COMPUTE_SHADER_DERIVATIVES_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceComputeShaderDerivativesPropertiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_PROFILE_INFO_KHR:
       return Struct::VkVideoDecodeAV1ProfileInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_CAPABILITIES_KHR:
       return Struct::VkVideoDecodeAV1CapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_SESSION_PARAMETERS_CREATE_INFO_KHR:
       return Struct::VkVideoDecodeAV1SessionParametersCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_PICTURE_INFO_KHR:
       return Struct::VkVideoDecodeAV1PictureInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_DPB_SLOT_INFO_KHR:
       return Struct::VkVideoDecodeAV1DpbSlotInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_AV1_FEATURES_KHR:
       return Struct::VkPhysicalDeviceVideoEncodeAV1FeaturesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_CAPABILITIES_KHR:
       return Struct::VkVideoEncodeAV1CapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_QUALITY_LEVEL_PROPERTIES_KHR:
       return Struct::VkVideoEncodeAV1QualityLevelPropertiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_SESSION_CREATE_INFO_KHR:
       return Struct::VkVideoEncodeAV1SessionCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_SESSION_PARAMETERS_CREATE_INFO_KHR:
       return Struct::VkVideoEncodeAV1SessionParametersCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_PICTURE_INFO_KHR:
       return Struct::VkVideoEncodeAV1PictureInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_DPB_SLOT_INFO_KHR:
       return Struct::VkVideoEncodeAV1DpbSlotInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_PROFILE_INFO_KHR:
       return Struct::VkVideoEncodeAV1ProfileInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_GOP_REMAINING_FRAME_INFO_KHR:
       return Struct::VkVideoEncodeAV1GopRemainingFrameInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_RATE_CONTROL_INFO_KHR:
       return Struct::VkVideoEncodeAV1RateControlInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_RATE_CONTROL_LAYER_INFO_KHR:
       return Struct::VkVideoEncodeAV1RateControlLayerInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_DECODE_VP9_FEATURES_KHR:
       return Struct::VkPhysicalDeviceVideoDecodeVP9FeaturesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_VP9_PROFILE_INFO_KHR:
       return Struct::VkVideoDecodeVP9ProfileInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_VP9_CAPABILITIES_KHR:
       return Struct::VkVideoDecodeVP9CapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_VP9_PICTURE_INFO_KHR:
       return Struct::VkVideoDecodeVP9PictureInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_MAINTENANCE_1_FEATURES_KHR:
       return Struct::VkPhysicalDeviceVideoMaintenance1FeaturesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_INLINE_QUERY_INFO_KHR:
       return Struct::VkVideoInlineQueryInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_UNIFIED_IMAGE_LAYOUTS_FEATURES_KHR:
       return Struct::VkPhysicalDeviceUnifiedImageLayoutsFeaturesKHR;
    case VK_STRUCTURE_TYPE_ATTACHMENT_FEEDBACK_LOOP_INFO_EXT:
       return Struct::VkAttachmentFeedbackLoopInfoEXT;
    case VK_STRUCTURE_TYPE_CALIBRATED_TIMESTAMP_INFO_KHR:
       return Struct::VkCalibratedTimestampInfoKHR;
    case VK_STRUCTURE_TYPE_SET_DESCRIPTOR_BUFFER_OFFSETS_INFO_EXT:
       return Struct::VkSetDescriptorBufferOffsetsInfoEXT;
    case VK_STRUCTURE_TYPE_BIND_DESCRIPTOR_BUFFER_EMBEDDED_SAMPLERS_INFO_EXT:
       return Struct::VkBindDescriptorBufferEmbeddedSamplersInfoEXT;
    case VK_STRUCTURE_TYPE_COPY_MEMORY_INDIRECT_INFO_KHR:
       return Struct::VkCopyMemoryIndirectInfoKHR;
    case VK_STRUCTURE_TYPE_COPY_MEMORY_TO_IMAGE_INDIRECT_INFO_KHR:
       return Struct::VkCopyMemoryToImageIndirectInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COPY_MEMORY_INDIRECT_FEATURES_KHR:
       return Struct::VkPhysicalDeviceCopyMemoryIndirectFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COPY_MEMORY_INDIRECT_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceCopyMemoryIndirectPropertiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_INTRA_REFRESH_CAPABILITIES_KHR:
       return Struct::VkVideoEncodeIntraRefreshCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_INTRA_REFRESH_CREATE_INFO_KHR:
       return Struct::VkVideoEncodeSessionIntraRefreshCreateInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_INTRA_REFRESH_INFO_KHR:
       return Struct::VkVideoEncodeIntraRefreshInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_REFERENCE_INTRA_REFRESH_INFO_KHR:
       return Struct::VkVideoReferenceIntraRefreshInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_INTRA_REFRESH_FEATURES_KHR:
       return Struct::VkPhysicalDeviceVideoEncodeIntraRefreshFeaturesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUANTIZATION_MAP_CAPABILITIES_KHR:
       return Struct::VkVideoEncodeQuantizationMapCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_FORMAT_QUANTIZATION_MAP_PROPERTIES_KHR:
       return Struct::VkVideoFormatQuantizationMapPropertiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUANTIZATION_MAP_INFO_KHR:
       return Struct::VkVideoEncodeQuantizationMapInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_QUANTIZATION_MAP_SESSION_PARAMETERS_CREATE_INFO_KHR:
       return Struct::VkVideoEncodeQuantizationMapSessionParametersCreateInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_QUANTIZATION_MAP_FEATURES_KHR:
       return Struct::VkPhysicalDeviceVideoEncodeQuantizationMapFeaturesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H264_QUANTIZATION_MAP_CAPABILITIES_KHR:
       return Struct::VkVideoEncodeH264QuantizationMapCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_H265_QUANTIZATION_MAP_CAPABILITIES_KHR:
       return Struct::VkVideoEncodeH265QuantizationMapCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_FORMAT_H265_QUANTIZATION_MAP_PROPERTIES_KHR:
       return Struct::VkVideoFormatH265QuantizationMapPropertiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_AV1_QUANTIZATION_MAP_CAPABILITIES_KHR:
       return Struct::VkVideoEncodeAV1QuantizationMapCapabilitiesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_FORMAT_AV1_QUANTIZATION_MAP_PROPERTIES_KHR:
       return Struct::VkVideoFormatAV1QuantizationMapPropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_RELAXED_EXTENDED_INSTRUCTION_FEATURES_KHR:
       return Struct::VkPhysicalDeviceShaderRelaxedExtendedInstructionFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_7_FEATURES_KHR:
       return Struct::VkPhysicalDeviceMaintenance7FeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_7_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceMaintenance7PropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_API_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceLayeredApiPropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_API_PROPERTIES_LIST_KHR:
       return Struct::VkPhysicalDeviceLayeredApiPropertiesListKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_API_VULKAN_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceLayeredApiVulkanPropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FAULT_FEATURES_KHR:
       return Struct::VkPhysicalDeviceFaultFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FAULT_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceFaultPropertiesKHR;
    case VK_STRUCTURE_TYPE_DEVICE_FAULT_INFO_KHR:
       return Struct::VkDeviceFaultInfoKHR;
    case VK_STRUCTURE_TYPE_DEVICE_FAULT_DEBUG_INFO_KHR:
       return Struct::VkDeviceFaultDebugInfoKHR;
    case VK_STRUCTURE_TYPE_MEMORY_BARRIER_ACCESS_FLAGS_3_KHR:
       return Struct::VkMemoryBarrierAccessFlags3KHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_8_FEATURES_KHR:
       return Struct::VkPhysicalDeviceMaintenance8FeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_FMA_FEATURES_KHR:
       return Struct::VkPhysicalDeviceShaderFmaFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_9_FEATURES_KHR:
       return Struct::VkPhysicalDeviceMaintenance9FeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_9_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceMaintenance9PropertiesKHR;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_OWNERSHIP_TRANSFER_PROPERTIES_KHR:
       return Struct::VkQueueFamilyOwnershipTransferPropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_MAINTENANCE_2_FEATURES_KHR:
       return Struct::VkPhysicalDeviceVideoMaintenance2FeaturesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H264_INLINE_SESSION_PARAMETERS_INFO_KHR:
       return Struct::VkVideoDecodeH264InlineSessionParametersInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_H265_INLINE_SESSION_PARAMETERS_INFO_KHR:
       return Struct::VkVideoDecodeH265InlineSessionParametersInfoKHR;
    case VK_STRUCTURE_TYPE_VIDEO_DECODE_AV1_INLINE_SESSION_PARAMETERS_INFO_KHR:
       return Struct::VkVideoDecodeAV1InlineSessionParametersInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_FEEDBACK_2_FEATURES_KHR:
       return Struct::VkPhysicalDeviceVideoEncodeFeedback2FeaturesKHR;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_FEEDBACK_2_CAPABILITIES_KHR:
       return Struct::VkVideoEncodeFeedback2CapabilitiesKHR;
    case VK_STRUCTURE_TYPE_QUERY_POOL_VIDEO_ENCODE_PER_PARTITION_FEEDBACK_CREATE_INFO_KHR:
       return Struct::VkQueryPoolVideoEncodePerPartitionFeedbackCreateInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLAMP_ZERO_ONE_FEATURES_KHR:
       return Struct::VkPhysicalDeviceDepthClampZeroOneFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ROBUSTNESS_2_FEATURES_KHR:
       return Struct::VkPhysicalDeviceRobustness2FeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ROBUSTNESS_2_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceRobustness2PropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_MODE_FIFO_LATEST_READY_FEATURES_KHR:
       return Struct::VkPhysicalDevicePresentModeFifoLatestReadyFeaturesKHR;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_MICROMAP_DATA_KHR:
       return Struct::VkAccelerationStructureGeometryMicromapDataKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_FEATURES_KHR:
       return Struct::VkPhysicalDeviceOpacityMicromapFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceOpacityMicromapPropertiesKHR;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_TRIANGLES_OPACITY_MICROMAP_KHR:
       return Struct::VkAccelerationStructureTrianglesOpacityMicromapKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_10_FEATURES_KHR:
       return Struct::VkPhysicalDeviceMaintenance10FeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_10_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceMaintenance10PropertiesKHR;
    case VK_STRUCTURE_TYPE_RENDERING_END_INFO_KHR:
       return Struct::VkRenderingEndInfoKHR;
    case VK_STRUCTURE_TYPE_RENDERING_ATTACHMENT_FLAGS_INFO_KHR:
       return Struct::VkRenderingAttachmentFlagsInfoKHR;
    case VK_STRUCTURE_TYPE_RESOLVE_IMAGE_MODE_INFO_KHR:
       return Struct::VkResolveImageModeInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_11_FEATURES_KHR:
       return Struct::VkPhysicalDeviceMaintenance11FeaturesKHR;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_OPTIMAL_IMAGE_TRANSFER_GRANULARITY_PROPERTIES_KHR:
       return Struct::VkQueueFamilyOptimalImageTransferGranularityPropertiesKHR;
    case VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_4_KHR:
       return Struct::VkFormatProperties4KHR;
    case VK_STRUCTURE_TYPE_IMAGE_USAGE_FLAGS_2_CREATE_INFO_KHR:
       return Struct::VkImageUsageFlags2CreateInfoKHR;
    case VK_STRUCTURE_TYPE_IMAGE_CREATE_FLAGS_2_CREATE_INFO_KHR:
       return Struct::VkImageCreateFlags2CreateInfoKHR;
    case VK_STRUCTURE_TYPE_IMAGE_VIEW_USAGE_2_CREATE_INFO_KHR:
       return Struct::VkImageViewUsage2CreateInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_FLAGS_FEATURES_KHR:
       return Struct::VkPhysicalDeviceExtendedFlagsFeaturesKHR;
    case VK_STRUCTURE_TYPE_IMAGE_STENCIL_USAGE_2_CREATE_INFO_KHR:
       return Struct::VkImageStencilUsage2CreateInfoKHR;
    case VK_STRUCTURE_TYPE_SHARED_PRESENT_SURFACE_CAPABILITIES_2_KHR:
       return Struct::VkSharedPresentSurfaceCapabilities2KHR;
    case VK_STRUCTURE_TYPE_DEBUG_REPORT_CALLBACK_CREATE_INFO_EXT:
       return Struct::VkDebugReportCallbackCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_RASTERIZATION_ORDER_AMD:
       return Struct::VkPipelineRasterizationStateRasterizationOrderAMD;
    case VK_STRUCTURE_TYPE_DEBUG_MARKER_OBJECT_NAME_INFO_EXT:
       return Struct::VkDebugMarkerObjectNameInfoEXT;
    case VK_STRUCTURE_TYPE_DEBUG_MARKER_OBJECT_TAG_INFO_EXT:
       return Struct::VkDebugMarkerObjectTagInfoEXT;
    case VK_STRUCTURE_TYPE_DEBUG_MARKER_MARKER_INFO_EXT:
       return Struct::VkDebugMarkerMarkerInfoEXT;
    case VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_IMAGE_CREATE_INFO_NV:
       return Struct::VkDedicatedAllocationImageCreateInfoNV;
    case VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_BUFFER_CREATE_INFO_NV:
       return Struct::VkDedicatedAllocationBufferCreateInfoNV;
    case VK_STRUCTURE_TYPE_DEDICATED_ALLOCATION_MEMORY_ALLOCATE_INFO_NV:
       return Struct::VkDedicatedAllocationMemoryAllocateInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TRANSFORM_FEEDBACK_FEATURES_EXT:
       return Struct::VkPhysicalDeviceTransformFeedbackFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TRANSFORM_FEEDBACK_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceTransformFeedbackPropertiesEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_STREAM_CREATE_INFO_EXT:
       return Struct::VkPipelineRasterizationStateStreamCreateInfoEXT;
    case VK_STRUCTURE_TYPE_CU_MODULE_CREATE_INFO_NVX:
       return Struct::VkCuModuleCreateInfoNVX;
    case VK_STRUCTURE_TYPE_CU_MODULE_TEXTURING_MODE_CREATE_INFO_NVX:
       return Struct::VkCuModuleTexturingModeCreateInfoNVX;
    case VK_STRUCTURE_TYPE_CU_FUNCTION_CREATE_INFO_NVX:
       return Struct::VkCuFunctionCreateInfoNVX;
    case VK_STRUCTURE_TYPE_CU_LAUNCH_INFO_NVX:
       return Struct::VkCuLaunchInfoNVX;
    case VK_STRUCTURE_TYPE_IMAGE_VIEW_HANDLE_INFO_NVX:
       return Struct::VkImageViewHandleInfoNVX;
    case VK_STRUCTURE_TYPE_IMAGE_VIEW_ADDRESS_PROPERTIES_NVX:
       return Struct::VkImageViewAddressPropertiesNVX;
    case VK_STRUCTURE_TYPE_TEXTURE_LOD_GATHER_FORMAT_PROPERTIES_AMD:
       return Struct::VkTextureLODGatherFormatPropertiesAMD;
    case VK_STRUCTURE_TYPE_STREAM_DESCRIPTOR_SURFACE_CREATE_INFO_GGP:
       return Struct::VkStreamDescriptorSurfaceCreateInfoGGP;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CORNER_SAMPLED_IMAGE_FEATURES_NV:
       return Struct::VkPhysicalDeviceCornerSampledImageFeaturesNV;
    case VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_IMAGE_CREATE_INFO_NV:
       return Struct::VkExternalMemoryImageCreateInfoNV;
    case VK_STRUCTURE_TYPE_EXPORT_MEMORY_ALLOCATE_INFO_NV:
       return Struct::VkExportMemoryAllocateInfoNV;
    case VK_STRUCTURE_TYPE_IMPORT_MEMORY_WIN32_HANDLE_INFO_NV:
       return Struct::VkImportMemoryWin32HandleInfoNV;
    case VK_STRUCTURE_TYPE_EXPORT_MEMORY_WIN32_HANDLE_INFO_NV:
       return Struct::VkExportMemoryWin32HandleInfoNV;
    case VK_STRUCTURE_TYPE_WIN32_KEYED_MUTEX_ACQUIRE_RELEASE_INFO_NV:
       return Struct::VkWin32KeyedMutexAcquireReleaseInfoNV;
    case VK_STRUCTURE_TYPE_VALIDATION_FLAGS_EXT:
       return Struct::VkValidationFlagsEXT;
    case VK_STRUCTURE_TYPE_VI_SURFACE_CREATE_INFO_NN:
       return Struct::VkViSurfaceCreateInfoNN;
    case VK_STRUCTURE_TYPE_IMAGE_VIEW_ASTC_DECODE_MODE_EXT:
       return Struct::VkImageViewASTCDecodeModeEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ASTC_DECODE_FEATURES_EXT:
       return Struct::VkPhysicalDeviceASTCDecodeFeaturesEXT;
    case VK_STRUCTURE_TYPE_CONDITIONAL_RENDERING_BEGIN_INFO_EXT:
       return Struct::VkConditionalRenderingBeginInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CONDITIONAL_RENDERING_FEATURES_EXT:
       return Struct::VkPhysicalDeviceConditionalRenderingFeaturesEXT;
    case VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_CONDITIONAL_RENDERING_INFO_EXT:
       return Struct::VkCommandBufferInheritanceConditionalRenderingInfoEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_W_SCALING_STATE_CREATE_INFO_NV:
       return Struct::VkPipelineViewportWScalingStateCreateInfoNV;
    case VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_2_EXT:
       return Struct::VkSurfaceCapabilities2EXT;
    case VK_STRUCTURE_TYPE_DISPLAY_POWER_INFO_EXT:
       return Struct::VkDisplayPowerInfoEXT;
    case VK_STRUCTURE_TYPE_DEVICE_EVENT_INFO_EXT:
       return Struct::VkDeviceEventInfoEXT;
    case VK_STRUCTURE_TYPE_DISPLAY_EVENT_INFO_EXT:
       return Struct::VkDisplayEventInfoEXT;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_COUNTER_CREATE_INFO_EXT:
       return Struct::VkSwapchainCounterCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PRESENT_TIMES_INFO_GOOGLE:
       return Struct::VkPresentTimesInfoGOOGLE;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_ATTRIBUTES_PROPERTIES_NVX:
       return Struct::VkPhysicalDeviceMultiviewPerViewAttributesPropertiesNVX;
    case VK_STRUCTURE_TYPE_MULTIVIEW_PER_VIEW_ATTRIBUTES_INFO_NVX:
       return Struct::VkMultiviewPerViewAttributesInfoNVX;
    case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_SWIZZLE_STATE_CREATE_INFO_NV:
       return Struct::VkPipelineViewportSwizzleStateCreateInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DISCARD_RECTANGLE_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceDiscardRectanglePropertiesEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_DISCARD_RECTANGLE_STATE_CREATE_INFO_EXT:
       return Struct::VkPipelineDiscardRectangleStateCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CONSERVATIVE_RASTERIZATION_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceConservativeRasterizationPropertiesEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_CONSERVATIVE_STATE_CREATE_INFO_EXT:
       return Struct::VkPipelineRasterizationConservativeStateCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLIP_ENABLE_FEATURES_EXT:
       return Struct::VkPhysicalDeviceDepthClipEnableFeaturesEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_DEPTH_CLIP_STATE_CREATE_INFO_EXT:
       return Struct::VkPipelineRasterizationDepthClipStateCreateInfoEXT;
    case VK_STRUCTURE_TYPE_HDR_METADATA_EXT:
       return Struct::VkHdrMetadataEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RELAXED_LINE_RASTERIZATION_FEATURES_IMG:
       return Struct::VkPhysicalDeviceRelaxedLineRasterizationFeaturesIMG;
    case VK_STRUCTURE_TYPE_IOS_SURFACE_CREATE_INFO_MVK:
       return Struct::VkIOSSurfaceCreateInfoMVK;
    case VK_STRUCTURE_TYPE_MACOS_SURFACE_CREATE_INFO_MVK:
       return Struct::VkMacOSSurfaceCreateInfoMVK;
    case VK_STRUCTURE_TYPE_DEBUG_UTILS_LABEL_EXT:
       return Struct::VkDebugUtilsLabelEXT;
    case VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_NAME_INFO_EXT:
       return Struct::VkDebugUtilsObjectNameInfoEXT;
    case VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CALLBACK_DATA_EXT:
       return Struct::VkDebugUtilsMessengerCallbackDataEXT;
    case VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT:
       return Struct::VkDebugUtilsMessengerCreateInfoEXT;
    case VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_TAG_INFO_EXT:
       return Struct::VkDebugUtilsObjectTagInfoEXT;
    case VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_USAGE_ANDROID:
       return Struct::VkAndroidHardwareBufferUsageANDROID;
    case VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_PROPERTIES_ANDROID:
       return Struct::VkAndroidHardwareBufferPropertiesANDROID;
    case VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_PROPERTIES_ANDROID:
       return Struct::VkAndroidHardwareBufferFormatPropertiesANDROID;
    case VK_STRUCTURE_TYPE_IMPORT_ANDROID_HARDWARE_BUFFER_INFO_ANDROID:
       return Struct::VkImportAndroidHardwareBufferInfoANDROID;
    case VK_STRUCTURE_TYPE_MEMORY_GET_ANDROID_HARDWARE_BUFFER_INFO_ANDROID:
       return Struct::VkMemoryGetAndroidHardwareBufferInfoANDROID;
    case VK_STRUCTURE_TYPE_EXTERNAL_FORMAT_ANDROID:
       return Struct::VkExternalFormatANDROID;
    case VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_PROPERTIES_2_ANDROID:
       return Struct::VkAndroidHardwareBufferFormatProperties2ANDROID;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GPA_FEATURES_AMD:
       return Struct::VkPhysicalDeviceGpaFeaturesAMD;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GPA_PROPERTIES_AMD:
       return Struct::VkPhysicalDeviceGpaPropertiesAMD;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GPA_PROPERTIES_2_AMD:
       return Struct::VkPhysicalDeviceGpaProperties2AMD;
    case VK_STRUCTURE_TYPE_GPA_SAMPLE_BEGIN_INFO_AMD:
       return Struct::VkGpaSampleBeginInfoAMD;
    case VK_STRUCTURE_TYPE_GPA_DEVICE_CLOCK_MODE_INFO_AMD:
       return Struct::VkGpaDeviceClockModeInfoAMD;
    case VK_STRUCTURE_TYPE_GPA_DEVICE_GET_CLOCK_INFO_AMD:
       return Struct::VkGpaDeviceGetClockInfoAMD;
    case VK_STRUCTURE_TYPE_GPA_SESSION_CREATE_INFO_AMD:
       return Struct::VkGpaSessionCreateInfoAMD;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ENQUEUE_FEATURES_AMDX:
       return Struct::VkPhysicalDeviceShaderEnqueueFeaturesAMDX;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ENQUEUE_PROPERTIES_AMDX:
       return Struct::VkPhysicalDeviceShaderEnqueuePropertiesAMDX;
    case VK_STRUCTURE_TYPE_EXECUTION_GRAPH_PIPELINE_SCRATCH_SIZE_AMDX:
       return Struct::VkExecutionGraphPipelineScratchSizeAMDX;
    case VK_STRUCTURE_TYPE_EXECUTION_GRAPH_PIPELINE_CREATE_INFO_AMDX:
       return Struct::VkExecutionGraphPipelineCreateInfoAMDX;
    case VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_NODE_CREATE_INFO_AMDX:
       return Struct::VkPipelineShaderStageNodeCreateInfoAMDX;
    case VK_STRUCTURE_TYPE_TEXEL_BUFFER_DESCRIPTOR_INFO_EXT:
       return Struct::VkTexelBufferDescriptorInfoEXT;
    case VK_STRUCTURE_TYPE_IMAGE_DESCRIPTOR_INFO_EXT:
       return Struct::VkImageDescriptorInfoEXT;
    case VK_STRUCTURE_TYPE_TENSOR_VIEW_CREATE_INFO_ARM:
       return Struct::VkTensorViewCreateInfoARM;
    case VK_STRUCTURE_TYPE_RESOURCE_DESCRIPTOR_INFO_EXT:
       return Struct::VkResourceDescriptorInfoEXT;
    case VK_STRUCTURE_TYPE_BIND_HEAP_INFO_EXT:
       return Struct::VkBindHeapInfoEXT;
    case VK_STRUCTURE_TYPE_PUSH_DATA_INFO_EXT:
       return Struct::VkPushDataInfoEXT;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_AND_BINDING_MAPPING_EXT:
       return Struct::VkDescriptorSetAndBindingMappingEXT;
    case VK_STRUCTURE_TYPE_SHADER_DESCRIPTOR_SET_AND_BINDING_MAPPING_INFO_EXT:
       return Struct::VkShaderDescriptorSetAndBindingMappingInfoEXT;
    case VK_STRUCTURE_TYPE_OPAQUE_CAPTURE_DATA_CREATE_INFO_EXT:
       return Struct::VkOpaqueCaptureDataCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_HEAP_FEATURES_EXT:
       return Struct::VkPhysicalDeviceDescriptorHeapFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_HEAP_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceDescriptorHeapPropertiesEXT;
    case VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_DESCRIPTOR_HEAP_INFO_EXT:
       return Struct::VkCommandBufferInheritanceDescriptorHeapInfoEXT;
    case VK_STRUCTURE_TYPE_SAMPLER_CUSTOM_BORDER_COLOR_INDEX_CREATE_INFO_EXT:
       return Struct::VkSamplerCustomBorderColorIndexCreateInfoEXT;
    case VK_STRUCTURE_TYPE_SAMPLER_CUSTOM_BORDER_COLOR_CREATE_INFO_EXT:
       return Struct::VkSamplerCustomBorderColorCreateInfoEXT;
    case VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_PUSH_DATA_TOKEN_NV:
       return Struct::VkIndirectCommandsLayoutPushDataTokenNV;
    case VK_STRUCTURE_TYPE_SUBSAMPLED_IMAGE_FORMAT_PROPERTIES_EXT:
       return Struct::VkSubsampledImageFormatPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_HEAP_TENSOR_PROPERTIES_ARM:
       return Struct::VkPhysicalDeviceDescriptorHeapTensorPropertiesARM;
    case VK_STRUCTURE_TYPE_ATTACHMENT_SAMPLE_COUNT_INFO_AMD:
       return Struct::VkAttachmentSampleCountInfoAMD;
    case VK_STRUCTURE_TYPE_SAMPLE_LOCATIONS_INFO_EXT:
       return Struct::VkSampleLocationsInfoEXT;
    case VK_STRUCTURE_TYPE_RENDER_PASS_SAMPLE_LOCATIONS_BEGIN_INFO_EXT:
       return Struct::VkRenderPassSampleLocationsBeginInfoEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_SAMPLE_LOCATIONS_STATE_CREATE_INFO_EXT:
       return Struct::VkPipelineSampleLocationsStateCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SAMPLE_LOCATIONS_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceSampleLocationsPropertiesEXT;
    case VK_STRUCTURE_TYPE_MULTISAMPLE_PROPERTIES_EXT:
       return Struct::VkMultisamplePropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BLEND_OPERATION_ADVANCED_FEATURES_EXT:
       return Struct::VkPhysicalDeviceBlendOperationAdvancedFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BLEND_OPERATION_ADVANCED_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceBlendOperationAdvancedPropertiesEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_ADVANCED_STATE_CREATE_INFO_EXT:
       return Struct::VkPipelineColorBlendAdvancedStateCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_TO_COLOR_STATE_CREATE_INFO_NV:
       return Struct::VkPipelineCoverageToColorStateCreateInfoNV;
    case VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_MODULATION_STATE_CREATE_INFO_NV:
       return Struct::VkPipelineCoverageModulationStateCreateInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SM_BUILTINS_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceShaderSMBuiltinsPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SM_BUILTINS_FEATURES_NV:
       return Struct::VkPhysicalDeviceShaderSMBuiltinsFeaturesNV;
    case VK_STRUCTURE_TYPE_DRM_FORMAT_MODIFIER_PROPERTIES_LIST_EXT:
       return Struct::VkDrmFormatModifierPropertiesListEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_DRM_FORMAT_MODIFIER_INFO_EXT:
       return Struct::VkPhysicalDeviceImageDrmFormatModifierInfoEXT;
    case VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_LIST_CREATE_INFO_EXT:
       return Struct::VkImageDrmFormatModifierListCreateInfoEXT;
    case VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_EXPLICIT_CREATE_INFO_EXT:
       return Struct::VkImageDrmFormatModifierExplicitCreateInfoEXT;
    case VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_PROPERTIES_EXT:
       return Struct::VkImageDrmFormatModifierPropertiesEXT;
    case VK_STRUCTURE_TYPE_DRM_FORMAT_MODIFIER_PROPERTIES_LIST_2_EXT:
       return Struct::VkDrmFormatModifierPropertiesList2EXT;
    case VK_STRUCTURE_TYPE_VALIDATION_CACHE_CREATE_INFO_EXT:
       return Struct::VkValidationCacheCreateInfoEXT;
    case VK_STRUCTURE_TYPE_SHADER_MODULE_VALIDATION_CACHE_CREATE_INFO_EXT:
       return Struct::VkShaderModuleValidationCacheCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_SHADING_RATE_IMAGE_STATE_CREATE_INFO_NV:
       return Struct::VkPipelineViewportShadingRateImageStateCreateInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADING_RATE_IMAGE_FEATURES_NV:
       return Struct::VkPhysicalDeviceShadingRateImageFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADING_RATE_IMAGE_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceShadingRateImagePropertiesNV;
    case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_COARSE_SAMPLE_ORDER_STATE_CREATE_INFO_NV:
       return Struct::VkPipelineViewportCoarseSampleOrderStateCreateInfoNV;
    case VK_STRUCTURE_TYPE_RAY_TRACING_SHADER_GROUP_CREATE_INFO_NV:
       return Struct::VkRayTracingShaderGroupCreateInfoNV;
    case VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_CREATE_INFO_NV:
       return Struct::VkRayTracingPipelineCreateInfoNV;
    case VK_STRUCTURE_TYPE_GEOMETRY_TRIANGLES_NV:
       return Struct::VkGeometryTrianglesNV;
    case VK_STRUCTURE_TYPE_GEOMETRY_AABB_NV:
       return Struct::VkGeometryAABBNV;
    case VK_STRUCTURE_TYPE_GEOMETRY_NV:
       return Struct::VkGeometryNV;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_INFO_NV:
       return Struct::VkAccelerationStructureInfoNV;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CREATE_INFO_NV:
       return Struct::VkAccelerationStructureCreateInfoNV;
    case VK_STRUCTURE_TYPE_BIND_ACCELERATION_STRUCTURE_MEMORY_INFO_NV:
       return Struct::VkBindAccelerationStructureMemoryInfoNV;
    case VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_ACCELERATION_STRUCTURE_NV:
       return Struct::VkWriteDescriptorSetAccelerationStructureNV;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_MEMORY_REQUIREMENTS_INFO_NV:
       return Struct::VkAccelerationStructureMemoryRequirementsInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceRayTracingPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_REPRESENTATIVE_FRAGMENT_TEST_FEATURES_NV:
       return Struct::VkPhysicalDeviceRepresentativeFragmentTestFeaturesNV;
    case VK_STRUCTURE_TYPE_PIPELINE_REPRESENTATIVE_FRAGMENT_TEST_STATE_CREATE_INFO_NV:
       return Struct::VkPipelineRepresentativeFragmentTestStateCreateInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_VIEW_IMAGE_FORMAT_INFO_EXT:
       return Struct::VkPhysicalDeviceImageViewImageFormatInfoEXT;
    case VK_STRUCTURE_TYPE_FILTER_CUBIC_IMAGE_VIEW_IMAGE_FORMAT_PROPERTIES_EXT:
       return Struct::VkFilterCubicImageViewImageFormatPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_CONVERSION_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceCooperativeMatrixConversionFeaturesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ELAPSED_TIMER_QUERY_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceElapsedTimerQueryFeaturesQCOM;
    case VK_STRUCTURE_TYPE_IMPORT_MEMORY_HOST_POINTER_INFO_EXT:
       return Struct::VkImportMemoryHostPointerInfoEXT;
    case VK_STRUCTURE_TYPE_MEMORY_HOST_POINTER_PROPERTIES_EXT:
       return Struct::VkMemoryHostPointerPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_HOST_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceExternalMemoryHostPropertiesEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_COMPILER_CONTROL_CREATE_INFO_AMD:
       return Struct::VkPipelineCompilerControlCreateInfoAMD;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_AMD:
       return Struct::VkPhysicalDeviceShaderCorePropertiesAMD;
    case VK_STRUCTURE_TYPE_DEVICE_MEMORY_OVERALLOCATION_CREATE_INFO_AMD:
       return Struct::VkDeviceMemoryOverallocationCreateInfoAMD;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_DIVISOR_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceVertexAttributeDivisorPropertiesEXT;
    case VK_STRUCTURE_TYPE_PRESENT_FRAME_TOKEN_GGP:
       return Struct::VkPresentFrameTokenGGP;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_FEATURES_NV:
       return Struct::VkPhysicalDeviceMeshShaderFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceMeshShaderPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_IMAGE_FOOTPRINT_FEATURES_NV:
       return Struct::VkPhysicalDeviceShaderImageFootprintFeaturesNV;
    case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_EXCLUSIVE_SCISSOR_STATE_CREATE_INFO_NV:
       return Struct::VkPipelineViewportExclusiveScissorStateCreateInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXCLUSIVE_SCISSOR_FEATURES_NV:
       return Struct::VkPhysicalDeviceExclusiveScissorFeaturesNV;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_CHECKPOINT_PROPERTIES_NV:
       return Struct::VkQueueFamilyCheckpointPropertiesNV;
    case VK_STRUCTURE_TYPE_CHECKPOINT_DATA_NV:
       return Struct::VkCheckpointDataNV;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_CHECKPOINT_PROPERTIES_2_NV:
       return Struct::VkQueueFamilyCheckpointProperties2NV;
    case VK_STRUCTURE_TYPE_CHECKPOINT_DATA_2_NV:
       return Struct::VkCheckpointData2NV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_TIMING_FEATURES_EXT:
       return Struct::VkPhysicalDevicePresentTimingFeaturesEXT;
    case VK_STRUCTURE_TYPE_PRESENT_TIMING_SURFACE_CAPABILITIES_EXT:
       return Struct::VkPresentTimingSurfaceCapabilitiesEXT;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_CALIBRATED_TIMESTAMP_INFO_EXT:
       return Struct::VkSwapchainCalibratedTimestampInfoEXT;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_TIMING_PROPERTIES_EXT:
       return Struct::VkSwapchainTimingPropertiesEXT;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_TIME_DOMAIN_PROPERTIES_EXT:
       return Struct::VkSwapchainTimeDomainPropertiesEXT;
    case VK_STRUCTURE_TYPE_PAST_PRESENTATION_TIMING_INFO_EXT:
       return Struct::VkPastPresentationTimingInfoEXT;
    case VK_STRUCTURE_TYPE_PAST_PRESENTATION_TIMING_EXT:
       return Struct::VkPastPresentationTimingEXT;
    case VK_STRUCTURE_TYPE_PAST_PRESENTATION_TIMING_PROPERTIES_EXT:
       return Struct::VkPastPresentationTimingPropertiesEXT;
    case VK_STRUCTURE_TYPE_PRESENT_TIMING_INFO_EXT:
       return Struct::VkPresentTimingInfoEXT;
    case VK_STRUCTURE_TYPE_PRESENT_TIMINGS_INFO_EXT:
       return Struct::VkPresentTimingsInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INTEGER_FUNCTIONS_2_FEATURES_INTEL:
       return Struct::VkPhysicalDeviceShaderIntegerFunctions2FeaturesINTEL;
    case VK_STRUCTURE_TYPE_INITIALIZE_PERFORMANCE_API_INFO_INTEL:
       return Struct::VkInitializePerformanceApiInfoINTEL;
    case VK_STRUCTURE_TYPE_QUERY_POOL_PERFORMANCE_QUERY_CREATE_INFO_INTEL:
       return Struct::VkQueryPoolPerformanceQueryCreateInfoINTEL;
    case VK_STRUCTURE_TYPE_PERFORMANCE_MARKER_INFO_INTEL:
       return Struct::VkPerformanceMarkerInfoINTEL;
    case VK_STRUCTURE_TYPE_PERFORMANCE_STREAM_MARKER_INFO_INTEL:
       return Struct::VkPerformanceStreamMarkerInfoINTEL;
    case VK_STRUCTURE_TYPE_PERFORMANCE_OVERRIDE_INFO_INTEL:
       return Struct::VkPerformanceOverrideInfoINTEL;
    case VK_STRUCTURE_TYPE_PERFORMANCE_CONFIGURATION_ACQUIRE_INFO_INTEL:
       return Struct::VkPerformanceConfigurationAcquireInfoINTEL;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PCI_BUS_INFO_PROPERTIES_EXT:
       return Struct::VkPhysicalDevicePCIBusInfoPropertiesEXT;
    case VK_STRUCTURE_TYPE_DISPLAY_NATIVE_HDR_SURFACE_CAPABILITIES_AMD:
       return Struct::VkDisplayNativeHdrSurfaceCapabilitiesAMD;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_DISPLAY_NATIVE_HDR_CREATE_INFO_AMD:
       return Struct::VkSwapchainDisplayNativeHdrCreateInfoAMD;
    case VK_STRUCTURE_TYPE_IMAGEPIPE_SURFACE_CREATE_INFO_FUCHSIA:
       return Struct::VkImagePipeSurfaceCreateInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_METAL_SURFACE_CREATE_INFO_EXT:
       return Struct::VkMetalSurfaceCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_FEATURES_EXT:
       return Struct::VkPhysicalDeviceFragmentDensityMapFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceFragmentDensityMapPropertiesEXT;
    case VK_STRUCTURE_TYPE_RENDER_PASS_FRAGMENT_DENSITY_MAP_CREATE_INFO_EXT:
       return Struct::VkRenderPassFragmentDensityMapCreateInfoEXT;
    case VK_STRUCTURE_TYPE_RENDERING_FRAGMENT_DENSITY_MAP_ATTACHMENT_INFO_EXT:
       return Struct::VkRenderingFragmentDensityMapAttachmentInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_2_AMD:
       return Struct::VkPhysicalDeviceShaderCoreProperties2AMD;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COHERENT_MEMORY_FEATURES_AMD:
       return Struct::VkPhysicalDeviceCoherentMemoryFeaturesAMD;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_IMAGE_ATOMIC_INT64_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderImageAtomicInt64FeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_BUDGET_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceMemoryBudgetPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_PRIORITY_FEATURES_EXT:
       return Struct::VkPhysicalDeviceMemoryPriorityFeaturesEXT;
    case VK_STRUCTURE_TYPE_MEMORY_PRIORITY_ALLOCATE_INFO_EXT:
       return Struct::VkMemoryPriorityAllocateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEDICATED_ALLOCATION_IMAGE_ALIASING_FEATURES_NV:
       return Struct::VkPhysicalDeviceDedicatedAllocationImageAliasingFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BUFFER_DEVICE_ADDRESS_FEATURES_EXT:
       return Struct::VkPhysicalDeviceBufferDeviceAddressFeaturesEXT;
    case VK_STRUCTURE_TYPE_BUFFER_DEVICE_ADDRESS_CREATE_INFO_EXT:
       return Struct::VkBufferDeviceAddressCreateInfoEXT;
    case VK_STRUCTURE_TYPE_VALIDATION_FEATURES_EXT:
       return Struct::VkValidationFeaturesEXT;
    case VK_STRUCTURE_TYPE_COOPERATIVE_MATRIX_PROPERTIES_NV:
       return Struct::VkCooperativeMatrixPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_FEATURES_NV:
       return Struct::VkPhysicalDeviceCooperativeMatrixFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceCooperativeMatrixPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COVERAGE_REDUCTION_MODE_FEATURES_NV:
       return Struct::VkPhysicalDeviceCoverageReductionModeFeaturesNV;
    case VK_STRUCTURE_TYPE_PIPELINE_COVERAGE_REDUCTION_STATE_CREATE_INFO_NV:
       return Struct::VkPipelineCoverageReductionStateCreateInfoNV;
    case VK_STRUCTURE_TYPE_FRAMEBUFFER_MIXED_SAMPLES_COMBINATION_NV:
       return Struct::VkFramebufferMixedSamplesCombinationNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADER_INTERLOCK_FEATURES_EXT:
       return Struct::VkPhysicalDeviceFragmentShaderInterlockFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_IMAGE_ARRAYS_FEATURES_EXT:
       return Struct::VkPhysicalDeviceYcbcrImageArraysFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROVOKING_VERTEX_FEATURES_EXT:
       return Struct::VkPhysicalDeviceProvokingVertexFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PROVOKING_VERTEX_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceProvokingVertexPropertiesEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_PROVOKING_VERTEX_STATE_CREATE_INFO_EXT:
       return Struct::VkPipelineRasterizationProvokingVertexStateCreateInfoEXT;
    case VK_STRUCTURE_TYPE_SURFACE_FULL_SCREEN_EXCLUSIVE_INFO_EXT:
       return Struct::VkSurfaceFullScreenExclusiveInfoEXT;
    case VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_FULL_SCREEN_EXCLUSIVE_EXT:
       return Struct::VkSurfaceCapabilitiesFullScreenExclusiveEXT;
    case VK_STRUCTURE_TYPE_SURFACE_FULL_SCREEN_EXCLUSIVE_WIN32_INFO_EXT:
       return Struct::VkSurfaceFullScreenExclusiveWin32InfoEXT;
    case VK_STRUCTURE_TYPE_HEADLESS_SURFACE_CREATE_INFO_EXT:
       return Struct::VkHeadlessSurfaceCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_FLOAT_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderAtomicFloatFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_FEATURES_EXT:
       return Struct::VkPhysicalDeviceExtendedDynamicStateFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAP_MEMORY_PLACED_FEATURES_EXT:
       return Struct::VkPhysicalDeviceMapMemoryPlacedFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAP_MEMORY_PLACED_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceMapMemoryPlacedPropertiesEXT;
    case VK_STRUCTURE_TYPE_MEMORY_MAP_PLACED_INFO_EXT:
       return Struct::VkMemoryMapPlacedInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_FLOAT_2_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderAtomicFloat2FeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceDeviceGeneratedCommandsPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_FEATURES_NV:
       return Struct::VkPhysicalDeviceDeviceGeneratedCommandsFeaturesNV;
    case VK_STRUCTURE_TYPE_GRAPHICS_SHADER_GROUP_CREATE_INFO_NV:
       return Struct::VkGraphicsShaderGroupCreateInfoNV;
    case VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_SHADER_GROUPS_CREATE_INFO_NV:
       return Struct::VkGraphicsPipelineShaderGroupsCreateInfoNV;
    case VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_TOKEN_NV:
       return Struct::VkIndirectCommandsLayoutTokenNV;
    case VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_CREATE_INFO_NV:
       return Struct::VkIndirectCommandsLayoutCreateInfoNV;
    case VK_STRUCTURE_TYPE_GENERATED_COMMANDS_INFO_NV:
       return Struct::VkGeneratedCommandsInfoNV;
    case VK_STRUCTURE_TYPE_GENERATED_COMMANDS_MEMORY_REQUIREMENTS_INFO_NV:
       return Struct::VkGeneratedCommandsMemoryRequirementsInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INHERITED_VIEWPORT_SCISSOR_FEATURES_NV:
       return Struct::VkPhysicalDeviceInheritedViewportScissorFeaturesNV;
    case VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_VIEWPORT_SCISSOR_INFO_NV:
       return Struct::VkCommandBufferInheritanceViewportScissorInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXEL_BUFFER_ALIGNMENT_FEATURES_EXT:
       return Struct::VkPhysicalDeviceTexelBufferAlignmentFeaturesEXT;
    case VK_STRUCTURE_TYPE_RENDER_PASS_TRANSFORM_BEGIN_INFO_QCOM:
       return Struct::VkRenderPassTransformBeginInfoQCOM;
    case VK_STRUCTURE_TYPE_COMMAND_BUFFER_INHERITANCE_RENDER_PASS_TRANSFORM_INFO_QCOM:
       return Struct::VkCommandBufferInheritanceRenderPassTransformInfoQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_BIAS_CONTROL_FEATURES_EXT:
       return Struct::VkPhysicalDeviceDepthBiasControlFeaturesEXT;
    case VK_STRUCTURE_TYPE_DEPTH_BIAS_INFO_EXT:
       return Struct::VkDepthBiasInfoEXT;
    case VK_STRUCTURE_TYPE_DEPTH_BIAS_REPRESENTATION_INFO_EXT:
       return Struct::VkDepthBiasRepresentationInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_MEMORY_REPORT_FEATURES_EXT:
       return Struct::VkPhysicalDeviceDeviceMemoryReportFeaturesEXT;
    case VK_STRUCTURE_TYPE_DEVICE_MEMORY_REPORT_CALLBACK_DATA_EXT:
       return Struct::VkDeviceMemoryReportCallbackDataEXT;
    case VK_STRUCTURE_TYPE_DEVICE_DEVICE_MEMORY_REPORT_CREATE_INFO_EXT:
       return Struct::VkDeviceDeviceMemoryReportCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUSTOM_BORDER_COLOR_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceCustomBorderColorPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUSTOM_BORDER_COLOR_FEATURES_EXT:
       return Struct::VkPhysicalDeviceCustomBorderColorFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TEXTURE_COMPRESSION_ASTC_3D_FEATURES_EXT:
       return Struct::VkPhysicalDeviceTextureCompressionASTC3DFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_BARRIER_FEATURES_NV:
       return Struct::VkPhysicalDevicePresentBarrierFeaturesNV;
    case VK_STRUCTURE_TYPE_SURFACE_CAPABILITIES_PRESENT_BARRIER_NV:
       return Struct::VkSurfaceCapabilitiesPresentBarrierNV;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_PRESENT_BARRIER_CREATE_INFO_NV:
       return Struct::VkSwapchainPresentBarrierCreateInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DIAGNOSTICS_CONFIG_FEATURES_NV:
       return Struct::VkPhysicalDeviceDiagnosticsConfigFeaturesNV;
    case VK_STRUCTURE_TYPE_DEVICE_DIAGNOSTICS_CONFIG_CREATE_INFO_NV:
       return Struct::VkDeviceDiagnosticsConfigCreateInfoNV;
    case VK_STRUCTURE_TYPE_PERF_HINT_INFO_QCOM:
       return Struct::VkPerfHintInfoQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_QUEUE_PERF_HINT_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceQueuePerfHintFeaturesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_QUEUE_PERF_HINT_PROPERTIES_QCOM:
       return Struct::VkPhysicalDeviceQueuePerfHintPropertiesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_3_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceImageProcessing3FeaturesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MULTIPLE_WAIT_QUEUES_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceShaderMultipleWaitQueuesFeaturesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MULTIPLE_WAIT_QUEUES_PROPERTIES_QCOM:
       return Struct::VkPhysicalDeviceShaderMultipleWaitQueuesPropertiesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SPLIT_BARRIER_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderSplitBarrierFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SPLIT_BARRIER_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceShaderSplitBarrierPropertiesEXT;
    case VK_STRUCTURE_TYPE_CUDA_MODULE_CREATE_INFO_NV:
       return Struct::VkCudaModuleCreateInfoNV;
    case VK_STRUCTURE_TYPE_CUDA_FUNCTION_CREATE_INFO_NV:
       return Struct::VkCudaFunctionCreateInfoNV;
    case VK_STRUCTURE_TYPE_CUDA_LAUNCH_INFO_NV:
       return Struct::VkCudaLaunchInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUDA_KERNEL_LAUNCH_FEATURES_NV:
       return Struct::VkPhysicalDeviceCudaKernelLaunchFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUDA_KERNEL_LAUNCH_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceCudaKernelLaunchPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_SHADING_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceTileShadingFeaturesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_SHADING_PROPERTIES_QCOM:
       return Struct::VkPhysicalDeviceTileShadingPropertiesQCOM;
    case VK_STRUCTURE_TYPE_RENDER_PASS_TILE_SHADING_CREATE_INFO_QCOM:
       return Struct::VkRenderPassTileShadingCreateInfoQCOM;
    case VK_STRUCTURE_TYPE_PER_TILE_BEGIN_INFO_QCOM:
       return Struct::VkPerTileBeginInfoQCOM;
    case VK_STRUCTURE_TYPE_PER_TILE_END_INFO_QCOM:
       return Struct::VkPerTileEndInfoQCOM;
    case VK_STRUCTURE_TYPE_DISPATCH_TILE_INFO_QCOM:
       return Struct::VkDispatchTileInfoQCOM;
    case VK_STRUCTURE_TYPE_QUERY_LOW_LATENCY_SUPPORT_NV:
       return Struct::VkQueryLowLatencySupportNV;
    case VK_STRUCTURE_TYPE_EXPORT_METAL_OBJECT_CREATE_INFO_EXT:
       return Struct::VkExportMetalObjectCreateInfoEXT;
    case VK_STRUCTURE_TYPE_EXPORT_METAL_OBJECTS_INFO_EXT:
       return Struct::VkExportMetalObjectsInfoEXT;
    case VK_STRUCTURE_TYPE_EXPORT_METAL_DEVICE_INFO_EXT:
       return Struct::VkExportMetalDeviceInfoEXT;
    case VK_STRUCTURE_TYPE_EXPORT_METAL_COMMAND_QUEUE_INFO_EXT:
       return Struct::VkExportMetalCommandQueueInfoEXT;
    case VK_STRUCTURE_TYPE_EXPORT_METAL_BUFFER_INFO_EXT:
       return Struct::VkExportMetalBufferInfoEXT;
    case VK_STRUCTURE_TYPE_IMPORT_METAL_BUFFER_INFO_EXT:
       return Struct::VkImportMetalBufferInfoEXT;
    case VK_STRUCTURE_TYPE_EXPORT_METAL_TEXTURE_INFO_EXT:
       return Struct::VkExportMetalTextureInfoEXT;
    case VK_STRUCTURE_TYPE_IMPORT_METAL_TEXTURE_INFO_EXT:
       return Struct::VkImportMetalTextureInfoEXT;
    case VK_STRUCTURE_TYPE_EXPORT_METAL_IO_SURFACE_INFO_EXT:
       return Struct::VkExportMetalIOSurfaceInfoEXT;
    case VK_STRUCTURE_TYPE_IMPORT_METAL_IO_SURFACE_INFO_EXT:
       return Struct::VkImportMetalIOSurfaceInfoEXT;
    case VK_STRUCTURE_TYPE_EXPORT_METAL_SHARED_EVENT_INFO_EXT:
       return Struct::VkExportMetalSharedEventInfoEXT;
    case VK_STRUCTURE_TYPE_IMPORT_METAL_SHARED_EVENT_INFO_EXT:
       return Struct::VkImportMetalSharedEventInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceDescriptorBufferPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_FEATURES_EXT:
       return Struct::VkPhysicalDeviceDescriptorBufferFeaturesEXT;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_ADDRESS_INFO_EXT:
       return Struct::VkDescriptorAddressInfoEXT;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_BUFFER_BINDING_INFO_EXT:
       return Struct::VkDescriptorBufferBindingInfoEXT;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_BUFFER_BINDING_PUSH_DESCRIPTOR_BUFFER_HANDLE_EXT:
       return Struct::VkDescriptorBufferBindingPushDescriptorBufferHandleEXT;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_GET_INFO_EXT:
       return Struct::VkDescriptorGetInfoEXT;
    case VK_STRUCTURE_TYPE_BUFFER_CAPTURE_DESCRIPTOR_DATA_INFO_EXT:
       return Struct::VkBufferCaptureDescriptorDataInfoEXT;
    case VK_STRUCTURE_TYPE_IMAGE_CAPTURE_DESCRIPTOR_DATA_INFO_EXT:
       return Struct::VkImageCaptureDescriptorDataInfoEXT;
    case VK_STRUCTURE_TYPE_IMAGE_VIEW_CAPTURE_DESCRIPTOR_DATA_INFO_EXT:
       return Struct::VkImageViewCaptureDescriptorDataInfoEXT;
    case VK_STRUCTURE_TYPE_SAMPLER_CAPTURE_DESCRIPTOR_DATA_INFO_EXT:
       return Struct::VkSamplerCaptureDescriptorDataInfoEXT;
    case VK_STRUCTURE_TYPE_OPAQUE_CAPTURE_DESCRIPTOR_DATA_CREATE_INFO_EXT:
       return Struct::VkOpaqueCaptureDescriptorDataCreateInfoEXT;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CAPTURE_DESCRIPTOR_DATA_INFO_EXT:
       return Struct::VkAccelerationStructureCaptureDescriptorDataInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_DENSITY_MAP_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceDescriptorBufferDensityMapPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GRAPHICS_PIPELINE_LIBRARY_FEATURES_EXT:
       return Struct::VkPhysicalDeviceGraphicsPipelineLibraryFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_GRAPHICS_PIPELINE_LIBRARY_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceGraphicsPipelineLibraryPropertiesEXT;
    case VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_LIBRARY_CREATE_INFO_EXT:
       return Struct::VkGraphicsPipelineLibraryCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_EARLY_AND_LATE_FRAGMENT_TESTS_FEATURES_AMD:
       return Struct::VkPhysicalDeviceShaderEarlyAndLateFragmentTestsFeaturesAMD;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_ENUMS_FEATURES_NV:
       return Struct::VkPhysicalDeviceFragmentShadingRateEnumsFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_SHADING_RATE_ENUMS_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceFragmentShadingRateEnumsPropertiesNV;
    case VK_STRUCTURE_TYPE_PIPELINE_FRAGMENT_SHADING_RATE_ENUM_STATE_CREATE_INFO_NV:
       return Struct::VkPipelineFragmentShadingRateEnumStateCreateInfoNV;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_MOTION_TRIANGLES_DATA_NV:
       return Struct::VkAccelerationStructureGeometryMotionTrianglesDataNV;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_MOTION_INFO_NV:
       return Struct::VkAccelerationStructureMotionInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_MOTION_BLUR_FEATURES_NV:
       return Struct::VkPhysicalDeviceRayTracingMotionBlurFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_2_PLANE_444_FORMATS_FEATURES_EXT:
       return Struct::VkPhysicalDeviceYcbcr2Plane444FormatsFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_2_FEATURES_EXT:
       return Struct::VkPhysicalDeviceFragmentDensityMap2FeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_2_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceFragmentDensityMap2PropertiesEXT;
    case VK_STRUCTURE_TYPE_COPY_COMMAND_TRANSFORM_INFO_QCOM:
       return Struct::VkCopyCommandTransformInfoQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_COMPRESSION_CONTROL_FEATURES_EXT:
       return Struct::VkPhysicalDeviceImageCompressionControlFeaturesEXT;
    case VK_STRUCTURE_TYPE_IMAGE_COMPRESSION_CONTROL_EXT:
       return Struct::VkImageCompressionControlEXT;
    case VK_STRUCTURE_TYPE_IMAGE_COMPRESSION_PROPERTIES_EXT:
       return Struct::VkImageCompressionPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ATTACHMENT_FEEDBACK_LOOP_LAYOUT_FEATURES_EXT:
       return Struct::VkPhysicalDeviceAttachmentFeedbackLoopLayoutFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_4444_FORMATS_FEATURES_EXT:
       return Struct::VkPhysicalDevice4444FormatsFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FAULT_FEATURES_EXT:
       return Struct::VkPhysicalDeviceFaultFeaturesEXT;
    case VK_STRUCTURE_TYPE_DEVICE_FAULT_COUNTS_EXT:
       return Struct::VkDeviceFaultCountsEXT;
    case VK_STRUCTURE_TYPE_DEVICE_FAULT_INFO_EXT:
       return Struct::VkDeviceFaultInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RASTERIZATION_ORDER_ATTACHMENT_ACCESS_FEATURES_EXT:
       return Struct::VkPhysicalDeviceRasterizationOrderAttachmentAccessFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RGBA10X6_FORMATS_FEATURES_EXT:
       return Struct::VkPhysicalDeviceRGBA10X6FormatsFeaturesEXT;
    case VK_STRUCTURE_TYPE_DIRECTFB_SURFACE_CREATE_INFO_EXT:
       return Struct::VkDirectFBSurfaceCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MUTABLE_DESCRIPTOR_TYPE_FEATURES_EXT:
       return Struct::VkPhysicalDeviceMutableDescriptorTypeFeaturesEXT;
    case VK_STRUCTURE_TYPE_MUTABLE_DESCRIPTOR_TYPE_CREATE_INFO_EXT:
       return Struct::VkMutableDescriptorTypeCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_INPUT_DYNAMIC_STATE_FEATURES_EXT:
       return Struct::VkPhysicalDeviceVertexInputDynamicStateFeaturesEXT;
    case VK_STRUCTURE_TYPE_VERTEX_INPUT_BINDING_DESCRIPTION_2_EXT:
       return Struct::VkVertexInputBindingDescription2EXT;
    case VK_STRUCTURE_TYPE_VERTEX_INPUT_ATTRIBUTE_DESCRIPTION_2_EXT:
       return Struct::VkVertexInputAttributeDescription2EXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DRM_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceDrmPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ADDRESS_BINDING_REPORT_FEATURES_EXT:
       return Struct::VkPhysicalDeviceAddressBindingReportFeaturesEXT;
    case VK_STRUCTURE_TYPE_DEVICE_ADDRESS_BINDING_CALLBACK_DATA_EXT:
       return Struct::VkDeviceAddressBindingCallbackDataEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLIP_CONTROL_FEATURES_EXT:
       return Struct::VkPhysicalDeviceDepthClipControlFeaturesEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_DEPTH_CLIP_CONTROL_CREATE_INFO_EXT:
       return Struct::VkPipelineViewportDepthClipControlCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIMITIVE_TOPOLOGY_LIST_RESTART_FEATURES_EXT:
       return Struct::VkPhysicalDevicePrimitiveTopologyListRestartFeaturesEXT;
    case VK_STRUCTURE_TYPE_IMPORT_MEMORY_ZIRCON_HANDLE_INFO_FUCHSIA:
       return Struct::VkImportMemoryZirconHandleInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_MEMORY_ZIRCON_HANDLE_PROPERTIES_FUCHSIA:
       return Struct::VkMemoryZirconHandlePropertiesFUCHSIA;
    case VK_STRUCTURE_TYPE_MEMORY_GET_ZIRCON_HANDLE_INFO_FUCHSIA:
       return Struct::VkMemoryGetZirconHandleInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_IMPORT_SEMAPHORE_ZIRCON_HANDLE_INFO_FUCHSIA:
       return Struct::VkImportSemaphoreZirconHandleInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_SEMAPHORE_GET_ZIRCON_HANDLE_INFO_FUCHSIA:
       return Struct::VkSemaphoreGetZirconHandleInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_BUFFER_COLLECTION_CREATE_INFO_FUCHSIA:
       return Struct::VkBufferCollectionCreateInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_IMPORT_MEMORY_BUFFER_COLLECTION_FUCHSIA:
       return Struct::VkImportMemoryBufferCollectionFUCHSIA;
    case VK_STRUCTURE_TYPE_BUFFER_COLLECTION_IMAGE_CREATE_INFO_FUCHSIA:
       return Struct::VkBufferCollectionImageCreateInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_BUFFER_COLLECTION_CONSTRAINTS_INFO_FUCHSIA:
       return Struct::VkBufferCollectionConstraintsInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_BUFFER_CONSTRAINTS_INFO_FUCHSIA:
       return Struct::VkBufferConstraintsInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_BUFFER_COLLECTION_BUFFER_CREATE_INFO_FUCHSIA:
       return Struct::VkBufferCollectionBufferCreateInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_SYSMEM_COLOR_SPACE_FUCHSIA:
       return Struct::VkSysmemColorSpaceFUCHSIA;
    case VK_STRUCTURE_TYPE_BUFFER_COLLECTION_PROPERTIES_FUCHSIA:
       return Struct::VkBufferCollectionPropertiesFUCHSIA;
    case VK_STRUCTURE_TYPE_IMAGE_FORMAT_CONSTRAINTS_INFO_FUCHSIA:
       return Struct::VkImageFormatConstraintsInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_IMAGE_CONSTRAINTS_INFO_FUCHSIA:
       return Struct::VkImageConstraintsInfoFUCHSIA;
    case VK_STRUCTURE_TYPE_SUBPASS_SHADING_PIPELINE_CREATE_INFO_HUAWEI:
       return Struct::VkSubpassShadingPipelineCreateInfoHUAWEI;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_SHADING_FEATURES_HUAWEI:
       return Struct::VkPhysicalDeviceSubpassShadingFeaturesHUAWEI;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_SHADING_PROPERTIES_HUAWEI:
       return Struct::VkPhysicalDeviceSubpassShadingPropertiesHUAWEI;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_INVOCATION_MASK_FEATURES_HUAWEI:
       return Struct::VkPhysicalDeviceInvocationMaskFeaturesHUAWEI;
    case VK_STRUCTURE_TYPE_MEMORY_GET_REMOTE_ADDRESS_INFO_NV:
       return Struct::VkMemoryGetRemoteAddressInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_RDMA_FEATURES_NV:
       return Struct::VkPhysicalDeviceExternalMemoryRDMAFeaturesNV;
    case VK_STRUCTURE_TYPE_PIPELINE_PROPERTIES_IDENTIFIER_EXT:
       return Struct::VkPipelinePropertiesIdentifierEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_PROPERTIES_FEATURES_EXT:
       return Struct::VkPhysicalDevicePipelinePropertiesFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAME_BOUNDARY_FEATURES_EXT:
       return Struct::VkPhysicalDeviceFrameBoundaryFeaturesEXT;
    case VK_STRUCTURE_TYPE_FRAME_BOUNDARY_EXT:
       return Struct::VkFrameBoundaryEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_FEATURES_EXT:
       return Struct::VkPhysicalDeviceMultisampledRenderToSingleSampledFeaturesEXT;
    case VK_STRUCTURE_TYPE_SUBPASS_RESOLVE_PERFORMANCE_QUERY_EXT:
       return Struct::VkSubpassResolvePerformanceQueryEXT;
    case VK_STRUCTURE_TYPE_MULTISAMPLED_RENDER_TO_SINGLE_SAMPLED_INFO_EXT:
       return Struct::VkMultisampledRenderToSingleSampledInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_2_FEATURES_EXT:
       return Struct::VkPhysicalDeviceExtendedDynamicState2FeaturesEXT;
    case VK_STRUCTURE_TYPE_SCREEN_SURFACE_CREATE_INFO_QNX:
       return Struct::VkScreenSurfaceCreateInfoQNX;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COLOR_WRITE_ENABLE_FEATURES_EXT:
       return Struct::VkPhysicalDeviceColorWriteEnableFeaturesEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_COLOR_WRITE_CREATE_INFO_EXT:
       return Struct::VkPipelineColorWriteCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIMITIVES_GENERATED_QUERY_FEATURES_EXT:
       return Struct::VkPhysicalDevicePrimitivesGeneratedQueryFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VIDEO_ENCODE_RGB_CONVERSION_FEATURES_VALVE:
       return Struct::VkPhysicalDeviceVideoEncodeRgbConversionFeaturesVALVE;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_RGB_CONVERSION_CAPABILITIES_VALVE:
       return Struct::VkVideoEncodeRgbConversionCapabilitiesVALVE;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_PROFILE_RGB_CONVERSION_INFO_VALVE:
       return Struct::VkVideoEncodeProfileRgbConversionInfoVALVE;
    case VK_STRUCTURE_TYPE_VIDEO_ENCODE_SESSION_RGB_CONVERSION_CREATE_INFO_VALVE:
       return Struct::VkVideoEncodeSessionRgbConversionCreateInfoVALVE;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_VIEW_MIN_LOD_FEATURES_EXT:
       return Struct::VkPhysicalDeviceImageViewMinLodFeaturesEXT;
    case VK_STRUCTURE_TYPE_IMAGE_VIEW_MIN_LOD_CREATE_INFO_EXT:
       return Struct::VkImageViewMinLodCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTI_DRAW_FEATURES_EXT:
       return Struct::VkPhysicalDeviceMultiDrawFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTI_DRAW_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceMultiDrawPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_2D_VIEW_OF_3D_FEATURES_EXT:
       return Struct::VkPhysicalDeviceImage2DViewOf3DFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TILE_IMAGE_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderTileImageFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_TILE_IMAGE_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceShaderTileImagePropertiesEXT;
    case VK_STRUCTURE_TYPE_MICROMAP_BUILD_INFO_EXT:
       return Struct::VkMicromapBuildInfoEXT;
    case VK_STRUCTURE_TYPE_MICROMAP_CREATE_INFO_EXT:
       return Struct::VkMicromapCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_FEATURES_EXT:
       return Struct::VkPhysicalDeviceOpacityMicromapFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPACITY_MICROMAP_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceOpacityMicromapPropertiesEXT;
    case VK_STRUCTURE_TYPE_MICROMAP_VERSION_INFO_EXT:
       return Struct::VkMicromapVersionInfoEXT;
    case VK_STRUCTURE_TYPE_COPY_MICROMAP_TO_MEMORY_INFO_EXT:
       return Struct::VkCopyMicromapToMemoryInfoEXT;
    case VK_STRUCTURE_TYPE_COPY_MEMORY_TO_MICROMAP_INFO_EXT:
       return Struct::VkCopyMemoryToMicromapInfoEXT;
    case VK_STRUCTURE_TYPE_COPY_MICROMAP_INFO_EXT:
       return Struct::VkCopyMicromapInfoEXT;
    case VK_STRUCTURE_TYPE_MICROMAP_BUILD_SIZES_INFO_EXT:
       return Struct::VkMicromapBuildSizesInfoEXT;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_TRIANGLES_OPACITY_MICROMAP_EXT:
       return Struct::VkAccelerationStructureTrianglesOpacityMicromapEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DISPLACEMENT_MICROMAP_FEATURES_NV:
       return Struct::VkPhysicalDeviceDisplacementMicromapFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DISPLACEMENT_MICROMAP_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceDisplacementMicromapPropertiesNV;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_TRIANGLES_DISPLACEMENT_MICROMAP_NV:
       return Struct::VkAccelerationStructureTrianglesDisplacementMicromapNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_FEATURES_HUAWEI:
       return Struct::VkPhysicalDeviceClusterCullingShaderFeaturesHUAWEI;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_PROPERTIES_HUAWEI:
       return Struct::VkPhysicalDeviceClusterCullingShaderPropertiesHUAWEI;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_CULLING_SHADER_VRS_FEATURES_HUAWEI:
       return Struct::VkPhysicalDeviceClusterCullingShaderVrsFeaturesHUAWEI;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_BORDER_COLOR_SWIZZLE_FEATURES_EXT:
       return Struct::VkPhysicalDeviceBorderColorSwizzleFeaturesEXT;
    case VK_STRUCTURE_TYPE_SAMPLER_BORDER_COLOR_COMPONENT_MAPPING_CREATE_INFO_EXT:
       return Struct::VkSamplerBorderColorComponentMappingCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PAGEABLE_DEVICE_LOCAL_MEMORY_FEATURES_EXT:
       return Struct::VkPhysicalDevicePageableDeviceLocalMemoryFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_PROPERTIES_ARM:
       return Struct::VkPhysicalDeviceShaderCorePropertiesARM;
    case VK_STRUCTURE_TYPE_DEVICE_QUEUE_SHADER_CORE_CONTROL_CREATE_INFO_ARM:
       return Struct::VkDeviceQueueShaderCoreControlCreateInfoARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_FEATURES_ARM:
       return Struct::VkPhysicalDeviceSchedulingControlsFeaturesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_PROPERTIES_ARM:
       return Struct::VkPhysicalDeviceSchedulingControlsPropertiesARM;
    case VK_STRUCTURE_TYPE_DISPATCH_PARAMETERS_ARM:
       return Struct::VkDispatchParametersARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SCHEDULING_CONTROLS_DISPATCH_PARAMETERS_PROPERTIES_ARM:
       return Struct::VkPhysicalDeviceSchedulingControlsDispatchParametersPropertiesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_SLICED_VIEW_OF_3D_FEATURES_EXT:
       return Struct::VkPhysicalDeviceImageSlicedViewOf3DFeaturesEXT;
    case VK_STRUCTURE_TYPE_IMAGE_VIEW_SLICED_CREATE_INFO_EXT:
       return Struct::VkImageViewSlicedCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_SET_HOST_MAPPING_FEATURES_VALVE:
       return Struct::VkPhysicalDeviceDescriptorSetHostMappingFeaturesVALVE;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_BINDING_REFERENCE_VALVE:
       return Struct::VkDescriptorSetBindingReferenceVALVE;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_HOST_MAPPING_INFO_VALVE:
       return Struct::VkDescriptorSetLayoutHostMappingInfoVALVE;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NON_SEAMLESS_CUBE_MAP_FEATURES_EXT:
       return Struct::VkPhysicalDeviceNonSeamlessCubeMapFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RENDER_PASS_STRIPED_FEATURES_ARM:
       return Struct::VkPhysicalDeviceRenderPassStripedFeaturesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RENDER_PASS_STRIPED_PROPERTIES_ARM:
       return Struct::VkPhysicalDeviceRenderPassStripedPropertiesARM;
    case VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_INFO_ARM:
       return Struct::VkRenderPassStripeInfoARM;
    case VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_BEGIN_INFO_ARM:
       return Struct::VkRenderPassStripeBeginInfoARM;
    case VK_STRUCTURE_TYPE_RENDER_PASS_STRIPE_SUBMIT_INFO_ARM:
       return Struct::VkRenderPassStripeSubmitInfoARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_OFFSET_FEATURES_EXT:
       return Struct::VkPhysicalDeviceFragmentDensityMapOffsetFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_OFFSET_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceFragmentDensityMapOffsetPropertiesEXT;
    case VK_STRUCTURE_TYPE_RENDER_PASS_FRAGMENT_DENSITY_MAP_OFFSET_END_INFO_EXT:
       return Struct::VkRenderPassFragmentDensityMapOffsetEndInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COPY_MEMORY_INDIRECT_FEATURES_NV:
       return Struct::VkPhysicalDeviceCopyMemoryIndirectFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_DECOMPRESSION_FEATURES_EXT:
       return Struct::VkPhysicalDeviceMemoryDecompressionFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MEMORY_DECOMPRESSION_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceMemoryDecompressionPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_COMPUTE_FEATURES_NV:
       return Struct::VkPhysicalDeviceDeviceGeneratedCommandsComputeFeaturesNV;
    case VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_INDIRECT_BUFFER_INFO_NV:
       return Struct::VkComputePipelineIndirectBufferInfoNV;
    case VK_STRUCTURE_TYPE_PIPELINE_INDIRECT_DEVICE_ADDRESS_INFO_NV:
       return Struct::VkPipelineIndirectDeviceAddressInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_LINEAR_SWEPT_SPHERES_FEATURES_NV:
       return Struct::VkPhysicalDeviceRayTracingLinearSweptSpheresFeaturesNV;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_LINEAR_SWEPT_SPHERES_DATA_NV:
       return Struct::VkAccelerationStructureGeometryLinearSweptSpheresDataNV;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_SPHERES_DATA_NV:
       return Struct::VkAccelerationStructureGeometrySpheresDataNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LINEAR_COLOR_ATTACHMENT_FEATURES_NV:
       return Struct::VkPhysicalDeviceLinearColorAttachmentFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_COMPRESSION_CONTROL_SWAPCHAIN_FEATURES_EXT:
       return Struct::VkPhysicalDeviceImageCompressionControlSwapchainFeaturesEXT;
    case VK_STRUCTURE_TYPE_IMAGE_VIEW_SAMPLE_WEIGHT_CREATE_INFO_QCOM:
       return Struct::VkImageViewSampleWeightCreateInfoQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceImageProcessingFeaturesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_PROPERTIES_QCOM:
       return Struct::VkPhysicalDeviceImageProcessingPropertiesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NESTED_COMMAND_BUFFER_FEATURES_EXT:
       return Struct::VkPhysicalDeviceNestedCommandBufferFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_NESTED_COMMAND_BUFFER_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceNestedCommandBufferPropertiesEXT;
    case VK_STRUCTURE_TYPE_NATIVE_BUFFER_USAGE_OHOS:
       return Struct::VkNativeBufferUsageOHOS;
    case VK_STRUCTURE_TYPE_NATIVE_BUFFER_PROPERTIES_OHOS:
       return Struct::VkNativeBufferPropertiesOHOS;
    case VK_STRUCTURE_TYPE_NATIVE_BUFFER_FORMAT_PROPERTIES_OHOS:
       return Struct::VkNativeBufferFormatPropertiesOHOS;
    case VK_STRUCTURE_TYPE_IMPORT_NATIVE_BUFFER_INFO_OHOS:
       return Struct::VkImportNativeBufferInfoOHOS;
    case VK_STRUCTURE_TYPE_MEMORY_GET_NATIVE_BUFFER_INFO_OHOS:
       return Struct::VkMemoryGetNativeBufferInfoOHOS;
    case VK_STRUCTURE_TYPE_EXTERNAL_FORMAT_OHOS:
       return Struct::VkExternalFormatOHOS;
    case VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_ACQUIRE_UNMODIFIED_EXT:
       return Struct::VkExternalMemoryAcquireUnmodifiedEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_3_FEATURES_EXT:
       return Struct::VkPhysicalDeviceExtendedDynamicState3FeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_DYNAMIC_STATE_3_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceExtendedDynamicState3PropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SUBPASS_MERGE_FEEDBACK_FEATURES_EXT:
       return Struct::VkPhysicalDeviceSubpassMergeFeedbackFeaturesEXT;
    case VK_STRUCTURE_TYPE_RENDER_PASS_CREATION_CONTROL_EXT:
       return Struct::VkRenderPassCreationControlEXT;
    case VK_STRUCTURE_TYPE_RENDER_PASS_CREATION_FEEDBACK_CREATE_INFO_EXT:
       return Struct::VkRenderPassCreationFeedbackCreateInfoEXT;
    case VK_STRUCTURE_TYPE_RENDER_PASS_SUBPASS_FEEDBACK_CREATE_INFO_EXT:
       return Struct::VkRenderPassSubpassFeedbackCreateInfoEXT;
    case VK_STRUCTURE_TYPE_DIRECT_DRIVER_LOADING_INFO_LUNARG:
       return Struct::VkDirectDriverLoadingInfoLUNARG;
    case VK_STRUCTURE_TYPE_DIRECT_DRIVER_LOADING_LIST_LUNARG:
       return Struct::VkDirectDriverLoadingListLUNARG;
    case VK_STRUCTURE_TYPE_TENSOR_DESCRIPTION_ARM:
       return Struct::VkTensorDescriptionARM;
    case VK_STRUCTURE_TYPE_TENSOR_CREATE_INFO_ARM:
       return Struct::VkTensorCreateInfoARM;
    case VK_STRUCTURE_TYPE_TENSOR_MEMORY_REQUIREMENTS_INFO_ARM:
       return Struct::VkTensorMemoryRequirementsInfoARM;
    case VK_STRUCTURE_TYPE_BIND_TENSOR_MEMORY_INFO_ARM:
       return Struct::VkBindTensorMemoryInfoARM;
    case VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_TENSOR_ARM:
       return Struct::VkWriteDescriptorSetTensorARM;
    case VK_STRUCTURE_TYPE_TENSOR_FORMAT_PROPERTIES_ARM:
       return Struct::VkTensorFormatPropertiesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TENSOR_PROPERTIES_ARM:
       return Struct::VkPhysicalDeviceTensorPropertiesARM;
    case VK_STRUCTURE_TYPE_TENSOR_MEMORY_BARRIER_ARM:
       return Struct::VkTensorMemoryBarrierARM;
    case VK_STRUCTURE_TYPE_TENSOR_DEPENDENCY_INFO_ARM:
       return Struct::VkTensorDependencyInfoARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TENSOR_FEATURES_ARM:
       return Struct::VkPhysicalDeviceTensorFeaturesARM;
    case VK_STRUCTURE_TYPE_DEVICE_TENSOR_MEMORY_REQUIREMENTS_ARM:
       return Struct::VkDeviceTensorMemoryRequirementsARM;
    case VK_STRUCTURE_TYPE_TENSOR_COPY_ARM:
       return Struct::VkTensorCopyARM;
    case VK_STRUCTURE_TYPE_COPY_TENSOR_INFO_ARM:
       return Struct::VkCopyTensorInfoARM;
    case VK_STRUCTURE_TYPE_MEMORY_DEDICATED_ALLOCATE_INFO_TENSOR_ARM:
       return Struct::VkMemoryDedicatedAllocateInfoTensorARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_TENSOR_INFO_ARM:
       return Struct::VkPhysicalDeviceExternalTensorInfoARM;
    case VK_STRUCTURE_TYPE_EXTERNAL_TENSOR_PROPERTIES_ARM:
       return Struct::VkExternalTensorPropertiesARM;
    case VK_STRUCTURE_TYPE_EXTERNAL_MEMORY_TENSOR_CREATE_INFO_ARM:
       return Struct::VkExternalMemoryTensorCreateInfoARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_TENSOR_FEATURES_ARM:
       return Struct::VkPhysicalDeviceDescriptorBufferTensorFeaturesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_BUFFER_TENSOR_PROPERTIES_ARM:
       return Struct::VkPhysicalDeviceDescriptorBufferTensorPropertiesARM;
    case VK_STRUCTURE_TYPE_DESCRIPTOR_GET_TENSOR_INFO_ARM:
       return Struct::VkDescriptorGetTensorInfoARM;
    case VK_STRUCTURE_TYPE_TENSOR_CAPTURE_DESCRIPTOR_DATA_INFO_ARM:
       return Struct::VkTensorCaptureDescriptorDataInfoARM;
    case VK_STRUCTURE_TYPE_TENSOR_VIEW_CAPTURE_DESCRIPTOR_DATA_INFO_ARM:
       return Struct::VkTensorViewCaptureDescriptorDataInfoARM;
    case VK_STRUCTURE_TYPE_FRAME_BOUNDARY_TENSORS_ARM:
       return Struct::VkFrameBoundaryTensorsARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MODULE_IDENTIFIER_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderModuleIdentifierFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MODULE_IDENTIFIER_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceShaderModuleIdentifierPropertiesEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_MODULE_IDENTIFIER_CREATE_INFO_EXT:
       return Struct::VkPipelineShaderStageModuleIdentifierCreateInfoEXT;
    case VK_STRUCTURE_TYPE_SHADER_MODULE_IDENTIFIER_EXT:
       return Struct::VkShaderModuleIdentifierEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPTICAL_FLOW_FEATURES_NV:
       return Struct::VkPhysicalDeviceOpticalFlowFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_OPTICAL_FLOW_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceOpticalFlowPropertiesNV;
    case VK_STRUCTURE_TYPE_OPTICAL_FLOW_IMAGE_FORMAT_INFO_NV:
       return Struct::VkOpticalFlowImageFormatInfoNV;
    case VK_STRUCTURE_TYPE_OPTICAL_FLOW_IMAGE_FORMAT_PROPERTIES_NV:
       return Struct::VkOpticalFlowImageFormatPropertiesNV;
    case VK_STRUCTURE_TYPE_OPTICAL_FLOW_SESSION_CREATE_INFO_NV:
       return Struct::VkOpticalFlowSessionCreateInfoNV;
    case VK_STRUCTURE_TYPE_OPTICAL_FLOW_SESSION_CREATE_PRIVATE_DATA_INFO_NV:
       return Struct::VkOpticalFlowSessionCreatePrivateDataInfoNV;
    case VK_STRUCTURE_TYPE_OPTICAL_FLOW_EXECUTE_INFO_NV:
       return Struct::VkOpticalFlowExecuteInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LEGACY_DITHERING_FEATURES_EXT:
       return Struct::VkPhysicalDeviceLegacyDitheringFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FORMAT_RESOLVE_FEATURES_ANDROID:
       return Struct::VkPhysicalDeviceExternalFormatResolveFeaturesANDROID;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_FORMAT_RESOLVE_PROPERTIES_ANDROID:
       return Struct::VkPhysicalDeviceExternalFormatResolvePropertiesANDROID;
    case VK_STRUCTURE_TYPE_ANDROID_HARDWARE_BUFFER_FORMAT_RESOLVE_PROPERTIES_ANDROID:
       return Struct::VkAndroidHardwareBufferFormatResolvePropertiesANDROID;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ANTI_LAG_FEATURES_AMD:
       return Struct::VkPhysicalDeviceAntiLagFeaturesAMD;
    case VK_STRUCTURE_TYPE_ANTI_LAG_PRESENTATION_INFO_AMD:
       return Struct::VkAntiLagPresentationInfoAMD;
    case VK_STRUCTURE_TYPE_ANTI_LAG_DATA_AMD:
       return Struct::VkAntiLagDataAMD;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DENSE_GEOMETRY_FORMAT_FEATURES_AMDX:
       return Struct::VkPhysicalDeviceDenseGeometryFormatFeaturesAMDX;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_DENSE_GEOMETRY_FORMAT_TRIANGLES_DATA_AMDX:
       return Struct::VkAccelerationStructureDenseGeometryFormatTrianglesDataAMDX;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_OBJECT_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderObjectFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_OBJECT_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceShaderObjectPropertiesEXT;
    case VK_STRUCTURE_TYPE_SHADER_CREATE_INFO_EXT:
       return Struct::VkShaderCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_PROPERTIES_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceTilePropertiesFeaturesQCOM;
    case VK_STRUCTURE_TYPE_TILE_PROPERTIES_QCOM:
       return Struct::VkTilePropertiesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_AMIGO_PROFILING_FEATURES_SEC:
       return Struct::VkPhysicalDeviceAmigoProfilingFeaturesSEC;
    case VK_STRUCTURE_TYPE_AMIGO_PROFILING_SUBMIT_INFO_SEC:
       return Struct::VkAmigoProfilingSubmitInfoSEC;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_VIEWPORTS_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceMultiviewPerViewViewportsFeaturesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceRayTracingInvocationReorderPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_FEATURES_NV:
       return Struct::VkPhysicalDeviceRayTracingInvocationReorderFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_VECTOR_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceCooperativeVectorPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_VECTOR_FEATURES_NV:
       return Struct::VkPhysicalDeviceCooperativeVectorFeaturesNV;
    case VK_STRUCTURE_TYPE_COOPERATIVE_VECTOR_PROPERTIES_NV:
       return Struct::VkCooperativeVectorPropertiesNV;
    case VK_STRUCTURE_TYPE_CONVERT_COOPERATIVE_VECTOR_MATRIX_INFO_NV:
       return Struct::VkConvertCooperativeVectorMatrixInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_SPARSE_ADDRESS_SPACE_FEATURES_NV:
       return Struct::VkPhysicalDeviceExtendedSparseAddressSpaceFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTENDED_SPARSE_ADDRESS_SPACE_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceExtendedSparseAddressSpacePropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LEGACY_VERTEX_ATTRIBUTES_FEATURES_EXT:
       return Struct::VkPhysicalDeviceLegacyVertexAttributesFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LEGACY_VERTEX_ATTRIBUTES_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceLegacyVertexAttributesPropertiesEXT;
    case VK_STRUCTURE_TYPE_LAYER_SETTINGS_CREATE_INFO_EXT:
       return Struct::VkLayerSettingsCreateInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_BUILTINS_FEATURES_ARM:
       return Struct::VkPhysicalDeviceShaderCoreBuiltinsFeaturesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_CORE_BUILTINS_PROPERTIES_ARM:
       return Struct::VkPhysicalDeviceShaderCoreBuiltinsPropertiesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_LIBRARY_GROUP_HANDLES_FEATURES_EXT:
       return Struct::VkPhysicalDevicePipelineLibraryGroupHandlesFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DYNAMIC_RENDERING_UNUSED_ATTACHMENTS_FEATURES_EXT:
       return Struct::VkPhysicalDeviceDynamicRenderingUnusedAttachmentsFeaturesEXT;
    case VK_STRUCTURE_TYPE_LATENCY_SLEEP_MODE_INFO_NV:
       return Struct::VkLatencySleepModeInfoNV;
    case VK_STRUCTURE_TYPE_LATENCY_SLEEP_INFO_NV:
       return Struct::VkLatencySleepInfoNV;
    case VK_STRUCTURE_TYPE_SET_LATENCY_MARKER_INFO_NV:
       return Struct::VkSetLatencyMarkerInfoNV;
    case VK_STRUCTURE_TYPE_LATENCY_TIMINGS_FRAME_REPORT_NV:
       return Struct::VkLatencyTimingsFrameReportNV;
    case VK_STRUCTURE_TYPE_GET_LATENCY_MARKER_INFO_NV:
       return Struct::VkGetLatencyMarkerInfoNV;
    case VK_STRUCTURE_TYPE_LATENCY_SUBMISSION_PRESENT_ID_NV:
       return Struct::VkLatencySubmissionPresentIdNV;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_LATENCY_CREATE_INFO_NV:
       return Struct::VkSwapchainLatencyCreateInfoNV;
    case VK_STRUCTURE_TYPE_OUT_OF_BAND_QUEUE_TYPE_INFO_NV:
       return Struct::VkOutOfBandQueueTypeInfoNV;
    case VK_STRUCTURE_TYPE_LATENCY_SURFACE_CAPABILITIES_NV:
       return Struct::VkLatencySurfaceCapabilitiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DATA_GRAPH_FEATURES_ARM:
       return Struct::VkPhysicalDeviceDataGraphFeaturesARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_CONSTANT_ARM:
       return Struct::VkDataGraphPipelineConstantARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_RESOURCE_INFO_ARM:
       return Struct::VkDataGraphPipelineResourceInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_COMPILER_CONTROL_CREATE_INFO_ARM:
       return Struct::VkDataGraphPipelineCompilerControlCreateInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_CREATE_INFO_ARM:
       return Struct::VkDataGraphPipelineCreateInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SHADER_MODULE_CREATE_INFO_ARM:
       return Struct::VkDataGraphPipelineShaderModuleCreateInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SESSION_CREATE_INFO_ARM:
       return Struct::VkDataGraphPipelineSessionCreateInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SESSION_BIND_POINT_REQUIREMENTS_INFO_ARM:
       return Struct::VkDataGraphPipelineSessionBindPointRequirementsInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SESSION_BIND_POINT_REQUIREMENT_ARM:
       return Struct::VkDataGraphPipelineSessionBindPointRequirementARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SESSION_MEMORY_REQUIREMENTS_INFO_ARM:
       return Struct::VkDataGraphPipelineSessionMemoryRequirementsInfoARM;
    case VK_STRUCTURE_TYPE_BIND_DATA_GRAPH_PIPELINE_SESSION_MEMORY_INFO_ARM:
       return Struct::VkBindDataGraphPipelineSessionMemoryInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_INFO_ARM:
       return Struct::VkDataGraphPipelineInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_PROPERTY_QUERY_RESULT_ARM:
       return Struct::VkDataGraphPipelinePropertyQueryResultARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_IDENTIFIER_CREATE_INFO_ARM:
       return Struct::VkDataGraphPipelineIdentifierCreateInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_DISPATCH_INFO_ARM:
       return Struct::VkDataGraphPipelineDispatchInfoARM;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_DATA_GRAPH_PROPERTIES_ARM:
       return Struct::VkQueueFamilyDataGraphPropertiesARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PROCESSING_ENGINE_CREATE_INFO_ARM:
       return Struct::VkDataGraphProcessingEngineCreateInfoARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_QUEUE_FAMILY_DATA_GRAPH_PROCESSING_ENGINE_INFO_ARM:
       return Struct::VkPhysicalDeviceQueueFamilyDataGraphProcessingEngineInfoARM;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_DATA_GRAPH_PROCESSING_ENGINE_PROPERTIES_ARM:
       return Struct::VkQueueFamilyDataGraphProcessingEnginePropertiesARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_CONSTANT_TENSOR_SEMI_STRUCTURED_SPARSITY_INFO_ARM:
       return Struct::VkDataGraphPipelineConstantTensorSemiStructuredSparsityInfoARM;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_DATA_GRAPH_TOSA_PROPERTIES_ARM:
       return Struct::VkQueueFamilyDataGraphTOSAPropertiesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTIVIEW_PER_VIEW_RENDER_AREAS_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceMultiviewPerViewRenderAreasFeaturesQCOM;
    case VK_STRUCTURE_TYPE_MULTIVIEW_PER_VIEW_RENDER_AREAS_RENDER_PASS_BEGIN_INFO_QCOM:
       return Struct::VkMultiviewPerViewRenderAreasRenderPassBeginInfoQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PER_STAGE_DESCRIPTOR_SET_FEATURES_NV:
       return Struct::VkPhysicalDevicePerStageDescriptorSetFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_2_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceImageProcessing2FeaturesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_PROCESSING_2_PROPERTIES_QCOM:
       return Struct::VkPhysicalDeviceImageProcessing2PropertiesQCOM;
    case VK_STRUCTURE_TYPE_SAMPLER_BLOCK_MATCH_WINDOW_CREATE_INFO_QCOM:
       return Struct::VkSamplerBlockMatchWindowCreateInfoQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUBIC_WEIGHTS_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceCubicWeightsFeaturesQCOM;
    case VK_STRUCTURE_TYPE_SAMPLER_CUBIC_WEIGHTS_CREATE_INFO_QCOM:
       return Struct::VkSamplerCubicWeightsCreateInfoQCOM;
    case VK_STRUCTURE_TYPE_BLIT_IMAGE_CUBIC_WEIGHTS_INFO_QCOM:
       return Struct::VkBlitImageCubicWeightsInfoQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_YCBCR_DEGAMMA_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceYcbcrDegammaFeaturesQCOM;
    case VK_STRUCTURE_TYPE_SAMPLER_YCBCR_CONVERSION_YCBCR_DEGAMMA_CREATE_INFO_QCOM:
       return Struct::VkSamplerYcbcrConversionYcbcrDegammaCreateInfoQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUBIC_CLAMP_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceCubicClampFeaturesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ATTACHMENT_FEEDBACK_LOOP_DYNAMIC_STATE_FEATURES_EXT:
       return Struct::VkPhysicalDeviceAttachmentFeedbackLoopDynamicStateFeaturesEXT;
    case VK_STRUCTURE_TYPE_SCREEN_BUFFER_PROPERTIES_QNX:
       return Struct::VkScreenBufferPropertiesQNX;
    case VK_STRUCTURE_TYPE_SCREEN_BUFFER_FORMAT_PROPERTIES_QNX:
       return Struct::VkScreenBufferFormatPropertiesQNX;
    case VK_STRUCTURE_TYPE_IMPORT_SCREEN_BUFFER_INFO_QNX:
       return Struct::VkImportScreenBufferInfoQNX;
    case VK_STRUCTURE_TYPE_EXTERNAL_FORMAT_QNX:
       return Struct::VkExternalFormatQNX;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_MEMORY_SCREEN_BUFFER_FEATURES_QNX:
       return Struct::VkPhysicalDeviceExternalMemoryScreenBufferFeaturesQNX;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_LAYERED_DRIVER_PROPERTIES_MSFT:
       return Struct::VkPhysicalDeviceLayeredDriverPropertiesMSFT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_POOL_OVERALLOCATION_FEATURES_NV:
       return Struct::VkPhysicalDeviceDescriptorPoolOverallocationFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_MEMORY_HEAP_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceTileMemoryHeapFeaturesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TILE_MEMORY_HEAP_PROPERTIES_QCOM:
       return Struct::VkPhysicalDeviceTileMemoryHeapPropertiesQCOM;
    case VK_STRUCTURE_TYPE_TILE_MEMORY_REQUIREMENTS_QCOM:
       return Struct::VkTileMemoryRequirementsQCOM;
    case VK_STRUCTURE_TYPE_TILE_MEMORY_BIND_INFO_QCOM:
       return Struct::VkTileMemoryBindInfoQCOM;
    case VK_STRUCTURE_TYPE_TILE_MEMORY_SIZE_INFO_QCOM:
       return Struct::VkTileMemorySizeInfoQCOM;
    case VK_STRUCTURE_TYPE_DECOMPRESS_MEMORY_INFO_EXT:
       return Struct::VkDecompressMemoryInfoEXT;
    case VK_STRUCTURE_TYPE_DISPLAY_SURFACE_STEREO_CREATE_INFO_NV:
       return Struct::VkDisplaySurfaceStereoCreateInfoNV;
    case VK_STRUCTURE_TYPE_DISPLAY_MODE_STEREO_PROPERTIES_NV:
       return Struct::VkDisplayModeStereoPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAW_ACCESS_CHAINS_FEATURES_NV:
       return Struct::VkPhysicalDeviceRawAccessChainsFeaturesNV;
    case VK_STRUCTURE_TYPE_EXTERNAL_COMPUTE_QUEUE_DEVICE_CREATE_INFO_NV:
       return Struct::VkExternalComputeQueueDeviceCreateInfoNV;
    case VK_STRUCTURE_TYPE_EXTERNAL_COMPUTE_QUEUE_CREATE_INFO_NV:
       return Struct::VkExternalComputeQueueCreateInfoNV;
    case VK_STRUCTURE_TYPE_EXTERNAL_COMPUTE_QUEUE_DATA_PARAMS_NV:
       return Struct::VkExternalComputeQueueDataParamsNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_EXTERNAL_COMPUTE_QUEUE_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceExternalComputeQueuePropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COMMAND_BUFFER_INHERITANCE_FEATURES_NV:
       return Struct::VkPhysicalDeviceCommandBufferInheritanceFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_ATOMIC_FLOAT16_VECTOR_FEATURES_NV:
       return Struct::VkPhysicalDeviceShaderAtomicFloat16VectorFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_REPLICATED_COMPOSITES_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderReplicatedCompositesFeaturesEXT;
    case VK_STRUCTURE_TYPE_TENSOR_ROLLING_BACKING_CREATE_INFO_ARM:
       return Struct::VkTensorRollingBackingCreateInfoARM;
    case VK_STRUCTURE_TYPE_TENSOR_EXPLICIT_TILING_FORMAT_PROPERTIES_ARM:
       return Struct::VkTensorExplicitTilingFormatPropertiesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_FLOAT8_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderFloat8FeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_VALIDATION_FEATURES_NV:
       return Struct::VkPhysicalDeviceRayTracingValidationFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_ACCELERATION_STRUCTURE_FEATURES_NV:
       return Struct::VkPhysicalDeviceClusterAccelerationStructureFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CLUSTER_ACCELERATION_STRUCTURE_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceClusterAccelerationStructurePropertiesNV;
    case VK_STRUCTURE_TYPE_CLUSTER_ACCELERATION_STRUCTURE_CLUSTERS_BOTTOM_LEVEL_INPUT_NV:
       return Struct::VkClusterAccelerationStructureClustersBottomLevelInputNV;
    case VK_STRUCTURE_TYPE_CLUSTER_ACCELERATION_STRUCTURE_TRIANGLE_CLUSTER_INPUT_NV:
       return Struct::VkClusterAccelerationStructureTriangleClusterInputNV;
    case VK_STRUCTURE_TYPE_CLUSTER_ACCELERATION_STRUCTURE_MOVE_OBJECTS_INPUT_NV:
       return Struct::VkClusterAccelerationStructureMoveObjectsInputNV;
    case VK_STRUCTURE_TYPE_CLUSTER_ACCELERATION_STRUCTURE_INPUT_INFO_NV:
       return Struct::VkClusterAccelerationStructureInputInfoNV;
    case VK_STRUCTURE_TYPE_CLUSTER_ACCELERATION_STRUCTURE_COMMANDS_INFO_NV:
       return Struct::VkClusterAccelerationStructureCommandsInfoNV;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_BUILD_SIZES_INFO_KHR:
       return Struct::VkAccelerationStructureBuildSizesInfoKHR;
    case VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_CLUSTER_ACCELERATION_STRUCTURE_CREATE_INFO_NV:
       return Struct::VkRayTracingPipelineClusterAccelerationStructureCreateInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PARTITIONED_ACCELERATION_STRUCTURE_FEATURES_NV:
       return Struct::VkPhysicalDevicePartitionedAccelerationStructureFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PARTITIONED_ACCELERATION_STRUCTURE_PROPERTIES_NV:
       return Struct::VkPhysicalDevicePartitionedAccelerationStructurePropertiesNV;
    case VK_STRUCTURE_TYPE_PARTITIONED_ACCELERATION_STRUCTURE_FLAGS_NV:
       return Struct::VkPartitionedAccelerationStructureFlagsNV;
    case VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_PARTITIONED_ACCELERATION_STRUCTURE_NV:
       return Struct::VkWriteDescriptorSetPartitionedAccelerationStructureNV;
    case VK_STRUCTURE_TYPE_PARTITIONED_ACCELERATION_STRUCTURE_INSTANCES_INPUT_NV:
       return Struct::VkPartitionedAccelerationStructureInstancesInputNV;
    case VK_STRUCTURE_TYPE_BUILD_PARTITIONED_ACCELERATION_STRUCTURE_INFO_NV:
       return Struct::VkBuildPartitionedAccelerationStructureInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_FEATURES_EXT:
       return Struct::VkPhysicalDeviceDeviceGeneratedCommandsFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEVICE_GENERATED_COMMANDS_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceDeviceGeneratedCommandsPropertiesEXT;
    case VK_STRUCTURE_TYPE_GENERATED_COMMANDS_MEMORY_REQUIREMENTS_INFO_EXT:
       return Struct::VkGeneratedCommandsMemoryRequirementsInfoEXT;
    case VK_STRUCTURE_TYPE_INDIRECT_EXECUTION_SET_PIPELINE_INFO_EXT:
       return Struct::VkIndirectExecutionSetPipelineInfoEXT;
    case VK_STRUCTURE_TYPE_INDIRECT_EXECUTION_SET_SHADER_LAYOUT_INFO_EXT:
       return Struct::VkIndirectExecutionSetShaderLayoutInfoEXT;
    case VK_STRUCTURE_TYPE_INDIRECT_EXECUTION_SET_SHADER_INFO_EXT:
       return Struct::VkIndirectExecutionSetShaderInfoEXT;
    case VK_STRUCTURE_TYPE_INDIRECT_EXECUTION_SET_CREATE_INFO_EXT:
       return Struct::VkIndirectExecutionSetCreateInfoEXT;
    case VK_STRUCTURE_TYPE_GENERATED_COMMANDS_INFO_EXT:
       return Struct::VkGeneratedCommandsInfoEXT;
    case VK_STRUCTURE_TYPE_WRITE_INDIRECT_EXECUTION_SET_PIPELINE_EXT:
       return Struct::VkWriteIndirectExecutionSetPipelineEXT;
    case VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_TOKEN_EXT:
       return Struct::VkIndirectCommandsLayoutTokenEXT;
    case VK_STRUCTURE_TYPE_INDIRECT_COMMANDS_LAYOUT_CREATE_INFO_EXT:
       return Struct::VkIndirectCommandsLayoutCreateInfoEXT;
    case VK_STRUCTURE_TYPE_GENERATED_COMMANDS_PIPELINE_INFO_EXT:
       return Struct::VkGeneratedCommandsPipelineInfoEXT;
    case VK_STRUCTURE_TYPE_GENERATED_COMMANDS_SHADER_INFO_EXT:
       return Struct::VkGeneratedCommandsShaderInfoEXT;
    case VK_STRUCTURE_TYPE_WRITE_INDIRECT_EXECUTION_SET_SHADER_EXT:
       return Struct::VkWriteIndirectExecutionSetShaderEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_ALIGNMENT_CONTROL_FEATURES_MESA:
       return Struct::VkPhysicalDeviceImageAlignmentControlFeaturesMESA;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_ALIGNMENT_CONTROL_PROPERTIES_MESA:
       return Struct::VkPhysicalDeviceImageAlignmentControlPropertiesMESA;
    case VK_STRUCTURE_TYPE_IMAGE_ALIGNMENT_CONTROL_CREATE_INFO_MESA:
       return Struct::VkImageAlignmentControlCreateInfoMESA;
    case VK_STRUCTURE_TYPE_PUSH_CONSTANT_BANK_INFO_NV:
       return Struct::VkPushConstantBankInfoNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PUSH_CONSTANT_BANK_FEATURES_NV:
       return Struct::VkPhysicalDevicePushConstantBankFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PUSH_CONSTANT_BANK_PROPERTIES_NV:
       return Struct::VkPhysicalDevicePushConstantBankPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceRayTracingInvocationReorderPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_INVOCATION_REORDER_FEATURES_EXT:
       return Struct::VkPhysicalDeviceRayTracingInvocationReorderFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DEPTH_CLAMP_CONTROL_FEATURES_EXT:
       return Struct::VkPhysicalDeviceDepthClampControlFeaturesEXT;
    case VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_DEPTH_CLAMP_CONTROL_CREATE_INFO_EXT:
       return Struct::VkPipelineViewportDepthClampControlCreateInfoEXT;
    case VK_STRUCTURE_TYPE_SURFACE_CREATE_INFO_OHOS:
       return Struct::VkSurfaceCreateInfoOHOS;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HDR_VIVID_FEATURES_HUAWEI:
       return Struct::VkPhysicalDeviceHdrVividFeaturesHUAWEI;
    case VK_STRUCTURE_TYPE_HDR_VIVID_DYNAMIC_METADATA_HUAWEI:
       return Struct::VkHdrVividDynamicMetadataHUAWEI;
    case VK_STRUCTURE_TYPE_COOPERATIVE_MATRIX_FLEXIBLE_DIMENSIONS_PROPERTIES_NV:
       return Struct::VkCooperativeMatrixFlexibleDimensionsPropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_2_FEATURES_NV:
       return Struct::VkPhysicalDeviceCooperativeMatrix2FeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_2_PROPERTIES_NV:
       return Struct::VkPhysicalDeviceCooperativeMatrix2PropertiesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_OPACITY_MICROMAP_FEATURES_ARM:
       return Struct::VkPhysicalDevicePipelineOpacityMicromapFeaturesARM;
    case VK_STRUCTURE_TYPE_IMPORT_MEMORY_METAL_HANDLE_INFO_EXT:
       return Struct::VkImportMemoryMetalHandleInfoEXT;
    case VK_STRUCTURE_TYPE_MEMORY_METAL_HANDLE_PROPERTIES_EXT:
       return Struct::VkMemoryMetalHandlePropertiesEXT;
    case VK_STRUCTURE_TYPE_MEMORY_GET_METAL_HANDLE_INFO_EXT:
       return Struct::VkMemoryGetMetalHandleInfoEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_COUNTERS_BY_REGION_FEATURES_ARM:
       return Struct::VkPhysicalDevicePerformanceCountersByRegionFeaturesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PERFORMANCE_COUNTERS_BY_REGION_PROPERTIES_ARM:
       return Struct::VkPhysicalDevicePerformanceCountersByRegionPropertiesARM;
    case VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_ARM:
       return Struct::VkPerformanceCounterARM;
    case VK_STRUCTURE_TYPE_PERFORMANCE_COUNTER_DESCRIPTION_ARM:
       return Struct::VkPerformanceCounterDescriptionARM;
    case VK_STRUCTURE_TYPE_RENDER_PASS_PERFORMANCE_COUNTERS_BY_REGION_BEGIN_INFO_ARM:
       return Struct::VkRenderPassPerformanceCountersByRegionBeginInfoARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INSTRUMENTATION_FEATURES_ARM:
       return Struct::VkPhysicalDeviceShaderInstrumentationFeaturesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_INSTRUMENTATION_PROPERTIES_ARM:
       return Struct::VkPhysicalDeviceShaderInstrumentationPropertiesARM;
    case VK_STRUCTURE_TYPE_SHADER_INSTRUMENTATION_CREATE_INFO_ARM:
       return Struct::VkShaderInstrumentationCreateInfoARM;
    case VK_STRUCTURE_TYPE_SHADER_INSTRUMENTATION_METRIC_DESCRIPTION_ARM:
       return Struct::VkShaderInstrumentationMetricDescriptionARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VERTEX_ATTRIBUTE_ROBUSTNESS_FEATURES_EXT:
       return Struct::VkPhysicalDeviceVertexAttributeRobustnessFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FORMAT_PACK_FEATURES_ARM:
       return Struct::VkPhysicalDeviceFormatPackFeaturesARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_LAYERED_FEATURES_VALVE:
       return Struct::VkPhysicalDeviceFragmentDensityMapLayeredFeaturesVALVE;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FRAGMENT_DENSITY_MAP_LAYERED_PROPERTIES_VALVE:
       return Struct::VkPhysicalDeviceFragmentDensityMapLayeredPropertiesVALVE;
    case VK_STRUCTURE_TYPE_PIPELINE_FRAGMENT_DENSITY_MAP_LAYERED_CREATE_INFO_VALVE:
       return Struct::VkPipelineFragmentDensityMapLayeredCreateInfoVALVE;
    case VK_STRUCTURE_TYPE_SET_PRESENT_CONFIG_NV:
       return Struct::VkSetPresentConfigNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRESENT_METERING_FEATURES_NV:
       return Struct::VkPhysicalDevicePresentMeteringFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MULTISAMPLED_RENDER_TO_SWAPCHAIN_FEATURES_EXT:
       return Struct::VkPhysicalDeviceMultisampledRenderToSwapchainFeaturesEXT;
    case VK_STRUCTURE_TYPE_SWAPCHAIN_FLAGS_SURFACE_CAPABILITIES_EXT:
       return Struct::VkSwapchainFlagsSurfaceCapabilitiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ZERO_INITIALIZE_DEVICE_MEMORY_FEATURES_EXT:
       return Struct::VkPhysicalDeviceZeroInitializeDeviceMemoryFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_64_BIT_INDEXING_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShader64BitIndexingFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_CUSTOM_RESOLVE_FEATURES_EXT:
       return Struct::VkPhysicalDeviceCustomResolveFeaturesEXT;
    case VK_STRUCTURE_TYPE_BEGIN_CUSTOM_RESOLVE_INFO_EXT:
       return Struct::VkBeginCustomResolveInfoEXT;
    case VK_STRUCTURE_TYPE_CUSTOM_RESOLVE_CREATE_INFO_EXT:
       return Struct::VkCustomResolveCreateInfoEXT;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_BUILTIN_MODEL_CREATE_INFO_QCOM:
       return Struct::VkDataGraphPipelineBuiltinModelCreateInfoQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DATA_GRAPH_MODEL_FEATURES_QCOM:
       return Struct::VkPhysicalDeviceDataGraphModelFeaturesQCOM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DATA_GRAPH_OPTICAL_FLOW_FEATURES_ARM:
       return Struct::VkPhysicalDeviceDataGraphOpticalFlowFeaturesARM;
    case VK_STRUCTURE_TYPE_QUEUE_FAMILY_DATA_GRAPH_OPTICAL_FLOW_PROPERTIES_ARM:
       return Struct::VkQueueFamilyDataGraphOpticalFlowPropertiesARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_OPTICAL_FLOW_CREATE_INFO_ARM:
       return Struct::VkDataGraphPipelineOpticalFlowCreateInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_OPTICAL_FLOW_IMAGE_FORMAT_PROPERTIES_ARM:
       return Struct::VkDataGraphOpticalFlowImageFormatPropertiesARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_OPTICAL_FLOW_IMAGE_FORMAT_INFO_ARM:
       return Struct::VkDataGraphOpticalFlowImageFormatInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_OPTICAL_FLOW_DISPATCH_INFO_ARM:
       return Struct::VkDataGraphPipelineOpticalFlowDispatchInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_RESOURCE_INFO_IMAGE_LAYOUT_ARM:
       return Struct::VkDataGraphPipelineResourceInfoImageLayoutARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SINGLE_NODE_CONNECTION_ARM:
       return Struct::VkDataGraphPipelineSingleNodeConnectionARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SINGLE_NODE_CREATE_INFO_ARM:
       return Struct::VkDataGraphPipelineSingleNodeCreateInfoARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_LONG_VECTOR_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderLongVectorFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_LONG_VECTOR_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceShaderLongVectorPropertiesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PIPELINE_CACHE_INCREMENTAL_MODE_FEATURES_SEC:
       return Struct::VkPhysicalDevicePipelineCacheIncrementalModeFeaturesSEC;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_UNIFORM_BUFFER_UNSIZED_ARRAY_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderUniformBufferUnsizedArrayFeaturesEXT;
    case VK_STRUCTURE_TYPE_COMPUTE_OCCUPANCY_PRIORITY_PARAMETERS_NV:
       return Struct::VkComputeOccupancyPriorityParametersNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COMPUTE_OCCUPANCY_PRIORITY_FEATURES_NV:
       return Struct::VkPhysicalDeviceComputeOccupancyPriorityFeaturesNV;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_SUBGROUP_PARTITIONED_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderSubgroupPartitionedFeaturesEXT;
    case VK_STRUCTURE_TYPE_UBM_SURFACE_CREATE_INFO_SEC:
       return Struct::VkUbmSurfaceCreateInfoSEC;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_OCP_MICROSCALING_TYPES_FEATURES_EXT:
       return Struct::VkPhysicalDeviceShaderOCPMicroscalingTypesFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_SHADER_MIXED_FLOAT_DOT_PRODUCT_FEATURES_VALVE:
       return Struct::VkPhysicalDeviceShaderMixedFloatDotProductFeaturesVALVE;
    case VK_STRUCTURE_TYPE_THROTTLE_HINT_SUBMIT_INFO_SEC:
       return Struct::VkThrottleHintSubmitInfoSEC;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_THROTTLE_HINT_FEATURES_SEC:
       return Struct::VkPhysicalDeviceThrottleHintFeaturesSEC;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DATA_GRAPH_NEURAL_ACCELERATOR_STATISTICS_FEATURES_ARM:
       return Struct::VkPhysicalDeviceDataGraphNeuralAcceleratorStatisticsFeaturesARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_NEURAL_STATISTICS_CREATE_INFO_ARM:
       return Struct::VkDataGraphPipelineNeuralStatisticsCreateInfoARM;
    case VK_STRUCTURE_TYPE_DATA_GRAPH_PIPELINE_SESSION_NEURAL_STATISTICS_CREATE_INFO_ARM:
       return Struct::VkDataGraphPipelineSessionNeuralStatisticsCreateInfoARM;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PRIMITIVE_RESTART_INDEX_FEATURES_EXT:
       return Struct::VkPhysicalDevicePrimitiveRestartIndexFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_COOPERATIVE_MATRIX_DECODE_VECTOR_FEATURES_NV:
       return Struct::VkPhysicalDeviceCooperativeMatrixDecodeVectorFeaturesNV;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_TRIANGLES_DATA_KHR:
       return Struct::VkAccelerationStructureGeometryTrianglesDataKHR;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_AABBS_DATA_KHR:
       return Struct::VkAccelerationStructureGeometryAabbsDataKHR;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_INSTANCES_DATA_KHR:
       return Struct::VkAccelerationStructureGeometryInstancesDataKHR;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_GEOMETRY_KHR:
       return Struct::VkAccelerationStructureGeometryKHR;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_BUILD_GEOMETRY_INFO_KHR:
       return Struct::VkAccelerationStructureBuildGeometryInfoKHR;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_CREATE_INFO_KHR:
       return Struct::VkAccelerationStructureCreateInfoKHR;
    case VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET_ACCELERATION_STRUCTURE_KHR:
       return Struct::VkWriteDescriptorSetAccelerationStructureKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ACCELERATION_STRUCTURE_FEATURES_KHR:
       return Struct::VkPhysicalDeviceAccelerationStructureFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_ACCELERATION_STRUCTURE_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceAccelerationStructurePropertiesKHR;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_DEVICE_ADDRESS_INFO_KHR:
       return Struct::VkAccelerationStructureDeviceAddressInfoKHR;
    case VK_STRUCTURE_TYPE_ACCELERATION_STRUCTURE_VERSION_INFO_KHR:
       return Struct::VkAccelerationStructureVersionInfoKHR;
    case VK_STRUCTURE_TYPE_COPY_ACCELERATION_STRUCTURE_TO_MEMORY_INFO_KHR:
       return Struct::VkCopyAccelerationStructureToMemoryInfoKHR;
    case VK_STRUCTURE_TYPE_COPY_MEMORY_TO_ACCELERATION_STRUCTURE_INFO_KHR:
       return Struct::VkCopyMemoryToAccelerationStructureInfoKHR;
    case VK_STRUCTURE_TYPE_COPY_ACCELERATION_STRUCTURE_INFO_KHR:
       return Struct::VkCopyAccelerationStructureInfoKHR;
    case VK_STRUCTURE_TYPE_RAY_TRACING_SHADER_GROUP_CREATE_INFO_KHR:
       return Struct::VkRayTracingShaderGroupCreateInfoKHR;
    case VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_INTERFACE_CREATE_INFO_KHR:
       return Struct::VkRayTracingPipelineInterfaceCreateInfoKHR;
    case VK_STRUCTURE_TYPE_RAY_TRACING_PIPELINE_CREATE_INFO_KHR:
       return Struct::VkRayTracingPipelineCreateInfoKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PIPELINE_FEATURES_KHR:
       return Struct::VkPhysicalDeviceRayTracingPipelineFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_TRACING_PIPELINE_PROPERTIES_KHR:
       return Struct::VkPhysicalDeviceRayTracingPipelinePropertiesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAY_QUERY_FEATURES_KHR:
       return Struct::VkPhysicalDeviceRayQueryFeaturesKHR;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_FEATURES_EXT:
       return Struct::VkPhysicalDeviceMeshShaderFeaturesEXT;
    case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MESH_SHADER_PROPERTIES_EXT:
       return Struct::VkPhysicalDeviceMeshShaderPropertiesEXT;

    default:
        break;
    }
    return Struct::Empty;
}

Func FindAlias(Func func) {
    switch (func) {
    case Func::vkCmdBeginRenderingKHR:
       return Func::vkCmdBeginRendering;
    case Func::vkCmdEndRenderingKHR:
       return Func::vkCmdEndRendering;
    case Func::vkGetPhysicalDeviceFeatures2KHR:
       return Func::vkGetPhysicalDeviceFeatures2;
    case Func::vkGetPhysicalDeviceProperties2KHR:
       return Func::vkGetPhysicalDeviceProperties2;
    case Func::vkGetPhysicalDeviceFormatProperties2KHR:
       return Func::vkGetPhysicalDeviceFormatProperties2;
    case Func::vkGetPhysicalDeviceImageFormatProperties2KHR:
       return Func::vkGetPhysicalDeviceImageFormatProperties2;
    case Func::vkGetPhysicalDeviceQueueFamilyProperties2KHR:
       return Func::vkGetPhysicalDeviceQueueFamilyProperties2;
    case Func::vkGetPhysicalDeviceMemoryProperties2KHR:
       return Func::vkGetPhysicalDeviceMemoryProperties2;
    case Func::vkGetPhysicalDeviceSparseImageFormatProperties2KHR:
       return Func::vkGetPhysicalDeviceSparseImageFormatProperties2;
    case Func::vkGetDeviceGroupPeerMemoryFeaturesKHR:
       return Func::vkGetDeviceGroupPeerMemoryFeatures;
    case Func::vkCmdSetDeviceMaskKHR:
       return Func::vkCmdSetDeviceMask;
    case Func::vkCmdDispatchBaseKHR:
       return Func::vkCmdDispatchBase;
    case Func::vkTrimCommandPoolKHR:
       return Func::vkTrimCommandPool;
    case Func::vkEnumeratePhysicalDeviceGroupsKHR:
       return Func::vkEnumeratePhysicalDeviceGroups;
    case Func::vkGetPhysicalDeviceExternalBufferPropertiesKHR:
       return Func::vkGetPhysicalDeviceExternalBufferProperties;
    case Func::vkGetPhysicalDeviceExternalSemaphorePropertiesKHR:
       return Func::vkGetPhysicalDeviceExternalSemaphoreProperties;
    case Func::vkCmdPushDescriptorSetKHR:
       return Func::vkCmdPushDescriptorSet;
    case Func::vkCmdPushDescriptorSetWithTemplateKHR:
       return Func::vkCmdPushDescriptorSetWithTemplate;
    case Func::vkCreateDescriptorUpdateTemplateKHR:
       return Func::vkCreateDescriptorUpdateTemplate;
    case Func::vkDestroyDescriptorUpdateTemplateKHR:
       return Func::vkDestroyDescriptorUpdateTemplate;
    case Func::vkUpdateDescriptorSetWithTemplateKHR:
       return Func::vkUpdateDescriptorSetWithTemplate;
    case Func::vkCreateRenderPass2KHR:
       return Func::vkCreateRenderPass2;
    case Func::vkCmdBeginRenderPass2KHR:
       return Func::vkCmdBeginRenderPass2;
    case Func::vkCmdNextSubpass2KHR:
       return Func::vkCmdNextSubpass2;
    case Func::vkCmdEndRenderPass2KHR:
       return Func::vkCmdEndRenderPass2;
    case Func::vkGetPhysicalDeviceExternalFencePropertiesKHR:
       return Func::vkGetPhysicalDeviceExternalFenceProperties;
    case Func::vkGetImageMemoryRequirements2KHR:
       return Func::vkGetImageMemoryRequirements2;
    case Func::vkGetBufferMemoryRequirements2KHR:
       return Func::vkGetBufferMemoryRequirements2;
    case Func::vkGetImageSparseMemoryRequirements2KHR:
       return Func::vkGetImageSparseMemoryRequirements2;
    case Func::vkCreateSamplerYcbcrConversionKHR:
       return Func::vkCreateSamplerYcbcrConversion;
    case Func::vkDestroySamplerYcbcrConversionKHR:
       return Func::vkDestroySamplerYcbcrConversion;
    case Func::vkBindBufferMemory2KHR:
       return Func::vkBindBufferMemory2;
    case Func::vkBindImageMemory2KHR:
       return Func::vkBindImageMemory2;
    case Func::vkGetDescriptorSetLayoutSupportKHR:
       return Func::vkGetDescriptorSetLayoutSupport;
    case Func::vkCmdDrawIndirectCountKHR:
       return Func::vkCmdDrawIndirectCount;
    case Func::vkCmdDrawIndexedIndirectCountKHR:
       return Func::vkCmdDrawIndexedIndirectCount;
    case Func::vkGetSemaphoreCounterValueKHR:
       return Func::vkGetSemaphoreCounterValue;
    case Func::vkWaitSemaphoresKHR:
       return Func::vkWaitSemaphores;
    case Func::vkSignalSemaphoreKHR:
       return Func::vkSignalSemaphore;
    case Func::vkCmdSetRenderingAttachmentLocationsKHR:
       return Func::vkCmdSetRenderingAttachmentLocations;
    case Func::vkCmdSetRenderingInputAttachmentIndicesKHR:
       return Func::vkCmdSetRenderingInputAttachmentIndices;
    case Func::vkGetBufferDeviceAddressKHR:
       return Func::vkGetBufferDeviceAddress;
    case Func::vkGetBufferOpaqueCaptureAddressKHR:
       return Func::vkGetBufferOpaqueCaptureAddress;
    case Func::vkGetDeviceMemoryOpaqueCaptureAddressKHR:
       return Func::vkGetDeviceMemoryOpaqueCaptureAddress;
    case Func::vkMapMemory2KHR:
       return Func::vkMapMemory2;
    case Func::vkUnmapMemory2KHR:
       return Func::vkUnmapMemory2;
    case Func::vkCmdSetEvent2KHR:
       return Func::vkCmdSetEvent2;
    case Func::vkCmdResetEvent2KHR:
       return Func::vkCmdResetEvent2;
    case Func::vkCmdWaitEvents2KHR:
       return Func::vkCmdWaitEvents2;
    case Func::vkCmdPipelineBarrier2KHR:
       return Func::vkCmdPipelineBarrier2;
    case Func::vkCmdWriteTimestamp2KHR:
       return Func::vkCmdWriteTimestamp2;
    case Func::vkQueueSubmit2KHR:
       return Func::vkQueueSubmit2;
    case Func::vkCmdCopyBuffer2KHR:
       return Func::vkCmdCopyBuffer2;
    case Func::vkCmdCopyImage2KHR:
       return Func::vkCmdCopyImage2;
    case Func::vkCmdCopyBufferToImage2KHR:
       return Func::vkCmdCopyBufferToImage2;
    case Func::vkCmdCopyImageToBuffer2KHR:
       return Func::vkCmdCopyImageToBuffer2;
    case Func::vkCmdBlitImage2KHR:
       return Func::vkCmdBlitImage2;
    case Func::vkCmdResolveImage2KHR:
       return Func::vkCmdResolveImage2;
    case Func::vkGetDeviceBufferMemoryRequirementsKHR:
       return Func::vkGetDeviceBufferMemoryRequirements;
    case Func::vkGetDeviceImageMemoryRequirementsKHR:
       return Func::vkGetDeviceImageMemoryRequirements;
    case Func::vkGetDeviceImageSparseMemoryRequirementsKHR:
       return Func::vkGetDeviceImageSparseMemoryRequirements;
    case Func::vkCmdBindIndexBuffer2KHR:
       return Func::vkCmdBindIndexBuffer2;
    case Func::vkGetRenderingAreaGranularityKHR:
       return Func::vkGetRenderingAreaGranularity;
    case Func::vkGetDeviceImageSubresourceLayoutKHR:
       return Func::vkGetDeviceImageSubresourceLayout;
    case Func::vkGetImageSubresourceLayout2KHR:
       return Func::vkGetImageSubresourceLayout2;
    case Func::vkCmdSetLineStippleKHR:
       return Func::vkCmdSetLineStipple;
    case Func::vkCmdBindDescriptorSets2KHR:
       return Func::vkCmdBindDescriptorSets2;
    case Func::vkCmdPushConstants2KHR:
       return Func::vkCmdPushConstants2;
    case Func::vkCmdPushDescriptorSet2KHR:
       return Func::vkCmdPushDescriptorSet2;
    case Func::vkCmdPushDescriptorSetWithTemplate2KHR:
       return Func::vkCmdPushDescriptorSetWithTemplate2;
    case Func::vkCmdDrawIndirectCountAMD:
       return Func::vkCmdDrawIndirectCount;
    case Func::vkCmdDrawIndexedIndirectCountAMD:
       return Func::vkCmdDrawIndexedIndirectCount;
    case Func::vkGetRayTracingShaderGroupHandlesNV:
       return Func::vkGetRayTracingShaderGroupHandlesKHR;
    case Func::vkGetPhysicalDeviceCalibrateableTimeDomainsEXT:
       return Func::vkGetPhysicalDeviceCalibrateableTimeDomainsKHR;
    case Func::vkGetCalibratedTimestampsEXT:
       return Func::vkGetCalibratedTimestampsKHR;
    case Func::vkGetBufferDeviceAddressEXT:
       return Func::vkGetBufferDeviceAddress;
    case Func::vkGetPhysicalDeviceToolPropertiesEXT:
       return Func::vkGetPhysicalDeviceToolProperties;
    case Func::vkCmdSetLineStippleEXT:
       return Func::vkCmdSetLineStipple;
    case Func::vkResetQueryPoolEXT:
       return Func::vkResetQueryPool;
    case Func::vkCmdSetCullModeEXT:
       return Func::vkCmdSetCullMode;
    case Func::vkCmdSetFrontFaceEXT:
       return Func::vkCmdSetFrontFace;
    case Func::vkCmdSetPrimitiveTopologyEXT:
       return Func::vkCmdSetPrimitiveTopology;
    case Func::vkCmdSetViewportWithCountEXT:
       return Func::vkCmdSetViewportWithCount;
    case Func::vkCmdSetScissorWithCountEXT:
       return Func::vkCmdSetScissorWithCount;
    case Func::vkCmdBindVertexBuffers2EXT:
       return Func::vkCmdBindVertexBuffers2;
    case Func::vkCmdSetDepthTestEnableEXT:
       return Func::vkCmdSetDepthTestEnable;
    case Func::vkCmdSetDepthWriteEnableEXT:
       return Func::vkCmdSetDepthWriteEnable;
    case Func::vkCmdSetDepthCompareOpEXT:
       return Func::vkCmdSetDepthCompareOp;
    case Func::vkCmdSetDepthBoundsTestEnableEXT:
       return Func::vkCmdSetDepthBoundsTestEnable;
    case Func::vkCmdSetStencilTestEnableEXT:
       return Func::vkCmdSetStencilTestEnable;
    case Func::vkCmdSetStencilOpEXT:
       return Func::vkCmdSetStencilOp;
    case Func::vkCopyMemoryToImageEXT:
       return Func::vkCopyMemoryToImage;
    case Func::vkCopyImageToMemoryEXT:
       return Func::vkCopyImageToMemory;
    case Func::vkCopyImageToImageEXT:
       return Func::vkCopyImageToImage;
    case Func::vkTransitionImageLayoutEXT:
       return Func::vkTransitionImageLayout;
    case Func::vkGetImageSubresourceLayout2EXT:
       return Func::vkGetImageSubresourceLayout2;
    case Func::vkReleaseSwapchainImagesEXT:
       return Func::vkReleaseSwapchainImagesKHR;
    case Func::vkCreatePrivateDataSlotEXT:
       return Func::vkCreatePrivateDataSlot;
    case Func::vkDestroyPrivateDataSlotEXT:
       return Func::vkDestroyPrivateDataSlot;
    case Func::vkSetPrivateDataEXT:
       return Func::vkSetPrivateData;
    case Func::vkGetPrivateDataEXT:
       return Func::vkGetPrivateData;
    case Func::vkCmdSetRasterizerDiscardEnableEXT:
       return Func::vkCmdSetRasterizerDiscardEnable;
    case Func::vkCmdSetDepthBiasEnableEXT:
       return Func::vkCmdSetDepthBiasEnable;
    case Func::vkCmdSetPrimitiveRestartEnableEXT:
       return Func::vkCmdSetPrimitiveRestartEnable;
    case Func::vkCmdEndRendering2EXT:
       return Func::vkCmdEndRendering2KHR;

    default:
        break;
    }
    return func;
}
// clang-format on

std::string String(const Extensions& extensions) {
    std::ostringstream out;
    for (size_t i = 0; i < extensions.size(); i++) {
        out << String(extensions[i]);
        if (i + 1 != extensions.size()) {
            out << " or ";
        }
    }
    return out.str();
}

std::string String(const Requirement& requirement) {
    if (requirement.extension == Extension::Empty) {
        APIVersion api_version(static_cast<uint32_t>(requirement.version));
        return StringAPIVersion(api_version);
    } else {
        return String(requirement.extension);
    }
}

std::string String(const Requirements& requirements) {
    std::ostringstream out;
    for (size_t i = 0; i < requirements.size(); i++) {
        out << String(requirements[i]);
        if (i + 1 != requirements.size()) {
            out << " or ";
        }
    }
    return out.str();
}

vvl::Field Field_VkDescriptorMappingSourceDataEXT(VkDescriptorMappingSourceEXT selector) {
    switch (selector) {
        case VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_CONSTANT_OFFSET_EXT:
            return vvl::Field::constantOffset;
        case VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_PUSH_INDEX_EXT:
            return vvl::Field::pushIndex;
        case VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_EXT:
            return vvl::Field::indirectIndex;
        case VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_INDIRECT_INDEX_ARRAY_EXT:
            return vvl::Field::indirectIndexArray;
        case VK_DESCRIPTOR_MAPPING_SOURCE_RESOURCE_HEAP_DATA_EXT:
            return vvl::Field::heapData;
        case VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_DATA_EXT:
            return vvl::Field::pushDataOffset;
        case VK_DESCRIPTOR_MAPPING_SOURCE_PUSH_ADDRESS_EXT:
            return vvl::Field::pushAddressOffset;
        case VK_DESCRIPTOR_MAPPING_SOURCE_INDIRECT_ADDRESS_EXT:
            return vvl::Field::indirectAddress;
        case VK_DESCRIPTOR_MAPPING_SOURCE_HEAP_WITH_SHADER_RECORD_INDEX_EXT:
            return vvl::Field::shaderRecordIndex;
        case VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_DATA_EXT:
            return vvl::Field::shaderRecordDataOffset;
        case VK_DESCRIPTOR_MAPPING_SOURCE_SHADER_RECORD_ADDRESS_EXT:
            return vvl::Field::shaderRecordAddressOffset;
        default:
            break;
    }
    return vvl::Field::Empty;
}

}  // namespace vvl
// NOLINTEND
